
/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <math.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "libzscanner/scanner.h"
#include "libzscanner/functions.h"
#include "libknot/descriptor.h"

/*! \brief Maximal length of rdata item. */
#define MAX_ITEM_LENGTH		255

/*! \brief Latitude value for equator (2^31). */
#define LOC_LAT_ZERO	(uint32_t)2147483648
/*! \brief Longitude value for meridian (2^31). */
#define LOC_LONG_ZERO	(uint32_t)2147483648
/*! \brief Zero level altitude value. */
#define LOC_ALT_ZERO	(uint32_t)10000000

/*! \brief Shorthand for setting warning data. */
#define WARN(err_code) { s->error.code = err_code; }
/*! \brief Shorthand for setting error data. */
#define ERR(err_code) { WARN(err_code); s->error.fatal = true; }
/*! \brief Shorthand for error reset. */
#define NOERR { WARN(ZS_OK); s->error.fatal = false; }

/*!
 * \brief Writes record type number to r_data.
 *
 * \param type		Type number.
 * \param rdata_tail	Position where to write type number to.
 */
static inline void type_num(const uint16_t type, uint8_t **rdata_tail)
{
	*((uint16_t *)*rdata_tail) = htons(type);
	*rdata_tail += 2;
}

/*!
 * \brief Sets bit to bitmap window.
 *
 * \param type		Type number.
 * \param s		Scanner context.
 */
static inline void window_add_bit(const uint16_t type, zs_scanner_t *s) {
	uint8_t win      = type / 256;
	uint8_t bit_pos  = type % 256;
	uint8_t byte_pos = bit_pos / 8;

	((s->windows[win]).bitmap)[byte_pos] |= 128 >> (bit_pos % 8);

	if ((s->windows[win]).length < byte_pos + 1) {
		(s->windows[win]).length = byte_pos + 1;
	}

	if (s->last_window < win) {
		s->last_window = win;
	}
}

// Include scanner file (in Ragel).






__attribute__((visibility("default")))
int zs_init(
	zs_scanner_t *s,
	const char *origin,
	const uint16_t rclass,
	const uint32_t ttl)
{
	if (s == NULL) {
		return -1;
	}

	memset(s, 0, sizeof(*s));

	// Nonzero initial scanner state.
	s->cs = 1227;

	// Reset the file descriptor.
	s->file.descriptor = -1;

	// Use the root zone as origin if not specified.
	if (origin == NULL || strlen(origin) == 0) {
		origin = ".";
	}
	size_t origin_len = strlen(origin);

	// Prepare a zone settings header.
	const char *format;
	if (origin[origin_len - 1] != '.') {
		format = "$ORIGIN %s.\n";
	} else {
		format = "$ORIGIN %s\n";
	}

	char settings[1024];
	int ret = snprintf(settings, sizeof(settings), format, origin);
	if (ret <= 0 || ret >= sizeof(settings)) {
		ERR(ZS_ENOMEM);
		return -1;
	}

	// Parse the settings to set up the scanner origin.
	if (zs_set_input_string(s, settings, ret) != 0 ||
	    zs_parse_all(s) != 0) {
		return -1;
	}

	// Set scanner defaults.
	s->path = strdup(".");
	if (s->path == NULL) {
		ERR(ZS_ENOMEM);
		return -1;
	}
	s->default_class = rclass;
	s->default_ttl = ttl;
	s->line_counter = 1;

	s->state = ZS_STATE_NONE;
	s->process.automatic = false;

	return 0;
}

static void input_deinit(
	zs_scanner_t *s,
	bool keep_filename)
{
	// Deinit the file input.
	if (s->file.descriptor != -1) {
		// Unmap the file content.
		if (s->input.start != NULL) {
			if (s->input.mmaped) {
				munmap((void *)s->input.start,
				       s->input.end - s->input.start);
			} else {
				free((void *)s->input.start);
			}
		}

		// Close the opened file.
		close(s->file.descriptor);
		s->file.descriptor = -1;
	}

	// Keep file name for possible trailing error report.
	if (!keep_filename) {
		free(s->file.name);
		s->file.name = NULL;
	}

	// Unset the input limits.
	s->input.start   = NULL;
	s->input.current = NULL;
	s->input.end     = NULL;
	s->input.eof     = false;
}

__attribute__((visibility("default")))
void zs_deinit(
	zs_scanner_t *s)
{
	if (s == NULL) {
		return;
	}

	input_deinit(s, false);
	free(s->path);
}

static int set_input_string(
	zs_scanner_t *s,
	const char *input,
	size_t size,
	bool final_block)
{
	if (s == NULL) {
		return -1;
	}

	if (input == NULL) {
		ERR(ZS_EINVAL);
		return -1;
	}

	// Deinit possibly opened file.
	input_deinit(s, final_block);

	// Set the scanner input limits.
	s->input.start   = input;
	s->input.current = input;
	s->input.end     = input + size;
	s->input.eof     = final_block;

	return 0;
}

static char *read_file_to_buf(
	int fd,
	size_t *bufsize)
{
	size_t bufs = 0, newbufs = 8192;
	char *buf = malloc(bufs + newbufs);
	int ret = 0;

	while (buf != NULL && (ret = read(fd, buf + bufs, newbufs)) == newbufs) {
		bufs += newbufs;
		newbufs = bufs;
		char *newbuf = realloc(buf, bufs + newbufs);
		if (newbuf == NULL) {
			free(buf);
		}
		buf = newbuf;
	}
	if (ret < 0) {
		free(buf);
		return NULL;
	}

	*bufsize = bufs + ret;
	return buf;
}

__attribute__((visibility("default")))
int zs_set_input_string(
	zs_scanner_t *s,
	const char *input,
	size_t size)
{
	s->state = ZS_STATE_NONE;

	return set_input_string(s, input, size, false);
}

__attribute__((visibility("default")))
int zs_set_input_file(
	zs_scanner_t *s,
	const char *file_name)
{
	if (s == NULL) {
		return -1;
	}

	if (file_name == NULL) {
		ERR(ZS_EINVAL);
		return -1;
	}

	// Deinit possibly opened file.
	input_deinit(s, false);

	// Try to open the file.
	s->file.descriptor = open(file_name, O_RDONLY);
	if (s->file.descriptor == -1) {
		ERR(ZS_FILE_OPEN);
		return -1;
	}

	char *start = NULL;
	size_t size = 0;

	// Check the input.
	struct stat file_stat;
	if (fstat(s->file.descriptor, &file_stat) == -1) {
		ERR(ZS_FILE_INVALID);
		input_deinit(s, false);
		return -1;
	} else if (S_ISCHR(file_stat.st_mode) ||
	           S_ISBLK(file_stat.st_mode) ||
	           S_ISFIFO(file_stat.st_mode)) {
		// Workaround if cannot mmap, read to memory.
		start = read_file_to_buf(s->file.descriptor, &size);
		if (start == NULL) {
			ERR(ZS_FILE_INVALID);
			input_deinit(s, false);
			return -1;
		}
	} else if (!S_ISREG(file_stat.st_mode)) { // Require regular file.
		ERR(ZS_FILE_INVALID);
		input_deinit(s, false);
		return -1;
	} else if (file_stat.st_size > 0) { // Mmap non-emtpy file.
		start = mmap(0, file_stat.st_size, PROT_READ, MAP_SHARED,
		             s->file.descriptor, 0);
		if (start == MAP_FAILED) {
			ERR(ZS_FILE_INVALID);
			input_deinit(s, false);
			return -1;
		}

		size = file_stat.st_size;
		s->input.mmaped = true;

		// Try to set the mapped memory advise to sequential.
#if defined(MADV_SEQUENTIAL) && !defined(__sun)
		(void)madvise(start, size, MADV_SEQUENTIAL);
#else
#ifdef POSIX_MADV_SEQUENTIAL
		(void)posix_madvise(start, size, POSIX_MADV_SEQUENTIAL);
#endif /* POSIX_MADV_SEQUENTIAL */
#endif /* MADV_SEQUENTIAL && !__sun */
	}

	// Set the scanner input limits.
	s->input.start   = start;
	s->input.current = start;
	s->input.end     = start + size;

	// Get absolute path of the zone file if possible.
	char *full_name = realpath(file_name, NULL);
	if (full_name != NULL) {
		free(s->path);
		s->path = strdup(dirname(full_name));
		free(full_name);
		if (s->path == NULL) {
			ERR(ZS_ENOMEM);
			input_deinit(s, false);
			return -1;
		}
	}

	s->file.name = strdup(file_name);
	if (s->file.name == NULL) {
		ERR(ZS_ENOMEM);
		input_deinit(s, false);
		return -1;
	}

	s->state = ZS_STATE_NONE;

	return 0;
}

__attribute__((visibility("default")))
int zs_set_processing(
	zs_scanner_t *s,
	void (*process_record)(zs_scanner_t *),
	void (*process_error)(zs_scanner_t *),
	void *data)
{
	if (s == NULL) {
		return -1;
	}

	s->process.record = process_record;
	s->process.error = process_error;
	s->process.data = data;

	return 0;
}

__attribute__((visibility("default")))
int zs_set_processing_comment(
	zs_scanner_t *s,
	void (*process_comment)(zs_scanner_t *))
{
	if (s == NULL) {
		return -1;
	}

	s->process.comment = process_comment;

	return 0;
}

typedef enum {
	WRAP_NONE,     // Initial state.
	WRAP_DETECTED, // Input block end is a first '\' in rdata.
	WRAP_PROCESS   // Parsing of auxiliary block = "\".
} wrap_t;

static void parse(
	zs_scanner_t *s,
	wrap_t *wrap)
{
	// Restore scanner input limits (Ragel internals).
	const char *p = s->input.current;
	const char *pe = s->input.end;
	const char *eof = s->input.eof ? pe : NULL;

	// Restore state variables (Ragel internals).
	int cs  = s->cs;
	int top = s->top;
	int stack[ZS_RAGEL_STACK_SIZE];
	memcpy(stack, s->stack, sizeof(stack));

	// Next 2 variables are for better performance.
	// Restoring r_data pointer to next free space.
	uint8_t *rdata_tail = s->r_data + s->r_data_tail;
	// Initialization of the last r_data byte.
	uint8_t *rdata_stop = s->r_data + ZS_MAX_RDATA_LENGTH - 1;

	// Write scanner body (in C).

	{
	short _widec;
	if ( p == pe )
		goto _test_eof;
	goto _resume;

_again:
	switch ( cs ) {
		case 1227: goto st1227;
		case 0: goto st0;
		case 1: goto st1;
		case 2: goto st2;
		case 3: goto st3;
		case 4: goto st4;
		case 5: goto st5;
		case 6: goto st6;
		case 7: goto st7;
		case 8: goto st8;
		case 9: goto st9;
		case 10: goto st10;
		case 11: goto st11;
		case 12: goto st12;
		case 13: goto st13;
		case 1228: goto st1228;
		case 14: goto st14;
		case 15: goto st15;
		case 16: goto st16;
		case 17: goto st17;
		case 18: goto st18;
		case 19: goto st19;
		case 20: goto st20;
		case 21: goto st21;
		case 22: goto st22;
		case 23: goto st23;
		case 24: goto st24;
		case 25: goto st25;
		case 26: goto st26;
		case 27: goto st27;
		case 28: goto st28;
		case 29: goto st29;
		case 30: goto st30;
		case 31: goto st31;
		case 32: goto st32;
		case 33: goto st33;
		case 34: goto st34;
		case 35: goto st35;
		case 36: goto st36;
		case 37: goto st37;
		case 38: goto st38;
		case 39: goto st39;
		case 40: goto st40;
		case 41: goto st41;
		case 42: goto st42;
		case 43: goto st43;
		case 44: goto st44;
		case 45: goto st45;
		case 46: goto st46;
		case 47: goto st47;
		case 48: goto st48;
		case 49: goto st49;
		case 50: goto st50;
		case 51: goto st51;
		case 52: goto st52;
		case 53: goto st53;
		case 54: goto st54;
		case 55: goto st55;
		case 56: goto st56;
		case 57: goto st57;
		case 58: goto st58;
		case 59: goto st59;
		case 60: goto st60;
		case 61: goto st61;
		case 62: goto st62;
		case 63: goto st63;
		case 64: goto st64;
		case 65: goto st65;
		case 66: goto st66;
		case 67: goto st67;
		case 68: goto st68;
		case 69: goto st69;
		case 70: goto st70;
		case 71: goto st71;
		case 72: goto st72;
		case 73: goto st73;
		case 74: goto st74;
		case 75: goto st75;
		case 76: goto st76;
		case 77: goto st77;
		case 78: goto st78;
		case 79: goto st79;
		case 80: goto st80;
		case 81: goto st81;
		case 82: goto st82;
		case 83: goto st83;
		case 84: goto st84;
		case 85: goto st85;
		case 86: goto st86;
		case 87: goto st87;
		case 88: goto st88;
		case 89: goto st89;
		case 90: goto st90;
		case 91: goto st91;
		case 92: goto st92;
		case 93: goto st93;
		case 94: goto st94;
		case 95: goto st95;
		case 96: goto st96;
		case 97: goto st97;
		case 98: goto st98;
		case 99: goto st99;
		case 100: goto st100;
		case 101: goto st101;
		case 102: goto st102;
		case 103: goto st103;
		case 104: goto st104;
		case 105: goto st105;
		case 106: goto st106;
		case 107: goto st107;
		case 108: goto st108;
		case 109: goto st109;
		case 110: goto st110;
		case 111: goto st111;
		case 112: goto st112;
		case 113: goto st113;
		case 114: goto st114;
		case 115: goto st115;
		case 116: goto st116;
		case 117: goto st117;
		case 118: goto st118;
		case 119: goto st119;
		case 120: goto st120;
		case 121: goto st121;
		case 122: goto st122;
		case 123: goto st123;
		case 124: goto st124;
		case 125: goto st125;
		case 126: goto st126;
		case 127: goto st127;
		case 128: goto st128;
		case 129: goto st129;
		case 130: goto st130;
		case 131: goto st131;
		case 132: goto st132;
		case 133: goto st133;
		case 134: goto st134;
		case 135: goto st135;
		case 136: goto st136;
		case 137: goto st137;
		case 138: goto st138;
		case 139: goto st139;
		case 140: goto st140;
		case 141: goto st141;
		case 142: goto st142;
		case 143: goto st143;
		case 144: goto st144;
		case 145: goto st145;
		case 146: goto st146;
		case 147: goto st147;
		case 148: goto st148;
		case 149: goto st149;
		case 150: goto st150;
		case 151: goto st151;
		case 152: goto st152;
		case 153: goto st153;
		case 154: goto st154;
		case 155: goto st155;
		case 156: goto st156;
		case 157: goto st157;
		case 158: goto st158;
		case 159: goto st159;
		case 160: goto st160;
		case 161: goto st161;
		case 162: goto st162;
		case 163: goto st163;
		case 164: goto st164;
		case 165: goto st165;
		case 166: goto st166;
		case 1229: goto st1229;
		case 167: goto st167;
		case 168: goto st168;
		case 169: goto st169;
		case 170: goto st170;
		case 171: goto st171;
		case 172: goto st172;
		case 173: goto st173;
		case 174: goto st174;
		case 175: goto st175;
		case 176: goto st176;
		case 1230: goto st1230;
		case 177: goto st177;
		case 178: goto st178;
		case 179: goto st179;
		case 180: goto st180;
		case 181: goto st181;
		case 182: goto st182;
		case 183: goto st183;
		case 184: goto st184;
		case 1231: goto st1231;
		case 185: goto st185;
		case 186: goto st186;
		case 187: goto st187;
		case 1232: goto st1232;
		case 188: goto st188;
		case 189: goto st189;
		case 190: goto st190;
		case 191: goto st191;
		case 192: goto st192;
		case 193: goto st193;
		case 194: goto st194;
		case 195: goto st195;
		case 196: goto st196;
		case 197: goto st197;
		case 198: goto st198;
		case 1233: goto st1233;
		case 199: goto st199;
		case 200: goto st200;
		case 201: goto st201;
		case 202: goto st202;
		case 1234: goto st1234;
		case 203: goto st203;
		case 204: goto st204;
		case 205: goto st205;
		case 206: goto st206;
		case 207: goto st207;
		case 208: goto st208;
		case 209: goto st209;
		case 210: goto st210;
		case 211: goto st211;
		case 212: goto st212;
		case 213: goto st213;
		case 214: goto st214;
		case 215: goto st215;
		case 216: goto st216;
		case 217: goto st217;
		case 218: goto st218;
		case 219: goto st219;
		case 220: goto st220;
		case 1235: goto st1235;
		case 221: goto st221;
		case 222: goto st222;
		case 223: goto st223;
		case 224: goto st224;
		case 225: goto st225;
		case 226: goto st226;
		case 227: goto st227;
		case 228: goto st228;
		case 229: goto st229;
		case 230: goto st230;
		case 231: goto st231;
		case 232: goto st232;
		case 233: goto st233;
		case 234: goto st234;
		case 235: goto st235;
		case 236: goto st236;
		case 1236: goto st1236;
		case 237: goto st237;
		case 238: goto st238;
		case 239: goto st239;
		case 240: goto st240;
		case 241: goto st241;
		case 242: goto st242;
		case 243: goto st243;
		case 244: goto st244;
		case 245: goto st245;
		case 246: goto st246;
		case 247: goto st247;
		case 248: goto st248;
		case 249: goto st249;
		case 250: goto st250;
		case 251: goto st251;
		case 252: goto st252;
		case 253: goto st253;
		case 254: goto st254;
		case 255: goto st255;
		case 256: goto st256;
		case 257: goto st257;
		case 258: goto st258;
		case 259: goto st259;
		case 260: goto st260;
		case 261: goto st261;
		case 262: goto st262;
		case 263: goto st263;
		case 264: goto st264;
		case 265: goto st265;
		case 266: goto st266;
		case 267: goto st267;
		case 268: goto st268;
		case 269: goto st269;
		case 270: goto st270;
		case 271: goto st271;
		case 272: goto st272;
		case 273: goto st273;
		case 274: goto st274;
		case 275: goto st275;
		case 276: goto st276;
		case 277: goto st277;
		case 278: goto st278;
		case 279: goto st279;
		case 280: goto st280;
		case 281: goto st281;
		case 282: goto st282;
		case 283: goto st283;
		case 284: goto st284;
		case 285: goto st285;
		case 286: goto st286;
		case 287: goto st287;
		case 288: goto st288;
		case 289: goto st289;
		case 290: goto st290;
		case 291: goto st291;
		case 292: goto st292;
		case 293: goto st293;
		case 294: goto st294;
		case 295: goto st295;
		case 296: goto st296;
		case 297: goto st297;
		case 298: goto st298;
		case 299: goto st299;
		case 300: goto st300;
		case 1237: goto st1237;
		case 301: goto st301;
		case 302: goto st302;
		case 1238: goto st1238;
		case 303: goto st303;
		case 304: goto st304;
		case 305: goto st305;
		case 306: goto st306;
		case 307: goto st307;
		case 308: goto st308;
		case 309: goto st309;
		case 310: goto st310;
		case 311: goto st311;
		case 1239: goto st1239;
		case 1240: goto st1240;
		case 312: goto st312;
		case 313: goto st313;
		case 314: goto st314;
		case 315: goto st315;
		case 316: goto st316;
		case 317: goto st317;
		case 318: goto st318;
		case 319: goto st319;
		case 320: goto st320;
		case 321: goto st321;
		case 322: goto st322;
		case 323: goto st323;
		case 324: goto st324;
		case 325: goto st325;
		case 1241: goto st1241;
		case 326: goto st326;
		case 327: goto st327;
		case 328: goto st328;
		case 329: goto st329;
		case 330: goto st330;
		case 331: goto st331;
		case 332: goto st332;
		case 333: goto st333;
		case 334: goto st334;
		case 335: goto st335;
		case 1242: goto st1242;
		case 336: goto st336;
		case 337: goto st337;
		case 338: goto st338;
		case 339: goto st339;
		case 340: goto st340;
		case 341: goto st341;
		case 342: goto st342;
		case 343: goto st343;
		case 344: goto st344;
		case 345: goto st345;
		case 346: goto st346;
		case 347: goto st347;
		case 348: goto st348;
		case 349: goto st349;
		case 1243: goto st1243;
		case 350: goto st350;
		case 351: goto st351;
		case 352: goto st352;
		case 353: goto st353;
		case 354: goto st354;
		case 355: goto st355;
		case 356: goto st356;
		case 1244: goto st1244;
		case 357: goto st357;
		case 358: goto st358;
		case 359: goto st359;
		case 360: goto st360;
		case 361: goto st361;
		case 362: goto st362;
		case 363: goto st363;
		case 364: goto st364;
		case 365: goto st365;
		case 1245: goto st1245;
		case 1246: goto st1246;
		case 1247: goto st1247;
		case 366: goto st366;
		case 367: goto st367;
		case 368: goto st368;
		case 369: goto st369;
		case 370: goto st370;
		case 371: goto st371;
		case 372: goto st372;
		case 373: goto st373;
		case 1248: goto st1248;
		case 1249: goto st1249;
		case 374: goto st374;
		case 375: goto st375;
		case 376: goto st376;
		case 1250: goto st1250;
		case 377: goto st377;
		case 378: goto st378;
		case 379: goto st379;
		case 380: goto st380;
		case 381: goto st381;
		case 382: goto st382;
		case 383: goto st383;
		case 384: goto st384;
		case 385: goto st385;
		case 386: goto st386;
		case 387: goto st387;
		case 388: goto st388;
		case 389: goto st389;
		case 390: goto st390;
		case 391: goto st391;
		case 392: goto st392;
		case 393: goto st393;
		case 394: goto st394;
		case 395: goto st395;
		case 396: goto st396;
		case 397: goto st397;
		case 398: goto st398;
		case 399: goto st399;
		case 400: goto st400;
		case 401: goto st401;
		case 402: goto st402;
		case 403: goto st403;
		case 404: goto st404;
		case 405: goto st405;
		case 406: goto st406;
		case 407: goto st407;
		case 408: goto st408;
		case 409: goto st409;
		case 410: goto st410;
		case 411: goto st411;
		case 412: goto st412;
		case 413: goto st413;
		case 414: goto st414;
		case 415: goto st415;
		case 416: goto st416;
		case 417: goto st417;
		case 418: goto st418;
		case 419: goto st419;
		case 420: goto st420;
		case 421: goto st421;
		case 422: goto st422;
		case 423: goto st423;
		case 424: goto st424;
		case 425: goto st425;
		case 426: goto st426;
		case 427: goto st427;
		case 428: goto st428;
		case 429: goto st429;
		case 430: goto st430;
		case 431: goto st431;
		case 432: goto st432;
		case 433: goto st433;
		case 434: goto st434;
		case 435: goto st435;
		case 436: goto st436;
		case 437: goto st437;
		case 438: goto st438;
		case 439: goto st439;
		case 440: goto st440;
		case 441: goto st441;
		case 442: goto st442;
		case 443: goto st443;
		case 444: goto st444;
		case 445: goto st445;
		case 446: goto st446;
		case 447: goto st447;
		case 448: goto st448;
		case 449: goto st449;
		case 450: goto st450;
		case 451: goto st451;
		case 452: goto st452;
		case 453: goto st453;
		case 454: goto st454;
		case 455: goto st455;
		case 456: goto st456;
		case 457: goto st457;
		case 458: goto st458;
		case 459: goto st459;
		case 460: goto st460;
		case 461: goto st461;
		case 462: goto st462;
		case 463: goto st463;
		case 464: goto st464;
		case 465: goto st465;
		case 466: goto st466;
		case 467: goto st467;
		case 468: goto st468;
		case 469: goto st469;
		case 470: goto st470;
		case 471: goto st471;
		case 472: goto st472;
		case 473: goto st473;
		case 474: goto st474;
		case 475: goto st475;
		case 476: goto st476;
		case 477: goto st477;
		case 478: goto st478;
		case 479: goto st479;
		case 480: goto st480;
		case 481: goto st481;
		case 482: goto st482;
		case 483: goto st483;
		case 484: goto st484;
		case 485: goto st485;
		case 486: goto st486;
		case 487: goto st487;
		case 488: goto st488;
		case 489: goto st489;
		case 490: goto st490;
		case 491: goto st491;
		case 492: goto st492;
		case 493: goto st493;
		case 494: goto st494;
		case 495: goto st495;
		case 496: goto st496;
		case 497: goto st497;
		case 498: goto st498;
		case 499: goto st499;
		case 500: goto st500;
		case 501: goto st501;
		case 502: goto st502;
		case 503: goto st503;
		case 504: goto st504;
		case 505: goto st505;
		case 506: goto st506;
		case 507: goto st507;
		case 508: goto st508;
		case 509: goto st509;
		case 510: goto st510;
		case 511: goto st511;
		case 512: goto st512;
		case 513: goto st513;
		case 514: goto st514;
		case 515: goto st515;
		case 516: goto st516;
		case 517: goto st517;
		case 518: goto st518;
		case 519: goto st519;
		case 520: goto st520;
		case 521: goto st521;
		case 522: goto st522;
		case 523: goto st523;
		case 524: goto st524;
		case 525: goto st525;
		case 526: goto st526;
		case 527: goto st527;
		case 528: goto st528;
		case 529: goto st529;
		case 530: goto st530;
		case 1251: goto st1251;
		case 1252: goto st1252;
		case 1253: goto st1253;
		case 531: goto st531;
		case 532: goto st532;
		case 533: goto st533;
		case 534: goto st534;
		case 535: goto st535;
		case 1254: goto st1254;
		case 536: goto st536;
		case 537: goto st537;
		case 538: goto st538;
		case 539: goto st539;
		case 1255: goto st1255;
		case 1256: goto st1256;
		case 1257: goto st1257;
		case 540: goto st540;
		case 541: goto st541;
		case 1258: goto st1258;
		case 542: goto st542;
		case 543: goto st543;
		case 544: goto st544;
		case 1259: goto st1259;
		case 545: goto st545;
		case 546: goto st546;
		case 547: goto st547;
		case 548: goto st548;
		case 549: goto st549;
		case 550: goto st550;
		case 551: goto st551;
		case 552: goto st552;
		case 553: goto st553;
		case 554: goto st554;
		case 555: goto st555;
		case 556: goto st556;
		case 557: goto st557;
		case 558: goto st558;
		case 559: goto st559;
		case 560: goto st560;
		case 561: goto st561;
		case 562: goto st562;
		case 563: goto st563;
		case 564: goto st564;
		case 565: goto st565;
		case 566: goto st566;
		case 567: goto st567;
		case 568: goto st568;
		case 569: goto st569;
		case 570: goto st570;
		case 571: goto st571;
		case 572: goto st572;
		case 573: goto st573;
		case 574: goto st574;
		case 575: goto st575;
		case 576: goto st576;
		case 577: goto st577;
		case 578: goto st578;
		case 579: goto st579;
		case 580: goto st580;
		case 581: goto st581;
		case 582: goto st582;
		case 583: goto st583;
		case 584: goto st584;
		case 585: goto st585;
		case 586: goto st586;
		case 587: goto st587;
		case 588: goto st588;
		case 589: goto st589;
		case 590: goto st590;
		case 591: goto st591;
		case 592: goto st592;
		case 593: goto st593;
		case 594: goto st594;
		case 595: goto st595;
		case 596: goto st596;
		case 597: goto st597;
		case 598: goto st598;
		case 599: goto st599;
		case 600: goto st600;
		case 601: goto st601;
		case 602: goto st602;
		case 603: goto st603;
		case 604: goto st604;
		case 605: goto st605;
		case 606: goto st606;
		case 607: goto st607;
		case 608: goto st608;
		case 609: goto st609;
		case 610: goto st610;
		case 611: goto st611;
		case 612: goto st612;
		case 613: goto st613;
		case 614: goto st614;
		case 615: goto st615;
		case 616: goto st616;
		case 617: goto st617;
		case 618: goto st618;
		case 619: goto st619;
		case 620: goto st620;
		case 621: goto st621;
		case 622: goto st622;
		case 623: goto st623;
		case 624: goto st624;
		case 625: goto st625;
		case 626: goto st626;
		case 627: goto st627;
		case 628: goto st628;
		case 629: goto st629;
		case 630: goto st630;
		case 631: goto st631;
		case 632: goto st632;
		case 633: goto st633;
		case 634: goto st634;
		case 635: goto st635;
		case 636: goto st636;
		case 637: goto st637;
		case 638: goto st638;
		case 639: goto st639;
		case 640: goto st640;
		case 641: goto st641;
		case 642: goto st642;
		case 643: goto st643;
		case 644: goto st644;
		case 645: goto st645;
		case 646: goto st646;
		case 647: goto st647;
		case 648: goto st648;
		case 1260: goto st1260;
		case 649: goto st649;
		case 650: goto st650;
		case 651: goto st651;
		case 652: goto st652;
		case 653: goto st653;
		case 654: goto st654;
		case 655: goto st655;
		case 656: goto st656;
		case 657: goto st657;
		case 658: goto st658;
		case 659: goto st659;
		case 660: goto st660;
		case 661: goto st661;
		case 662: goto st662;
		case 663: goto st663;
		case 664: goto st664;
		case 665: goto st665;
		case 666: goto st666;
		case 667: goto st667;
		case 668: goto st668;
		case 669: goto st669;
		case 670: goto st670;
		case 671: goto st671;
		case 672: goto st672;
		case 673: goto st673;
		case 674: goto st674;
		case 675: goto st675;
		case 676: goto st676;
		case 677: goto st677;
		case 678: goto st678;
		case 679: goto st679;
		case 680: goto st680;
		case 681: goto st681;
		case 682: goto st682;
		case 683: goto st683;
		case 684: goto st684;
		case 685: goto st685;
		case 686: goto st686;
		case 687: goto st687;
		case 688: goto st688;
		case 689: goto st689;
		case 1261: goto st1261;
		case 690: goto st690;
		case 691: goto st691;
		case 1262: goto st1262;
		case 692: goto st692;
		case 693: goto st693;
		case 694: goto st694;
		case 695: goto st695;
		case 696: goto st696;
		case 697: goto st697;
		case 698: goto st698;
		case 699: goto st699;
		case 700: goto st700;
		case 701: goto st701;
		case 702: goto st702;
		case 703: goto st703;
		case 704: goto st704;
		case 705: goto st705;
		case 1263: goto st1263;
		case 706: goto st706;
		case 707: goto st707;
		case 708: goto st708;
		case 709: goto st709;
		case 710: goto st710;
		case 711: goto st711;
		case 712: goto st712;
		case 713: goto st713;
		case 714: goto st714;
		case 715: goto st715;
		case 716: goto st716;
		case 717: goto st717;
		case 718: goto st718;
		case 719: goto st719;
		case 720: goto st720;
		case 721: goto st721;
		case 722: goto st722;
		case 723: goto st723;
		case 724: goto st724;
		case 725: goto st725;
		case 726: goto st726;
		case 727: goto st727;
		case 1264: goto st1264;
		case 728: goto st728;
		case 729: goto st729;
		case 730: goto st730;
		case 731: goto st731;
		case 732: goto st732;
		case 1265: goto st1265;
		case 733: goto st733;
		case 734: goto st734;
		case 735: goto st735;
		case 736: goto st736;
		case 737: goto st737;
		case 1266: goto st1266;
		case 738: goto st738;
		case 739: goto st739;
		case 740: goto st740;
		case 741: goto st741;
		case 742: goto st742;
		case 1267: goto st1267;
		case 1268: goto st1268;
		case 1269: goto st1269;
		case 743: goto st743;
		case 744: goto st744;
		case 1270: goto st1270;
		case 745: goto st745;
		case 746: goto st746;
		case 747: goto st747;
		case 748: goto st748;
		case 749: goto st749;
		case 750: goto st750;
		case 751: goto st751;
		case 752: goto st752;
		case 753: goto st753;
		case 754: goto st754;
		case 755: goto st755;
		case 756: goto st756;
		case 757: goto st757;
		case 758: goto st758;
		case 759: goto st759;
		case 760: goto st760;
		case 761: goto st761;
		case 762: goto st762;
		case 763: goto st763;
		case 764: goto st764;
		case 765: goto st765;
		case 766: goto st766;
		case 767: goto st767;
		case 768: goto st768;
		case 769: goto st769;
		case 770: goto st770;
		case 771: goto st771;
		case 1271: goto st1271;
		case 1272: goto st1272;
		case 1273: goto st1273;
		case 772: goto st772;
		case 773: goto st773;
		case 774: goto st774;
		case 1274: goto st1274;
		case 1275: goto st1275;
		case 775: goto st775;
		case 776: goto st776;
		case 777: goto st777;
		case 778: goto st778;
		case 1276: goto st1276;
		case 1277: goto st1277;
		case 779: goto st779;
		case 780: goto st780;
		case 781: goto st781;
		case 782: goto st782;
		case 1278: goto st1278;
		case 1279: goto st1279;
		case 783: goto st783;
		case 784: goto st784;
		case 785: goto st785;
		case 786: goto st786;
		case 787: goto st787;
		case 788: goto st788;
		case 789: goto st789;
		case 790: goto st790;
		case 791: goto st791;
		case 792: goto st792;
		case 793: goto st793;
		case 794: goto st794;
		case 795: goto st795;
		case 796: goto st796;
		case 797: goto st797;
		case 798: goto st798;
		case 799: goto st799;
		case 800: goto st800;
		case 801: goto st801;
		case 802: goto st802;
		case 803: goto st803;
		case 804: goto st804;
		case 805: goto st805;
		case 806: goto st806;
		case 807: goto st807;
		case 1280: goto st1280;
		case 808: goto st808;
		case 809: goto st809;
		case 810: goto st810;
		case 811: goto st811;
		case 812: goto st812;
		case 813: goto st813;
		case 814: goto st814;
		case 815: goto st815;
		case 816: goto st816;
		case 817: goto st817;
		case 818: goto st818;
		case 819: goto st819;
		case 820: goto st820;
		case 821: goto st821;
		case 822: goto st822;
		case 1281: goto st1281;
		case 823: goto st823;
		case 824: goto st824;
		case 825: goto st825;
		case 826: goto st826;
		case 827: goto st827;
		case 828: goto st828;
		case 829: goto st829;
		case 830: goto st830;
		case 831: goto st831;
		case 832: goto st832;
		case 833: goto st833;
		case 834: goto st834;
		case 835: goto st835;
		case 1282: goto st1282;
		case 836: goto st836;
		case 837: goto st837;
		case 838: goto st838;
		case 839: goto st839;
		case 840: goto st840;
		case 841: goto st841;
		case 842: goto st842;
		case 843: goto st843;
		case 844: goto st844;
		case 845: goto st845;
		case 846: goto st846;
		case 1283: goto st1283;
		case 1284: goto st1284;
		case 847: goto st847;
		case 848: goto st848;
		case 849: goto st849;
		case 1285: goto st1285;
		case 850: goto st850;
		case 851: goto st851;
		case 852: goto st852;
		case 853: goto st853;
		case 854: goto st854;
		case 855: goto st855;
		case 856: goto st856;
		case 857: goto st857;
		case 858: goto st858;
		case 859: goto st859;
		case 1286: goto st1286;
		case 1287: goto st1287;
		case 1288: goto st1288;
		case 860: goto st860;
		case 861: goto st861;
		case 862: goto st862;
		case 863: goto st863;
		case 864: goto st864;
		case 865: goto st865;
		case 866: goto st866;
		case 867: goto st867;
		case 868: goto st868;
		case 869: goto st869;
		case 870: goto st870;
		case 1289: goto st1289;
		case 1290: goto st1290;
		case 1291: goto st1291;
		case 871: goto st871;
		case 872: goto st872;
		case 873: goto st873;
		case 874: goto st874;
		case 875: goto st875;
		case 876: goto st876;
		case 877: goto st877;
		case 878: goto st878;
		case 879: goto st879;
		case 880: goto st880;
		case 881: goto st881;
		case 882: goto st882;
		case 1292: goto st1292;
		case 883: goto st883;
		case 884: goto st884;
		case 885: goto st885;
		case 1293: goto st1293;
		case 1294: goto st1294;
		case 886: goto st886;
		case 1295: goto st1295;
		case 1296: goto st1296;
		case 887: goto st887;
		case 1297: goto st1297;
		case 1298: goto st1298;
		case 888: goto st888;
		case 889: goto st889;
		case 890: goto st890;
		case 891: goto st891;
		case 892: goto st892;
		case 893: goto st893;
		case 894: goto st894;
		case 895: goto st895;
		case 896: goto st896;
		case 897: goto st897;
		case 898: goto st898;
		case 899: goto st899;
		case 900: goto st900;
		case 901: goto st901;
		case 902: goto st902;
		case 903: goto st903;
		case 904: goto st904;
		case 905: goto st905;
		case 906: goto st906;
		case 907: goto st907;
		case 908: goto st908;
		case 909: goto st909;
		case 910: goto st910;
		case 911: goto st911;
		case 912: goto st912;
		case 913: goto st913;
		case 914: goto st914;
		case 915: goto st915;
		case 916: goto st916;
		case 917: goto st917;
		case 918: goto st918;
		case 919: goto st919;
		case 920: goto st920;
		case 921: goto st921;
		case 922: goto st922;
		case 923: goto st923;
		case 924: goto st924;
		case 925: goto st925;
		case 926: goto st926;
		case 927: goto st927;
		case 928: goto st928;
		case 929: goto st929;
		case 1299: goto st1299;
		case 930: goto st930;
		case 931: goto st931;
		case 932: goto st932;
		case 933: goto st933;
		case 934: goto st934;
		case 935: goto st935;
		case 936: goto st936;
		case 937: goto st937;
		case 938: goto st938;
		case 939: goto st939;
		case 940: goto st940;
		case 941: goto st941;
		case 942: goto st942;
		case 943: goto st943;
		case 944: goto st944;
		case 945: goto st945;
		case 946: goto st946;
		case 947: goto st947;
		case 948: goto st948;
		case 949: goto st949;
		case 950: goto st950;
		case 951: goto st951;
		case 952: goto st952;
		case 953: goto st953;
		case 954: goto st954;
		case 955: goto st955;
		case 956: goto st956;
		case 957: goto st957;
		case 958: goto st958;
		case 959: goto st959;
		case 960: goto st960;
		case 961: goto st961;
		case 962: goto st962;
		case 963: goto st963;
		case 964: goto st964;
		case 965: goto st965;
		case 966: goto st966;
		case 967: goto st967;
		case 968: goto st968;
		case 969: goto st969;
		case 970: goto st970;
		case 971: goto st971;
		case 972: goto st972;
		case 973: goto st973;
		case 974: goto st974;
		case 975: goto st975;
		case 976: goto st976;
		case 977: goto st977;
		case 978: goto st978;
		case 979: goto st979;
		case 980: goto st980;
		case 981: goto st981;
		case 982: goto st982;
		case 983: goto st983;
		case 984: goto st984;
		case 985: goto st985;
		case 986: goto st986;
		case 987: goto st987;
		case 988: goto st988;
		case 989: goto st989;
		case 990: goto st990;
		case 991: goto st991;
		case 992: goto st992;
		case 993: goto st993;
		case 994: goto st994;
		case 995: goto st995;
		case 996: goto st996;
		case 997: goto st997;
		case 998: goto st998;
		case 999: goto st999;
		case 1000: goto st1000;
		case 1001: goto st1001;
		case 1002: goto st1002;
		case 1003: goto st1003;
		case 1004: goto st1004;
		case 1005: goto st1005;
		case 1006: goto st1006;
		case 1007: goto st1007;
		case 1008: goto st1008;
		case 1009: goto st1009;
		case 1010: goto st1010;
		case 1011: goto st1011;
		case 1012: goto st1012;
		case 1013: goto st1013;
		case 1014: goto st1014;
		case 1015: goto st1015;
		case 1016: goto st1016;
		case 1017: goto st1017;
		case 1018: goto st1018;
		case 1019: goto st1019;
		case 1020: goto st1020;
		case 1021: goto st1021;
		case 1022: goto st1022;
		case 1023: goto st1023;
		case 1024: goto st1024;
		case 1025: goto st1025;
		case 1026: goto st1026;
		case 1027: goto st1027;
		case 1028: goto st1028;
		case 1029: goto st1029;
		case 1030: goto st1030;
		case 1031: goto st1031;
		case 1032: goto st1032;
		case 1033: goto st1033;
		case 1034: goto st1034;
		case 1035: goto st1035;
		case 1036: goto st1036;
		case 1037: goto st1037;
		case 1038: goto st1038;
		case 1039: goto st1039;
		case 1040: goto st1040;
		case 1041: goto st1041;
		case 1042: goto st1042;
		case 1043: goto st1043;
		case 1044: goto st1044;
		case 1045: goto st1045;
		case 1046: goto st1046;
		case 1047: goto st1047;
		case 1048: goto st1048;
		case 1049: goto st1049;
		case 1050: goto st1050;
		case 1051: goto st1051;
		case 1052: goto st1052;
		case 1053: goto st1053;
		case 1054: goto st1054;
		case 1055: goto st1055;
		case 1056: goto st1056;
		case 1057: goto st1057;
		case 1058: goto st1058;
		case 1059: goto st1059;
		case 1060: goto st1060;
		case 1061: goto st1061;
		case 1062: goto st1062;
		case 1063: goto st1063;
		case 1064: goto st1064;
		case 1065: goto st1065;
		case 1066: goto st1066;
		case 1067: goto st1067;
		case 1068: goto st1068;
		case 1069: goto st1069;
		case 1070: goto st1070;
		case 1071: goto st1071;
		case 1072: goto st1072;
		case 1073: goto st1073;
		case 1074: goto st1074;
		case 1075: goto st1075;
		case 1076: goto st1076;
		case 1077: goto st1077;
		case 1078: goto st1078;
		case 1079: goto st1079;
		case 1080: goto st1080;
		case 1081: goto st1081;
		case 1082: goto st1082;
		case 1083: goto st1083;
		case 1084: goto st1084;
		case 1085: goto st1085;
		case 1086: goto st1086;
		case 1087: goto st1087;
		case 1088: goto st1088;
		case 1089: goto st1089;
		case 1090: goto st1090;
		case 1091: goto st1091;
		case 1092: goto st1092;
		case 1093: goto st1093;
		case 1300: goto st1300;
		case 1094: goto st1094;
		case 1095: goto st1095;
		case 1096: goto st1096;
		case 1097: goto st1097;
		case 1098: goto st1098;
		case 1099: goto st1099;
		case 1100: goto st1100;
		case 1101: goto st1101;
		case 1301: goto st1301;
		case 1102: goto st1102;
		case 1103: goto st1103;
		case 1104: goto st1104;
		case 1105: goto st1105;
		case 1106: goto st1106;
		case 1302: goto st1302;
		case 1107: goto st1107;
		case 1108: goto st1108;
		case 1109: goto st1109;
		case 1110: goto st1110;
		case 1111: goto st1111;
		case 1112: goto st1112;
		case 1113: goto st1113;
		case 1114: goto st1114;
		case 1115: goto st1115;
		case 1116: goto st1116;
		case 1117: goto st1117;
		case 1118: goto st1118;
		case 1119: goto st1119;
		case 1120: goto st1120;
		case 1121: goto st1121;
		case 1122: goto st1122;
		case 1123: goto st1123;
		case 1124: goto st1124;
		case 1303: goto st1303;
		case 1125: goto st1125;
		case 1126: goto st1126;
		case 1127: goto st1127;
		case 1128: goto st1128;
		case 1129: goto st1129;
		case 1130: goto st1130;
		case 1131: goto st1131;
		case 1132: goto st1132;
		case 1133: goto st1133;
		case 1134: goto st1134;
		case 1135: goto st1135;
		case 1136: goto st1136;
		case 1137: goto st1137;
		case 1138: goto st1138;
		case 1139: goto st1139;
		case 1140: goto st1140;
		case 1141: goto st1141;
		case 1142: goto st1142;
		case 1143: goto st1143;
		case 1304: goto st1304;
		case 1144: goto st1144;
		case 1145: goto st1145;
		case 1146: goto st1146;
		case 1147: goto st1147;
		case 1148: goto st1148;
		case 1149: goto st1149;
		case 1150: goto st1150;
		case 1151: goto st1151;
		case 1152: goto st1152;
		case 1153: goto st1153;
		case 1154: goto st1154;
		case 1155: goto st1155;
		case 1156: goto st1156;
		case 1157: goto st1157;
		case 1158: goto st1158;
		case 1305: goto st1305;
		case 1306: goto st1306;
		case 1307: goto st1307;
		case 1159: goto st1159;
		case 1160: goto st1160;
		case 1161: goto st1161;
		case 1162: goto st1162;
		case 1163: goto st1163;
		case 1164: goto st1164;
		case 1165: goto st1165;
		case 1166: goto st1166;
		case 1308: goto st1308;
		case 1167: goto st1167;
		case 1168: goto st1168;
		case 1169: goto st1169;
		case 1170: goto st1170;
		case 1171: goto st1171;
		case 1172: goto st1172;
		case 1173: goto st1173;
		case 1174: goto st1174;
		case 1175: goto st1175;
		case 1176: goto st1176;
		case 1177: goto st1177;
		case 1309: goto st1309;
		case 1310: goto st1310;
		case 1311: goto st1311;
		case 1178: goto st1178;
		case 1179: goto st1179;
		case 1180: goto st1180;
		case 1181: goto st1181;
		case 1182: goto st1182;
		case 1183: goto st1183;
		case 1184: goto st1184;
		case 1312: goto st1312;
		case 1185: goto st1185;
		case 1186: goto st1186;
		case 1187: goto st1187;
		case 1188: goto st1188;
		case 1189: goto st1189;
		case 1190: goto st1190;
		case 1191: goto st1191;
		case 1192: goto st1192;
		case 1193: goto st1193;
		case 1194: goto st1194;
		case 1195: goto st1195;
		case 1196: goto st1196;
		case 1197: goto st1197;
		case 1313: goto st1313;
		case 1198: goto st1198;
		case 1199: goto st1199;
		case 1200: goto st1200;
		case 1201: goto st1201;
		case 1202: goto st1202;
		case 1203: goto st1203;
		case 1204: goto st1204;
		case 1314: goto st1314;
		case 1205: goto st1205;
		case 1206: goto st1206;
		case 1207: goto st1207;
		case 1208: goto st1208;
		case 1209: goto st1209;
		case 1210: goto st1210;
		case 1315: goto st1315;
		case 1211: goto st1211;
		case 1212: goto st1212;
		case 1213: goto st1213;
		case 1214: goto st1214;
		case 1215: goto st1215;
		case 1216: goto st1216;
		case 1316: goto st1316;
		case 1217: goto st1217;
		case 1218: goto st1218;
		case 1219: goto st1219;
		case 1220: goto st1220;
		case 1221: goto st1221;
		case 1222: goto st1222;
		case 1223: goto st1223;
		case 1224: goto st1224;
		case 1317: goto st1317;
		case 1225: goto st1225;
		case 1226: goto st1226;
	default: break;
	}

	if ( ++p == pe )
		goto _test_eof;
_resume:
	switch ( cs )
	{
tr22:
	{
		s->line_counter++;
	}
	goto st1227;
tr87:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1227; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1227; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1227;
tr93:
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1227; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1227; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1227;
tr96:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1227; goto _out;}
			}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1227; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1227; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1227;
tr731:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1227; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1227;
tr825:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1227; goto _out;}
			}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1227; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1227; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1227;
tr943:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1227; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1227; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1227; goto _out;}
			}
		}
	}
	goto st1227;
tr947:
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1227; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1227; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1227; goto _out;}
			}
		}
	}
	goto st1227;
tr3857:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->line_counter++;
	}
	goto st1227;
tr3921:
	{
		NOERR;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->line_counter++;
	}
	goto st1227;
st1227:
	if ( ++p == pe )
		goto _test_eof1227;
case 1227:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3852;
		case 32: goto tr3852;
		case 36: goto st177;
		case 40: goto tr3854;
		case 41: goto tr3855;
		case 42: goto tr3856;
		case 92: goto tr3856;
		case 95: goto tr3856;
		case 778: goto tr3857;
		case 827: goto tr3858;
		case 1034: goto tr3859;
		case 1083: goto tr3860;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3856;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3856;
	} else
		goto tr3856;
	goto tr3851;
tr0:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr26:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	goto st0;
tr38:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	goto st0;
tr61:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr75:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr83:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr89:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr118:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr145:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr151:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr633:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	goto st0;
tr635:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr732:
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st299;}
	}
	goto st0;
tr743:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	goto st0;
tr757:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	goto st0;
tr785:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr800:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr838:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st299;}
	}
	goto st0;
tr848:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr854:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr867:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr881:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr973:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	goto st0;
tr984:
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st299;}
	}
	goto st0;
tr997:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st299;}
	}
	goto st0;
tr1005:
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st299;}
	}
	goto st0;
tr1011:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st299;}
	}
	goto st0;
tr1022:
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st299;}
	}
	goto st0;
tr1029:
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st299;}
	}
	goto st0;
tr1035:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st299;}
	}
	goto st0;
tr1037:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st299;}
	}
	goto st0;
tr1049:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st299;}
	}
	goto st0;
tr1058:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st299;}
	}
	goto st0;
tr1073:
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st299;}
	}
	goto st0;
tr1085:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st299;}
	}
	goto st0;
tr1101:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st299;}
	}
	goto st0;
tr1115:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st299;}
	}
	goto st0;
tr1116:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr1123:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr1133:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st299;}
	}
	goto st0;
tr1150:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st299;}
	}
	goto st0;
tr1167:
	{
		WARN(ZS_BAD_BASE64_CHAR);
		p--; {goto st299;}
	}
	goto st0;
tr1183:
	{
		WARN(ZS_BAD_BITMAP);
		p--; {goto st299;}
	}
	goto st0;
tr1733:
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr1739:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr1747:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr1809:
	{
		WARN(ZS_BAD_ALGORITHM);
		p--; {goto st299;}
	}
	goto st0;
tr1931:
	{
		WARN(ZS_BAD_CERT_TYPE);
		p--; {goto st299;}
	}
	goto st0;
tr1984:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2007:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2026:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2192:
	{
		s->long_string = false;
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2215:
	{
		WARN(ZS_BAD_LOC_DATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2618:
	{
		WARN(ZS_BAD_APL);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2633:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_APL);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2678:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2749:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2764:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2778:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2801:
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2825:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2832:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr2951:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3007:
	{
		WARN(ZS_BAD_TIMESTAMP_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3419:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3530:
	{
		WARN(ZS_BAD_BASE32HEX_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3769:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3775:
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3781:
	{
		WARN(ZS_BAD_CHAR_DASH);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	goto st0;
tr3851:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr3873:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr3896:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr3915:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
tr3939:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st299;}
	}
	goto st0;
st0:
cs = 0;
	goto _out;
tr2:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1;
tr3:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1;
tr3852:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3854:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3855:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3925:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3926:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3916:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3918:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3919:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
st1:
	if ( ++p == pe )
		goto _test_eof1;
case 1:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1;
		case 32: goto st1;
		case 40: goto tr2;
		case 41: goto tr3;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 778: goto tr22;
		case 827: goto tr23;
		case 1034: goto tr24;
		case 1083: goto tr25;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr4:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st2;
tr30:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st2;
st2:
	if ( ++p == pe )
		goto _test_eof2;
case 2:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr27;
		case 32: goto tr27;
		case 40: goto tr28;
		case 41: goto tr29;
		case 68: goto tr31;
		case 72: goto tr32;
		case 77: goto tr33;
		case 83: goto st191;
		case 87: goto tr35;
		case 100: goto tr31;
		case 104: goto tr32;
		case 109: goto tr33;
		case 115: goto st191;
		case 119: goto tr35;
		case 1034: goto tr36;
		case 1083: goto tr37;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr30;
	goto tr26;
tr40:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st3;
tr41:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st3;
tr59:
	{
		s->line_counter++;
	}
	goto st3;
tr27:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st3;
tr28:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st3;
tr29:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st3;
tr36:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
tr784:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 3; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
tr777:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st3;
tr778:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st3;
tr779:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st3;
tr781:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
st3:
	if ( ++p == pe )
		goto _test_eof3;
case 3:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st3;
		case 32: goto st3;
		case 40: goto tr40;
		case 41: goto tr41;
		case 65: goto tr42;
		case 67: goto tr43;
		case 68: goto tr44;
		case 69: goto tr45;
		case 72: goto tr46;
		case 73: goto tr47;
		case 75: goto tr48;
		case 76: goto tr49;
		case 77: goto tr50;
		case 78: goto tr51;
		case 79: goto tr52;
		case 80: goto tr53;
		case 82: goto tr54;
		case 83: goto tr55;
		case 84: goto tr56;
		case 85: goto tr57;
		case 90: goto tr58;
		case 97: goto tr42;
		case 99: goto tr43;
		case 100: goto tr44;
		case 101: goto tr45;
		case 104: goto tr46;
		case 105: goto tr47;
		case 107: goto tr48;
		case 108: goto tr49;
		case 109: goto tr50;
		case 110: goto tr51;
		case 111: goto tr52;
		case 112: goto tr53;
		case 114: goto tr54;
		case 115: goto tr55;
		case 116: goto tr56;
		case 117: goto tr57;
		case 122: goto tr58;
		case 1034: goto tr59;
		case 1083: goto tr60;
	}
	goto tr38;
tr5:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
tr42:
	{
		s->r_class = s->default_class;
	}
	goto st4;
tr685:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
st4:
	if ( ++p == pe )
		goto _test_eof4;
case 4:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr62;
		case 32: goto tr62;
		case 40: goto tr63;
		case 41: goto tr64;
		case 65: goto st259;
		case 70: goto st262;
		case 80: goto st266;
		case 97: goto st259;
		case 102: goto st262;
		case 112: goto st266;
		case 2058: goto tr68;
		case 2107: goto tr69;
		case 2314: goto tr70;
		case 2363: goto tr70;
		case 2570: goto tr71;
		case 2619: goto tr72;
	}
	goto tr61;
tr76:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr77:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr79:
	{
		s->line_counter++;
	}
	goto st5;
tr62:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr63:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr64:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr68:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr100:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 5; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st5;
tr163:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr164:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr165:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr166:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr185:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr186:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr187:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr188:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr193:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr194:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr195:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr196:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr203:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr204:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr205:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr206:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr214:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr215:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr216:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr217:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr225:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr226:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr227:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr228:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr239:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr240:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr241:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr242:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr251:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr252:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr253:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr254:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr262:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr263:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr264:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr265:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr270:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr271:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr272:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr273:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr283:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr284:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr285:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr286:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr292:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr293:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr294:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr295:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr304:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr305:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr306:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr307:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr347:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr348:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr349:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr350:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr358:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr359:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr360:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr361:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr366:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr367:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr368:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr369:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr379:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr380:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr381:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr382:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr388:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr389:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr390:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr391:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr397:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr398:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr399:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr400:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr405:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr406:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr407:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr408:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr418:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr419:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr420:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr421:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr426:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr427:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr428:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr429:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr440:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr441:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr442:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr443:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr449:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr450:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr451:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr452:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr457:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr458:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr459:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr461:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr467:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr468:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr469:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr471:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr476:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr477:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr478:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr480:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr489:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr490:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr491:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr492:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr506:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr507:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr508:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr509:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr516:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr517:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr518:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr519:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr527:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr528:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr529:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr530:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr538:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr539:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr540:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr541:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr546:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr547:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr548:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr549:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr563:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr564:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr565:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr566:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr572:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr573:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr574:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr575:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr581:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr582:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr583:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr584:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr590:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr591:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr592:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr593:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr601:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr602:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr603:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr604:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr614:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr615:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr616:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr617:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr623:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr624:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr625:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr626:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr636:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr637:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr638:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr640:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr647:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr648:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr649:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr650:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr660:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr661:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr662:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr663:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr910:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr911:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr912:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr913:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr921:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr922:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr923:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr924:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr930:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr931:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st5;
tr932:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st5;
tr933:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
st5:
	if ( ++p == pe )
		goto _test_eof5;
case 5:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st5;
		case 32: goto st5;
		case 40: goto tr76;
		case 41: goto tr77;
		case 92: goto tr78;
		case 2058: goto tr79;
		case 2107: goto tr80;
		case 2314: goto tr73;
		case 2363: goto tr73;
		case 2570: goto tr81;
		case 2619: goto tr82;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr73;
	} else
		goto tr73;
	goto tr75;
tr70:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr73:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr97:
	{
		if (*wrap == WRAP_NONE) {
			p--;
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr98:
	{
		switch (s->r_type) {
		// Next types must not have empty rdata.
		case KNOT_RRTYPE_A:
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
		case KNOT_RRTYPE_SOA:
		case KNOT_RRTYPE_HINFO:
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
		case KNOT_RRTYPE_RP:
		case KNOT_RRTYPE_AAAA:
		case KNOT_RRTYPE_LOC:
		case KNOT_RRTYPE_SRV:
		case KNOT_RRTYPE_NAPTR:
		case KNOT_RRTYPE_CERT:
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_SSHFP:
		case KNOT_RRTYPE_IPSECKEY:
		case KNOT_RRTYPE_RRSIG:
		case KNOT_RRTYPE_NSEC:
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_NSEC3:
		case KNOT_RRTYPE_NSEC3PARAM:
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
		case KNOT_RRTYPE_CDS:
		case KNOT_RRTYPE_CDNSKEY:
		case KNOT_RRTYPE_OPENPGPKEY:
		case KNOT_RRTYPE_CSYNC:
		case KNOT_RRTYPE_ZONEMD:
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L32:
		case KNOT_RRTYPE_L64:
		case KNOT_RRTYPE_LP:
		case KNOT_RRTYPE_EUI48:
		case KNOT_RRTYPE_EUI64:
		case KNOT_RRTYPE_URI:
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st524;}
		// Next types can have empty rdata.
		case KNOT_RRTYPE_APL:
		default:
			{stack[top++] = 6;goto st533;}
		}
	}
	goto st6;
tr168:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr190:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr198:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr208:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr219:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr230:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr244:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr256:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr267:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr275:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr288:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr297:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr309:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr352:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr363:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr371:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr384:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr393:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr402:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr410:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr423:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr431:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr445:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr454:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr463:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr473:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr482:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr494:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr511:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr521:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr532:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr543:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr551:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr568:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr577:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr586:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr595:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr606:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr619:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr628:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr642:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr652:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr665:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr915:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr926:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
tr935:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 6;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 6;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 6;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 6;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st6;
st6:
	if ( ++p == pe )
		goto _test_eof6;
case 6:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr83;
tr91:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st7;
tr92:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st7;
tr84:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st7;
tr85:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st7;
tr86:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st7;
st7:
	if ( ++p == pe )
		goto _test_eof7;
case 7:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr91;
		case 41: goto tr92;
		case 778: goto tr93;
		case 827: goto tr94;
		case 1034: goto tr93;
		case 1083: goto tr94;
	}
	goto tr89;
tr94:
	{
		s->buffer_length = 0;
	}
	goto st8;
tr88:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st8;
tr95:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st8;
tr826:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st8;
tr830:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st8;
st8:
	if ( ++p == pe )
		goto _test_eof8;
case 8:
	if ( (*p) == 10 )
		goto tr96;
	goto tr95;
tr78:
	{
		if (pe - p == 1) {
			*wrap = WRAP_DETECTED;
		}
	}
	goto st9;
st9:
	if ( ++p == pe )
		goto _test_eof9;
case 9:
	if ( (*p) == 35 )
		goto tr98;
	goto tr97;
tr80:
	{
		s->buffer_length = 0;
	}
	goto st10;
tr69:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr99:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st10;
tr167:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr189:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr197:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr207:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr218:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr229:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr243:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr255:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr266:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr274:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr287:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr296:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr308:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr351:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr362:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr370:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr383:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr392:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr401:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr409:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr422:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr430:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr444:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr453:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr462:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr472:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr481:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr493:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr510:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr520:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr531:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr542:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr550:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr567:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr576:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr585:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr594:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr605:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr618:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr627:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr641:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr651:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr664:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr914:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr925:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
tr934:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	goto st10;
st10:
	if ( ++p == pe )
		goto _test_eof10;
case 10:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr100;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr99;
	goto tr75;
tr71:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr81:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr169:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr191:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr199:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr209:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr220:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr231:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr245:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr257:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr268:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr276:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr289:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr298:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr310:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr353:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr364:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr372:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr385:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr394:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr403:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr411:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr424:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr432:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr446:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr455:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr464:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr474:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr483:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr495:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr512:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr522:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr533:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr544:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr552:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr569:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr578:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr587:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr596:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr607:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr620:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr629:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr643:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr653:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr666:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr916:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr927:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
tr936:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 11;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 11;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 11;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 11;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st11;
st11:
	if ( ++p == pe )
		goto _test_eof11;
case 11:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr101;
		case 32: goto tr101;
		case 40: goto tr102;
		case 41: goto tr103;
		case 92: goto tr78;
		case 1802: goto tr87;
		case 1851: goto tr88;
		case 2058: goto tr87;
		case 2107: goto tr104;
		case 2314: goto tr105;
		case 2363: goto tr106;
		case 2570: goto tr87;
		case 2619: goto tr107;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr73;
	} else
		goto tr73;
	goto tr83;
tr109:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st12;
tr110:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st12;
tr101:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st12;
tr102:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st12;
tr103:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st12;
st12:
	if ( ++p == pe )
		goto _test_eof12;
case 12:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st12;
		case 32: goto st12;
		case 40: goto tr109;
		case 41: goto tr110;
		case 92: goto tr78;
		case 1802: goto tr93;
		case 1851: goto tr94;
		case 2058: goto tr93;
		case 2107: goto tr111;
		case 2314: goto tr112;
		case 2363: goto tr113;
		case 2570: goto tr93;
		case 2619: goto tr114;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr73;
	} else
		goto tr73;
	goto tr83;
tr111:
	{
		s->buffer_length = 0;
	}
	goto st13;
tr115:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st13;
tr104:
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st13;
tr174:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st13;
tr178:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st13;
tr805:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st13;
st13:
	if ( ++p == pe )
		goto _test_eof13;
case 13:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr96;
		case 1034: goto tr116;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr115;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr83;
tr116:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1228; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1228; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1228; goto _out;}
		}
	}
	goto st1228;
tr813:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1228; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1228; goto _out;}
		}
	}
	goto st1228;
tr804:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1228; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1228; goto _out;}
		}
	}
	goto st1228;
st1228:
	if ( ++p == pe )
		goto _test_eof1228;
case 1228:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3861;
		case 32: goto tr3861;
		case 36: goto tr3862;
		case 40: goto tr3863;
		case 41: goto tr3864;
		case 42: goto tr3865;
		case 58: goto tr73;
		case 92: goto tr3866;
		case 95: goto tr3865;
		case 1802: goto tr3857;
		case 1851: goto tr3858;
		case 2058: goto tr3867;
		case 2107: goto tr3868;
		case 2314: goto tr3869;
		case 2363: goto tr3870;
		case 2570: goto tr3871;
		case 2619: goto tr3872;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr73;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3865;
		} else
			goto tr73;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3865;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr73;
			} else if ( _widec >= 97 )
				goto tr3865;
		} else
			goto tr73;
	} else
		goto tr73;
	goto tr848;
tr119:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st14;
tr120:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st14;
tr3861:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st14;
tr3863:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st14;
tr3864:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st14;
st14:
	if ( ++p == pe )
		goto _test_eof14;
case 14:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st14;
		case 32: goto st14;
		case 40: goto tr119;
		case 41: goto tr120;
		case 58: goto tr73;
		case 65: goto tr122;
		case 67: goto tr123;
		case 68: goto tr124;
		case 69: goto tr125;
		case 72: goto tr126;
		case 73: goto tr127;
		case 75: goto tr128;
		case 76: goto tr129;
		case 77: goto tr130;
		case 78: goto tr131;
		case 79: goto tr132;
		case 80: goto tr133;
		case 82: goto tr134;
		case 83: goto tr135;
		case 84: goto tr136;
		case 85: goto tr137;
		case 90: goto tr138;
		case 92: goto tr78;
		case 97: goto tr122;
		case 99: goto tr123;
		case 100: goto tr124;
		case 101: goto tr125;
		case 104: goto tr126;
		case 105: goto tr127;
		case 107: goto tr128;
		case 108: goto tr129;
		case 109: goto tr130;
		case 110: goto tr131;
		case 111: goto tr132;
		case 112: goto tr133;
		case 114: goto tr134;
		case 115: goto tr135;
		case 116: goto tr136;
		case 117: goto tr137;
		case 122: goto tr138;
		case 1802: goto tr22;
		case 1851: goto tr23;
		case 2058: goto tr139;
		case 2107: goto tr140;
		case 2314: goto tr141;
		case 2363: goto tr142;
		case 2570: goto tr143;
		case 2619: goto tr144;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr73;
		} else if ( _widec >= 48 )
			goto tr121;
	} else
		goto tr73;
	goto tr118;
tr121:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 15;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 15;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 15;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 15;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 15;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 15;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 15;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 15;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 15;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 15;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 15;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 15;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 15;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 15;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 15;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 15;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 15;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 15;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 15;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 15;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 15;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 15;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 15;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 15;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 15;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 15;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 15;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 15;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 15;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 15;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 15;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st15;
st15:
	if ( ++p == pe )
		goto _test_eof15;
case 15:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr146;
		case 32: goto tr146;
		case 40: goto tr147;
		case 41: goto tr148;
		case 68: goto tr31;
		case 72: goto tr32;
		case 77: goto tr33;
		case 83: goto st191;
		case 87: goto tr35;
		case 100: goto tr31;
		case 104: goto tr32;
		case 109: goto tr33;
		case 115: goto st191;
		case 119: goto tr35;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr149;
		case 1083: goto tr150;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr30;
	goto tr145;
tr153:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st16;
tr154:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st16;
tr146:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st16;
tr147:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st16;
tr148:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st16;
st16:
	if ( ++p == pe )
		goto _test_eof16;
case 16:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st16;
		case 32: goto st16;
		case 40: goto tr153;
		case 41: goto tr154;
		case 65: goto tr42;
		case 67: goto tr43;
		case 68: goto tr44;
		case 69: goto tr45;
		case 72: goto tr46;
		case 73: goto tr47;
		case 75: goto tr48;
		case 76: goto tr49;
		case 77: goto tr50;
		case 78: goto tr51;
		case 79: goto tr52;
		case 80: goto tr53;
		case 82: goto tr54;
		case 83: goto tr55;
		case 84: goto tr56;
		case 85: goto tr57;
		case 90: goto tr58;
		case 97: goto tr42;
		case 99: goto tr43;
		case 100: goto tr44;
		case 101: goto tr45;
		case 104: goto tr46;
		case 105: goto tr47;
		case 107: goto tr48;
		case 108: goto tr49;
		case 109: goto tr50;
		case 110: goto tr51;
		case 111: goto tr52;
		case 112: goto tr53;
		case 114: goto tr54;
		case 115: goto tr55;
		case 116: goto tr56;
		case 117: goto tr57;
		case 122: goto tr58;
		case 778: goto tr93;
		case 827: goto tr94;
		case 1034: goto tr155;
		case 1083: goto tr156;
	}
	goto tr151;
tr6:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st17;
tr43:
	{
		s->r_class = s->default_class;
	}
	goto st17;
tr686:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st17;
st17:
	if ( ++p == pe )
		goto _test_eof17;
case 17:
	switch( (*p) ) {
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 83: goto st36;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 115: goto st36;
	}
	goto tr38;
st18:
	if ( ++p == pe )
		goto _test_eof18;
case 18:
	switch( (*p) ) {
		case 65: goto st19;
		case 97: goto st19;
	}
	goto tr38;
st19:
	if ( ++p == pe )
		goto _test_eof19;
case 19:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr163;
		case 32: goto tr163;
		case 40: goto tr164;
		case 41: goto tr165;
		case 2058: goto tr166;
		case 2107: goto tr167;
		case 2314: goto tr168;
		case 2363: goto tr168;
		case 2570: goto tr169;
		case 2619: goto tr170;
	}
	goto tr61;
tr72:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr82:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr170:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr192:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr200:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr210:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr221:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr232:
	{ s->r_type = KNOT_RRTYPE_CSYNC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr246:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr258:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr269:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr277:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr290:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr299:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr311:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr354:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr365:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr373:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr386:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr395:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr404:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr412:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr425:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr433:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr447:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr456:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr465:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr475:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr484:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr496:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr513:
	{ s->r_type = KNOT_RRTYPE_OPENPGPKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr523:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr534:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr545:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr553:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr570:
	{ s->r_type = KNOT_RRTYPE_SMIMEA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr579:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr588:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr597:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr608:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr621:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr630:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr644:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr654:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr667:
	{ s->r_type = KNOT_RRTYPE_ZONEMD; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr917:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr928:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
tr937:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 20;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 20;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 20;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 20;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st20;
st20:
	if ( ++p == pe )
		goto _test_eof20;
case 20:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr84;
		case 778: goto tr87;
		case 800: goto tr84;
		case 808: goto tr85;
		case 809: goto tr86;
		case 827: goto tr88;
		case 1033: goto tr171;
		case 1034: goto tr87;
		case 1056: goto tr171;
		case 1064: goto tr172;
		case 1065: goto tr173;
		case 1083: goto tr174;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr99;
	goto tr83;
tr175:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st21;
tr171:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st21;
tr172:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st21;
tr173:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st21;
tr176:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st21;
tr177:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st21;
st21:
	if ( ++p == pe )
		goto _test_eof21;
case 21:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr93;
		case 800: goto st7;
		case 808: goto tr91;
		case 809: goto tr92;
		case 827: goto tr94;
		case 1033: goto tr175;
		case 1034: goto tr93;
		case 1056: goto tr175;
		case 1064: goto tr176;
		case 1065: goto tr177;
		case 1083: goto tr178;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr99;
	goto tr83;
st22:
	if ( ++p == pe )
		goto _test_eof22;
case 22:
	switch( (*p) ) {
		case 78: goto st23;
		case 83: goto st28;
		case 110: goto st23;
		case 115: goto st28;
	}
	goto tr38;
st23:
	if ( ++p == pe )
		goto _test_eof23;
case 23:
	switch( (*p) ) {
		case 83: goto st24;
		case 115: goto st24;
	}
	goto tr38;
st24:
	if ( ++p == pe )
		goto _test_eof24;
case 24:
	switch( (*p) ) {
		case 75: goto st25;
		case 107: goto st25;
	}
	goto tr38;
st25:
	if ( ++p == pe )
		goto _test_eof25;
case 25:
	switch( (*p) ) {
		case 69: goto st26;
		case 101: goto st26;
	}
	goto tr38;
st26:
	if ( ++p == pe )
		goto _test_eof26;
case 26:
	switch( (*p) ) {
		case 89: goto st27;
		case 121: goto st27;
	}
	goto tr38;
st27:
	if ( ++p == pe )
		goto _test_eof27;
case 27:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr185;
		case 32: goto tr185;
		case 40: goto tr186;
		case 41: goto tr187;
		case 2058: goto tr188;
		case 2107: goto tr189;
		case 2314: goto tr190;
		case 2363: goto tr190;
		case 2570: goto tr191;
		case 2619: goto tr192;
	}
	goto tr61;
st28:
	if ( ++p == pe )
		goto _test_eof28;
case 28:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr193;
		case 32: goto tr193;
		case 40: goto tr194;
		case 41: goto tr195;
		case 2058: goto tr196;
		case 2107: goto tr197;
		case 2314: goto tr198;
		case 2363: goto tr198;
		case 2570: goto tr199;
		case 2619: goto tr200;
	}
	goto tr61;
st29:
	if ( ++p == pe )
		goto _test_eof29;
case 29:
	switch( (*p) ) {
		case 82: goto st30;
		case 114: goto st30;
	}
	goto tr38;
st30:
	if ( ++p == pe )
		goto _test_eof30;
case 30:
	switch( (*p) ) {
		case 84: goto st31;
		case 116: goto st31;
	}
	goto tr38;
st31:
	if ( ++p == pe )
		goto _test_eof31;
case 31:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr203;
		case 32: goto tr203;
		case 40: goto tr204;
		case 41: goto tr205;
		case 2058: goto tr206;
		case 2107: goto tr207;
		case 2314: goto tr208;
		case 2363: goto tr208;
		case 2570: goto tr209;
		case 2619: goto tr210;
	}
	goto tr61;
st32:
	if ( ++p == pe )
		goto _test_eof32;
case 32:
	switch( (*p) ) {
		case 65: goto st33;
		case 97: goto st33;
	}
	goto tr38;
st33:
	if ( ++p == pe )
		goto _test_eof33;
case 33:
	switch( (*p) ) {
		case 77: goto st34;
		case 109: goto st34;
	}
	goto tr38;
st34:
	if ( ++p == pe )
		goto _test_eof34;
case 34:
	switch( (*p) ) {
		case 69: goto st35;
		case 101: goto st35;
	}
	goto tr38;
st35:
	if ( ++p == pe )
		goto _test_eof35;
case 35:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr214;
		case 32: goto tr214;
		case 40: goto tr215;
		case 41: goto tr216;
		case 2058: goto tr217;
		case 2107: goto tr218;
		case 2314: goto tr219;
		case 2363: goto tr219;
		case 2570: goto tr220;
		case 2619: goto tr221;
	}
	goto tr61;
st36:
	if ( ++p == pe )
		goto _test_eof36;
case 36:
	switch( (*p) ) {
		case 89: goto st37;
		case 121: goto st37;
	}
	goto tr38;
st37:
	if ( ++p == pe )
		goto _test_eof37;
case 37:
	switch( (*p) ) {
		case 78: goto st38;
		case 110: goto st38;
	}
	goto tr38;
st38:
	if ( ++p == pe )
		goto _test_eof38;
case 38:
	switch( (*p) ) {
		case 67: goto st39;
		case 99: goto st39;
	}
	goto tr38;
st39:
	if ( ++p == pe )
		goto _test_eof39;
case 39:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr225;
		case 32: goto tr225;
		case 40: goto tr226;
		case 41: goto tr227;
		case 2058: goto tr228;
		case 2107: goto tr229;
		case 2314: goto tr230;
		case 2363: goto tr230;
		case 2570: goto tr231;
		case 2619: goto tr232;
	}
	goto tr61;
tr7:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st40;
tr44:
	{
		s->r_class = s->default_class;
	}
	goto st40;
tr687:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st40;
st40:
	if ( ++p == pe )
		goto _test_eof40;
case 40:
	switch( (*p) ) {
		case 72: goto st41;
		case 78: goto st45;
		case 83: goto st53;
		case 104: goto st41;
		case 110: goto st45;
		case 115: goto st53;
	}
	goto tr38;
st41:
	if ( ++p == pe )
		goto _test_eof41;
case 41:
	switch( (*p) ) {
		case 67: goto st42;
		case 99: goto st42;
	}
	goto tr38;
st42:
	if ( ++p == pe )
		goto _test_eof42;
case 42:
	switch( (*p) ) {
		case 73: goto st43;
		case 105: goto st43;
	}
	goto tr38;
st43:
	if ( ++p == pe )
		goto _test_eof43;
case 43:
	switch( (*p) ) {
		case 68: goto st44;
		case 100: goto st44;
	}
	goto tr38;
st44:
	if ( ++p == pe )
		goto _test_eof44;
case 44:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr239;
		case 32: goto tr239;
		case 40: goto tr240;
		case 41: goto tr241;
		case 2058: goto tr242;
		case 2107: goto tr243;
		case 2314: goto tr244;
		case 2363: goto tr244;
		case 2570: goto tr245;
		case 2619: goto tr246;
	}
	goto tr61;
st45:
	if ( ++p == pe )
		goto _test_eof45;
case 45:
	switch( (*p) ) {
		case 65: goto st46;
		case 83: goto st49;
		case 97: goto st46;
		case 115: goto st49;
	}
	goto tr38;
st46:
	if ( ++p == pe )
		goto _test_eof46;
case 46:
	switch( (*p) ) {
		case 77: goto st47;
		case 109: goto st47;
	}
	goto tr38;
st47:
	if ( ++p == pe )
		goto _test_eof47;
case 47:
	switch( (*p) ) {
		case 69: goto st48;
		case 101: goto st48;
	}
	goto tr38;
st48:
	if ( ++p == pe )
		goto _test_eof48;
case 48:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr251;
		case 32: goto tr251;
		case 40: goto tr252;
		case 41: goto tr253;
		case 2058: goto tr254;
		case 2107: goto tr255;
		case 2314: goto tr256;
		case 2363: goto tr256;
		case 2570: goto tr257;
		case 2619: goto tr258;
	}
	goto tr61;
st49:
	if ( ++p == pe )
		goto _test_eof49;
case 49:
	switch( (*p) ) {
		case 75: goto st50;
		case 107: goto st50;
	}
	goto tr38;
st50:
	if ( ++p == pe )
		goto _test_eof50;
case 50:
	switch( (*p) ) {
		case 69: goto st51;
		case 101: goto st51;
	}
	goto tr38;
st51:
	if ( ++p == pe )
		goto _test_eof51;
case 51:
	switch( (*p) ) {
		case 89: goto st52;
		case 121: goto st52;
	}
	goto tr38;
st52:
	if ( ++p == pe )
		goto _test_eof52;
case 52:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr262;
		case 32: goto tr262;
		case 40: goto tr263;
		case 41: goto tr264;
		case 2058: goto tr265;
		case 2107: goto tr266;
		case 2314: goto tr267;
		case 2363: goto tr267;
		case 2570: goto tr268;
		case 2619: goto tr269;
	}
	goto tr61;
st53:
	if ( ++p == pe )
		goto _test_eof53;
case 53:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr270;
		case 32: goto tr270;
		case 40: goto tr271;
		case 41: goto tr272;
		case 2058: goto tr273;
		case 2107: goto tr274;
		case 2314: goto tr275;
		case 2363: goto tr275;
		case 2570: goto tr276;
		case 2619: goto tr277;
	}
	goto tr61;
tr8:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st54;
tr45:
	{
		s->r_class = s->default_class;
	}
	goto st54;
tr688:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st54;
st54:
	if ( ++p == pe )
		goto _test_eof54;
case 54:
	switch( (*p) ) {
		case 85: goto st55;
		case 117: goto st55;
	}
	goto tr38;
st55:
	if ( ++p == pe )
		goto _test_eof55;
case 55:
	switch( (*p) ) {
		case 73: goto st56;
		case 105: goto st56;
	}
	goto tr38;
st56:
	if ( ++p == pe )
		goto _test_eof56;
case 56:
	switch( (*p) ) {
		case 52: goto st57;
		case 54: goto st59;
	}
	goto tr38;
st57:
	if ( ++p == pe )
		goto _test_eof57;
case 57:
	if ( (*p) == 56 )
		goto st58;
	goto tr38;
st58:
	if ( ++p == pe )
		goto _test_eof58;
case 58:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr283;
		case 32: goto tr283;
		case 40: goto tr284;
		case 41: goto tr285;
		case 2058: goto tr286;
		case 2107: goto tr287;
		case 2314: goto tr288;
		case 2363: goto tr288;
		case 2570: goto tr289;
		case 2619: goto tr290;
	}
	goto tr61;
st59:
	if ( ++p == pe )
		goto _test_eof59;
case 59:
	if ( (*p) == 52 )
		goto st60;
	goto tr38;
st60:
	if ( ++p == pe )
		goto _test_eof60;
case 60:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr292;
		case 32: goto tr292;
		case 40: goto tr293;
		case 41: goto tr294;
		case 2058: goto tr295;
		case 2107: goto tr296;
		case 2314: goto tr297;
		case 2363: goto tr297;
		case 2570: goto tr298;
		case 2619: goto tr299;
	}
	goto tr61;
tr9:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st61;
tr46:
	{
		s->r_class = s->default_class;
	}
	goto st61;
tr689:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st61;
st61:
	if ( ++p == pe )
		goto _test_eof61;
case 61:
	switch( (*p) ) {
		case 73: goto st62;
		case 105: goto st62;
	}
	goto tr38;
st62:
	if ( ++p == pe )
		goto _test_eof62;
case 62:
	switch( (*p) ) {
		case 78: goto st63;
		case 110: goto st63;
	}
	goto tr38;
st63:
	if ( ++p == pe )
		goto _test_eof63;
case 63:
	switch( (*p) ) {
		case 70: goto st64;
		case 102: goto st64;
	}
	goto tr38;
st64:
	if ( ++p == pe )
		goto _test_eof64;
case 64:
	switch( (*p) ) {
		case 79: goto st65;
		case 111: goto st65;
	}
	goto tr38;
st65:
	if ( ++p == pe )
		goto _test_eof65;
case 65:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr304;
		case 32: goto tr304;
		case 40: goto tr305;
		case 41: goto tr306;
		case 2058: goto tr307;
		case 2107: goto tr308;
		case 2314: goto tr309;
		case 2363: goto tr309;
		case 2570: goto tr310;
		case 2619: goto tr311;
	}
	goto tr61;
tr47:
	{
		s->r_class = s->default_class;
	}
	goto st66;
st66:
	if ( ++p == pe )
		goto _test_eof66;
case 66:
	switch( (*p) ) {
		case 78: goto st67;
		case 80: goto st70;
		case 110: goto st67;
		case 112: goto st70;
	}
	goto tr38;
st67:
	if ( ++p == pe )
		goto _test_eof67;
case 67:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr314;
		case 32: goto tr314;
		case 40: goto tr315;
		case 41: goto tr316;
		case 1034: goto tr317;
		case 1083: goto tr318;
	}
	goto tr38;
tr320:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st68;
tr321:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st68;
tr339:
	{
		s->line_counter++;
	}
	goto st68;
tr704:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st68;
tr705:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st68;
tr706:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st68;
tr713:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st68;
tr669:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 68; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st68;
tr314:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st68;
tr315:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st68;
tr316:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st68;
tr317:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->line_counter++;
	}
	goto st68;
tr722:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st68;
tr723:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st68;
tr724:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st68;
tr726:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st68;
st68:
	if ( ++p == pe )
		goto _test_eof68;
case 68:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st68;
		case 32: goto st68;
		case 40: goto tr320;
		case 41: goto tr321;
		case 65: goto st4;
		case 67: goto st17;
		case 68: goto st40;
		case 69: goto st54;
		case 72: goto st61;
		case 73: goto st69;
		case 75: goto st77;
		case 76: goto st81;
		case 77: goto st89;
		case 78: goto st95;
		case 79: goto st111;
		case 80: goto st121;
		case 82: goto st124;
		case 83: goto st131;
		case 84: goto st147;
		case 85: goto st157;
		case 90: goto st160;
		case 97: goto st4;
		case 99: goto st17;
		case 100: goto st40;
		case 101: goto st54;
		case 104: goto st61;
		case 105: goto st69;
		case 107: goto st77;
		case 108: goto st81;
		case 109: goto st89;
		case 110: goto st95;
		case 111: goto st111;
		case 112: goto st121;
		case 114: goto st124;
		case 115: goto st131;
		case 116: goto st147;
		case 117: goto st157;
		case 122: goto st160;
		case 1034: goto tr339;
		case 1083: goto tr340;
	}
	goto tr38;
tr690:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st69;
st69:
	if ( ++p == pe )
		goto _test_eof69;
case 69:
	switch( (*p) ) {
		case 80: goto st70;
		case 112: goto st70;
	}
	goto tr38;
st70:
	if ( ++p == pe )
		goto _test_eof70;
case 70:
	switch( (*p) ) {
		case 83: goto st71;
		case 115: goto st71;
	}
	goto tr38;
st71:
	if ( ++p == pe )
		goto _test_eof71;
case 71:
	switch( (*p) ) {
		case 69: goto st72;
		case 101: goto st72;
	}
	goto tr38;
st72:
	if ( ++p == pe )
		goto _test_eof72;
case 72:
	switch( (*p) ) {
		case 67: goto st73;
		case 99: goto st73;
	}
	goto tr38;
st73:
	if ( ++p == pe )
		goto _test_eof73;
case 73:
	switch( (*p) ) {
		case 75: goto st74;
		case 107: goto st74;
	}
	goto tr38;
st74:
	if ( ++p == pe )
		goto _test_eof74;
case 74:
	switch( (*p) ) {
		case 69: goto st75;
		case 101: goto st75;
	}
	goto tr38;
st75:
	if ( ++p == pe )
		goto _test_eof75;
case 75:
	switch( (*p) ) {
		case 89: goto st76;
		case 121: goto st76;
	}
	goto tr38;
st76:
	if ( ++p == pe )
		goto _test_eof76;
case 76:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr347;
		case 32: goto tr347;
		case 40: goto tr348;
		case 41: goto tr349;
		case 2058: goto tr350;
		case 2107: goto tr351;
		case 2314: goto tr352;
		case 2363: goto tr352;
		case 2570: goto tr353;
		case 2619: goto tr354;
	}
	goto tr61;
tr11:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st77;
tr48:
	{
		s->r_class = s->default_class;
	}
	goto st77;
tr691:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st77;
st77:
	if ( ++p == pe )
		goto _test_eof77;
case 77:
	switch( (*p) ) {
		case 69: goto st78;
		case 88: goto st80;
		case 101: goto st78;
		case 120: goto st80;
	}
	goto tr38;
st78:
	if ( ++p == pe )
		goto _test_eof78;
case 78:
	switch( (*p) ) {
		case 89: goto st79;
		case 121: goto st79;
	}
	goto tr38;
st79:
	if ( ++p == pe )
		goto _test_eof79;
case 79:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr358;
		case 32: goto tr358;
		case 40: goto tr359;
		case 41: goto tr360;
		case 2058: goto tr361;
		case 2107: goto tr362;
		case 2314: goto tr363;
		case 2363: goto tr363;
		case 2570: goto tr364;
		case 2619: goto tr365;
	}
	goto tr61;
st80:
	if ( ++p == pe )
		goto _test_eof80;
case 80:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr366;
		case 32: goto tr366;
		case 40: goto tr367;
		case 41: goto tr368;
		case 2058: goto tr369;
		case 2107: goto tr370;
		case 2314: goto tr371;
		case 2363: goto tr371;
		case 2570: goto tr372;
		case 2619: goto tr373;
	}
	goto tr61;
tr12:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st81;
tr49:
	{
		s->r_class = s->default_class;
	}
	goto st81;
tr692:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st81;
st81:
	if ( ++p == pe )
		goto _test_eof81;
case 81:
	switch( (*p) ) {
		case 51: goto st82;
		case 54: goto st84;
		case 79: goto st86;
		case 80: goto st88;
		case 111: goto st86;
		case 112: goto st88;
	}
	goto tr38;
st82:
	if ( ++p == pe )
		goto _test_eof82;
case 82:
	if ( (*p) == 50 )
		goto st83;
	goto tr38;
st83:
	if ( ++p == pe )
		goto _test_eof83;
case 83:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr379;
		case 32: goto tr379;
		case 40: goto tr380;
		case 41: goto tr381;
		case 2058: goto tr382;
		case 2107: goto tr383;
		case 2314: goto tr384;
		case 2363: goto tr384;
		case 2570: goto tr385;
		case 2619: goto tr386;
	}
	goto tr61;
st84:
	if ( ++p == pe )
		goto _test_eof84;
case 84:
	if ( (*p) == 52 )
		goto st85;
	goto tr38;
st85:
	if ( ++p == pe )
		goto _test_eof85;
case 85:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr388;
		case 32: goto tr388;
		case 40: goto tr389;
		case 41: goto tr390;
		case 2058: goto tr391;
		case 2107: goto tr392;
		case 2314: goto tr393;
		case 2363: goto tr393;
		case 2570: goto tr394;
		case 2619: goto tr395;
	}
	goto tr61;
st86:
	if ( ++p == pe )
		goto _test_eof86;
case 86:
	switch( (*p) ) {
		case 67: goto st87;
		case 99: goto st87;
	}
	goto tr38;
st87:
	if ( ++p == pe )
		goto _test_eof87;
case 87:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr397;
		case 32: goto tr397;
		case 40: goto tr398;
		case 41: goto tr399;
		case 2058: goto tr400;
		case 2107: goto tr401;
		case 2314: goto tr402;
		case 2363: goto tr402;
		case 2570: goto tr403;
		case 2619: goto tr404;
	}
	goto tr61;
st88:
	if ( ++p == pe )
		goto _test_eof88;
case 88:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr405;
		case 32: goto tr405;
		case 40: goto tr406;
		case 41: goto tr407;
		case 2058: goto tr408;
		case 2107: goto tr409;
		case 2314: goto tr410;
		case 2363: goto tr410;
		case 2570: goto tr411;
		case 2619: goto tr412;
	}
	goto tr61;
tr13:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st89;
tr50:
	{
		s->r_class = s->default_class;
	}
	goto st89;
tr693:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st89;
st89:
	if ( ++p == pe )
		goto _test_eof89;
case 89:
	switch( (*p) ) {
		case 73: goto st90;
		case 88: goto st94;
		case 105: goto st90;
		case 120: goto st94;
	}
	goto tr38;
st90:
	if ( ++p == pe )
		goto _test_eof90;
case 90:
	switch( (*p) ) {
		case 78: goto st91;
		case 110: goto st91;
	}
	goto tr38;
st91:
	if ( ++p == pe )
		goto _test_eof91;
case 91:
	switch( (*p) ) {
		case 70: goto st92;
		case 102: goto st92;
	}
	goto tr38;
st92:
	if ( ++p == pe )
		goto _test_eof92;
case 92:
	switch( (*p) ) {
		case 79: goto st93;
		case 111: goto st93;
	}
	goto tr38;
st93:
	if ( ++p == pe )
		goto _test_eof93;
case 93:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr418;
		case 32: goto tr418;
		case 40: goto tr419;
		case 41: goto tr420;
		case 2058: goto tr421;
		case 2107: goto tr422;
		case 2314: goto tr423;
		case 2363: goto tr423;
		case 2570: goto tr424;
		case 2619: goto tr425;
	}
	goto tr61;
st94:
	if ( ++p == pe )
		goto _test_eof94;
case 94:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr426;
		case 32: goto tr426;
		case 40: goto tr427;
		case 41: goto tr428;
		case 2058: goto tr429;
		case 2107: goto tr430;
		case 2314: goto tr431;
		case 2363: goto tr431;
		case 2570: goto tr432;
		case 2619: goto tr433;
	}
	goto tr61;
tr14:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st95;
tr51:
	{
		s->r_class = s->default_class;
	}
	goto st95;
tr694:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st95;
st95:
	if ( ++p == pe )
		goto _test_eof95;
case 95:
	switch( (*p) ) {
		case 65: goto st96;
		case 73: goto st100;
		case 83: goto st102;
		case 97: goto st96;
		case 105: goto st100;
		case 115: goto st102;
	}
	goto tr38;
st96:
	if ( ++p == pe )
		goto _test_eof96;
case 96:
	switch( (*p) ) {
		case 80: goto st97;
		case 112: goto st97;
	}
	goto tr38;
st97:
	if ( ++p == pe )
		goto _test_eof97;
case 97:
	switch( (*p) ) {
		case 84: goto st98;
		case 116: goto st98;
	}
	goto tr38;
st98:
	if ( ++p == pe )
		goto _test_eof98;
case 98:
	switch( (*p) ) {
		case 82: goto st99;
		case 114: goto st99;
	}
	goto tr38;
st99:
	if ( ++p == pe )
		goto _test_eof99;
case 99:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr440;
		case 32: goto tr440;
		case 40: goto tr441;
		case 41: goto tr442;
		case 2058: goto tr443;
		case 2107: goto tr444;
		case 2314: goto tr445;
		case 2363: goto tr445;
		case 2570: goto tr446;
		case 2619: goto tr447;
	}
	goto tr61;
st100:
	if ( ++p == pe )
		goto _test_eof100;
case 100:
	switch( (*p) ) {
		case 68: goto st101;
		case 100: goto st101;
	}
	goto tr38;
st101:
	if ( ++p == pe )
		goto _test_eof101;
case 101:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr449;
		case 32: goto tr449;
		case 40: goto tr450;
		case 41: goto tr451;
		case 2058: goto tr452;
		case 2107: goto tr453;
		case 2314: goto tr454;
		case 2363: goto tr454;
		case 2570: goto tr455;
		case 2619: goto tr456;
	}
	goto tr61;
st102:
	if ( ++p == pe )
		goto _test_eof102;
case 102:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr457;
		case 32: goto tr457;
		case 40: goto tr458;
		case 41: goto tr459;
		case 69: goto st103;
		case 101: goto st103;
		case 2058: goto tr461;
		case 2107: goto tr462;
		case 2314: goto tr463;
		case 2363: goto tr463;
		case 2570: goto tr464;
		case 2619: goto tr465;
	}
	goto tr61;
st103:
	if ( ++p == pe )
		goto _test_eof103;
case 103:
	switch( (*p) ) {
		case 67: goto st104;
		case 99: goto st104;
	}
	goto tr38;
st104:
	if ( ++p == pe )
		goto _test_eof104;
case 104:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr467;
		case 32: goto tr467;
		case 40: goto tr468;
		case 41: goto tr469;
		case 51: goto st105;
		case 2058: goto tr471;
		case 2107: goto tr472;
		case 2314: goto tr473;
		case 2363: goto tr473;
		case 2570: goto tr474;
		case 2619: goto tr475;
	}
	goto tr61;
st105:
	if ( ++p == pe )
		goto _test_eof105;
case 105:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr476;
		case 32: goto tr476;
		case 40: goto tr477;
		case 41: goto tr478;
		case 80: goto st106;
		case 112: goto st106;
		case 2058: goto tr480;
		case 2107: goto tr481;
		case 2314: goto tr482;
		case 2363: goto tr482;
		case 2570: goto tr483;
		case 2619: goto tr484;
	}
	goto tr61;
st106:
	if ( ++p == pe )
		goto _test_eof106;
case 106:
	switch( (*p) ) {
		case 65: goto st107;
		case 97: goto st107;
	}
	goto tr38;
st107:
	if ( ++p == pe )
		goto _test_eof107;
case 107:
	switch( (*p) ) {
		case 82: goto st108;
		case 114: goto st108;
	}
	goto tr38;
st108:
	if ( ++p == pe )
		goto _test_eof108;
case 108:
	switch( (*p) ) {
		case 65: goto st109;
		case 97: goto st109;
	}
	goto tr38;
st109:
	if ( ++p == pe )
		goto _test_eof109;
case 109:
	switch( (*p) ) {
		case 77: goto st110;
		case 109: goto st110;
	}
	goto tr38;
st110:
	if ( ++p == pe )
		goto _test_eof110;
case 110:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr489;
		case 32: goto tr489;
		case 40: goto tr490;
		case 41: goto tr491;
		case 2058: goto tr492;
		case 2107: goto tr493;
		case 2314: goto tr494;
		case 2363: goto tr494;
		case 2570: goto tr495;
		case 2619: goto tr496;
	}
	goto tr61;
tr15:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st111;
tr52:
	{
		s->r_class = s->default_class;
	}
	goto st111;
tr695:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st111;
st111:
	if ( ++p == pe )
		goto _test_eof111;
case 111:
	switch( (*p) ) {
		case 80: goto st112;
		case 112: goto st112;
	}
	goto tr38;
st112:
	if ( ++p == pe )
		goto _test_eof112;
case 112:
	switch( (*p) ) {
		case 69: goto st113;
		case 101: goto st113;
	}
	goto tr38;
st113:
	if ( ++p == pe )
		goto _test_eof113;
case 113:
	switch( (*p) ) {
		case 78: goto st114;
		case 110: goto st114;
	}
	goto tr38;
st114:
	if ( ++p == pe )
		goto _test_eof114;
case 114:
	switch( (*p) ) {
		case 80: goto st115;
		case 112: goto st115;
	}
	goto tr38;
st115:
	if ( ++p == pe )
		goto _test_eof115;
case 115:
	switch( (*p) ) {
		case 71: goto st116;
		case 103: goto st116;
	}
	goto tr38;
st116:
	if ( ++p == pe )
		goto _test_eof116;
case 116:
	switch( (*p) ) {
		case 80: goto st117;
		case 112: goto st117;
	}
	goto tr38;
st117:
	if ( ++p == pe )
		goto _test_eof117;
case 117:
	switch( (*p) ) {
		case 75: goto st118;
		case 107: goto st118;
	}
	goto tr38;
st118:
	if ( ++p == pe )
		goto _test_eof118;
case 118:
	switch( (*p) ) {
		case 69: goto st119;
		case 101: goto st119;
	}
	goto tr38;
st119:
	if ( ++p == pe )
		goto _test_eof119;
case 119:
	switch( (*p) ) {
		case 89: goto st120;
		case 121: goto st120;
	}
	goto tr38;
st120:
	if ( ++p == pe )
		goto _test_eof120;
case 120:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr506;
		case 32: goto tr506;
		case 40: goto tr507;
		case 41: goto tr508;
		case 2058: goto tr509;
		case 2107: goto tr510;
		case 2314: goto tr511;
		case 2363: goto tr511;
		case 2570: goto tr512;
		case 2619: goto tr513;
	}
	goto tr61;
tr16:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st121;
tr53:
	{
		s->r_class = s->default_class;
	}
	goto st121;
tr696:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st121;
st121:
	if ( ++p == pe )
		goto _test_eof121;
case 121:
	switch( (*p) ) {
		case 84: goto st122;
		case 116: goto st122;
	}
	goto tr38;
st122:
	if ( ++p == pe )
		goto _test_eof122;
case 122:
	switch( (*p) ) {
		case 82: goto st123;
		case 114: goto st123;
	}
	goto tr38;
st123:
	if ( ++p == pe )
		goto _test_eof123;
case 123:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr516;
		case 32: goto tr516;
		case 40: goto tr517;
		case 41: goto tr518;
		case 2058: goto tr519;
		case 2107: goto tr520;
		case 2314: goto tr521;
		case 2363: goto tr521;
		case 2570: goto tr522;
		case 2619: goto tr523;
	}
	goto tr61;
tr17:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st124;
tr54:
	{
		s->r_class = s->default_class;
	}
	goto st124;
tr697:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st124;
st124:
	if ( ++p == pe )
		goto _test_eof124;
case 124:
	switch( (*p) ) {
		case 80: goto st125;
		case 82: goto st126;
		case 84: goto st130;
		case 112: goto st125;
		case 114: goto st126;
		case 116: goto st130;
	}
	goto tr38;
st125:
	if ( ++p == pe )
		goto _test_eof125;
case 125:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr527;
		case 32: goto tr527;
		case 40: goto tr528;
		case 41: goto tr529;
		case 2058: goto tr530;
		case 2107: goto tr531;
		case 2314: goto tr532;
		case 2363: goto tr532;
		case 2570: goto tr533;
		case 2619: goto tr534;
	}
	goto tr61;
st126:
	if ( ++p == pe )
		goto _test_eof126;
case 126:
	switch( (*p) ) {
		case 83: goto st127;
		case 115: goto st127;
	}
	goto tr38;
st127:
	if ( ++p == pe )
		goto _test_eof127;
case 127:
	switch( (*p) ) {
		case 73: goto st128;
		case 105: goto st128;
	}
	goto tr38;
st128:
	if ( ++p == pe )
		goto _test_eof128;
case 128:
	switch( (*p) ) {
		case 71: goto st129;
		case 103: goto st129;
	}
	goto tr38;
st129:
	if ( ++p == pe )
		goto _test_eof129;
case 129:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr538;
		case 32: goto tr538;
		case 40: goto tr539;
		case 41: goto tr540;
		case 2058: goto tr541;
		case 2107: goto tr542;
		case 2314: goto tr543;
		case 2363: goto tr543;
		case 2570: goto tr544;
		case 2619: goto tr545;
	}
	goto tr61;
st130:
	if ( ++p == pe )
		goto _test_eof130;
case 130:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr546;
		case 32: goto tr546;
		case 40: goto tr547;
		case 41: goto tr548;
		case 2058: goto tr549;
		case 2107: goto tr550;
		case 2314: goto tr551;
		case 2363: goto tr551;
		case 2570: goto tr552;
		case 2619: goto tr553;
	}
	goto tr61;
tr18:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st131;
tr55:
	{
		s->r_class = s->default_class;
	}
	goto st131;
tr698:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st131;
st131:
	if ( ++p == pe )
		goto _test_eof131;
case 131:
	switch( (*p) ) {
		case 77: goto st132;
		case 79: goto st137;
		case 80: goto st139;
		case 82: goto st141;
		case 83: goto st143;
		case 109: goto st132;
		case 111: goto st137;
		case 112: goto st139;
		case 114: goto st141;
		case 115: goto st143;
	}
	goto tr38;
st132:
	if ( ++p == pe )
		goto _test_eof132;
case 132:
	switch( (*p) ) {
		case 73: goto st133;
		case 105: goto st133;
	}
	goto tr38;
st133:
	if ( ++p == pe )
		goto _test_eof133;
case 133:
	switch( (*p) ) {
		case 77: goto st134;
		case 109: goto st134;
	}
	goto tr38;
st134:
	if ( ++p == pe )
		goto _test_eof134;
case 134:
	switch( (*p) ) {
		case 69: goto st135;
		case 101: goto st135;
	}
	goto tr38;
st135:
	if ( ++p == pe )
		goto _test_eof135;
case 135:
	switch( (*p) ) {
		case 65: goto st136;
		case 97: goto st136;
	}
	goto tr38;
st136:
	if ( ++p == pe )
		goto _test_eof136;
case 136:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr563;
		case 32: goto tr563;
		case 40: goto tr564;
		case 41: goto tr565;
		case 2058: goto tr566;
		case 2107: goto tr567;
		case 2314: goto tr568;
		case 2363: goto tr568;
		case 2570: goto tr569;
		case 2619: goto tr570;
	}
	goto tr61;
st137:
	if ( ++p == pe )
		goto _test_eof137;
case 137:
	switch( (*p) ) {
		case 65: goto st138;
		case 97: goto st138;
	}
	goto tr38;
st138:
	if ( ++p == pe )
		goto _test_eof138;
case 138:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr572;
		case 32: goto tr572;
		case 40: goto tr573;
		case 41: goto tr574;
		case 2058: goto tr575;
		case 2107: goto tr576;
		case 2314: goto tr577;
		case 2363: goto tr577;
		case 2570: goto tr578;
		case 2619: goto tr579;
	}
	goto tr61;
st139:
	if ( ++p == pe )
		goto _test_eof139;
case 139:
	switch( (*p) ) {
		case 70: goto st140;
		case 102: goto st140;
	}
	goto tr38;
st140:
	if ( ++p == pe )
		goto _test_eof140;
case 140:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr581;
		case 32: goto tr581;
		case 40: goto tr582;
		case 41: goto tr583;
		case 2058: goto tr584;
		case 2107: goto tr585;
		case 2314: goto tr586;
		case 2363: goto tr586;
		case 2570: goto tr587;
		case 2619: goto tr588;
	}
	goto tr61;
st141:
	if ( ++p == pe )
		goto _test_eof141;
case 141:
	switch( (*p) ) {
		case 86: goto st142;
		case 118: goto st142;
	}
	goto tr38;
st142:
	if ( ++p == pe )
		goto _test_eof142;
case 142:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr590;
		case 32: goto tr590;
		case 40: goto tr591;
		case 41: goto tr592;
		case 2058: goto tr593;
		case 2107: goto tr594;
		case 2314: goto tr595;
		case 2363: goto tr595;
		case 2570: goto tr596;
		case 2619: goto tr597;
	}
	goto tr61;
st143:
	if ( ++p == pe )
		goto _test_eof143;
case 143:
	switch( (*p) ) {
		case 72: goto st144;
		case 104: goto st144;
	}
	goto tr38;
st144:
	if ( ++p == pe )
		goto _test_eof144;
case 144:
	switch( (*p) ) {
		case 70: goto st145;
		case 102: goto st145;
	}
	goto tr38;
st145:
	if ( ++p == pe )
		goto _test_eof145;
case 145:
	switch( (*p) ) {
		case 80: goto st146;
		case 112: goto st146;
	}
	goto tr38;
st146:
	if ( ++p == pe )
		goto _test_eof146;
case 146:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr601;
		case 32: goto tr601;
		case 40: goto tr602;
		case 41: goto tr603;
		case 2058: goto tr604;
		case 2107: goto tr605;
		case 2314: goto tr606;
		case 2363: goto tr606;
		case 2570: goto tr607;
		case 2619: goto tr608;
	}
	goto tr61;
tr19:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st147;
tr56:
	{
		s->r_class = s->default_class;
	}
	goto st147;
tr699:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st147;
st147:
	if ( ++p == pe )
		goto _test_eof147;
case 147:
	switch( (*p) ) {
		case 76: goto st148;
		case 88: goto st151;
		case 89: goto st153;
		case 108: goto st148;
		case 120: goto st151;
		case 121: goto st153;
	}
	goto tr38;
st148:
	if ( ++p == pe )
		goto _test_eof148;
case 148:
	switch( (*p) ) {
		case 83: goto st149;
		case 115: goto st149;
	}
	goto tr38;
st149:
	if ( ++p == pe )
		goto _test_eof149;
case 149:
	switch( (*p) ) {
		case 65: goto st150;
		case 97: goto st150;
	}
	goto tr38;
st150:
	if ( ++p == pe )
		goto _test_eof150;
case 150:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr614;
		case 32: goto tr614;
		case 40: goto tr615;
		case 41: goto tr616;
		case 2058: goto tr617;
		case 2107: goto tr618;
		case 2314: goto tr619;
		case 2363: goto tr619;
		case 2570: goto tr620;
		case 2619: goto tr621;
	}
	goto tr61;
st151:
	if ( ++p == pe )
		goto _test_eof151;
case 151:
	switch( (*p) ) {
		case 84: goto st152;
		case 116: goto st152;
	}
	goto tr38;
st152:
	if ( ++p == pe )
		goto _test_eof152;
case 152:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr623;
		case 32: goto tr623;
		case 40: goto tr624;
		case 41: goto tr625;
		case 2058: goto tr626;
		case 2107: goto tr627;
		case 2314: goto tr628;
		case 2363: goto tr628;
		case 2570: goto tr629;
		case 2619: goto tr630;
	}
	goto tr61;
st153:
	if ( ++p == pe )
		goto _test_eof153;
case 153:
	switch( (*p) ) {
		case 80: goto st154;
		case 112: goto st154;
	}
	goto tr38;
st154:
	if ( ++p == pe )
		goto _test_eof154;
case 154:
	switch( (*p) ) {
		case 69: goto st155;
		case 101: goto st155;
	}
	goto tr38;
st155:
	if ( ++p == pe )
		goto _test_eof155;
case 155:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr634;
	goto tr633;
tr634:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st156;
tr639:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st156;
st156:
	if ( ++p == pe )
		goto _test_eof156;
case 156:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr636;
		case 32: goto tr636;
		case 40: goto tr637;
		case 41: goto tr638;
		case 2058: goto tr640;
		case 2107: goto tr641;
		case 2314: goto tr642;
		case 2363: goto tr642;
		case 2570: goto tr643;
		case 2619: goto tr644;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr639;
	goto tr635;
tr20:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st157;
tr57:
	{
		s->r_class = s->default_class;
	}
	goto st157;
tr700:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st157;
st157:
	if ( ++p == pe )
		goto _test_eof157;
case 157:
	switch( (*p) ) {
		case 82: goto st158;
		case 114: goto st158;
	}
	goto tr38;
st158:
	if ( ++p == pe )
		goto _test_eof158;
case 158:
	switch( (*p) ) {
		case 73: goto st159;
		case 105: goto st159;
	}
	goto tr38;
st159:
	if ( ++p == pe )
		goto _test_eof159;
case 159:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr647;
		case 32: goto tr647;
		case 40: goto tr648;
		case 41: goto tr649;
		case 2058: goto tr650;
		case 2107: goto tr651;
		case 2314: goto tr652;
		case 2363: goto tr652;
		case 2570: goto tr653;
		case 2619: goto tr654;
	}
	goto tr61;
tr21:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st160;
tr58:
	{
		s->r_class = s->default_class;
	}
	goto st160;
tr701:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st160;
st160:
	if ( ++p == pe )
		goto _test_eof160;
case 160:
	switch( (*p) ) {
		case 79: goto st161;
		case 111: goto st161;
	}
	goto tr38;
st161:
	if ( ++p == pe )
		goto _test_eof161;
case 161:
	switch( (*p) ) {
		case 78: goto st162;
		case 110: goto st162;
	}
	goto tr38;
st162:
	if ( ++p == pe )
		goto _test_eof162;
case 162:
	switch( (*p) ) {
		case 69: goto st163;
		case 101: goto st163;
	}
	goto tr38;
st163:
	if ( ++p == pe )
		goto _test_eof163;
case 163:
	switch( (*p) ) {
		case 77: goto st164;
		case 109: goto st164;
	}
	goto tr38;
st164:
	if ( ++p == pe )
		goto _test_eof164;
case 164:
	switch( (*p) ) {
		case 68: goto st165;
		case 100: goto st165;
	}
	goto tr38;
st165:
	if ( ++p == pe )
		goto _test_eof165;
case 165:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr660;
		case 32: goto tr660;
		case 40: goto tr661;
		case 41: goto tr662;
		case 2058: goto tr663;
		case 2107: goto tr664;
		case 2314: goto tr665;
		case 2363: goto tr665;
		case 2570: goto tr666;
		case 2619: goto tr667;
	}
	goto tr61;
tr340:
	{
		s->buffer_length = 0;
	}
	goto st166;
tr714:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st166;
tr668:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st166;
tr318:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->buffer_length = 0;
	}
	goto st166;
tr727:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st166;
st166:
	if ( ++p == pe )
		goto _test_eof166;
case 166:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr669;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr668;
	goto tr38;
tr979:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1229; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1229; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1229; goto _out;}
		}
	}
	goto st1229;
tr149:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1229; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1229; goto _out;}
		}
	}
	goto st1229;
tr155:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1229; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1229; goto _out;}
		}
	}
	goto st1229;
st1229:
	if ( ++p == pe )
		goto _test_eof1229;
case 1229:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3874;
		case 32: goto tr3874;
		case 36: goto st177;
		case 40: goto tr3875;
		case 41: goto tr3876;
		case 42: goto tr3856;
		case 65: goto tr3877;
		case 67: goto tr3878;
		case 68: goto tr3879;
		case 69: goto tr3880;
		case 72: goto tr3881;
		case 73: goto tr3882;
		case 75: goto tr3883;
		case 76: goto tr3884;
		case 77: goto tr3885;
		case 78: goto tr3886;
		case 79: goto tr3887;
		case 80: goto tr3888;
		case 82: goto tr3889;
		case 83: goto tr3890;
		case 84: goto tr3891;
		case 85: goto tr3892;
		case 90: goto tr3893;
		case 92: goto tr3856;
		case 95: goto tr3856;
		case 97: goto tr3877;
		case 99: goto tr3878;
		case 100: goto tr3879;
		case 101: goto tr3880;
		case 104: goto tr3881;
		case 105: goto tr3882;
		case 107: goto tr3883;
		case 108: goto tr3884;
		case 109: goto tr3885;
		case 110: goto tr3886;
		case 111: goto tr3887;
		case 112: goto tr3888;
		case 114: goto tr3889;
		case 115: goto tr3890;
		case 116: goto tr3891;
		case 117: goto tr3892;
		case 122: goto tr3893;
		case 778: goto tr3857;
		case 827: goto tr3858;
		case 1034: goto tr3894;
		case 1083: goto tr3895;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3856;
	} else if ( _widec > 89 ) {
		if ( 98 <= _widec && _widec <= 121 )
			goto tr3856;
	} else
		goto tr3856;
	goto tr3873;
tr671:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st167;
tr672:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st167;
tr3874:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st167;
tr3875:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st167;
tr3876:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st167;
st167:
	if ( ++p == pe )
		goto _test_eof167;
case 167:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st167;
		case 32: goto st167;
		case 40: goto tr671;
		case 41: goto tr672;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 778: goto tr22;
		case 827: goto tr23;
		case 1034: goto tr673;
		case 1083: goto tr674;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr10:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st168;
st168:
	if ( ++p == pe )
		goto _test_eof168;
case 168:
	switch( (*p) ) {
		case 78: goto st169;
		case 80: goto st70;
		case 110: goto st169;
		case 112: goto st70;
	}
	goto tr38;
st169:
	if ( ++p == pe )
		goto _test_eof169;
case 169:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr676;
		case 32: goto tr676;
		case 40: goto tr677;
		case 41: goto tr678;
		case 1034: goto tr679;
		case 1083: goto tr680;
	}
	goto tr38;
tr682:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st170;
tr683:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st170;
tr702:
	{
		s->line_counter++;
	}
	goto st170;
tr729:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 170; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st170;
tr676:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st170;
tr677:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st170;
tr678:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st170;
tr679:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->line_counter++;
	}
	goto st170;
st170:
	if ( ++p == pe )
		goto _test_eof170;
case 170:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st170;
		case 32: goto st170;
		case 40: goto tr682;
		case 41: goto tr683;
		case 65: goto tr685;
		case 67: goto tr686;
		case 68: goto tr687;
		case 69: goto tr688;
		case 72: goto tr689;
		case 73: goto tr690;
		case 75: goto tr691;
		case 76: goto tr692;
		case 77: goto tr693;
		case 78: goto tr694;
		case 79: goto tr695;
		case 80: goto tr696;
		case 82: goto tr697;
		case 83: goto tr698;
		case 84: goto tr699;
		case 85: goto tr700;
		case 90: goto tr701;
		case 97: goto tr685;
		case 99: goto tr686;
		case 100: goto tr687;
		case 101: goto tr688;
		case 104: goto tr689;
		case 105: goto tr690;
		case 107: goto tr691;
		case 108: goto tr692;
		case 109: goto tr693;
		case 110: goto tr694;
		case 111: goto tr695;
		case 112: goto tr696;
		case 114: goto tr697;
		case 115: goto tr698;
		case 116: goto tr699;
		case 117: goto tr700;
		case 122: goto tr701;
		case 1034: goto tr702;
		case 1083: goto tr703;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr684;
	goto tr633;
tr684:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st171;
tr707:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st171;
st171:
	if ( ++p == pe )
		goto _test_eof171;
case 171:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr704;
		case 32: goto tr704;
		case 40: goto tr705;
		case 41: goto tr706;
		case 68: goto tr708;
		case 72: goto tr709;
		case 77: goto tr710;
		case 83: goto st172;
		case 87: goto tr712;
		case 100: goto tr708;
		case 104: goto tr709;
		case 109: goto tr710;
		case 115: goto st172;
		case 119: goto tr712;
		case 1034: goto tr713;
		case 1083: goto tr714;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr707;
	goto tr26;
tr708:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st172;
tr709:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st172;
tr710:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st172;
tr712:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st172;
st172:
	if ( ++p == pe )
		goto _test_eof172;
case 172:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr704;
		case 32: goto tr704;
		case 40: goto tr705;
		case 41: goto tr706;
		case 1034: goto tr713;
		case 1083: goto tr714;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr715;
	goto tr26;
tr716:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st173;
tr715:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st173;
tr725:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st173;
st173:
	if ( ++p == pe )
		goto _test_eof173;
case 173:
	switch( (*p) ) {
		case 68: goto tr717;
		case 72: goto tr718;
		case 77: goto tr719;
		case 83: goto st174;
		case 87: goto tr721;
		case 100: goto tr717;
		case 104: goto tr718;
		case 109: goto tr719;
		case 115: goto st174;
		case 119: goto tr721;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr716;
	goto tr26;
tr717:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st174;
tr718:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st174;
tr719:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st174;
tr721:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st174;
st174:
	if ( ++p == pe )
		goto _test_eof174;
case 174:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr722;
		case 32: goto tr722;
		case 40: goto tr723;
		case 41: goto tr724;
		case 1034: goto tr726;
		case 1083: goto tr727;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr725;
	goto tr26;
tr703:
	{
		s->buffer_length = 0;
	}
	goto st175;
tr728:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st175;
tr680:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->buffer_length = 0;
	}
	goto st175;
st175:
	if ( ++p == pe )
		goto _test_eof175;
case 175:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr729;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr728;
	goto tr38;
tr23:
	{
		s->buffer_length = 0;
	}
	goto st176;
tr3858:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st176;
tr730:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st176;
tr3922:
	{
		NOERR;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st176;
st176:
	if ( ++p == pe )
		goto _test_eof176;
case 176:
	if ( (*p) == 10 )
		goto tr731;
	goto tr730;
tr673:
	{
		s->line_counter++;
	}
	goto st1230;
tr977:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1230; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1230;
tr866:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1230; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1230; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1230; goto _out;}
		}
	}
	goto st1230;
tr863:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1230; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1230; goto _out;}
		}
	}
	goto st1230;
tr858:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1230; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1230; goto _out;}
		}
	}
	goto st1230;
tr3894:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1230;
st1230:
	if ( ++p == pe )
		goto _test_eof1230;
case 1230:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3874;
		case 32: goto tr3874;
		case 36: goto st177;
		case 40: goto tr3875;
		case 41: goto tr3876;
		case 42: goto tr3856;
		case 65: goto tr3898;
		case 67: goto tr3899;
		case 68: goto tr3900;
		case 69: goto tr3901;
		case 72: goto tr3902;
		case 73: goto tr3903;
		case 75: goto tr3904;
		case 76: goto tr3905;
		case 77: goto tr3906;
		case 78: goto tr3907;
		case 79: goto tr3908;
		case 80: goto tr3909;
		case 82: goto tr3910;
		case 83: goto tr3911;
		case 84: goto tr3912;
		case 85: goto tr3913;
		case 90: goto tr3914;
		case 92: goto tr3856;
		case 95: goto tr3856;
		case 97: goto tr3898;
		case 99: goto tr3899;
		case 100: goto tr3900;
		case 101: goto tr3901;
		case 104: goto tr3902;
		case 105: goto tr3903;
		case 107: goto tr3904;
		case 108: goto tr3905;
		case 109: goto tr3906;
		case 110: goto tr3907;
		case 111: goto tr3908;
		case 112: goto tr3909;
		case 114: goto tr3910;
		case 115: goto tr3911;
		case 116: goto tr3912;
		case 117: goto tr3913;
		case 122: goto tr3914;
		case 778: goto tr3857;
		case 827: goto tr3858;
		case 1034: goto tr3894;
		case 1083: goto tr3895;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3856;
	} else if ( _widec > 57 ) {
		if ( _widec > 89 ) {
			if ( 98 <= _widec && _widec <= 121 )
				goto tr3856;
		} else if ( _widec >= 64 )
			goto tr3856;
	} else
		goto tr3897;
	goto tr3896;
tr3917:
	{
		NOERR;
	}
	goto st177;
st177:
	if ( ++p == pe )
		goto _test_eof177;
case 177:
	switch( (*p) ) {
		case 73: goto tr733;
		case 79: goto tr734;
		case 84: goto tr735;
		case 105: goto tr733;
		case 111: goto tr734;
		case 116: goto tr735;
	}
	goto tr732;
tr733:
	{
		ERR(ZS_OK);
	}
	goto st178;
st178:
	if ( ++p == pe )
		goto _test_eof178;
case 178:
	switch( (*p) ) {
		case 78: goto st179;
		case 110: goto st179;
	}
	goto tr732;
st179:
	if ( ++p == pe )
		goto _test_eof179;
case 179:
	switch( (*p) ) {
		case 67: goto st180;
		case 99: goto st180;
	}
	goto tr732;
st180:
	if ( ++p == pe )
		goto _test_eof180;
case 180:
	switch( (*p) ) {
		case 76: goto st181;
		case 108: goto st181;
	}
	goto tr732;
st181:
	if ( ++p == pe )
		goto _test_eof181;
case 181:
	switch( (*p) ) {
		case 85: goto st182;
		case 117: goto st182;
	}
	goto tr732;
st182:
	if ( ++p == pe )
		goto _test_eof182;
case 182:
	switch( (*p) ) {
		case 68: goto st183;
		case 100: goto st183;
	}
	goto tr732;
st183:
	if ( ++p == pe )
		goto _test_eof183;
case 183:
	switch( (*p) ) {
		case 69: goto st184;
		case 101: goto st184;
	}
	goto tr732;
st184:
	if ( ++p == pe )
		goto _test_eof184;
case 184:
	switch( (*p) ) {
		case 32: goto tr742;
		case 59: goto tr742;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr742;
	} else if ( (*p) >= 9 )
		goto tr742;
	goto tr732;
tr742:
	{ p--; {stack[top++] = 1231;goto st343;} }
	goto st1231;
tr844:
	{ p--; {stack[top++] = 1231;goto st331;} }
	goto st1231;
tr847:
	{ p--; {stack[top++] = 1231;goto st322;} }
	goto st1231;
st1231:
	if ( ++p == pe )
		goto _test_eof1231;
case 1231:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3916;
		case 32: goto tr3916;
		case 36: goto tr3917;
		case 40: goto tr3918;
		case 41: goto tr3919;
		case 42: goto tr3920;
		case 92: goto tr3920;
		case 95: goto tr3920;
		case 778: goto tr3921;
		case 827: goto tr3922;
		case 1034: goto tr3923;
		case 1083: goto tr3924;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3920;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3920;
	} else
		goto tr3920;
	goto tr3915;
tr3856:
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 185;goto st301;} }
	goto st185;
tr3920:
	{
		NOERR;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 185;goto st301;} }
	goto st185;
st185:
	if ( ++p == pe )
		goto _test_eof185;
case 185:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr743;
tr750:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st186;
tr751:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st186;
tr752:
	{
		s->line_counter++;
	}
	goto st186;
tr756:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 186; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st186;
tr744:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st186;
tr745:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st186;
tr746:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st186;
tr747:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st186;
st186:
	if ( ++p == pe )
		goto _test_eof186;
case 186:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st186;
		case 32: goto st186;
		case 40: goto tr750;
		case 41: goto tr751;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 1034: goto tr752;
		case 1083: goto tr753;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr633;
tr753:
	{
		s->buffer_length = 0;
	}
	goto st187;
tr754:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st187;
tr748:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st187;
st187:
	if ( ++p == pe )
		goto _test_eof187;
case 187:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr756;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr754;
	goto st0;
tr24:
	{
		s->line_counter++;
	}
	goto st1232;
tr975:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1232; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1232;
tr821:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1232; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	goto st1232;
tr818:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	goto st1232;
tr852:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	goto st1232;
tr889:
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	goto st1232;
tr892:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1232; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	goto st1232;
tr3859:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1232;
tr3927:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1232;
tr3923:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1232;
tr3932:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st1232;
tr3958:
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1232; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1232; goto _out;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st1232;
st1232:
	if ( ++p == pe )
		goto _test_eof1232;
case 1232:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3852;
		case 32: goto tr3852;
		case 36: goto st177;
		case 40: goto tr3925;
		case 41: goto tr3926;
		case 42: goto tr3856;
		case 65: goto tr3898;
		case 67: goto tr3899;
		case 68: goto tr3900;
		case 69: goto tr3901;
		case 72: goto tr3902;
		case 73: goto tr3903;
		case 75: goto tr3904;
		case 76: goto tr3905;
		case 77: goto tr3906;
		case 78: goto tr3907;
		case 79: goto tr3908;
		case 80: goto tr3909;
		case 82: goto tr3910;
		case 83: goto tr3911;
		case 84: goto tr3912;
		case 85: goto tr3913;
		case 90: goto tr3914;
		case 92: goto tr3856;
		case 95: goto tr3856;
		case 97: goto tr3898;
		case 99: goto tr3899;
		case 100: goto tr3900;
		case 101: goto tr3901;
		case 104: goto tr3902;
		case 105: goto tr3903;
		case 107: goto tr3904;
		case 108: goto tr3905;
		case 109: goto tr3906;
		case 110: goto tr3907;
		case 111: goto tr3908;
		case 112: goto tr3909;
		case 114: goto tr3910;
		case 115: goto tr3911;
		case 116: goto tr3912;
		case 117: goto tr3913;
		case 122: goto tr3914;
		case 778: goto tr3857;
		case 827: goto tr3858;
		case 1034: goto tr3927;
		case 1083: goto tr3928;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3856;
	} else if ( _widec > 57 ) {
		if ( _widec > 89 ) {
			if ( 98 <= _widec && _widec <= 121 )
				goto tr3856;
		} else if ( _widec >= 64 )
			goto tr3856;
	} else
		goto tr3897;
	goto tr3896;
tr3897:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 188;goto st301;} }
	goto st188;
st188:
	if ( ++p == pe )
		goto _test_eof188;
case 188:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr758;
		case 32: goto tr758;
		case 40: goto tr759;
		case 41: goto tr760;
		case 68: goto tr31;
		case 72: goto tr32;
		case 77: goto tr33;
		case 83: goto st191;
		case 87: goto tr35;
		case 100: goto tr31;
		case 104: goto tr32;
		case 109: goto tr33;
		case 115: goto st191;
		case 119: goto tr35;
		case 1034: goto tr761;
		case 1083: goto tr762;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr30;
	goto tr757;
tr764:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st189;
tr765:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st189;
tr766:
	{
		s->line_counter++;
	}
	goto st189;
tr769:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 189; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st189;
tr758:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st189;
tr759:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st189;
tr760:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st189;
tr761:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st189;
st189:
	if ( ++p == pe )
		goto _test_eof189;
case 189:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st189;
		case 32: goto st189;
		case 40: goto tr764;
		case 41: goto tr765;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 1034: goto tr766;
		case 1083: goto tr767;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr633;
tr767:
	{
		s->buffer_length = 0;
	}
	goto st190;
tr768:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st190;
tr762:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st190;
st190:
	if ( ++p == pe )
		goto _test_eof190;
case 190:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr769;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr768;
	goto tr38;
tr31:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st191;
tr32:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st191;
tr33:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st191;
tr35:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st191;
st191:
	if ( ++p == pe )
		goto _test_eof191;
case 191:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr27;
		case 32: goto tr27;
		case 40: goto tr28;
		case 41: goto tr29;
		case 1034: goto tr36;
		case 1083: goto tr37;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr770;
	goto tr26;
tr771:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st192;
tr770:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st192;
tr780:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st192;
st192:
	if ( ++p == pe )
		goto _test_eof192;
case 192:
	switch( (*p) ) {
		case 68: goto tr772;
		case 72: goto tr773;
		case 77: goto tr774;
		case 83: goto st193;
		case 87: goto tr776;
		case 100: goto tr772;
		case 104: goto tr773;
		case 109: goto tr774;
		case 115: goto st193;
		case 119: goto tr776;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr771;
	goto tr26;
tr772:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st193;
tr773:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st193;
tr774:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st193;
tr776:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st193;
st193:
	if ( ++p == pe )
		goto _test_eof193;
case 193:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr777;
		case 32: goto tr777;
		case 40: goto tr778;
		case 41: goto tr779;
		case 1034: goto tr781;
		case 1083: goto tr782;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr780;
	goto tr26;
tr60:
	{
		s->buffer_length = 0;
	}
	goto st194;
tr37:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st194;
tr783:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st194;
tr782:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st194;
st194:
	if ( ++p == pe )
		goto _test_eof194;
case 194:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr784;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr783;
	goto tr38;
tr3877:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 195;goto st301;} }
	goto st195;
tr3898:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 195;goto st301;} }
	goto st195;
st195:
	if ( ++p == pe )
		goto _test_eof195;
case 195:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr786;
		case 32: goto tr786;
		case 40: goto tr787;
		case 41: goto tr788;
		case 65: goto st259;
		case 70: goto st262;
		case 80: goto st266;
		case 97: goto st259;
		case 102: goto st262;
		case 112: goto st266;
		case 2058: goto tr789;
		case 2107: goto tr790;
		case 2314: goto tr70;
		case 2363: goto tr70;
		case 2570: goto tr791;
		case 2619: goto tr792;
	}
	goto tr785;
tr794:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st196;
tr795:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st196;
tr796:
	{
		s->line_counter++;
	}
	goto st196;
tr961:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 196; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st196;
tr786:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st196;
tr787:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st196;
tr788:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st196;
tr789:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st196;
st196:
	if ( ++p == pe )
		goto _test_eof196;
case 196:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st196;
		case 32: goto st196;
		case 40: goto tr794;
		case 41: goto tr795;
		case 58: goto tr73;
		case 65: goto tr122;
		case 67: goto tr123;
		case 68: goto tr124;
		case 69: goto tr125;
		case 72: goto tr126;
		case 73: goto tr127;
		case 75: goto tr128;
		case 76: goto tr129;
		case 77: goto tr130;
		case 78: goto tr131;
		case 79: goto tr132;
		case 80: goto tr133;
		case 82: goto tr134;
		case 83: goto tr135;
		case 84: goto tr136;
		case 85: goto tr137;
		case 90: goto tr138;
		case 92: goto tr78;
		case 97: goto tr122;
		case 99: goto tr123;
		case 100: goto tr124;
		case 101: goto tr125;
		case 104: goto tr126;
		case 105: goto tr127;
		case 107: goto tr128;
		case 108: goto tr129;
		case 109: goto tr130;
		case 110: goto tr131;
		case 111: goto tr132;
		case 112: goto tr133;
		case 114: goto tr134;
		case 115: goto tr135;
		case 116: goto tr136;
		case 117: goto tr137;
		case 122: goto tr138;
		case 2058: goto tr796;
		case 2107: goto tr797;
		case 2314: goto tr73;
		case 2363: goto tr73;
		case 2570: goto tr798;
		case 2619: goto tr799;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr73;
		} else if ( _widec >= 48 )
			goto tr121;
	} else
		goto tr73;
	goto tr635;
tr122:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 197;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 197;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 197;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 197;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 197;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 197;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 197;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 197;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 197;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 197;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 197;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 197;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 197;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 197;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 197;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 197;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 197;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 197;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 197;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 197;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 197;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 197;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 197;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 197;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 197;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 197;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 197;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 197;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 197;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 197;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 197;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st197;
st197:
	if ( ++p == pe )
		goto _test_eof197;
case 197:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr801;
		case 32: goto tr801;
		case 40: goto tr802;
		case 41: goto tr803;
		case 65: goto st259;
		case 70: goto st262;
		case 80: goto st266;
		case 97: goto st259;
		case 102: goto st262;
		case 112: goto st266;
		case 1802: goto tr87;
		case 1851: goto tr88;
		case 2058: goto tr804;
		case 2107: goto tr805;
		case 2314: goto tr806;
		case 2363: goto tr807;
		case 2570: goto tr808;
		case 2619: goto tr809;
	}
	goto tr800;
tr811:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st198;
tr812:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st198;
tr801:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st198;
tr802:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st198;
tr803:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st198;
st198:
	if ( ++p == pe )
		goto _test_eof198;
case 198:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st198;
		case 32: goto st198;
		case 40: goto tr811;
		case 41: goto tr812;
		case 92: goto tr78;
		case 1802: goto tr93;
		case 1851: goto tr94;
		case 2058: goto tr813;
		case 2107: goto tr111;
		case 2314: goto tr112;
		case 2363: goto tr113;
		case 2570: goto tr814;
		case 2619: goto tr114;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr73;
	} else
		goto tr73;
	goto tr83;
tr105:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1233;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1233;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1233;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1233;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1233;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1233;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1233;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1233;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1233;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1233;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1233;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1233;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1233;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1233;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1233;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1233;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1233;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1233;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1233;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1233;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1233;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1233;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1233;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1233;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1233;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1233;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1233;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1233;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1233;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1233;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1233;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1233; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1233; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1233;
tr112:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1233;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1233;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1233;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1233;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1233;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1233;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1233;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1233;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1233;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1233;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1233;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1233;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1233;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1233;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1233;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1233;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1233;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1233;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1233;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1233;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1233;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1233;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1233;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1233;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1233;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1233;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1233;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1233;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1233;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1233;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1233;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1233; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1233; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1233;
tr141:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1233;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1233;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1233;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1233;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1233;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1233;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1233;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1233;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1233;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1233;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1233;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1233;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1233;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1233;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1233;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1233;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1233;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1233;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1233;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1233;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1233;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1233;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1233;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1233;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1233;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1233;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1233;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1233;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1233;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1233;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1233;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1233;
tr806:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1233;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1233;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1233;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1233;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1233;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1233;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1233;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1233;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1233;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1233;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1233;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1233;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1233;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1233;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1233;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1233;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1233;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1233;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1233;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1233;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1233;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1233;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1233;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1233;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1233;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1233;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1233;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1233;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1233;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1233;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1233;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1233; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1233; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1233;
tr3869:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1233;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1233;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1233;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1233;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1233;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1233;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1233;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1233;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1233;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1233;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1233;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1233;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1233;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1233;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1233;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1233;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1233;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1233;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1233;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1233;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1233;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1233;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1233;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1233;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1233;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1233;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1233;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1233;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1233;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1233;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1233;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->line_counter++;
	}
	goto st1233;
st1233:
	if ( ++p == pe )
		goto _test_eof1233;
case 1233:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3929;
		case 32: goto tr3929;
		case 36: goto st177;
		case 40: goto tr3930;
		case 41: goto tr3931;
		case 42: goto tr3856;
		case 92: goto tr3856;
		case 95: goto tr3856;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr3932;
		case 1083: goto tr3933;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3856;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3856;
	} else
		goto tr3856;
	goto tr848;
tr816:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st199;
tr817:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st199;
tr886:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st199;
tr887:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st199;
tr888:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st199;
tr849:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st199;
tr850:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st199;
tr851:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st199;
tr3929:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st199;
tr3930:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st199;
tr3931:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st199;
st199:
	if ( ++p == pe )
		goto _test_eof199;
case 199:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st199;
		case 32: goto st199;
		case 40: goto tr816;
		case 41: goto tr817;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 778: goto tr93;
		case 827: goto tr94;
		case 1034: goto tr818;
		case 1083: goto tr819;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr819:
	{
		s->buffer_length = 0;
	}
	goto st200;
tr820:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st200;
tr890:
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st200;
tr895:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st200;
tr899:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st200;
tr853:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st200;
tr3933:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st200;
st200:
	if ( ++p == pe )
		goto _test_eof200;
case 200:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr96;
		case 1034: goto tr821;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr820;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr89;
tr106:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 201;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 201;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 201;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 201;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 201;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 201;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 201;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 201;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 201;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 201;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 201;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 201;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 201;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 201;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 201;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 201;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 201;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 201;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 201;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 201;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 201;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 201;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 201;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 201;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 201;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 201;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 201;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 201;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 201;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 201;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 201;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st201;
tr113:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 201;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 201;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 201;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 201;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 201;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 201;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 201;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 201;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 201;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 201;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 201;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 201;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 201;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 201;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 201;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 201;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 201;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 201;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 201;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 201;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 201;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 201;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 201;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 201;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 201;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 201;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 201;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 201;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 201;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 201;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 201;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st201;
tr807:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 201;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 201;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 201;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 201;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 201;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 201;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 201;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 201;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 201;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 201;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 201;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 201;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 201;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 201;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 201;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 201;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 201;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 201;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 201;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 201;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 201;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 201;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 201;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 201;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 201;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 201;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 201;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 201;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 201;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 201;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 201;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st201;
st201:
	if ( ++p == pe )
		goto _test_eof201;
case 201:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr822;
		case 32: goto tr822;
		case 40: goto tr823;
		case 41: goto tr824;
		case 778: goto tr825;
		case 827: goto tr826;
		case 1034: goto tr825;
		case 1083: goto tr826;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr95;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr95;
	} else
		goto tr95;
	goto tr83;
tr827:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st202;
tr822:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st202;
tr823:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st202;
tr824:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st202;
tr828:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st202;
tr829:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st202;
st202:
	if ( ++p == pe )
		goto _test_eof202;
case 202:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr827;
		case 32: goto tr827;
		case 40: goto tr828;
		case 41: goto tr829;
		case 778: goto tr96;
		case 827: goto tr830;
		case 1034: goto tr96;
		case 1083: goto tr830;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr95;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr95;
	} else
		goto tr95;
	goto tr89;
tr808:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1234;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1234;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1234;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1234;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1234;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1234;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1234;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1234;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1234;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1234;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1234;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1234;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1234;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1234;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1234;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1234;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1234;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1234;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1234;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1234;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1234;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1234;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1234;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1234;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1234;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1234;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1234;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1234;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1234;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1234;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1234;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1234; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1234; goto _out;}
		}
	}
	goto st1234;
tr814:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1234;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1234;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1234;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1234;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1234;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1234;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1234;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1234;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1234;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1234;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1234;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1234;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1234;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1234;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1234;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1234;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1234;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1234;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1234;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1234;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1234;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1234;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1234;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1234;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1234;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1234;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1234;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1234;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1234;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1234;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1234;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1234; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1234; goto _out;}
		}
	}
	goto st1234;
st1234:
	if ( ++p == pe )
		goto _test_eof1234;
case 1234:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3934;
		case 32: goto tr3934;
		case 36: goto tr3862;
		case 40: goto tr3935;
		case 41: goto tr3936;
		case 42: goto tr3865;
		case 58: goto tr73;
		case 92: goto tr3866;
		case 95: goto tr3865;
		case 1802: goto tr87;
		case 1851: goto tr88;
		case 2058: goto tr3932;
		case 2107: goto tr3937;
		case 2314: goto tr105;
		case 2363: goto tr106;
		case 2570: goto tr3932;
		case 2619: goto tr3938;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr73;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3865;
		} else
			goto tr73;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3865;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr73;
			} else if ( _widec >= 97 )
				goto tr3865;
		} else
			goto tr73;
	} else
		goto tr73;
	goto tr848;
tr832:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st203;
tr833:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st203;
tr962:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st203;
tr963:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st203;
tr964:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st203;
tr3934:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st203;
tr3935:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st203;
tr3936:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st203;
st203:
	if ( ++p == pe )
		goto _test_eof203;
case 203:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st203;
		case 32: goto st203;
		case 40: goto tr832;
		case 41: goto tr833;
		case 58: goto tr73;
		case 65: goto tr122;
		case 67: goto tr123;
		case 68: goto tr124;
		case 69: goto tr125;
		case 72: goto tr126;
		case 73: goto tr127;
		case 75: goto tr128;
		case 76: goto tr129;
		case 77: goto tr130;
		case 78: goto tr131;
		case 79: goto tr132;
		case 80: goto tr133;
		case 82: goto tr134;
		case 83: goto tr135;
		case 84: goto tr136;
		case 85: goto tr137;
		case 90: goto tr138;
		case 92: goto tr78;
		case 97: goto tr122;
		case 99: goto tr123;
		case 100: goto tr124;
		case 101: goto tr125;
		case 104: goto tr126;
		case 105: goto tr127;
		case 107: goto tr128;
		case 108: goto tr129;
		case 109: goto tr130;
		case 110: goto tr131;
		case 111: goto tr132;
		case 112: goto tr133;
		case 114: goto tr134;
		case 115: goto tr135;
		case 116: goto tr136;
		case 117: goto tr137;
		case 122: goto tr138;
		case 1802: goto tr93;
		case 1851: goto tr94;
		case 2058: goto tr818;
		case 2107: goto tr834;
		case 2314: goto tr112;
		case 2363: goto tr113;
		case 2570: goto tr818;
		case 2619: goto tr835;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr73;
		} else if ( _widec >= 48 )
			goto tr121;
	} else
		goto tr73;
	goto tr118;
tr123:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 204;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 204;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 204;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 204;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 204;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 204;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 204;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 204;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 204;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 204;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 204;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 204;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 204;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 204;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 204;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 204;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 204;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 204;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 204;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 204;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 204;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 204;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 204;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 204;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 204;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 204;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 204;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 204;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 204;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 204;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 204;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st204;
st204:
	if ( ++p == pe )
		goto _test_eof204;
case 204:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 83: goto st36;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 115: goto st36;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr124:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 205;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 205;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 205;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 205;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 205;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 205;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 205;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 205;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 205;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 205;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 205;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 205;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 205;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 205;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 205;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 205;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 205;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 205;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 205;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 205;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 205;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 205;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 205;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 205;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 205;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 205;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 205;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 205;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 205;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 205;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 205;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st205;
st205:
	if ( ++p == pe )
		goto _test_eof205;
case 205:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 72: goto st41;
		case 78: goto st45;
		case 83: goto st53;
		case 104: goto st41;
		case 110: goto st45;
		case 115: goto st53;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr125:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 206;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 206;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 206;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 206;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 206;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 206;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 206;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 206;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 206;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 206;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 206;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 206;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 206;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 206;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 206;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 206;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 206;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 206;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 206;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 206;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 206;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 206;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 206;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 206;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 206;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 206;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 206;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 206;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 206;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 206;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 206;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st206;
st206:
	if ( ++p == pe )
		goto _test_eof206;
case 206:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 85: goto st55;
		case 117: goto st55;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr126:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 207;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 207;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 207;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 207;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 207;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 207;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 207;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 207;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 207;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 207;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 207;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 207;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 207;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 207;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 207;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 207;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 207;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 207;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 207;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 207;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 207;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 207;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 207;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 207;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 207;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 207;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 207;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 207;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 207;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 207;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 207;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st207;
st207:
	if ( ++p == pe )
		goto _test_eof207;
case 207:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 73: goto st62;
		case 105: goto st62;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr127:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 208;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 208;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 208;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 208;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 208;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 208;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 208;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 208;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 208;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 208;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 208;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 208;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 208;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 208;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 208;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 208;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 208;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 208;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 208;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 208;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 208;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 208;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 208;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 208;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 208;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 208;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 208;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 208;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 208;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 208;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 208;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st208;
st208:
	if ( ++p == pe )
		goto _test_eof208;
case 208:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 78: goto st169;
		case 80: goto st70;
		case 110: goto st169;
		case 112: goto st70;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr128:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 209;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 209;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 209;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 209;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 209;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 209;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 209;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 209;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 209;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 209;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 209;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 209;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 209;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 209;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 209;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 209;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 209;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 209;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 209;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 209;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 209;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 209;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 209;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 209;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 209;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 209;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 209;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 209;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 209;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 209;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 209;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st209;
st209:
	if ( ++p == pe )
		goto _test_eof209;
case 209:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 69: goto st78;
		case 88: goto st80;
		case 101: goto st78;
		case 120: goto st80;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr129:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 210;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 210;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 210;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 210;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 210;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 210;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 210;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 210;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st210;
st210:
	if ( ++p == pe )
		goto _test_eof210;
case 210:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 51: goto st82;
		case 54: goto st84;
		case 79: goto st86;
		case 80: goto st88;
		case 111: goto st86;
		case 112: goto st88;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr130:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 211;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 211;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 211;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 211;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 211;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 211;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 211;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 211;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 211;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 211;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 211;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 211;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 211;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 211;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 211;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 211;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 211;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 211;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 211;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 211;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 211;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 211;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 211;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 211;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 211;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 211;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 211;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 211;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 211;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 211;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 211;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st211;
st211:
	if ( ++p == pe )
		goto _test_eof211;
case 211:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 73: goto st90;
		case 88: goto st94;
		case 105: goto st90;
		case 120: goto st94;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr131:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 212;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 212;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 212;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 212;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 212;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 212;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 212;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 212;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 212;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 212;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 212;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 212;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 212;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 212;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 212;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 212;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 212;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 212;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 212;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 212;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 212;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 212;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 212;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 212;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 212;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 212;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 212;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 212;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 212;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 212;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 212;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st212;
st212:
	if ( ++p == pe )
		goto _test_eof212;
case 212:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 65: goto st96;
		case 73: goto st100;
		case 83: goto st102;
		case 97: goto st96;
		case 105: goto st100;
		case 115: goto st102;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr132:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 213;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 213;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 213;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 213;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 213;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 213;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 213;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 213;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 213;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 213;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 213;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 213;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 213;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 213;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 213;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 213;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 213;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 213;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 213;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 213;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 213;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 213;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 213;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 213;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 213;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 213;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 213;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 213;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 213;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 213;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 213;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st213;
st213:
	if ( ++p == pe )
		goto _test_eof213;
case 213:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 80: goto st112;
		case 112: goto st112;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr133:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 214;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 214;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 214;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 214;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 214;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 214;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 214;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 214;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 214;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 214;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 214;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 214;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 214;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 214;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 214;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 214;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 214;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 214;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 214;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 214;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 214;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 214;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 214;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 214;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 214;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 214;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 214;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 214;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 214;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 214;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 214;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st214;
st214:
	if ( ++p == pe )
		goto _test_eof214;
case 214:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 84: goto st122;
		case 116: goto st122;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr134:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 215;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 215;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 215;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 215;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 215;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 215;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 215;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 215;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 215;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 215;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 215;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 215;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 215;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 215;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 215;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 215;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 215;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 215;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 215;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 215;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 215;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 215;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 215;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 215;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 215;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 215;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 215;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 215;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 215;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 215;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 215;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st215;
st215:
	if ( ++p == pe )
		goto _test_eof215;
case 215:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 80: goto st125;
		case 82: goto st126;
		case 84: goto st130;
		case 112: goto st125;
		case 114: goto st126;
		case 116: goto st130;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr135:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 216;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 216;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 216;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 216;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 216;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 216;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 216;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 216;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 216;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 216;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 216;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 216;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 216;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 216;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 216;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 216;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 216;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 216;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 216;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 216;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 216;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 216;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 216;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 216;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 216;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 216;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 216;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 216;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 216;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 216;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 216;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st216;
st216:
	if ( ++p == pe )
		goto _test_eof216;
case 216:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 77: goto st132;
		case 79: goto st137;
		case 80: goto st139;
		case 82: goto st141;
		case 83: goto st143;
		case 109: goto st132;
		case 111: goto st137;
		case 112: goto st139;
		case 114: goto st141;
		case 115: goto st143;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr136:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 217;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 217;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 217;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 217;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 217;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 217;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 217;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 217;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 217;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 217;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 217;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 217;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 217;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 217;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 217;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 217;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 217;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 217;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 217;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 217;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 217;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 217;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 217;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 217;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 217;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 217;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 217;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 217;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 217;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 217;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 217;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st217;
st217:
	if ( ++p == pe )
		goto _test_eof217;
case 217:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 76: goto st148;
		case 88: goto st151;
		case 89: goto st153;
		case 108: goto st148;
		case 120: goto st151;
		case 121: goto st153;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr137:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 218;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 218;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 218;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 218;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 218;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 218;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 218;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 218;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 218;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 218;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 218;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 218;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 218;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 218;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 218;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 218;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 218;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 218;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 218;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 218;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 218;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 218;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 218;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 218;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 218;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 218;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 218;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 218;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 218;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 218;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 218;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st218;
st218:
	if ( ++p == pe )
		goto _test_eof218;
case 218:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 82: goto st158;
		case 114: goto st158;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr138:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 219;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 219;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 219;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 219;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 219;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 219;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 219;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 219;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 219;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 219;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 219;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 219;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 219;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 219;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 219;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 219;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 219;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 219;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 219;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 219;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 219;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 219;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 219;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 219;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 219;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 219;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 219;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 219;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 219;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 219;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 219;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st219;
st219:
	if ( ++p == pe )
		goto _test_eof219;
case 219:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 79: goto st161;
		case 111: goto st161;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr800;
tr834:
	{
		s->buffer_length = 0;
	}
	goto st220;
tr836:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st220;
tr965:
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st220;
tr903:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st220;
tr907:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st220;
tr872:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st220;
tr3937:
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st220;
st220:
	if ( ++p == pe )
		goto _test_eof220;
case 220:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr96;
		case 1034: goto tr837;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr836;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr83;
tr139:
	{
		s->line_counter++;
	}
	goto st1235;
tr939:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1235; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1235;
tr837:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1235; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1235; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1235; goto _out;}
		}
	}
	goto st1235;
tr878:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1235; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1235; goto _out;}
		}
	}
	goto st1235;
tr871:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1235; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1235; goto _out;}
		}
	}
	goto st1235;
tr3867:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1235;
st1235:
	if ( ++p == pe )
		goto _test_eof1235;
case 1235:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3861;
		case 32: goto tr3861;
		case 36: goto tr3862;
		case 40: goto tr3863;
		case 41: goto tr3864;
		case 42: goto tr3865;
		case 58: goto tr73;
		case 65: goto tr3941;
		case 67: goto tr3942;
		case 68: goto tr3943;
		case 69: goto tr3944;
		case 72: goto tr3945;
		case 73: goto tr3946;
		case 75: goto tr3947;
		case 76: goto tr3948;
		case 77: goto tr3949;
		case 78: goto tr3950;
		case 79: goto tr3951;
		case 80: goto tr3952;
		case 82: goto tr3953;
		case 83: goto tr3954;
		case 84: goto tr3955;
		case 85: goto tr3956;
		case 90: goto tr3957;
		case 92: goto tr3866;
		case 95: goto tr3865;
		case 97: goto tr3941;
		case 99: goto tr3942;
		case 100: goto tr3943;
		case 101: goto tr3944;
		case 104: goto tr3945;
		case 105: goto tr3946;
		case 107: goto tr3947;
		case 108: goto tr3948;
		case 109: goto tr3949;
		case 110: goto tr3950;
		case 111: goto tr3951;
		case 112: goto tr3952;
		case 114: goto tr3953;
		case 115: goto tr3954;
		case 116: goto tr3955;
		case 117: goto tr3956;
		case 122: goto tr3957;
		case 1802: goto tr3857;
		case 1851: goto tr3858;
		case 2058: goto tr3867;
		case 2107: goto tr3868;
		case 2314: goto tr3869;
		case 2363: goto tr3870;
		case 2570: goto tr3871;
		case 2619: goto tr3872;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr73;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3940;
			} else if ( _widec >= 45 )
				goto tr3865;
		} else
			goto tr73;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 89 )
				goto tr3865;
		} else if ( _widec > 96 ) {
			if ( _widec > 121 ) {
				if ( 123 <= _widec )
					goto tr73;
			} else if ( _widec >= 98 )
				goto tr3865;
		} else
			goto tr73;
	} else
		goto tr73;
	goto tr3939;
tr3862:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 221;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 221;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 221;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 221;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 221;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 221;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 221;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 221;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 221;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 221;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 221;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 221;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 221;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 221;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 221;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 221;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 221;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 221;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 221;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 221;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 221;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 221;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 221;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 221;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 221;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 221;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 221;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 221;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 221;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 221;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 221;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st221;
st221:
	if ( ++p == pe )
		goto _test_eof221;
case 221:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr84;
		case 32: goto tr84;
		case 40: goto tr85;
		case 41: goto tr86;
		case 73: goto tr733;
		case 79: goto tr734;
		case 84: goto tr735;
		case 105: goto tr733;
		case 111: goto tr734;
		case 116: goto tr735;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr87;
		case 1083: goto tr88;
	}
	goto tr838;
tr734:
	{
		ERR(ZS_OK);
	}
	goto st222;
st222:
	if ( ++p == pe )
		goto _test_eof222;
case 222:
	switch( (*p) ) {
		case 82: goto st223;
		case 114: goto st223;
	}
	goto tr732;
st223:
	if ( ++p == pe )
		goto _test_eof223;
case 223:
	switch( (*p) ) {
		case 73: goto st224;
		case 105: goto st224;
	}
	goto tr732;
st224:
	if ( ++p == pe )
		goto _test_eof224;
case 224:
	switch( (*p) ) {
		case 71: goto st225;
		case 103: goto st225;
	}
	goto tr732;
st225:
	if ( ++p == pe )
		goto _test_eof225;
case 225:
	switch( (*p) ) {
		case 73: goto st226;
		case 105: goto st226;
	}
	goto tr732;
st226:
	if ( ++p == pe )
		goto _test_eof226;
case 226:
	switch( (*p) ) {
		case 78: goto st227;
		case 110: goto st227;
	}
	goto tr732;
st227:
	if ( ++p == pe )
		goto _test_eof227;
case 227:
	switch( (*p) ) {
		case 32: goto tr844;
		case 59: goto tr844;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr844;
	} else if ( (*p) >= 9 )
		goto tr844;
	goto tr732;
tr735:
	{
		ERR(ZS_OK);
	}
	goto st228;
st228:
	if ( ++p == pe )
		goto _test_eof228;
case 228:
	switch( (*p) ) {
		case 84: goto st229;
		case 116: goto st229;
	}
	goto tr732;
st229:
	if ( ++p == pe )
		goto _test_eof229;
case 229:
	switch( (*p) ) {
		case 76: goto st230;
		case 108: goto st230;
	}
	goto tr732;
st230:
	if ( ++p == pe )
		goto _test_eof230;
case 230:
	switch( (*p) ) {
		case 32: goto tr847;
		case 59: goto tr847;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr847;
	} else if ( (*p) >= 9 )
		goto tr847;
	goto tr732;
tr3865:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 231;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 231;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 231;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 231;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 231;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 231;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 231;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 231;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 231;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 231;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 231;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 231;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 231;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 231;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 231;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 231;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 231;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 231;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 231;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 231;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 231;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 231;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 231;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 231;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 231;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 231;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 231;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 231;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 231;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 231;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 231;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 231;goto st301;} }
	goto st231;
st231:
	if ( ++p == pe )
		goto _test_eof231;
case 231:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr848;
tr3940:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 232;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 232;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 232;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 232;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 232;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 232;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 232;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 232;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 232;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 232;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 232;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 232;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 232;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 232;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 232;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 232;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 232;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 232;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 232;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 232;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 232;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 232;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 232;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 232;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 232;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 232;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 232;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 232;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 232;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 232;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 232;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 232;goto st301;} }
	goto st232;
st232:
	if ( ++p == pe )
		goto _test_eof232;
case 232:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr855;
		case 32: goto tr855;
		case 40: goto tr856;
		case 41: goto tr857;
		case 68: goto tr31;
		case 72: goto tr32;
		case 77: goto tr33;
		case 83: goto st191;
		case 87: goto tr35;
		case 100: goto tr31;
		case 104: goto tr32;
		case 109: goto tr33;
		case 115: goto st191;
		case 119: goto tr35;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr858;
		case 1083: goto tr859;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr30;
	goto tr854;
tr861:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st233;
tr862:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st233;
tr855:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st233;
tr856:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st233;
tr857:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st233;
st233:
	if ( ++p == pe )
		goto _test_eof233;
case 233:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st233;
		case 32: goto st233;
		case 40: goto tr861;
		case 41: goto tr862;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 778: goto tr93;
		case 827: goto tr94;
		case 1034: goto tr863;
		case 1083: goto tr864;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr864:
	{
		s->buffer_length = 0;
	}
	goto st234;
tr865:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st234;
tr859:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st234;
st234:
	if ( ++p == pe )
		goto _test_eof234;
case 234:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr96;
		case 1034: goto tr866;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr865;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr151;
tr3941:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 235;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 235;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 235;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 235;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 235;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 235;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 235;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 235;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 235;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 235;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 235;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 235;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 235;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 235;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 235;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 235;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 235;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 235;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 235;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 235;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 235;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 235;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 235;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 235;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 235;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 235;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 235;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 235;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 235;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 235;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 235;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 235;goto st301;} }
	goto st235;
st235:
	if ( ++p == pe )
		goto _test_eof235;
case 235:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr868;
		case 32: goto tr868;
		case 40: goto tr869;
		case 41: goto tr870;
		case 65: goto st259;
		case 70: goto st262;
		case 80: goto st266;
		case 97: goto st259;
		case 102: goto st262;
		case 112: goto st266;
		case 1802: goto tr87;
		case 1851: goto tr88;
		case 2058: goto tr871;
		case 2107: goto tr872;
		case 2314: goto tr806;
		case 2363: goto tr807;
		case 2570: goto tr873;
		case 2619: goto tr874;
	}
	goto tr867;
tr876:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st236;
tr877:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st236;
tr868:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st236;
tr869:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st236;
tr870:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st236;
st236:
	if ( ++p == pe )
		goto _test_eof236;
case 236:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st236;
		case 32: goto st236;
		case 40: goto tr876;
		case 41: goto tr877;
		case 58: goto tr73;
		case 65: goto tr122;
		case 67: goto tr123;
		case 68: goto tr124;
		case 69: goto tr125;
		case 72: goto tr126;
		case 73: goto tr127;
		case 75: goto tr128;
		case 76: goto tr129;
		case 77: goto tr130;
		case 78: goto tr131;
		case 79: goto tr132;
		case 80: goto tr133;
		case 82: goto tr134;
		case 83: goto tr135;
		case 84: goto tr136;
		case 85: goto tr137;
		case 90: goto tr138;
		case 92: goto tr78;
		case 97: goto tr122;
		case 99: goto tr123;
		case 100: goto tr124;
		case 101: goto tr125;
		case 104: goto tr126;
		case 105: goto tr127;
		case 107: goto tr128;
		case 108: goto tr129;
		case 109: goto tr130;
		case 110: goto tr131;
		case 111: goto tr132;
		case 112: goto tr133;
		case 114: goto tr134;
		case 115: goto tr135;
		case 116: goto tr136;
		case 117: goto tr137;
		case 122: goto tr138;
		case 1802: goto tr93;
		case 1851: goto tr94;
		case 2058: goto tr878;
		case 2107: goto tr834;
		case 2314: goto tr112;
		case 2363: goto tr113;
		case 2570: goto tr879;
		case 2619: goto tr835;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr73;
		} else if ( _widec >= 48 )
			goto tr121;
	} else
		goto tr73;
	goto tr118;
tr143:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1236;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1236;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1236;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1236;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1236;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1236;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1236;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1236;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1236;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1236;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1236;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1236;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1236;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1236;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1236;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1236;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1236;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1236;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1236;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1236;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1236;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1236;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1236;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1236;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1236;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1236;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1236;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1236;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1236;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1236;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1236;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st1236;
tr879:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1236;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1236;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1236;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1236;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1236;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1236;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1236;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1236;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1236;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1236;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1236;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1236;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1236;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1236;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1236;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1236;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1236;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1236;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1236;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1236;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1236;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1236;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1236;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1236;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1236;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1236;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1236;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1236;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1236;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1236;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1236;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1236; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1236; goto _out;}
		}
	}
	goto st1236;
tr873:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1236;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1236;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1236;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1236;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1236;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1236;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1236;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1236;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1236;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1236;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1236;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1236;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1236;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1236;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1236;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1236;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1236;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1236;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1236;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1236;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1236;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1236;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1236;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1236;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1236;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1236;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1236;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1236;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1236;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1236;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1236;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1236; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1236; goto _out;}
		}
	}
	goto st1236;
tr3871:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1236;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1236;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1236;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1236;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1236;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1236;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1236;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1236;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1236;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1236;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1236;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1236;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1236;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1236;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1236;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1236;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1236;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1236;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1236;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 1236;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1236;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1236;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 1236;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 1236;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 1236;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1236;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1236;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1236;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1236;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1236;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1236;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1236;
st1236:
	if ( ++p == pe )
		goto _test_eof1236;
case 1236:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3934;
		case 32: goto tr3934;
		case 36: goto tr3862;
		case 40: goto tr3935;
		case 41: goto tr3936;
		case 42: goto tr3865;
		case 58: goto tr73;
		case 65: goto tr3941;
		case 67: goto tr3942;
		case 68: goto tr3943;
		case 69: goto tr3944;
		case 72: goto tr3945;
		case 73: goto tr3946;
		case 75: goto tr3947;
		case 76: goto tr3948;
		case 77: goto tr3949;
		case 78: goto tr3950;
		case 79: goto tr3951;
		case 80: goto tr3952;
		case 82: goto tr3953;
		case 83: goto tr3954;
		case 84: goto tr3955;
		case 85: goto tr3956;
		case 90: goto tr3957;
		case 92: goto tr3866;
		case 95: goto tr3865;
		case 97: goto tr3941;
		case 99: goto tr3942;
		case 100: goto tr3943;
		case 101: goto tr3944;
		case 104: goto tr3945;
		case 105: goto tr3946;
		case 107: goto tr3947;
		case 108: goto tr3948;
		case 109: goto tr3949;
		case 110: goto tr3950;
		case 111: goto tr3951;
		case 112: goto tr3952;
		case 114: goto tr3953;
		case 115: goto tr3954;
		case 116: goto tr3955;
		case 117: goto tr3956;
		case 122: goto tr3957;
		case 1802: goto tr87;
		case 1851: goto tr88;
		case 2058: goto tr3958;
		case 2107: goto tr3937;
		case 2314: goto tr105;
		case 2363: goto tr106;
		case 2570: goto tr3958;
		case 2619: goto tr3938;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr73;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3940;
			} else if ( _widec >= 45 )
				goto tr3865;
		} else
			goto tr73;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 89 )
				goto tr3865;
		} else if ( _widec > 96 ) {
			if ( _widec > 121 ) {
				if ( 123 <= _widec )
					goto tr73;
			} else if ( _widec >= 98 )
				goto tr3865;
		} else
			goto tr73;
	} else
		goto tr73;
	goto tr3939;
tr3942:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 237;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 237;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 237;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 237;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 237;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 237;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 237;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 237;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 237;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 237;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 237;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 237;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 237;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 237;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 237;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 237;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 237;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 237;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 237;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 237;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 237;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 237;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 237;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 237;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 237;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 237;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 237;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 237;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 237;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 237;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 237;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 237;goto st301;} }
	goto st237;
st237:
	if ( ++p == pe )
		goto _test_eof237;
case 237:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 83: goto st36;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 115: goto st36;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3943:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 238;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 238;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 238;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 238;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 238;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 238;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 238;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 238;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 238;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 238;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 238;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 238;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 238;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 238;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 238;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 238;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 238;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 238;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 238;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 238;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 238;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 238;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 238;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 238;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 238;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 238;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 238;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 238;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 238;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 238;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 238;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 238;goto st301;} }
	goto st238;
st238:
	if ( ++p == pe )
		goto _test_eof238;
case 238:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 72: goto st41;
		case 78: goto st45;
		case 83: goto st53;
		case 104: goto st41;
		case 110: goto st45;
		case 115: goto st53;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3944:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 239;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 239;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 239;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 239;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 239;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 239;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 239;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 239;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 239;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 239;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 239;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 239;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 239;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 239;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 239;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 239;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 239;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 239;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 239;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 239;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 239;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 239;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 239;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 239;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 239;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 239;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 239;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 239;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 239;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 239;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 239;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 239;goto st301;} }
	goto st239;
st239:
	if ( ++p == pe )
		goto _test_eof239;
case 239:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 85: goto st55;
		case 117: goto st55;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3945:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 240;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 240;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 240;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 240;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 240;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 240;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 240;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 240;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 240;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 240;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 240;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 240;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 240;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 240;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 240;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 240;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 240;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 240;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 240;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 240;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 240;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 240;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 240;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 240;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 240;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 240;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 240;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 240;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 240;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 240;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 240;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 240;goto st301;} }
	goto st240;
st240:
	if ( ++p == pe )
		goto _test_eof240;
case 240:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 73: goto st62;
		case 105: goto st62;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3946:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 241;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 241;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 241;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 241;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 241;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 241;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 241;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 241;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 241;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 241;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 241;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 241;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 241;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 241;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 241;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 241;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 241;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 241;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 241;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 241;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 241;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 241;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 241;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 241;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 241;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 241;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 241;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 241;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 241;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 241;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 241;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 241;goto st301;} }
	goto st241;
st241:
	if ( ++p == pe )
		goto _test_eof241;
case 241:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 78: goto st169;
		case 80: goto st70;
		case 110: goto st169;
		case 112: goto st70;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3947:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 242;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 242;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 242;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 242;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 242;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 242;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 242;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 242;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 242;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 242;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 242;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 242;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 242;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 242;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 242;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 242;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 242;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 242;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 242;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 242;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 242;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 242;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 242;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 242;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 242;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 242;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 242;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 242;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 242;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 242;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 242;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 242;goto st301;} }
	goto st242;
st242:
	if ( ++p == pe )
		goto _test_eof242;
case 242:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 69: goto st78;
		case 88: goto st80;
		case 101: goto st78;
		case 120: goto st80;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3948:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 243;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 243;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 243;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 243;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 243;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 243;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 243;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 243;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 243;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 243;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 243;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 243;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 243;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 243;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 243;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 243;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 243;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 243;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 243;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 243;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 243;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 243;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 243;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 243;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 243;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 243;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 243;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 243;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 243;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 243;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 243;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 243;goto st301;} }
	goto st243;
st243:
	if ( ++p == pe )
		goto _test_eof243;
case 243:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 51: goto st82;
		case 54: goto st84;
		case 79: goto st86;
		case 80: goto st88;
		case 111: goto st86;
		case 112: goto st88;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3949:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 244;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 244;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 244;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 244;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 244;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 244;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 244;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 244;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 244;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 244;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 244;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 244;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 244;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 244;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 244;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 244;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 244;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 244;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 244;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 244;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 244;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 244;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 244;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 244;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 244;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 244;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 244;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 244;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 244;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 244;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 244;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 244;goto st301;} }
	goto st244;
st244:
	if ( ++p == pe )
		goto _test_eof244;
case 244:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 73: goto st90;
		case 88: goto st94;
		case 105: goto st90;
		case 120: goto st94;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3950:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 245;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 245;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 245;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 245;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 245;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 245;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 245;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 245;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 245;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 245;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 245;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 245;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 245;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 245;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 245;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 245;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 245;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 245;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 245;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 245;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 245;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 245;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 245;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 245;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 245;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 245;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 245;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 245;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 245;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 245;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 245;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 245;goto st301;} }
	goto st245;
st245:
	if ( ++p == pe )
		goto _test_eof245;
case 245:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 65: goto st96;
		case 73: goto st100;
		case 83: goto st102;
		case 97: goto st96;
		case 105: goto st100;
		case 115: goto st102;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3951:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 246;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 246;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 246;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 246;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 246;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 246;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 246;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 246;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 246;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 246;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 246;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 246;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 246;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 246;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 246;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 246;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 246;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 246;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 246;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 246;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 246;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 246;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 246;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 246;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 246;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 246;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 246;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 246;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 246;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 246;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 246;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 246;goto st301;} }
	goto st246;
st246:
	if ( ++p == pe )
		goto _test_eof246;
case 246:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 80: goto st112;
		case 112: goto st112;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3952:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 247;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 247;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 247;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 247;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 247;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 247;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 247;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 247;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 247;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 247;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 247;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 247;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 247;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 247;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 247;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 247;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 247;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 247;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 247;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 247;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 247;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 247;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 247;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 247;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 247;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 247;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 247;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 247;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 247;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 247;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 247;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 247;goto st301;} }
	goto st247;
st247:
	if ( ++p == pe )
		goto _test_eof247;
case 247:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 84: goto st122;
		case 116: goto st122;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3953:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 248;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 248;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 248;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 248;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 248;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 248;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 248;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 248;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 248;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 248;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 248;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 248;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 248;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 248;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 248;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 248;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 248;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 248;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 248;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 248;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 248;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 248;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 248;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 248;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 248;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 248;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 248;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 248;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 248;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 248;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 248;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 248;goto st301;} }
	goto st248;
st248:
	if ( ++p == pe )
		goto _test_eof248;
case 248:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 80: goto st125;
		case 82: goto st126;
		case 84: goto st130;
		case 112: goto st125;
		case 114: goto st126;
		case 116: goto st130;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3954:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 249;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 249;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 249;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 249;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 249;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 249;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 249;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 249;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 249;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 249;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 249;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 249;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 249;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 249;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 249;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 249;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 249;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 249;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 249;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 249;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 249;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 249;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 249;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 249;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 249;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 249;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 249;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 249;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 249;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 249;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 249;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 249;goto st301;} }
	goto st249;
st249:
	if ( ++p == pe )
		goto _test_eof249;
case 249:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 77: goto st132;
		case 79: goto st137;
		case 80: goto st139;
		case 82: goto st141;
		case 83: goto st143;
		case 109: goto st132;
		case 111: goto st137;
		case 112: goto st139;
		case 114: goto st141;
		case 115: goto st143;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3955:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 250;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 250;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 250;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 250;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 250;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 250;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 250;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 250;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 250;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 250;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 250;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 250;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 250;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 250;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 250;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 250;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 250;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 250;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 250;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 250;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 250;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 250;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 250;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 250;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 250;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 250;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 250;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 250;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 250;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 250;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 250;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 250;goto st301;} }
	goto st250;
st250:
	if ( ++p == pe )
		goto _test_eof250;
case 250:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 76: goto st148;
		case 88: goto st151;
		case 89: goto st153;
		case 108: goto st148;
		case 120: goto st151;
		case 121: goto st153;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3956:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 251;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 251;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 251;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 251;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 251;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 251;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 251;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 251;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 251;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 251;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 251;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 251;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 251;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 251;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 251;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 251;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 251;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 251;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 251;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 251;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 251;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 251;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 251;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 251;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 251;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 251;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 251;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 251;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 251;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 251;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 251;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 251;goto st301;} }
	goto st251;
st251:
	if ( ++p == pe )
		goto _test_eof251;
case 251:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 82: goto st158;
		case 114: goto st158;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3957:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 252;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 252;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 252;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 252;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 252;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 252;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 252;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 252;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 252;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 252;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 252;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 252;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 252;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 252;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 252;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 252;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 252;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 252;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 252;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 252;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 252;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 252;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 252;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 252;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 252;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 252;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 252;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 252;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 252;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 252;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 252;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 252;goto st301;} }
	goto st252;
st252:
	if ( ++p == pe )
		goto _test_eof252;
case 252:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr849;
		case 32: goto tr849;
		case 40: goto tr850;
		case 41: goto tr851;
		case 79: goto st161;
		case 111: goto st161;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr852;
		case 1083: goto tr853;
	}
	goto tr867;
tr3866:
	{
		if (pe - p == 1) {
			*wrap = WRAP_DETECTED;
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 253;goto st301;} }
	goto st253;
st253:
	if ( ++p == pe )
		goto _test_eof253;
case 253:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr880;
		case 32: goto tr880;
		case 35: goto tr98;
		case 40: goto tr882;
		case 41: goto tr883;
		case 778: goto tr97;
		case 827: goto tr97;
		case 1034: goto tr884;
		case 1083: goto tr885;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr97;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr97;
	} else
		goto tr97;
	goto tr881;
tr880:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (*wrap == WRAP_NONE) {
			p--;
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 254;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 254;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 254;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 254;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 254;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 254;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 254;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 254;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 254;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 254;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 254;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 254;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 254;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 254;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 254;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 254;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 254;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 254;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 254;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 254;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 254;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 254;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 254;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 254;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 254;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 254;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 254;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 254;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 254;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 254;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 254;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st254;
tr882:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (*wrap == WRAP_NONE) {
			p--;
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 254;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 254;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 254;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 254;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 254;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 254;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 254;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 254;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 254;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 254;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 254;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 254;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 254;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 254;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 254;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 254;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 254;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 254;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 254;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 254;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 254;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 254;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 254;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 254;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 254;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 254;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 254;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 254;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 254;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 254;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 254;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st254;
tr883:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (*wrap == WRAP_NONE) {
			p--;
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 254;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 254;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 254;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 254;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 254;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 254;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 254;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 254;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 254;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 254;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 254;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 254;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 254;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 254;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 254;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 254;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 254;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 254;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 254;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 254;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 254;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 254;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 254;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 254;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 254;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 254;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 254;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 254;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 254;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 254;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 254;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st254;
tr884:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (*wrap == WRAP_NONE) {
			p--;
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 254;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 254;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 254;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 254;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 254;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 254;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 254;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 254;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 254;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 254;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 254;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 254;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 254;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 254;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 254;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 254;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 254;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 254;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 254;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 254;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 254;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 254;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 254;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 254;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 254;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 254;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 254;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 254;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 254;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 254;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 254;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st254;
st254:
	if ( ++p == pe )
		goto _test_eof254;
case 254:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr886;
		case 32: goto tr886;
		case 40: goto tr887;
		case 41: goto tr888;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 79: goto tr15;
		case 80: goto tr16;
		case 82: goto tr17;
		case 83: goto tr18;
		case 84: goto tr19;
		case 85: goto tr20;
		case 90: goto tr21;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 111: goto tr15;
		case 112: goto tr16;
		case 114: goto tr17;
		case 115: goto tr18;
		case 116: goto tr19;
		case 117: goto tr20;
		case 122: goto tr21;
		case 778: goto tr87;
		case 827: goto tr88;
		case 1034: goto tr889;
		case 1083: goto tr890;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr118;
tr885:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		if (*wrap == WRAP_NONE) {
			p--;
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 255;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 255;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 255;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 255;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 255;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 255;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 255;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 255;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 255;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 255;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 255;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 255;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 255;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 255;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 255;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 255;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 255;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 255;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 255;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 255;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 255;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 255;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 255;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 255;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 255;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 255;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 255;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 255;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 255;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 255;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 255;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st255;
st255:
	if ( ++p == pe )
		goto _test_eof255;
case 255:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr84;
		case 778: goto tr87;
		case 800: goto tr84;
		case 808: goto tr85;
		case 809: goto tr86;
		case 827: goto tr88;
		case 1033: goto tr891;
		case 1034: goto tr892;
		case 1056: goto tr891;
		case 1064: goto tr893;
		case 1065: goto tr894;
		case 1083: goto tr895;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr754;
	goto tr83;
tr896:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st256;
tr891:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st256;
tr893:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st256;
tr894:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st256;
tr897:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st256;
tr898:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st256;
st256:
	if ( ++p == pe )
		goto _test_eof256;
case 256:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr93;
		case 800: goto st7;
		case 808: goto tr91;
		case 809: goto tr92;
		case 827: goto tr94;
		case 1033: goto tr896;
		case 1034: goto tr821;
		case 1056: goto tr896;
		case 1064: goto tr897;
		case 1065: goto tr898;
		case 1083: goto tr899;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr754;
	goto tr89;
tr835:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 257;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 257;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 257;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 257;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 257;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 257;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 257;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 257;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 257;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 257;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 257;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 257;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 257;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 257;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 257;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 257;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 257;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 257;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 257;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 257;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 257;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 257;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 257;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 257;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 257;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 257;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 257;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 257;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 257;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 257;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 257;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st257;
tr966:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 257;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 257;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 257;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 257;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 257;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 257;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 257;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 257;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 257;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 257;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 257;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 257;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 257;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 257;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 257;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 257;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 257;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 257;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 257;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 257;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 257;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 257;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 257;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 257;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 257;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 257;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 257;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 257;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 257;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 257;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 257;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st257;
tr874:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 257;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 257;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 257;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 257;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 257;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 257;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 257;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 257;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 257;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 257;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 257;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 257;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 257;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 257;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 257;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 257;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 257;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 257;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 257;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 257;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 257;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 257;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 257;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 257;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 257;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 257;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 257;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 257;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 257;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 257;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 257;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st257;
tr3938:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 257;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 257;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 257;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 257;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 257;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 257;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 257;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 257;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 257;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 257;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 257;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 257;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 257;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 257;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 257;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 257;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 257;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 257;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 257;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 257;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 257;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 257;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 257;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 257;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 257;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 257;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 257;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 257;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 257;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 257;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 257;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	goto st257;
st257:
	if ( ++p == pe )
		goto _test_eof257;
case 257:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr822;
		case 778: goto tr825;
		case 800: goto tr822;
		case 808: goto tr823;
		case 809: goto tr824;
		case 827: goto tr826;
		case 1033: goto tr900;
		case 1034: goto tr892;
		case 1056: goto tr900;
		case 1064: goto tr901;
		case 1065: goto tr902;
		case 1083: goto tr903;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr836;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr83;
tr904:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st258;
tr900:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st258;
tr901:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st258;
tr902:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st258;
tr905:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st258;
tr906:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st258;
st258:
	if ( ++p == pe )
		goto _test_eof258;
case 258:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr827;
		case 778: goto tr96;
		case 800: goto tr827;
		case 808: goto tr828;
		case 809: goto tr829;
		case 827: goto tr830;
		case 1033: goto tr904;
		case 1034: goto tr821;
		case 1056: goto tr904;
		case 1064: goto tr905;
		case 1065: goto tr906;
		case 1083: goto tr907;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr836;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr83;
st259:
	if ( ++p == pe )
		goto _test_eof259;
case 259:
	switch( (*p) ) {
		case 65: goto st260;
		case 97: goto st260;
	}
	goto tr38;
st260:
	if ( ++p == pe )
		goto _test_eof260;
case 260:
	switch( (*p) ) {
		case 65: goto st261;
		case 97: goto st261;
	}
	goto tr38;
st261:
	if ( ++p == pe )
		goto _test_eof261;
case 261:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr910;
		case 32: goto tr910;
		case 40: goto tr911;
		case 41: goto tr912;
		case 2058: goto tr913;
		case 2107: goto tr914;
		case 2314: goto tr915;
		case 2363: goto tr915;
		case 2570: goto tr916;
		case 2619: goto tr917;
	}
	goto tr61;
st262:
	if ( ++p == pe )
		goto _test_eof262;
case 262:
	switch( (*p) ) {
		case 83: goto st263;
		case 115: goto st263;
	}
	goto tr38;
st263:
	if ( ++p == pe )
		goto _test_eof263;
case 263:
	switch( (*p) ) {
		case 68: goto st264;
		case 100: goto st264;
	}
	goto tr38;
st264:
	if ( ++p == pe )
		goto _test_eof264;
case 264:
	switch( (*p) ) {
		case 66: goto st265;
		case 98: goto st265;
	}
	goto tr38;
st265:
	if ( ++p == pe )
		goto _test_eof265;
case 265:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr921;
		case 32: goto tr921;
		case 40: goto tr922;
		case 41: goto tr923;
		case 2058: goto tr924;
		case 2107: goto tr925;
		case 2314: goto tr926;
		case 2363: goto tr926;
		case 2570: goto tr927;
		case 2619: goto tr928;
	}
	goto tr61;
st266:
	if ( ++p == pe )
		goto _test_eof266;
case 266:
	switch( (*p) ) {
		case 76: goto st267;
		case 108: goto st267;
	}
	goto tr38;
st267:
	if ( ++p == pe )
		goto _test_eof267;
case 267:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr930;
		case 32: goto tr930;
		case 40: goto tr931;
		case 41: goto tr932;
		case 2058: goto tr933;
		case 2107: goto tr934;
		case 2314: goto tr935;
		case 2363: goto tr935;
		case 2570: goto tr936;
		case 2619: goto tr937;
	}
	goto tr61;
tr140:
	{
		s->buffer_length = 0;
	}
	goto st268;
tr938:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st268;
tr3868:
	{
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st268;
st268:
	if ( ++p == pe )
		goto _test_eof268;
case 268:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr731;
		case 1034: goto tr939;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr938;
	} else if ( _widec >= 640 )
		goto tr730;
	goto tr83;
tr3870:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 269;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 269;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 269;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 269;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 269;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 269;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 269;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 269;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 269;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 269;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 269;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 269;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 269;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 269;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 269;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 269;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 269;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 269;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 269;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 269;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 269;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 269;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 269;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 269;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 269;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 269;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 269;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 269;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 269;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 269;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 269;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st269;
tr142:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 269;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 269;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 269;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 269;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 269;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 269;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 269;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 269;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 269;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 269;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 269;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 269;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 269;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 269;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 269;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 269;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 269;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 269;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 269;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 269;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 269;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 269;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 269;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 269;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 269;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 269;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 269;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 269;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 269;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 269;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 269;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st269;
st269:
	if ( ++p == pe )
		goto _test_eof269;
case 269:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr940;
		case 32: goto tr940;
		case 40: goto tr941;
		case 41: goto tr942;
		case 778: goto tr943;
		case 827: goto tr826;
		case 1034: goto tr943;
		case 1083: goto tr826;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr730;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr730;
	} else
		goto tr730;
	goto tr83;
tr944:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st270;
tr940:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st270;
tr941:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st270;
tr942:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st270;
tr945:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st270;
tr946:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st270;
st270:
	if ( ++p == pe )
		goto _test_eof270;
case 270:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr944;
		case 32: goto tr944;
		case 40: goto tr945;
		case 41: goto tr946;
		case 778: goto tr947;
		case 827: goto tr830;
		case 1034: goto tr947;
		case 1083: goto tr830;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr730;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr730;
	} else
		goto tr730;
	goto tr89;
tr144:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 271;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 271;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 271;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 271;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 271;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 271;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 271;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 271;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 271;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 271;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 271;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 271;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 271;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 271;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 271;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 271;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 271;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 271;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 271;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 271;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 271;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 271;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 271;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 271;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 271;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 271;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 271;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 271;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 271;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 271;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 271;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st271;
tr3872:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 271;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 271;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 271;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 271;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 271;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 271;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 271;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 271;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 271;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 271;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 271;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 271;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 271;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 271;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 271;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 271;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 271;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 271;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 271;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 271;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 271;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 271;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 271;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 271;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 271;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 271;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 271;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 271;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 271;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 271;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 271;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st271;
st271:
	if ( ++p == pe )
		goto _test_eof271;
case 271:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr940;
		case 778: goto tr943;
		case 800: goto tr940;
		case 808: goto tr941;
		case 809: goto tr942;
		case 827: goto tr826;
		case 1033: goto tr948;
		case 1034: goto tr892;
		case 1056: goto tr948;
		case 1064: goto tr949;
		case 1065: goto tr950;
		case 1083: goto tr903;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr938;
	} else if ( _widec >= 640 )
		goto tr730;
	goto tr83;
tr951:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st272;
tr948:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st272;
tr949:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st272;
tr950:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st272;
tr952:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st272;
tr953:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st272;
st272:
	if ( ++p == pe )
		goto _test_eof272;
case 272:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr944;
		case 778: goto tr947;
		case 800: goto tr944;
		case 808: goto tr945;
		case 809: goto tr946;
		case 827: goto tr830;
		case 1033: goto tr951;
		case 1034: goto tr821;
		case 1056: goto tr951;
		case 1064: goto tr952;
		case 1065: goto tr953;
		case 1083: goto tr907;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr938;
	} else if ( _widec >= 640 )
		goto tr730;
	goto tr83;
tr114:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 273;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 273;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 273;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 273;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 273;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 273;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 273;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 273;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 273;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 273;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 273;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 273;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 273;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 273;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 273;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 273;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 273;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 273;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 273;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 273;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 273;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 273;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 273;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 273;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 273;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 273;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 273;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 273;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 273;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 273;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 273;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st273;
tr107:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 273;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 273;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 273;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 273;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 273;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 273;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 273;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 273;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 273;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 273;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 273;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 273;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 273;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 273;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 273;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 273;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 273;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 273;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 273;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 273;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 273;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 273;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 273;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 273;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 273;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 273;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 273;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 273;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 273;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 273;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 273;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st273;
tr809:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 273;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 273;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 273;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 273;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 273;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 273;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 273;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 273;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 273;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 273;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 273;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 273;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 273;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 273;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 273;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 273;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 273;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 273;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 273;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 273;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 273;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 273;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 273;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 273;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 273;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 273;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 273;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 273;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 273;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 273;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 273;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st273;
st273:
	if ( ++p == pe )
		goto _test_eof273;
case 273:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr822;
		case 778: goto tr825;
		case 800: goto tr822;
		case 808: goto tr823;
		case 809: goto tr824;
		case 827: goto tr826;
		case 1033: goto tr954;
		case 1034: goto tr825;
		case 1056: goto tr954;
		case 1064: goto tr955;
		case 1065: goto tr956;
		case 1083: goto tr174;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr115;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr83;
tr957:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st274;
tr954:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st274;
tr955:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st274;
tr956:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st274;
tr958:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st274;
tr959:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st274;
st274:
	if ( ++p == pe )
		goto _test_eof274;
case 274:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr827;
		case 778: goto tr96;
		case 800: goto tr827;
		case 808: goto tr828;
		case 809: goto tr829;
		case 827: goto tr830;
		case 1033: goto tr957;
		case 1034: goto tr96;
		case 1056: goto tr957;
		case 1064: goto tr958;
		case 1065: goto tr959;
		case 1083: goto tr178;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr115;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr83;
tr797:
	{
		s->buffer_length = 0;
	}
	goto st275;
tr960:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st275;
tr790:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st275;
st275:
	if ( ++p == pe )
		goto _test_eof275;
case 275:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr961;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr960;
	goto tr75;
tr798:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 276;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 276;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 276;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 276;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 276;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 276;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 276;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 276;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 276;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 276;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 276;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 276;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 276;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 276;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 276;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 276;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 276;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 276;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 276;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 276;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 276;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 276;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 276;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 276;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 276;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 276;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 276;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 276;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 276;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 276;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 276;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st276;
tr791:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 276;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 276;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 276;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 276;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 276;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 276;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 276;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 276;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 276;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 276;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 276;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 276;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 276;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 276;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 276;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 276;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 276;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 276;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 276;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 276;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 276;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 276;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 276;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 276;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 276;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 276;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 276;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 276;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 276;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 276;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 276;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st276;
st276:
	if ( ++p == pe )
		goto _test_eof276;
case 276:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr962;
		case 32: goto tr962;
		case 40: goto tr963;
		case 41: goto tr964;
		case 58: goto tr73;
		case 65: goto tr122;
		case 67: goto tr123;
		case 68: goto tr124;
		case 69: goto tr125;
		case 72: goto tr126;
		case 73: goto tr127;
		case 75: goto tr128;
		case 76: goto tr129;
		case 77: goto tr130;
		case 78: goto tr131;
		case 79: goto tr132;
		case 80: goto tr133;
		case 82: goto tr134;
		case 83: goto tr135;
		case 84: goto tr136;
		case 85: goto tr137;
		case 90: goto tr138;
		case 92: goto tr78;
		case 97: goto tr122;
		case 99: goto tr123;
		case 100: goto tr124;
		case 101: goto tr125;
		case 104: goto tr126;
		case 105: goto tr127;
		case 107: goto tr128;
		case 108: goto tr129;
		case 109: goto tr130;
		case 110: goto tr131;
		case 111: goto tr132;
		case 112: goto tr133;
		case 114: goto tr134;
		case 115: goto tr135;
		case 116: goto tr136;
		case 117: goto tr137;
		case 122: goto tr138;
		case 1802: goto tr87;
		case 1851: goto tr88;
		case 2058: goto tr889;
		case 2107: goto tr965;
		case 2314: goto tr105;
		case 2363: goto tr106;
		case 2570: goto tr889;
		case 2619: goto tr966;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr73;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr73;
		} else if ( _widec >= 48 )
			goto tr121;
	} else
		goto tr73;
	goto tr118;
tr799:
	{
		s->buffer_length = 0;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 277;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 277;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 277;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 277;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 277;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 277;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 277;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 277;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 277;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 277;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 277;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 277;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 277;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 277;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 277;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 277;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 277;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 277;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 277;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 277;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 277;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 277;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 277;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 277;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 277;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 277;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 277;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 277;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 277;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 277;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 277;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st277;
tr792:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 277;goto st688;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 277;goto st690;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 277;goto st692;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 277;goto st724;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 277;goto st729;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 277;goto st734;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 277;goto st739;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 277;goto st743;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 277;goto st745;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 277;goto st800;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 277;goto st811;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 277;goto st828;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 277;goto st839;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 277;goto st850;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 277;goto st863;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 277;goto st873;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 277;goto st912;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 277;goto st1091;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 277;goto st1094;}
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_OPENPGPKEY:
			{stack[top++] = 277;goto st1105;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 277;goto st1107;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 277;goto st1136;}
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_SMIMEA:
			{stack[top++] = 277;goto st1149;}
		case KNOT_RRTYPE_CSYNC:
			{stack[top++] = 277;goto st1162;}
		case KNOT_RRTYPE_ZONEMD:
			{stack[top++] = 277;goto st1168;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 277;goto st1186;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 277;goto st1181;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 277;goto st1199;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 277;goto st1205;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 277;goto st1211;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 277;goto st1219;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st299;}
		}
	}
	goto st277;
st277:
	if ( ++p == pe )
		goto _test_eof277;
case 277:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr84;
		case 778: goto tr87;
		case 800: goto tr84;
		case 808: goto tr85;
		case 809: goto tr86;
		case 827: goto tr88;
		case 1033: goto tr967;
		case 1034: goto tr892;
		case 1056: goto tr967;
		case 1064: goto tr968;
		case 1065: goto tr969;
		case 1083: goto tr903;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr960;
	goto tr83;
tr970:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st278;
tr967:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st278;
tr968:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st278;
tr969:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st278;
tr971:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st278;
tr972:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st278;
st278:
	if ( ++p == pe )
		goto _test_eof278;
case 278:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr93;
		case 800: goto st7;
		case 808: goto tr91;
		case 809: goto tr92;
		case 827: goto tr94;
		case 1033: goto tr970;
		case 1034: goto tr821;
		case 1056: goto tr970;
		case 1064: goto tr971;
		case 1065: goto tr972;
		case 1083: goto tr907;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr960;
	goto tr83;
tr3878:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 279;goto st301;} }
	goto st279;
tr3899:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 279;goto st301;} }
	goto st279;
st279:
	if ( ++p == pe )
		goto _test_eof279;
case 279:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 83: goto st36;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 115: goto st36;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3879:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 280;goto st301;} }
	goto st280;
tr3900:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 280;goto st301;} }
	goto st280;
st280:
	if ( ++p == pe )
		goto _test_eof280;
case 280:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 72: goto st41;
		case 78: goto st45;
		case 83: goto st53;
		case 104: goto st41;
		case 110: goto st45;
		case 115: goto st53;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3880:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 281;goto st301;} }
	goto st281;
tr3901:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 281;goto st301;} }
	goto st281;
st281:
	if ( ++p == pe )
		goto _test_eof281;
case 281:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 85: goto st55;
		case 117: goto st55;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3881:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 282;goto st301;} }
	goto st282;
tr3902:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 282;goto st301;} }
	goto st282;
st282:
	if ( ++p == pe )
		goto _test_eof282;
case 282:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 73: goto st62;
		case 105: goto st62;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3903:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 283;goto st301;} }
	goto st283;
st283:
	if ( ++p == pe )
		goto _test_eof283;
case 283:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 78: goto st169;
		case 80: goto st70;
		case 110: goto st169;
		case 112: goto st70;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3883:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 284;goto st301;} }
	goto st284;
tr3904:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 284;goto st301;} }
	goto st284;
st284:
	if ( ++p == pe )
		goto _test_eof284;
case 284:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 69: goto st78;
		case 88: goto st80;
		case 101: goto st78;
		case 120: goto st80;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3884:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 285;goto st301;} }
	goto st285;
tr3905:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 285;goto st301;} }
	goto st285;
st285:
	if ( ++p == pe )
		goto _test_eof285;
case 285:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 51: goto st82;
		case 54: goto st84;
		case 79: goto st86;
		case 80: goto st88;
		case 111: goto st86;
		case 112: goto st88;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3885:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 286;goto st301;} }
	goto st286;
tr3906:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 286;goto st301;} }
	goto st286;
st286:
	if ( ++p == pe )
		goto _test_eof286;
case 286:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 73: goto st90;
		case 88: goto st94;
		case 105: goto st90;
		case 120: goto st94;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3886:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 287;goto st301;} }
	goto st287;
tr3907:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 287;goto st301;} }
	goto st287;
st287:
	if ( ++p == pe )
		goto _test_eof287;
case 287:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 65: goto st96;
		case 73: goto st100;
		case 83: goto st102;
		case 97: goto st96;
		case 105: goto st100;
		case 115: goto st102;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3887:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 288;goto st301;} }
	goto st288;
tr3908:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 288;goto st301;} }
	goto st288;
st288:
	if ( ++p == pe )
		goto _test_eof288;
case 288:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 80: goto st112;
		case 112: goto st112;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3888:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 289;goto st301;} }
	goto st289;
tr3909:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 289;goto st301;} }
	goto st289;
st289:
	if ( ++p == pe )
		goto _test_eof289;
case 289:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 84: goto st122;
		case 116: goto st122;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3889:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 290;goto st301;} }
	goto st290;
tr3910:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 290;goto st301;} }
	goto st290;
st290:
	if ( ++p == pe )
		goto _test_eof290;
case 290:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 80: goto st125;
		case 82: goto st126;
		case 84: goto st130;
		case 112: goto st125;
		case 114: goto st126;
		case 116: goto st130;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3890:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 291;goto st301;} }
	goto st291;
tr3911:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 291;goto st301;} }
	goto st291;
st291:
	if ( ++p == pe )
		goto _test_eof291;
case 291:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 77: goto st132;
		case 79: goto st137;
		case 80: goto st139;
		case 82: goto st141;
		case 83: goto st143;
		case 109: goto st132;
		case 111: goto st137;
		case 112: goto st139;
		case 114: goto st141;
		case 115: goto st143;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3891:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 292;goto st301;} }
	goto st292;
tr3912:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 292;goto st301;} }
	goto st292;
st292:
	if ( ++p == pe )
		goto _test_eof292;
case 292:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 76: goto st148;
		case 88: goto st151;
		case 89: goto st153;
		case 108: goto st148;
		case 120: goto st151;
		case 121: goto st153;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3892:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 293;goto st301;} }
	goto st293;
tr3913:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 293;goto st301;} }
	goto st293;
st293:
	if ( ++p == pe )
		goto _test_eof293;
case 293:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 82: goto st158;
		case 114: goto st158;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr3893:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 294;goto st301;} }
	goto st294;
tr3914:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 294;goto st301;} }
	goto st294;
st294:
	if ( ++p == pe )
		goto _test_eof294;
case 294:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 79: goto st161;
		case 111: goto st161;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr25:
	{
		s->buffer_length = 0;
	}
	goto st295;
tr974:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st295;
tr3860:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st295;
tr3928:
	{
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st295;
tr3924:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st295;
st295:
	if ( ++p == pe )
		goto _test_eof295;
case 295:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr731;
		case 1034: goto tr975;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr974;
	} else if ( _widec >= 640 )
		goto tr730;
	goto tr89;
tr674:
	{
		s->buffer_length = 0;
	}
	goto st296;
tr976:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st296;
tr3895:
	{
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st296;
st296:
	if ( ++p == pe )
		goto _test_eof296;
case 296:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr731;
		case 1034: goto tr977;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr976;
	} else if ( _widec >= 640 )
		goto tr730;
	goto tr151;
tr3882:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 297;goto st301;} }
	goto st297;
st297:
	if ( ++p == pe )
		goto _test_eof297;
case 297:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr744;
		case 32: goto tr744;
		case 40: goto tr745;
		case 41: goto tr746;
		case 78: goto st67;
		case 80: goto st70;
		case 110: goto st67;
		case 112: goto st70;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	goto tr973;
tr156:
	{
		s->buffer_length = 0;
	}
	goto st298;
tr978:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st298;
tr150:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st298;
st298:
	if ( ++p == pe )
		goto _test_eof298;
case 298:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr96;
		case 1034: goto tr979;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr978;
	} else if ( _widec >= 640 )
		goto tr95;
	goto tr151;
st299:
	if ( ++p == pe )
		goto _test_eof299;
case 299:
	if ( (*p) == 10 )
		goto tr981;
	goto tr980;
tr980:
	{
		s->buffer_length = 0;
	}
	{
		if ((*p) == '\r') {
			ERR(ZS_DOS_NEWLINE);
		}

		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st300;
tr982:
	{
		if ((*p) == '\r') {
			ERR(ZS_DOS_NEWLINE);
		}

		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st300;
st300:
	if ( ++p == pe )
		goto _test_eof300;
case 300:
	if ( (*p) == 10 )
		goto tr983;
	goto tr982;
tr981:
	cs = 1237;
	{
		s->buffer_length = 0;
	}
	{
		// Terminate the error context string.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error.counter++;

		// Initialize the fcall stack.
		top = 0;

		// Reset the multiline context.
		s->multiline = false;

		s->state = ZS_STATE_ERROR;

		// Execute the error callback.
		if (s->process.automatic) {
			if (s->process.error != NULL) {
				s->process.error(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; goto _out;}
				}
			}

			// Stop the scanner if fatal error.
			if (s->error.fatal) {
				{p++; goto _out;}
			}
			{goto st1227;}
		} else {
			// Return if external processing.
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr983:
	cs = 1237;
	{
		// Terminate the error context string.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error.counter++;

		// Initialize the fcall stack.
		top = 0;

		// Reset the multiline context.
		s->multiline = false;

		s->state = ZS_STATE_ERROR;

		// Execute the error callback.
		if (s->process.automatic) {
			if (s->process.error != NULL) {
				s->process.error(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; goto _out;}
				}
			}

			// Stop the scanner if fatal error.
			if (s->error.fatal) {
				{p++; goto _out;}
			}
			{goto st1227;}
		} else {
			// Return if external processing.
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto _again;
st1237:
	if ( ++p == pe )
		goto _test_eof1237;
case 1237:
	goto st0;
st301:
	if ( ++p == pe )
		goto _test_eof301;
case 301:
	switch( (*p) ) {
		case 42: goto tr985;
		case 46: goto tr986;
		case 64: goto st309;
		case 92: goto tr988;
		case 95: goto tr985;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr985;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr985;
	} else
		goto tr985;
	goto tr984;
tr985:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st302;
tr990:
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st302;
tr994:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st302;
tr1001:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st302;
st302:
	if ( ++p == pe )
		goto _test_eof302;
case 302:
	switch( (*p) ) {
		case 32: goto tr989;
		case 42: goto tr990;
		case 46: goto tr991;
		case 59: goto tr989;
		case 92: goto st304;
		case 95: goto tr990;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr989;
		} else if ( (*p) >= 9 )
			goto tr989;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr990;
		} else if ( (*p) >= 65 )
			goto tr990;
	} else
		goto tr990;
	goto tr984;
tr989:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		// Check for (relative + origin) dname length overflow.
		if (s->dname_tmp_length + s->zone_origin_length <= ZS_MAX_DNAME_LENGTH) {
			memcpy(s->dname + s->dname_tmp_length,
			       s->zone_origin,
			       s->zone_origin_length);

			s->dname_tmp_length += s->zone_origin_length;
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1238;
tr993:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1238;
tr1000:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		// Check for (relative + origin) dname length overflow.
		if (s->dname_tmp_length + s->zone_origin_length <= ZS_MAX_DNAME_LENGTH) {
			memcpy(s->dname + s->dname_tmp_length,
			       s->zone_origin,
			       s->zone_origin_length);

			s->dname_tmp_length += s->zone_origin_length;
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1238;
tr1004:
	{
		// Copy already verified zone origin.
		memcpy(s->dname,
		       s->zone_origin,
		       s->zone_origin_length);

		s->dname_tmp_length = s->zone_origin_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1238;
st1238:
	if ( ++p == pe )
		goto _test_eof1238;
case 1238:
	goto st0;
tr991:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st303;
tr1002:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st303;
st303:
	if ( ++p == pe )
		goto _test_eof303;
case 303:
	switch( (*p) ) {
		case 32: goto tr993;
		case 42: goto tr994;
		case 45: goto tr994;
		case 59: goto tr993;
		case 92: goto tr995;
		case 95: goto tr994;
	}
	if ( (*p) < 47 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr993;
		} else if ( (*p) >= 9 )
			goto tr993;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr994;
		} else if ( (*p) >= 65 )
			goto tr994;
	} else
		goto tr994;
	goto tr984;
tr988:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st304;
tr995:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st304;
tr1003:
	{
		s->dname_tmp_length++;
	}
	goto st304;
st304:
	if ( ++p == pe )
		goto _test_eof304;
case 304:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr996;
	goto tr990;
tr996:
	{
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st305;
st305:
	if ( ++p == pe )
		goto _test_eof305;
case 305:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr998;
	goto tr997;
tr998:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st306;
st306:
	if ( ++p == pe )
		goto _test_eof306;
case 306:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr999;
	goto tr997;
tr999:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st307;
st307:
	if ( ++p == pe )
		goto _test_eof307;
case 307:
	switch( (*p) ) {
		case 32: goto tr1000;
		case 42: goto tr1001;
		case 46: goto tr1002;
		case 59: goto tr1000;
		case 92: goto tr1003;
		case 95: goto tr1001;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr1000;
		} else if ( (*p) >= 9 )
			goto tr1000;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr1001;
		} else if ( (*p) >= 65 )
			goto tr1001;
	} else
		goto tr1001;
	goto tr997;
tr986:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st308;
st308:
	if ( ++p == pe )
		goto _test_eof308;
case 308:
	switch( (*p) ) {
		case 32: goto tr993;
		case 59: goto tr993;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr993;
	} else if ( (*p) >= 9 )
		goto tr993;
	goto tr984;
st309:
	if ( ++p == pe )
		goto _test_eof309;
case 309:
	switch( (*p) ) {
		case 32: goto tr1004;
		case 59: goto tr1004;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1004;
	} else if ( (*p) >= 9 )
		goto tr1004;
	goto tr984;
st310:
	if ( ++p == pe )
		goto _test_eof310;
case 310:
	switch( (*p) ) {
		case 34: goto st316;
		case 92: goto st312;
	}
	if ( (*p) > 58 ) {
		if ( 60 <= (*p) && (*p) <= 126 )
			goto tr1006;
	} else if ( (*p) >= 33 )
		goto tr1006;
	goto tr1005;
tr1006:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st311;
tr1016:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st311;
st311:
	if ( ++p == pe )
		goto _test_eof311;
case 311:
	switch( (*p) ) {
		case 32: goto tr1009;
		case 33: goto tr1006;
		case 59: goto tr1009;
		case 92: goto st312;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1009;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr1006;
		} else if ( (*p) >= 40 )
			goto tr1010;
	} else
		goto tr1006;
	goto tr1005;
tr1009:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1239;
tr1015:
	{
		rdata_tail++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1239;
st1239:
	if ( ++p == pe )
		goto _test_eof1239;
case 1239:
	goto st0;
tr1010:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1240;
tr1017:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1240;
st1240:
	if ( ++p == pe )
		goto _test_eof1240;
case 1240:
	switch( (*p) ) {
		case 32: goto tr1009;
		case 33: goto tr1006;
		case 59: goto tr1009;
		case 92: goto st312;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1009;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr1006;
		} else if ( (*p) >= 40 )
			goto tr1010;
	} else
		goto tr1006;
	goto tr1005;
tr1018:
	{
		rdata_tail++;
	}
	goto st312;
st312:
	if ( ++p == pe )
		goto _test_eof312;
case 312:
	if ( (*p) < 48 ) {
		if ( 32 <= (*p) && (*p) <= 47 )
			goto tr1006;
	} else if ( (*p) > 57 ) {
		if ( 58 <= (*p) && (*p) <= 126 )
			goto tr1006;
	} else
		goto tr1012;
	goto tr1011;
tr1012:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*rdata_tail = 0;
			s->item_length++;
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st313;
st313:
	if ( ++p == pe )
		goto _test_eof313;
case 313:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1013;
	goto tr1011;
tr1013:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st314;
st314:
	if ( ++p == pe )
		goto _test_eof314;
case 314:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1014;
	goto tr1011;
tr1014:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st315;
st315:
	if ( ++p == pe )
		goto _test_eof315;
case 315:
	switch( (*p) ) {
		case 32: goto tr1015;
		case 33: goto tr1016;
		case 59: goto tr1015;
		case 92: goto tr1018;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1015;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr1016;
		} else if ( (*p) >= 40 )
			goto tr1017;
	} else
		goto tr1016;
	goto tr1011;
tr1019:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st316;
tr1026:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st316;
st316:
	if ( ++p == pe )
		goto _test_eof316;
case 316:
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1019;
		case 34: goto st317;
		case 92: goto st318;
		case 522: goto tr1019;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr1019;
	goto tr1005;
tr1027:
	{
		rdata_tail++;
	}
	goto st317;
st317:
	if ( ++p == pe )
		goto _test_eof317;
case 317:
	switch( (*p) ) {
		case 32: goto tr1009;
		case 59: goto tr1009;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1009;
	} else if ( (*p) >= 9 )
		goto tr1009;
	goto tr1022;
tr1028:
	{
		rdata_tail++;
	}
	goto st318;
st318:
	if ( ++p == pe )
		goto _test_eof318;
case 318:
	if ( (*p) < 48 ) {
		if ( 32 <= (*p) && (*p) <= 47 )
			goto tr1019;
	} else if ( (*p) > 57 ) {
		if ( 58 <= (*p) && (*p) <= 126 )
			goto tr1019;
	} else
		goto tr1023;
	goto tr1011;
tr1023:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st299;}
				}
			}

			*rdata_tail = 0;
			s->item_length++;
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st319;
st319:
	if ( ++p == pe )
		goto _test_eof319;
case 319:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1024;
	goto tr1011;
tr1024:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st320;
st320:
	if ( ++p == pe )
		goto _test_eof320;
case 320:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1025;
	goto tr1011;
tr1025:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st321;
st321:
	if ( ++p == pe )
		goto _test_eof321;
case 321:
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1026;
		case 34: goto tr1027;
		case 92: goto tr1028;
		case 522: goto tr1026;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr1026;
	goto tr1011;
st322:
	if ( ++p == pe )
		goto _test_eof322;
case 322:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st323;
		case 32: goto st323;
		case 40: goto tr1031;
		case 41: goto tr1032;
		case 1034: goto tr1033;
		case 1083: goto tr1034;
	}
	goto tr1029;
tr1031:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st323;
tr1032:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st323;
tr1033:
	{
		s->line_counter++;
	}
	goto st323;
tr1072:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 323; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st323;
st323:
	if ( ++p == pe )
		goto _test_eof323;
case 323:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st323;
		case 32: goto st323;
		case 40: goto tr1031;
		case 41: goto tr1032;
		case 1034: goto tr1033;
		case 1083: goto tr1034;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1036;
	goto tr1035;
tr1036:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st324;
tr1041:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st324;
st324:
	if ( ++p == pe )
		goto _test_eof324;
case 324:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1038;
		case 32: goto tr1038;
		case 40: goto tr1039;
		case 41: goto tr1040;
		case 68: goto tr1042;
		case 72: goto tr1043;
		case 77: goto tr1044;
		case 83: goto st327;
		case 87: goto tr1046;
		case 100: goto tr1042;
		case 104: goto tr1043;
		case 109: goto tr1044;
		case 115: goto st327;
		case 119: goto tr1046;
		case 778: goto tr1047;
		case 827: goto tr1048;
		case 1034: goto tr1047;
		case 1083: goto tr1048;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1041;
	goto tr1037;
tr1051:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st325;
tr1052:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st325;
tr1038:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st325;
tr1039:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st325;
tr1040:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st325;
tr1065:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st325;
tr1066:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st325;
tr1067:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st325;
st325:
	if ( ++p == pe )
		goto _test_eof325;
case 325:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st325;
		case 32: goto st325;
		case 40: goto tr1051;
		case 41: goto tr1052;
		case 778: goto tr1053;
		case 827: goto tr1054;
		case 1034: goto tr1053;
		case 1083: goto tr1054;
	}
	goto tr1049;
tr1047:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1241;
tr1053:
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1241;
tr1056:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1241; goto _out;}
			}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1241;
tr1069:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1241;
st1241:
	if ( ++p == pe )
		goto _test_eof1241;
case 1241:
	goto st0;
tr1054:
	{
		s->buffer_length = 0;
	}
	goto st326;
tr1048:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st326;
tr1055:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st326;
tr1070:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st326;
st326:
	if ( ++p == pe )
		goto _test_eof326;
case 326:
	if ( (*p) == 10 )
		goto tr1056;
	goto tr1055;
tr1042:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st327;
tr1043:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st327;
tr1044:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st327;
tr1046:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st327;
st327:
	if ( ++p == pe )
		goto _test_eof327;
case 327:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1038;
		case 32: goto tr1038;
		case 40: goto tr1039;
		case 41: goto tr1040;
		case 778: goto tr1047;
		case 827: goto tr1048;
		case 1034: goto tr1047;
		case 1083: goto tr1048;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1057;
	goto tr1037;
tr1059:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st328;
tr1057:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st328;
tr1068:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st328;
st328:
	if ( ++p == pe )
		goto _test_eof328;
case 328:
	switch( (*p) ) {
		case 68: goto tr1060;
		case 72: goto tr1061;
		case 77: goto tr1062;
		case 83: goto st329;
		case 87: goto tr1064;
		case 100: goto tr1060;
		case 104: goto tr1061;
		case 109: goto tr1062;
		case 115: goto st329;
		case 119: goto tr1064;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1059;
	goto tr1058;
tr1060:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st329;
tr1061:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st329;
tr1062:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st329;
tr1064:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st329;
st329:
	if ( ++p == pe )
		goto _test_eof329;
case 329:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1065;
		case 32: goto tr1065;
		case 40: goto tr1066;
		case 41: goto tr1067;
		case 778: goto tr1069;
		case 827: goto tr1070;
		case 1034: goto tr1069;
		case 1083: goto tr1070;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1068;
	goto tr1037;
tr1034:
	{
		s->buffer_length = 0;
	}
	goto st330;
tr1071:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st330;
st330:
	if ( ++p == pe )
		goto _test_eof330;
case 330:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1072;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1071;
	goto tr1029;
st331:
	if ( ++p == pe )
		goto _test_eof331;
case 331:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st332;
		case 32: goto st332;
		case 40: goto tr1075;
		case 41: goto tr1076;
		case 1034: goto tr1077;
		case 1083: goto tr1078;
	}
	goto tr1073;
tr1075:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st332;
tr1076:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st332;
tr1077:
	{
		s->line_counter++;
	}
	goto st332;
tr1108:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 332; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st332;
st332:
	if ( ++p == pe )
		goto _test_eof332;
case 332:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st332;
		case 32: goto st332;
		case 40: goto tr1075;
		case 41: goto tr1076;
		case 42: goto tr1079;
		case 46: goto tr1080;
		case 92: goto tr1081;
		case 95: goto tr1079;
		case 1034: goto tr1077;
		case 1083: goto tr1078;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1079;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1079;
	} else
		goto tr1079;
	goto tr1073;
tr1082:
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st333;
tr1089:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st333;
tr1104:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st333;
tr1079:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st333;
st333:
	if ( ++p == pe )
		goto _test_eof333;
case 333:
	switch( (*p) ) {
		case 42: goto tr1082;
		case 46: goto tr1083;
		case 92: goto st337;
		case 95: goto tr1082;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1082;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1082;
	} else
		goto tr1082;
	goto tr1073;
tr1083:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st334;
tr1105:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st334;
st334:
	if ( ++p == pe )
		goto _test_eof334;
case 334:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1086;
		case 32: goto tr1086;
		case 40: goto tr1087;
		case 41: goto tr1088;
		case 42: goto tr1089;
		case 45: goto tr1089;
		case 92: goto tr1090;
		case 95: goto tr1089;
		case 778: goto tr1091;
		case 827: goto tr1092;
		case 1034: goto tr1091;
		case 1083: goto tr1092;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1089;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1089;
	} else
		goto tr1089;
	goto tr1085;
tr1094:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st335;
tr1095:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st335;
tr1086:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st335;
tr1087:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st335;
tr1088:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st335;
st335:
	if ( ++p == pe )
		goto _test_eof335;
case 335:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st335;
		case 32: goto st335;
		case 40: goto tr1094;
		case 41: goto tr1095;
		case 778: goto tr1096;
		case 827: goto tr1097;
		case 1034: goto tr1096;
		case 1083: goto tr1097;
	}
	goto tr1085;
tr1091:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1242;
tr1096:
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1242;
tr1099:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1242; goto _out;}
			}
		}
	}
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1242;
st1242:
	if ( ++p == pe )
		goto _test_eof1242;
case 1242:
	goto st0;
tr1097:
	{
		s->buffer_length = 0;
	}
	goto st336;
tr1098:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st336;
tr1092:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st336;
st336:
	if ( ++p == pe )
		goto _test_eof336;
case 336:
	if ( (*p) == 10 )
		goto tr1099;
	goto tr1098;
tr1090:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st337;
tr1106:
	{
		s->dname_tmp_length++;
	}
	goto st337;
tr1081:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st337;
st337:
	if ( ++p == pe )
		goto _test_eof337;
case 337:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1100;
	goto tr1082;
tr1100:
	{
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st338;
st338:
	if ( ++p == pe )
		goto _test_eof338;
case 338:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1102;
	goto tr1101;
tr1102:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st339;
st339:
	if ( ++p == pe )
		goto _test_eof339;
case 339:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1103;
	goto tr1101;
tr1103:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st340;
st340:
	if ( ++p == pe )
		goto _test_eof340;
case 340:
	switch( (*p) ) {
		case 42: goto tr1104;
		case 46: goto tr1105;
		case 92: goto tr1106;
		case 95: goto tr1104;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1104;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1104;
	} else
		goto tr1104;
	goto tr1101;
tr1080:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st341;
st341:
	if ( ++p == pe )
		goto _test_eof341;
case 341:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1086;
		case 32: goto tr1086;
		case 40: goto tr1087;
		case 41: goto tr1088;
		case 778: goto tr1091;
		case 827: goto tr1092;
		case 1034: goto tr1091;
		case 1083: goto tr1092;
	}
	goto tr1085;
tr1078:
	{
		s->buffer_length = 0;
	}
	goto st342;
tr1107:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st342;
st342:
	if ( ++p == pe )
		goto _test_eof342;
case 342:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1108;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1107;
	goto tr1073;
st343:
	if ( ++p == pe )
		goto _test_eof343;
case 343:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st344;
		case 32: goto st344;
		case 40: goto tr1110;
		case 41: goto tr1111;
		case 1034: goto tr1112;
		case 1083: goto tr1113;
	}
	goto st0;
tr1110:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st344;
tr1111:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st344;
tr1112:
	{
		s->line_counter++;
	}
	goto st344;
tr1166:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 344; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st344;
st344:
	if ( ++p == pe )
		goto _test_eof344;
case 344:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st344;
		case 32: goto st344;
		case 40: goto tr1110;
		case 41: goto tr1111;
		case 1034: goto tr1112;
		case 1083: goto tr1113;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1114;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1114;
	} else
		goto tr1114;
	goto tr1115;
tr1114:
	{
		rdata_tail = s->r_data;
	}
	{ p--; {stack[top++] = 345;goto st310;} }
	goto st345;
st345:
	if ( ++p == pe )
		goto _test_eof345;
case 345:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1117;
		case 32: goto tr1117;
		case 40: goto tr1118;
		case 41: goto tr1119;
		case 778: goto tr1120;
		case 827: goto tr1121;
		case 1034: goto tr1120;
		case 1083: goto tr1122;
	}
	goto tr1116;
tr1125:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st346;
tr1126:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st346;
tr1117:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st299;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st346;
tr1118:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st299;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st346;
tr1119:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st299;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st346;
st346:
	if ( ++p == pe )
		goto _test_eof346;
case 346:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st346;
		case 32: goto st346;
		case 40: goto tr1125;
		case 41: goto tr1126;
		case 42: goto tr1127;
		case 46: goto tr1128;
		case 92: goto tr1129;
		case 95: goto tr1127;
		case 778: goto tr1130;
		case 827: goto tr1131;
		case 1034: goto tr1130;
		case 1083: goto tr1132;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1127;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1127;
	} else
		goto tr1127;
	goto tr1123;
tr1134:
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st347;
tr1140:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st347;
tr1153:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st347;
tr1127:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st347;
st347:
	if ( ++p == pe )
		goto _test_eof347;
case 347:
	switch( (*p) ) {
		case 42: goto tr1134;
		case 46: goto tr1135;
		case 92: goto st351;
		case 95: goto tr1134;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1134;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1134;
	} else
		goto tr1134;
	goto tr1133;
tr1135:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st348;
tr1154:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < ZS_MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st348;
st348:
	if ( ++p == pe )
		goto _test_eof348;
case 348:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1137;
		case 32: goto tr1137;
		case 40: goto tr1138;
		case 41: goto tr1139;
		case 42: goto tr1140;
		case 45: goto tr1140;
		case 92: goto tr1141;
		case 95: goto tr1140;
		case 778: goto tr1142;
		case 827: goto tr1143;
		case 1034: goto tr1142;
		case 1083: goto tr1143;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1140;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1140;
	} else
		goto tr1140;
	goto tr1123;
tr1145:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st349;
tr1146:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st349;
tr1137:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st349;
tr1138:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st349;
tr1139:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st349;
st349:
	if ( ++p == pe )
		goto _test_eof349;
case 349:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st349;
		case 32: goto st349;
		case 40: goto tr1145;
		case 41: goto tr1146;
		case 778: goto tr1130;
		case 827: goto tr1131;
		case 1034: goto tr1130;
		case 1083: goto tr1131;
	}
	goto tr89;
tr1120:
	cs = 1243;
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st299;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			int ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			                   "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret >= sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st299;}
			}
			memcpy(s->include_filename, s->buffer, ret + 1);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st299;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st299;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr1130:
	cs = 1243;
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			int ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			                   "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret >= sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st299;}
			}
			memcpy(s->include_filename, s->buffer, ret + 1);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st299;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st299;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr1142:
	cs = 1243;
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			int ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			                   "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret >= sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st299;}
			}
			memcpy(s->include_filename, s->buffer, ret + 1);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st299;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st299;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr1148:
	cs = 1243;
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; goto _out;}
			}
		}
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			int ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			                   "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret >= sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st299;}
			}
			memcpy(s->include_filename, s->buffer, ret + 1);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st299;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st299;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
st1243:
	if ( ++p == pe )
		goto _test_eof1243;
case 1243:
	goto st0;
tr1131:
	{
		s->buffer_length = 0;
	}
	goto st350;
tr1147:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st350;
tr1121:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st299;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st350;
tr1143:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer_length = 0;
	}
	goto st350;
st350:
	if ( ++p == pe )
		goto _test_eof350;
case 350:
	if ( (*p) == 10 )
		goto tr1148;
	goto tr1147;
tr1141:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st351;
tr1155:
	{
		s->dname_tmp_length++;
	}
	goto st351;
tr1129:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st351;
st351:
	if ( ++p == pe )
		goto _test_eof351;
case 351:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1149;
	goto tr1134;
tr1149:
	{
		if (s->item_length < ZS_MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st352;
st352:
	if ( ++p == pe )
		goto _test_eof352;
case 352:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1151;
	goto tr1150;
tr1151:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st353;
st353:
	if ( ++p == pe )
		goto _test_eof353;
case 353:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1152;
	goto tr1150;
tr1152:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st354;
st354:
	if ( ++p == pe )
		goto _test_eof354;
case 354:
	switch( (*p) ) {
		case 42: goto tr1153;
		case 46: goto tr1154;
		case 92: goto tr1155;
		case 95: goto tr1153;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1153;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1153;
	} else
		goto tr1153;
	goto tr1150;
tr1128:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st355;
st355:
	if ( ++p == pe )
		goto _test_eof355;
case 355:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1137;
		case 32: goto tr1137;
		case 40: goto tr1138;
		case 41: goto tr1139;
		case 778: goto tr1142;
		case 827: goto tr1143;
		case 1034: goto tr1142;
		case 1083: goto tr1143;
	}
	goto tr1123;
tr1132:
	{
		s->buffer_length = 0;
	}
	goto st356;
tr1156:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st356;
tr1122:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st299;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer_length = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st356;
st356:
	if ( ++p == pe )
		goto _test_eof356;
case 356:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr1148;
		case 1034: goto tr1157;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr1156;
	} else if ( _widec >= 640 )
		goto tr1147;
	goto tr89;
tr1157:
	cs = 1244;
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			int ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			                   "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret >= sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st299;}
			}
			memcpy(s->include_filename, s->buffer, ret + 1);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st299;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st299;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1227; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto _again;
st1244:
	if ( ++p == pe )
		goto _test_eof1244;
case 1244:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st357;
		case 32: goto st357;
		case 40: goto tr1159;
		case 41: goto tr1160;
		case 42: goto tr1127;
		case 46: goto tr1128;
		case 92: goto tr1129;
		case 95: goto tr1127;
		case 1034: goto tr1161;
		case 1083: goto tr1162;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1127;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1127;
	} else
		goto tr1127;
	goto tr1133;
tr1159:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st357;
tr1160:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st357;
tr1161:
	{
		s->line_counter++;
	}
	goto st357;
tr1164:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 357; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st357;
st357:
	if ( ++p == pe )
		goto _test_eof357;
case 357:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st357;
		case 32: goto st357;
		case 40: goto tr1159;
		case 41: goto tr1160;
		case 42: goto tr1127;
		case 46: goto tr1128;
		case 92: goto tr1129;
		case 95: goto tr1127;
		case 1034: goto tr1161;
		case 1083: goto tr1162;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1127;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1127;
	} else
		goto tr1127;
	goto tr1133;
tr1162:
	{
		s->buffer_length = 0;
	}
	goto st358;
tr1163:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st358;
st358:
	if ( ++p == pe )
		goto _test_eof358;
case 358:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1164;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1163;
	goto st0;
tr1113:
	{
		s->buffer_length = 0;
	}
	goto st359;
tr1165:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st359;
st359:
	if ( ++p == pe )
		goto _test_eof359;
case 359:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1166;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1165;
	goto st0;
st360:
	if ( ++p == pe )
		goto _test_eof360;
case 360:
	if ( (*p) == 43 )
		goto tr1168;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1168;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1168;
	} else
		goto tr1168;
	goto tr1167;
tr1168:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st361;
st361:
	if ( ++p == pe )
		goto _test_eof361;
case 361:
	if ( (*p) == 43 )
		goto tr1169;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1169;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1169;
	} else
		goto tr1169;
	goto tr1167;
tr1169:
	{
		*(rdata_tail++) += second_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st362;
st362:
	if ( ++p == pe )
		goto _test_eof362;
case 362:
	switch( (*p) ) {
		case 43: goto tr1170;
		case 61: goto st366;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1170;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1170;
	} else
		goto tr1170;
	goto tr1167;
tr1170:
	{
		*(rdata_tail++) += third_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = third_right_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st363;
st363:
	if ( ++p == pe )
		goto _test_eof363;
case 363:
	switch( (*p) ) {
		case 43: goto tr1172;
		case 61: goto st364;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1172;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1172;
	} else
		goto tr1172;
	goto tr1167;
tr1174:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st364;
tr1175:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st364;
tr1176:
	{
		s->line_counter++;
	}
	goto st364;
tr1182:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 364; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st364;
tr1172:
	{
		*(rdata_tail++) += fourth_base64_to_num[(uint8_t)(*p)];
	}
	goto st364;
st364:
	if ( ++p == pe )
		goto _test_eof364;
case 364:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st364;
		case 32: goto st364;
		case 40: goto tr1174;
		case 41: goto tr1175;
		case 43: goto tr1168;
		case 2058: goto tr1176;
		case 2107: goto tr1177;
		case 2314: goto tr1178;
		case 2363: goto tr1178;
		case 2570: goto tr1179;
		case 2619: goto tr1180;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1168;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1168;
	} else
		goto tr1168;
	goto tr1167;
tr1177:
	{
		s->buffer_length = 0;
	}
	goto st365;
tr1181:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st365;
st365:
	if ( ++p == pe )
		goto _test_eof365;
case 365:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1182;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1181;
	goto tr1167;
tr1178:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1245;
st1245:
	if ( ++p == pe )
		goto _test_eof1245;
case 1245:
	goto st0;
tr1179:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1246;
st1246:
	if ( ++p == pe )
		goto _test_eof1246;
case 1246:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st364;
		case 32: goto st364;
		case 40: goto tr1174;
		case 41: goto tr1175;
		case 43: goto tr1168;
		case 2058: goto tr1176;
		case 2107: goto tr1177;
		case 2314: goto tr1178;
		case 2363: goto tr1178;
		case 2570: goto tr1179;
		case 2619: goto tr1180;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1168;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1168;
	} else
		goto tr1168;
	goto tr1167;
tr1180:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1247;
st1247:
	if ( ++p == pe )
		goto _test_eof1247;
case 1247:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1182;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1181;
	goto tr1167;
st366:
	if ( ++p == pe )
		goto _test_eof366;
case 366:
	if ( (*p) == 61 )
		goto st364;
	goto tr1167;
st367:
	if ( ++p == pe )
		goto _test_eof367;
case 367:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1184;
		case 32: goto tr1184;
		case 40: goto tr1185;
		case 41: goto tr1186;
		case 2058: goto tr1187;
		case 2107: goto tr1188;
		case 2314: goto tr1189;
		case 2363: goto tr1189;
		case 2570: goto tr1190;
		case 2619: goto tr1191;
	}
	goto tr1183;
tr1193:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1194:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1212:
	{
		s->line_counter++;
	}
	goto st368;
tr1239:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 368; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st368;
tr1184:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st368;
tr1185:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1186:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1187:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->line_counter++;
	}
	goto st368;
tr1217:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st368;
tr1218:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1219:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1223:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1230:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st368;
tr1231:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1232:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1233:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1246:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	goto st368;
tr1247:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1248:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1249:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1260:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	goto st368;
tr1261:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1262:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1263:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1268:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	goto st368;
tr1269:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1270:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1271:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1278:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st368;
tr1279:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1280:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1281:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1289:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st368;
tr1290:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1291:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1292:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1300:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	goto st368;
tr1301:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1302:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1303:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1314:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st368;
tr1315:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1316:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1317:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1326:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st368;
tr1327:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1328:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1329:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1337:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st368;
tr1338:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1339:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1340:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1345:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st368;
tr1346:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1347:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1348:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1358:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st368;
tr1359:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1360:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1361:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1367:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st368;
tr1368:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1369:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1370:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1379:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st368;
tr1380:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1381:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1382:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1394:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st368;
tr1395:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1396:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1397:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1405:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st368;
tr1406:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1407:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1408:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1413:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st368;
tr1414:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1415:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1416:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1426:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st368;
tr1427:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1428:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1429:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1435:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st368;
tr1436:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1437:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1438:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1444:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st368;
tr1445:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1446:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1447:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1452:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st368;
tr1453:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1454:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1455:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1465:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st368;
tr1466:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1467:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1468:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1473:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st368;
tr1474:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1475:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1476:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1487:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st368;
tr1488:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1489:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1490:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1496:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st368;
tr1497:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1498:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1499:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1504:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st368;
tr1505:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1506:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1508:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1514:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st368;
tr1515:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1516:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1518:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1523:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st368;
tr1524:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1525:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1527:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1536:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st368;
tr1537:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1538:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1539:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1553:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	goto st368;
tr1554:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1555:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1556:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1563:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st368;
tr1564:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1565:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1566:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1574:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st368;
tr1575:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1576:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1577:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1585:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st368;
tr1586:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1587:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1588:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1593:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st368;
tr1594:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1595:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1596:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1610:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	goto st368;
tr1611:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1612:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1613:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1619:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st368;
tr1620:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1621:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1622:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1628:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st368;
tr1629:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1630:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1631:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1637:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st368;
tr1638:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1639:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1640:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1648:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st368;
tr1649:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1650:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1651:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1661:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st368;
tr1662:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1663:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1664:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1670:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st368;
tr1671:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1672:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1673:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1681:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st368;
tr1682:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1683:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1685:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st368;
tr1692:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	goto st368;
tr1693:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1694:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1695:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1705:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	goto st368;
tr1706:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1707:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1708:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1716:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st368;
tr1717:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1718:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1719:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->line_counter++;
	}
	goto st368;
tr1725:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st368;
tr1726:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st368;
tr1727:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st368;
tr1728:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->line_counter++;
	}
	goto st368;
st368:
	if ( ++p == pe )
		goto _test_eof368;
case 368:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st368;
		case 32: goto st368;
		case 40: goto tr1193;
		case 41: goto tr1194;
		case 65: goto st369;
		case 67: goto st374;
		case 68: goto st395;
		case 69: goto st409;
		case 72: goto st416;
		case 73: goto st421;
		case 75: goto st429;
		case 76: goto st433;
		case 77: goto st441;
		case 78: goto st447;
		case 79: goto st463;
		case 80: goto st473;
		case 82: goto st476;
		case 83: goto st483;
		case 84: goto st499;
		case 85: goto st509;
		case 90: goto st512;
		case 97: goto st369;
		case 99: goto st374;
		case 100: goto st395;
		case 101: goto st409;
		case 104: goto st416;
		case 105: goto st421;
		case 107: goto st429;
		case 108: goto st433;
		case 109: goto st441;
		case 110: goto st447;
		case 111: goto st463;
		case 112: goto st473;
		case 114: goto st476;
		case 115: goto st483;
		case 116: goto st499;
		case 117: goto st509;
		case 122: goto st512;
		case 2058: goto tr1212;
		case 2107: goto tr1213;
		case 2314: goto tr1214;
		case 2363: goto tr1214;
		case 2570: goto tr1215;
		case 2619: goto tr1216;
	}
	goto tr1183;
st369:
	if ( ++p == pe )
		goto _test_eof369;
case 369:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1217;
		case 32: goto tr1217;
		case 40: goto tr1218;
		case 41: goto tr1219;
		case 65: goto st370;
		case 70: goto st518;
		case 80: goto st522;
		case 97: goto st370;
		case 102: goto st518;
		case 112: goto st522;
		case 2058: goto tr1223;
		case 2107: goto tr1224;
		case 2314: goto tr1225;
		case 2363: goto tr1225;
		case 2570: goto tr1226;
		case 2619: goto tr1227;
	}
	goto tr1183;
st370:
	if ( ++p == pe )
		goto _test_eof370;
case 370:
	switch( (*p) ) {
		case 65: goto st371;
		case 97: goto st371;
	}
	goto tr1183;
st371:
	if ( ++p == pe )
		goto _test_eof371;
case 371:
	switch( (*p) ) {
		case 65: goto st372;
		case 97: goto st372;
	}
	goto tr1183;
st372:
	if ( ++p == pe )
		goto _test_eof372;
case 372:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1230;
		case 32: goto tr1230;
		case 40: goto tr1231;
		case 41: goto tr1232;
		case 2058: goto tr1233;
		case 2107: goto tr1234;
		case 2314: goto tr1235;
		case 2363: goto tr1235;
		case 2570: goto tr1236;
		case 2619: goto tr1237;
	}
	goto tr1183;
tr1213:
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1238:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st373;
tr1188:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1224:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1234:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1250:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1264:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1272:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1282:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1293:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1304:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1318:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1330:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1341:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1349:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1362:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1371:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1383:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1398:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1409:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1417:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1430:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1439:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1448:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1456:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1469:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1477:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1491:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1500:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1509:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1519:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1528:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1540:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1557:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1567:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1578:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1589:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1597:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1614:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1623:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1632:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1641:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1652:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1665:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1674:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1686:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1696:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1709:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1720:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
tr1729:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->buffer_length = 0;
	}
	goto st373;
st373:
	if ( ++p == pe )
		goto _test_eof373;
case 373:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1239;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1238;
	goto tr1183;
tr1189:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1214:
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1225:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1235:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1251:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1265:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1273:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1283:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1294:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1305:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1319:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1331:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1342:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1350:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1363:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1372:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1384:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1399:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1410:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1418:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1431:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1440:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1449:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1457:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1470:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1478:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1492:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1501:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1510:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1520:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1529:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1541:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1558:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1568:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1579:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1590:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1598:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1615:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1624:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1633:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1642:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1653:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1666:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1675:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1687:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1697:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1710:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1721:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
tr1730:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1248;
st1248:
	if ( ++p == pe )
		goto _test_eof1248;
case 1248:
	goto st0;
tr1190:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1215:
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1226:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1236:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1252:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1266:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1274:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1284:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1295:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1306:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1320:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1332:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1343:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1351:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1364:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1373:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1385:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1400:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1411:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1419:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1432:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1441:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1450:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1458:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1471:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1479:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1493:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1502:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1511:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1521:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1530:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1542:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1559:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1569:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1580:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1591:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1599:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1616:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1625:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1634:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1643:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1654:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1667:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1676:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1688:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1698:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1711:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1722:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
tr1731:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->line_counter++;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1249;
st1249:
	if ( ++p == pe )
		goto _test_eof1249;
case 1249:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st368;
		case 32: goto st368;
		case 40: goto tr1193;
		case 41: goto tr1194;
		case 65: goto st369;
		case 67: goto st374;
		case 68: goto st395;
		case 69: goto st409;
		case 72: goto st416;
		case 73: goto st421;
		case 75: goto st429;
		case 76: goto st433;
		case 77: goto st441;
		case 78: goto st447;
		case 79: goto st463;
		case 80: goto st473;
		case 82: goto st476;
		case 83: goto st483;
		case 84: goto st499;
		case 85: goto st509;
		case 90: goto st512;
		case 97: goto st369;
		case 99: goto st374;
		case 100: goto st395;
		case 101: goto st409;
		case 104: goto st416;
		case 105: goto st421;
		case 107: goto st429;
		case 108: goto st433;
		case 109: goto st441;
		case 110: goto st447;
		case 111: goto st463;
		case 112: goto st473;
		case 114: goto st476;
		case 115: goto st483;
		case 116: goto st499;
		case 117: goto st509;
		case 122: goto st512;
		case 2058: goto tr1212;
		case 2107: goto tr1213;
		case 2314: goto tr1214;
		case 2363: goto tr1214;
		case 2570: goto tr1215;
		case 2619: goto tr1216;
	}
	goto tr1183;
st374:
	if ( ++p == pe )
		goto _test_eof374;
case 374:
	switch( (*p) ) {
		case 65: goto st375;
		case 68: goto st377;
		case 69: goto st384;
		case 78: goto st387;
		case 83: goto st391;
		case 97: goto st375;
		case 100: goto st377;
		case 101: goto st384;
		case 110: goto st387;
		case 115: goto st391;
	}
	goto tr1183;
st375:
	if ( ++p == pe )
		goto _test_eof375;
case 375:
	switch( (*p) ) {
		case 65: goto st376;
		case 97: goto st376;
	}
	goto tr1183;
st376:
	if ( ++p == pe )
		goto _test_eof376;
case 376:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1246;
		case 32: goto tr1246;
		case 40: goto tr1247;
		case 41: goto tr1248;
		case 2058: goto tr1249;
		case 2107: goto tr1250;
		case 2314: goto tr1251;
		case 2363: goto tr1251;
		case 2570: goto tr1252;
		case 2619: goto tr1253;
	}
	goto tr1183;
tr1191:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1216:
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1227:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1237:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1253:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1267:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1275:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1285:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1296:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1307:
	{ window_add_bit(KNOT_RRTYPE_CSYNC, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1321:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1333:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1344:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1352:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1365:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1374:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1386:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1401:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1412:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1420:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1433:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1442:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1451:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1459:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1472:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1480:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1494:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1503:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1512:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1522:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1531:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1543:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1560:
	{ window_add_bit(KNOT_RRTYPE_OPENPGPKEY, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1570:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1581:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1592:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1600:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1617:
	{ window_add_bit(KNOT_RRTYPE_SMIMEA, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1626:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1635:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1644:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1655:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1668:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1677:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1689:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1699:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1712:
	{ window_add_bit(KNOT_RRTYPE_ZONEMD, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1723:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
tr1732:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->buffer_length = 0;
	}
	{
		for (uint16_t window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st299;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1250;
st1250:
	if ( ++p == pe )
		goto _test_eof1250;
case 1250:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1239;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1238;
	goto tr1183;
st377:
	if ( ++p == pe )
		goto _test_eof377;
case 377:
	switch( (*p) ) {
		case 78: goto st378;
		case 83: goto st383;
		case 110: goto st378;
		case 115: goto st383;
	}
	goto tr1183;
st378:
	if ( ++p == pe )
		goto _test_eof378;
case 378:
	switch( (*p) ) {
		case 83: goto st379;
		case 115: goto st379;
	}
	goto tr1183;
st379:
	if ( ++p == pe )
		goto _test_eof379;
case 379:
	switch( (*p) ) {
		case 75: goto st380;
		case 107: goto st380;
	}
	goto tr1183;
st380:
	if ( ++p == pe )
		goto _test_eof380;
case 380:
	switch( (*p) ) {
		case 69: goto st381;
		case 101: goto st381;
	}
	goto tr1183;
st381:
	if ( ++p == pe )
		goto _test_eof381;
case 381:
	switch( (*p) ) {
		case 89: goto st382;
		case 121: goto st382;
	}
	goto tr1183;
st382:
	if ( ++p == pe )
		goto _test_eof382;
case 382:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1260;
		case 32: goto tr1260;
		case 40: goto tr1261;
		case 41: goto tr1262;
		case 2058: goto tr1263;
		case 2107: goto tr1264;
		case 2314: goto tr1265;
		case 2363: goto tr1265;
		case 2570: goto tr1266;
		case 2619: goto tr1267;
	}
	goto tr1183;
st383:
	if ( ++p == pe )
		goto _test_eof383;
case 383:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1268;
		case 32: goto tr1268;
		case 40: goto tr1269;
		case 41: goto tr1270;
		case 2058: goto tr1271;
		case 2107: goto tr1272;
		case 2314: goto tr1273;
		case 2363: goto tr1273;
		case 2570: goto tr1274;
		case 2619: goto tr1275;
	}
	goto tr1183;
st384:
	if ( ++p == pe )
		goto _test_eof384;
case 384:
	switch( (*p) ) {
		case 82: goto st385;
		case 114: goto st385;
	}
	goto tr1183;
st385:
	if ( ++p == pe )
		goto _test_eof385;
case 385:
	switch( (*p) ) {
		case 84: goto st386;
		case 116: goto st386;
	}
	goto tr1183;
st386:
	if ( ++p == pe )
		goto _test_eof386;
case 386:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1278;
		case 32: goto tr1278;
		case 40: goto tr1279;
		case 41: goto tr1280;
		case 2058: goto tr1281;
		case 2107: goto tr1282;
		case 2314: goto tr1283;
		case 2363: goto tr1283;
		case 2570: goto tr1284;
		case 2619: goto tr1285;
	}
	goto tr1183;
st387:
	if ( ++p == pe )
		goto _test_eof387;
case 387:
	switch( (*p) ) {
		case 65: goto st388;
		case 97: goto st388;
	}
	goto tr1183;
st388:
	if ( ++p == pe )
		goto _test_eof388;
case 388:
	switch( (*p) ) {
		case 77: goto st389;
		case 109: goto st389;
	}
	goto tr1183;
st389:
	if ( ++p == pe )
		goto _test_eof389;
case 389:
	switch( (*p) ) {
		case 69: goto st390;
		case 101: goto st390;
	}
	goto tr1183;
st390:
	if ( ++p == pe )
		goto _test_eof390;
case 390:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1289;
		case 32: goto tr1289;
		case 40: goto tr1290;
		case 41: goto tr1291;
		case 2058: goto tr1292;
		case 2107: goto tr1293;
		case 2314: goto tr1294;
		case 2363: goto tr1294;
		case 2570: goto tr1295;
		case 2619: goto tr1296;
	}
	goto tr1183;
st391:
	if ( ++p == pe )
		goto _test_eof391;
case 391:
	switch( (*p) ) {
		case 89: goto st392;
		case 121: goto st392;
	}
	goto tr1183;
st392:
	if ( ++p == pe )
		goto _test_eof392;
case 392:
	switch( (*p) ) {
		case 78: goto st393;
		case 110: goto st393;
	}
	goto tr1183;
st393:
	if ( ++p == pe )
		goto _test_eof393;
case 393:
	switch( (*p) ) {
		case 67: goto st394;
		case 99: goto st394;
	}
	goto tr1183;
st394:
	if ( ++p == pe )
		goto _test_eof394;
case 394:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1300;
		case 32: goto tr1300;
		case 40: goto tr1301;
		case 41: goto tr1302;
		case 2058: goto tr1303;
		case 2107: goto tr1304;
		case 2314: goto tr1305;
		case 2363: goto tr1305;
		case 2570: goto tr1306;
		case 2619: goto tr1307;
	}
	goto tr1183;
st395:
	if ( ++p == pe )
		goto _test_eof395;
case 395:
	switch( (*p) ) {
		case 72: goto st396;
		case 78: goto st400;
		case 83: goto st408;
		case 104: goto st396;
		case 110: goto st400;
		case 115: goto st408;
	}
	goto tr1183;
st396:
	if ( ++p == pe )
		goto _test_eof396;
case 396:
	switch( (*p) ) {
		case 67: goto st397;
		case 99: goto st397;
	}
	goto tr1183;
st397:
	if ( ++p == pe )
		goto _test_eof397;
case 397:
	switch( (*p) ) {
		case 73: goto st398;
		case 105: goto st398;
	}
	goto tr1183;
st398:
	if ( ++p == pe )
		goto _test_eof398;
case 398:
	switch( (*p) ) {
		case 68: goto st399;
		case 100: goto st399;
	}
	goto tr1183;
st399:
	if ( ++p == pe )
		goto _test_eof399;
case 399:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1314;
		case 32: goto tr1314;
		case 40: goto tr1315;
		case 41: goto tr1316;
		case 2058: goto tr1317;
		case 2107: goto tr1318;
		case 2314: goto tr1319;
		case 2363: goto tr1319;
		case 2570: goto tr1320;
		case 2619: goto tr1321;
	}
	goto tr1183;
st400:
	if ( ++p == pe )
		goto _test_eof400;
case 400:
	switch( (*p) ) {
		case 65: goto st401;
		case 83: goto st404;
		case 97: goto st401;
		case 115: goto st404;
	}
	goto tr1183;
st401:
	if ( ++p == pe )
		goto _test_eof401;
case 401:
	switch( (*p) ) {
		case 77: goto st402;
		case 109: goto st402;
	}
	goto tr1183;
st402:
	if ( ++p == pe )
		goto _test_eof402;
case 402:
	switch( (*p) ) {
		case 69: goto st403;
		case 101: goto st403;
	}
	goto tr1183;
st403:
	if ( ++p == pe )
		goto _test_eof403;
case 403:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1326;
		case 32: goto tr1326;
		case 40: goto tr1327;
		case 41: goto tr1328;
		case 2058: goto tr1329;
		case 2107: goto tr1330;
		case 2314: goto tr1331;
		case 2363: goto tr1331;
		case 2570: goto tr1332;
		case 2619: goto tr1333;
	}
	goto tr1183;
st404:
	if ( ++p == pe )
		goto _test_eof404;
case 404:
	switch( (*p) ) {
		case 75: goto st405;
		case 107: goto st405;
	}
	goto tr1183;
st405:
	if ( ++p == pe )
		goto _test_eof405;
case 405:
	switch( (*p) ) {
		case 69: goto st406;
		case 101: goto st406;
	}
	goto tr1183;
st406:
	if ( ++p == pe )
		goto _test_eof406;
case 406:
	switch( (*p) ) {
		case 89: goto st407;
		case 121: goto st407;
	}
	goto tr1183;
st407:
	if ( ++p == pe )
		goto _test_eof407;
case 407:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1337;
		case 32: goto tr1337;
		case 40: goto tr1338;
		case 41: goto tr1339;
		case 2058: goto tr1340;
		case 2107: goto tr1341;
		case 2314: goto tr1342;
		case 2363: goto tr1342;
		case 2570: goto tr1343;
		case 2619: goto tr1344;
	}
	goto tr1183;
st408:
	if ( ++p == pe )
		goto _test_eof408;
case 408:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1345;
		case 32: goto tr1345;
		case 40: goto tr1346;
		case 41: goto tr1347;
		case 2058: goto tr1348;
		case 2107: goto tr1349;
		case 2314: goto tr1350;
		case 2363: goto tr1350;
		case 2570: goto tr1351;
		case 2619: goto tr1352;
	}
	goto tr1183;
st409:
	if ( ++p == pe )
		goto _test_eof409;
case 409:
	switch( (*p) ) {
		case 85: goto st410;
		case 117: goto st410;
	}
	goto tr1183;
st410:
	if ( ++p == pe )
		goto _test_eof410;
case 410:
	switch( (*p) ) {
		case 73: goto st411;
		case 105: goto st411;
	}
	goto tr1183;
st411:
	if ( ++p == pe )
		goto _test_eof411;
case 411:
	switch( (*p) ) {
		case 52: goto st412;
		case 54: goto st414;
	}
	goto tr1183;
st412:
	if ( ++p == pe )
		goto _test_eof412;
case 412:
	if ( (*p) == 56 )
		goto st413;
	goto tr1183;
st413:
	if ( ++p == pe )
		goto _test_eof413;
case 413:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1358;
		case 32: goto tr1358;
		case 40: goto tr1359;
		case 41: goto tr1360;
		case 2058: goto tr1361;
		case 2107: goto tr1362;
		case 2314: goto tr1363;
		case 2363: goto tr1363;
		case 2570: goto tr1364;
		case 2619: goto tr1365;
	}
	goto tr1183;
st414:
	if ( ++p == pe )
		goto _test_eof414;
case 414:
	if ( (*p) == 52 )
		goto st415;
	goto tr1183;
st415:
	if ( ++p == pe )
		goto _test_eof415;
case 415:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1367;
		case 32: goto tr1367;
		case 40: goto tr1368;
		case 41: goto tr1369;
		case 2058: goto tr1370;
		case 2107: goto tr1371;
		case 2314: goto tr1372;
		case 2363: goto tr1372;
		case 2570: goto tr1373;
		case 2619: goto tr1374;
	}
	goto tr1183;
st416:
	if ( ++p == pe )
		goto _test_eof416;
case 416:
	switch( (*p) ) {
		case 73: goto st417;
		case 105: goto st417;
	}
	goto tr1183;
st417:
	if ( ++p == pe )
		goto _test_eof417;
case 417:
	switch( (*p) ) {
		case 78: goto st418;
		case 110: goto st418;
	}
	goto tr1183;
st418:
	if ( ++p == pe )
		goto _test_eof418;
case 418:
	switch( (*p) ) {
		case 70: goto st419;
		case 102: goto st419;
	}
	goto tr1183;
st419:
	if ( ++p == pe )
		goto _test_eof419;
case 419:
	switch( (*p) ) {
		case 79: goto st420;
		case 111: goto st420;
	}
	goto tr1183;
st420:
	if ( ++p == pe )
		goto _test_eof420;
case 420:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1379;
		case 32: goto tr1379;
		case 40: goto tr1380;
		case 41: goto tr1381;
		case 2058: goto tr1382;
		case 2107: goto tr1383;
		case 2314: goto tr1384;
		case 2363: goto tr1384;
		case 2570: goto tr1385;
		case 2619: goto tr1386;
	}
	goto tr1183;
st421:
	if ( ++p == pe )
		goto _test_eof421;
case 421:
	switch( (*p) ) {
		case 80: goto st422;
		case 112: goto st422;
	}
	goto tr1183;
st422:
	if ( ++p == pe )
		goto _test_eof422;
case 422:
	switch( (*p) ) {
		case 83: goto st423;
		case 115: goto st423;
	}
	goto tr1183;
st423:
	if ( ++p == pe )
		goto _test_eof423;
case 423:
	switch( (*p) ) {
		case 69: goto st424;
		case 101: goto st424;
	}
	goto tr1183;
st424:
	if ( ++p == pe )
		goto _test_eof424;
case 424:
	switch( (*p) ) {
		case 67: goto st425;
		case 99: goto st425;
	}
	goto tr1183;
st425:
	if ( ++p == pe )
		goto _test_eof425;
case 425:
	switch( (*p) ) {
		case 75: goto st426;
		case 107: goto st426;
	}
	goto tr1183;
st426:
	if ( ++p == pe )
		goto _test_eof426;
case 426:
	switch( (*p) ) {
		case 69: goto st427;
		case 101: goto st427;
	}
	goto tr1183;
st427:
	if ( ++p == pe )
		goto _test_eof427;
case 427:
	switch( (*p) ) {
		case 89: goto st428;
		case 121: goto st428;
	}
	goto tr1183;
st428:
	if ( ++p == pe )
		goto _test_eof428;
case 428:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1394;
		case 32: goto tr1394;
		case 40: goto tr1395;
		case 41: goto tr1396;
		case 2058: goto tr1397;
		case 2107: goto tr1398;
		case 2314: goto tr1399;
		case 2363: goto tr1399;
		case 2570: goto tr1400;
		case 2619: goto tr1401;
	}
	goto tr1183;
st429:
	if ( ++p == pe )
		goto _test_eof429;
case 429:
	switch( (*p) ) {
		case 69: goto st430;
		case 88: goto st432;
		case 101: goto st430;
		case 120: goto st432;
	}
	goto tr1183;
st430:
	if ( ++p == pe )
		goto _test_eof430;
case 430:
	switch( (*p) ) {
		case 89: goto st431;
		case 121: goto st431;
	}
	goto tr1183;
st431:
	if ( ++p == pe )
		goto _test_eof431;
case 431:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1405;
		case 32: goto tr1405;
		case 40: goto tr1406;
		case 41: goto tr1407;
		case 2058: goto tr1408;
		case 2107: goto tr1409;
		case 2314: goto tr1410;
		case 2363: goto tr1410;
		case 2570: goto tr1411;
		case 2619: goto tr1412;
	}
	goto tr1183;
st432:
	if ( ++p == pe )
		goto _test_eof432;
case 432:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1413;
		case 32: goto tr1413;
		case 40: goto tr1414;
		case 41: goto tr1415;
		case 2058: goto tr1416;
		case 2107: goto tr1417;
		case 2314: goto tr1418;
		case 2363: goto tr1418;
		case 2570: goto tr1419;
		case 2619: goto tr1420;
	}
	goto tr1183;
st433:
	if ( ++p == pe )
		goto _test_eof433;
case 433:
	switch( (*p) ) {
		case 51: goto st434;
		case 54: goto st436;
		case 79: goto st438;
		case 80: goto st440;
		case 111: goto st438;
		case 112: goto st440;
	}
	goto tr1183;
st434:
	if ( ++p == pe )
		goto _test_eof434;
case 434:
	if ( (*p) == 50 )
		goto st435;
	goto tr1183;
st435:
	if ( ++p == pe )
		goto _test_eof435;
case 435:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1426;
		case 32: goto tr1426;
		case 40: goto tr1427;
		case 41: goto tr1428;
		case 2058: goto tr1429;
		case 2107: goto tr1430;
		case 2314: goto tr1431;
		case 2363: goto tr1431;
		case 2570: goto tr1432;
		case 2619: goto tr1433;
	}
	goto tr1183;
st436:
	if ( ++p == pe )
		goto _test_eof436;
case 436:
	if ( (*p) == 52 )
		goto st437;
	goto tr1183;
st437:
	if ( ++p == pe )
		goto _test_eof437;
case 437:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1435;
		case 32: goto tr1435;
		case 40: goto tr1436;
		case 41: goto tr1437;
		case 2058: goto tr1438;
		case 2107: goto tr1439;
		case 2314: goto tr1440;
		case 2363: goto tr1440;
		case 2570: goto tr1441;
		case 2619: goto tr1442;
	}
	goto tr1183;
st438:
	if ( ++p == pe )
		goto _test_eof438;
case 438:
	switch( (*p) ) {
		case 67: goto st439;
		case 99: goto st439;
	}
	goto tr1183;
st439:
	if ( ++p == pe )
		goto _test_eof439;
case 439:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1444;
		case 32: goto tr1444;
		case 40: goto tr1445;
		case 41: goto tr1446;
		case 2058: goto tr1447;
		case 2107: goto tr1448;
		case 2314: goto tr1449;
		case 2363: goto tr1449;
		case 2570: goto tr1450;
		case 2619: goto tr1451;
	}
	goto tr1183;
st440:
	if ( ++p == pe )
		goto _test_eof440;
case 440:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1452;
		case 32: goto tr1452;
		case 40: goto tr1453;
		case 41: goto tr1454;
		case 2058: goto tr1455;
		case 2107: goto tr1456;
		case 2314: goto tr1457;
		case 2363: goto tr1457;
		case 2570: goto tr1458;
		case 2619: goto tr1459;
	}
	goto tr1183;
st441:
	if ( ++p == pe )
		goto _test_eof441;
case 441:
	switch( (*p) ) {
		case 73: goto st442;
		case 88: goto st446;
		case 105: goto st442;
		case 120: goto st446;
	}
	goto tr1183;
st442:
	if ( ++p == pe )
		goto _test_eof442;
case 442:
	switch( (*p) ) {
		case 78: goto st443;
		case 110: goto st443;
	}
	goto tr1183;
st443:
	if ( ++p == pe )
		goto _test_eof443;
case 443:
	switch( (*p) ) {
		case 70: goto st444;
		case 102: goto st444;
	}
	goto tr1183;
st444:
	if ( ++p == pe )
		goto _test_eof444;
case 444:
	switch( (*p) ) {
		case 79: goto st445;
		case 111: goto st445;
	}
	goto tr1183;
st445:
	if ( ++p == pe )
		goto _test_eof445;
case 445:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1465;
		case 32: goto tr1465;
		case 40: goto tr1466;
		case 41: goto tr1467;
		case 2058: goto tr1468;
		case 2107: goto tr1469;
		case 2314: goto tr1470;
		case 2363: goto tr1470;
		case 2570: goto tr1471;
		case 2619: goto tr1472;
	}
	goto tr1183;
st446:
	if ( ++p == pe )
		goto _test_eof446;
case 446:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1473;
		case 32: goto tr1473;
		case 40: goto tr1474;
		case 41: goto tr1475;
		case 2058: goto tr1476;
		case 2107: goto tr1477;
		case 2314: goto tr1478;
		case 2363: goto tr1478;
		case 2570: goto tr1479;
		case 2619: goto tr1480;
	}
	goto tr1183;
st447:
	if ( ++p == pe )
		goto _test_eof447;
case 447:
	switch( (*p) ) {
		case 65: goto st448;
		case 73: goto st452;
		case 83: goto st454;
		case 97: goto st448;
		case 105: goto st452;
		case 115: goto st454;
	}
	goto tr1183;
st448:
	if ( ++p == pe )
		goto _test_eof448;
case 448:
	switch( (*p) ) {
		case 80: goto st449;
		case 112: goto st449;
	}
	goto tr1183;
st449:
	if ( ++p == pe )
		goto _test_eof449;
case 449:
	switch( (*p) ) {
		case 84: goto st450;
		case 116: goto st450;
	}
	goto tr1183;
st450:
	if ( ++p == pe )
		goto _test_eof450;
case 450:
	switch( (*p) ) {
		case 82: goto st451;
		case 114: goto st451;
	}
	goto tr1183;
st451:
	if ( ++p == pe )
		goto _test_eof451;
case 451:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1487;
		case 32: goto tr1487;
		case 40: goto tr1488;
		case 41: goto tr1489;
		case 2058: goto tr1490;
		case 2107: goto tr1491;
		case 2314: goto tr1492;
		case 2363: goto tr1492;
		case 2570: goto tr1493;
		case 2619: goto tr1494;
	}
	goto tr1183;
st452:
	if ( ++p == pe )
		goto _test_eof452;
case 452:
	switch( (*p) ) {
		case 68: goto st453;
		case 100: goto st453;
	}
	goto tr1183;
st453:
	if ( ++p == pe )
		goto _test_eof453;
case 453:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1496;
		case 32: goto tr1496;
		case 40: goto tr1497;
		case 41: goto tr1498;
		case 2058: goto tr1499;
		case 2107: goto tr1500;
		case 2314: goto tr1501;
		case 2363: goto tr1501;
		case 2570: goto tr1502;
		case 2619: goto tr1503;
	}
	goto tr1183;
st454:
	if ( ++p == pe )
		goto _test_eof454;
case 454:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1504;
		case 32: goto tr1504;
		case 40: goto tr1505;
		case 41: goto tr1506;
		case 69: goto st455;
		case 101: goto st455;
		case 2058: goto tr1508;
		case 2107: goto tr1509;
		case 2314: goto tr1510;
		case 2363: goto tr1510;
		case 2570: goto tr1511;
		case 2619: goto tr1512;
	}
	goto tr1183;
st455:
	if ( ++p == pe )
		goto _test_eof455;
case 455:
	switch( (*p) ) {
		case 67: goto st456;
		case 99: goto st456;
	}
	goto tr1183;
st456:
	if ( ++p == pe )
		goto _test_eof456;
case 456:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1514;
		case 32: goto tr1514;
		case 40: goto tr1515;
		case 41: goto tr1516;
		case 51: goto st457;
		case 2058: goto tr1518;
		case 2107: goto tr1519;
		case 2314: goto tr1520;
		case 2363: goto tr1520;
		case 2570: goto tr1521;
		case 2619: goto tr1522;
	}
	goto tr1183;
st457:
	if ( ++p == pe )
		goto _test_eof457;
case 457:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1523;
		case 32: goto tr1523;
		case 40: goto tr1524;
		case 41: goto tr1525;
		case 80: goto st458;
		case 112: goto st458;
		case 2058: goto tr1527;
		case 2107: goto tr1528;
		case 2314: goto tr1529;
		case 2363: goto tr1529;
		case 2570: goto tr1530;
		case 2619: goto tr1531;
	}
	goto tr1183;
st458:
	if ( ++p == pe )
		goto _test_eof458;
case 458:
	switch( (*p) ) {
		case 65: goto st459;
		case 97: goto st459;
	}
	goto tr1183;
st459:
	if ( ++p == pe )
		goto _test_eof459;
case 459:
	switch( (*p) ) {
		case 82: goto st460;
		case 114: goto st460;
	}
	goto tr1183;
st460:
	if ( ++p == pe )
		goto _test_eof460;
case 460:
	switch( (*p) ) {
		case 65: goto st461;
		case 97: goto st461;
	}
	goto tr1183;
st461:
	if ( ++p == pe )
		goto _test_eof461;
case 461:
	switch( (*p) ) {
		case 77: goto st462;
		case 109: goto st462;
	}
	goto tr1183;
st462:
	if ( ++p == pe )
		goto _test_eof462;
case 462:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1536;
		case 32: goto tr1536;
		case 40: goto tr1537;
		case 41: goto tr1538;
		case 2058: goto tr1539;
		case 2107: goto tr1540;
		case 2314: goto tr1541;
		case 2363: goto tr1541;
		case 2570: goto tr1542;
		case 2619: goto tr1543;
	}
	goto tr1183;
st463:
	if ( ++p == pe )
		goto _test_eof463;
case 463:
	switch( (*p) ) {
		case 80: goto st464;
		case 112: goto st464;
	}
	goto tr1183;
st464:
	if ( ++p == pe )
		goto _test_eof464;
case 464:
	switch( (*p) ) {
		case 69: goto st465;
		case 101: goto st465;
	}
	goto tr1183;
st465:
	if ( ++p == pe )
		goto _test_eof465;
case 465:
	switch( (*p) ) {
		case 78: goto st466;
		case 110: goto st466;
	}
	goto tr1183;
st466:
	if ( ++p == pe )
		goto _test_eof466;
case 466:
	switch( (*p) ) {
		case 80: goto st467;
		case 112: goto st467;
	}
	goto tr1183;
st467:
	if ( ++p == pe )
		goto _test_eof467;
case 467:
	switch( (*p) ) {
		case 71: goto st468;
		case 103: goto st468;
	}
	goto tr1183;
st468:
	if ( ++p == pe )
		goto _test_eof468;
case 468:
	switch( (*p) ) {
		case 80: goto st469;
		case 112: goto st469;
	}
	goto tr1183;
st469:
	if ( ++p == pe )
		goto _test_eof469;
case 469:
	switch( (*p) ) {
		case 75: goto st470;
		case 107: goto st470;
	}
	goto tr1183;
st470:
	if ( ++p == pe )
		goto _test_eof470;
case 470:
	switch( (*p) ) {
		case 69: goto st471;
		case 101: goto st471;
	}
	goto tr1183;
st471:
	if ( ++p == pe )
		goto _test_eof471;
case 471:
	switch( (*p) ) {
		case 89: goto st472;
		case 121: goto st472;
	}
	goto tr1183;
st472:
	if ( ++p == pe )
		goto _test_eof472;
case 472:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1553;
		case 32: goto tr1553;
		case 40: goto tr1554;
		case 41: goto tr1555;
		case 2058: goto tr1556;
		case 2107: goto tr1557;
		case 2314: goto tr1558;
		case 2363: goto tr1558;
		case 2570: goto tr1559;
		case 2619: goto tr1560;
	}
	goto tr1183;
st473:
	if ( ++p == pe )
		goto _test_eof473;
case 473:
	switch( (*p) ) {
		case 84: goto st474;
		case 116: goto st474;
	}
	goto tr1183;
st474:
	if ( ++p == pe )
		goto _test_eof474;
case 474:
	switch( (*p) ) {
		case 82: goto st475;
		case 114: goto st475;
	}
	goto tr1183;
st475:
	if ( ++p == pe )
		goto _test_eof475;
case 475:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1563;
		case 32: goto tr1563;
		case 40: goto tr1564;
		case 41: goto tr1565;
		case 2058: goto tr1566;
		case 2107: goto tr1567;
		case 2314: goto tr1568;
		case 2363: goto tr1568;
		case 2570: goto tr1569;
		case 2619: goto tr1570;
	}
	goto tr1183;
st476:
	if ( ++p == pe )
		goto _test_eof476;
case 476:
	switch( (*p) ) {
		case 80: goto st477;
		case 82: goto st478;
		case 84: goto st482;
		case 112: goto st477;
		case 114: goto st478;
		case 116: goto st482;
	}
	goto tr1183;
st477:
	if ( ++p == pe )
		goto _test_eof477;
case 477:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1574;
		case 32: goto tr1574;
		case 40: goto tr1575;
		case 41: goto tr1576;
		case 2058: goto tr1577;
		case 2107: goto tr1578;
		case 2314: goto tr1579;
		case 2363: goto tr1579;
		case 2570: goto tr1580;
		case 2619: goto tr1581;
	}
	goto tr1183;
st478:
	if ( ++p == pe )
		goto _test_eof478;
case 478:
	switch( (*p) ) {
		case 83: goto st479;
		case 115: goto st479;
	}
	goto tr1183;
st479:
	if ( ++p == pe )
		goto _test_eof479;
case 479:
	switch( (*p) ) {
		case 73: goto st480;
		case 105: goto st480;
	}
	goto tr1183;
st480:
	if ( ++p == pe )
		goto _test_eof480;
case 480:
	switch( (*p) ) {
		case 71: goto st481;
		case 103: goto st481;
	}
	goto tr1183;
st481:
	if ( ++p == pe )
		goto _test_eof481;
case 481:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1585;
		case 32: goto tr1585;
		case 40: goto tr1586;
		case 41: goto tr1587;
		case 2058: goto tr1588;
		case 2107: goto tr1589;
		case 2314: goto tr1590;
		case 2363: goto tr1590;
		case 2570: goto tr1591;
		case 2619: goto tr1592;
	}
	goto tr1183;
st482:
	if ( ++p == pe )
		goto _test_eof482;
case 482:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1593;
		case 32: goto tr1593;
		case 40: goto tr1594;
		case 41: goto tr1595;
		case 2058: goto tr1596;
		case 2107: goto tr1597;
		case 2314: goto tr1598;
		case 2363: goto tr1598;
		case 2570: goto tr1599;
		case 2619: goto tr1600;
	}
	goto tr1183;
st483:
	if ( ++p == pe )
		goto _test_eof483;
case 483:
	switch( (*p) ) {
		case 77: goto st484;
		case 79: goto st489;
		case 80: goto st491;
		case 82: goto st493;
		case 83: goto st495;
		case 109: goto st484;
		case 111: goto st489;
		case 112: goto st491;
		case 114: goto st493;
		case 115: goto st495;
	}
	goto tr1183;
st484:
	if ( ++p == pe )
		goto _test_eof484;
case 484:
	switch( (*p) ) {
		case 73: goto st485;
		case 105: goto st485;
	}
	goto tr1183;
st485:
	if ( ++p == pe )
		goto _test_eof485;
case 485:
	switch( (*p) ) {
		case 77: goto st486;
		case 109: goto st486;
	}
	goto tr1183;
st486:
	if ( ++p == pe )
		goto _test_eof486;
case 486:
	switch( (*p) ) {
		case 69: goto st487;
		case 101: goto st487;
	}
	goto tr1183;
st487:
	if ( ++p == pe )
		goto _test_eof487;
case 487:
	switch( (*p) ) {
		case 65: goto st488;
		case 97: goto st488;
	}
	goto tr1183;
st488:
	if ( ++p == pe )
		goto _test_eof488;
case 488:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1610;
		case 32: goto tr1610;
		case 40: goto tr1611;
		case 41: goto tr1612;
		case 2058: goto tr1613;
		case 2107: goto tr1614;
		case 2314: goto tr1615;
		case 2363: goto tr1615;
		case 2570: goto tr1616;
		case 2619: goto tr1617;
	}
	goto tr1183;
st489:
	if ( ++p == pe )
		goto _test_eof489;
case 489:
	switch( (*p) ) {
		case 65: goto st490;
		case 97: goto st490;
	}
	goto tr1183;
st490:
	if ( ++p == pe )
		goto _test_eof490;
case 490:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1619;
		case 32: goto tr1619;
		case 40: goto tr1620;
		case 41: goto tr1621;
		case 2058: goto tr1622;
		case 2107: goto tr1623;
		case 2314: goto tr1624;
		case 2363: goto tr1624;
		case 2570: goto tr1625;
		case 2619: goto tr1626;
	}
	goto tr1183;
st491:
	if ( ++p == pe )
		goto _test_eof491;
case 491:
	switch( (*p) ) {
		case 70: goto st492;
		case 102: goto st492;
	}
	goto tr1183;
st492:
	if ( ++p == pe )
		goto _test_eof492;
case 492:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1628;
		case 32: goto tr1628;
		case 40: goto tr1629;
		case 41: goto tr1630;
		case 2058: goto tr1631;
		case 2107: goto tr1632;
		case 2314: goto tr1633;
		case 2363: goto tr1633;
		case 2570: goto tr1634;
		case 2619: goto tr1635;
	}
	goto tr1183;
st493:
	if ( ++p == pe )
		goto _test_eof493;
case 493:
	switch( (*p) ) {
		case 86: goto st494;
		case 118: goto st494;
	}
	goto tr1183;
st494:
	if ( ++p == pe )
		goto _test_eof494;
case 494:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1637;
		case 32: goto tr1637;
		case 40: goto tr1638;
		case 41: goto tr1639;
		case 2058: goto tr1640;
		case 2107: goto tr1641;
		case 2314: goto tr1642;
		case 2363: goto tr1642;
		case 2570: goto tr1643;
		case 2619: goto tr1644;
	}
	goto tr1183;
st495:
	if ( ++p == pe )
		goto _test_eof495;
case 495:
	switch( (*p) ) {
		case 72: goto st496;
		case 104: goto st496;
	}
	goto tr1183;
st496:
	if ( ++p == pe )
		goto _test_eof496;
case 496:
	switch( (*p) ) {
		case 70: goto st497;
		case 102: goto st497;
	}
	goto tr1183;
st497:
	if ( ++p == pe )
		goto _test_eof497;
case 497:
	switch( (*p) ) {
		case 80: goto st498;
		case 112: goto st498;
	}
	goto tr1183;
st498:
	if ( ++p == pe )
		goto _test_eof498;
case 498:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1648;
		case 32: goto tr1648;
		case 40: goto tr1649;
		case 41: goto tr1650;
		case 2058: goto tr1651;
		case 2107: goto tr1652;
		case 2314: goto tr1653;
		case 2363: goto tr1653;
		case 2570: goto tr1654;
		case 2619: goto tr1655;
	}
	goto tr1183;
st499:
	if ( ++p == pe )
		goto _test_eof499;
case 499:
	switch( (*p) ) {
		case 76: goto st500;
		case 88: goto st503;
		case 89: goto st505;
		case 108: goto st500;
		case 120: goto st503;
		case 121: goto st505;
	}
	goto tr1183;
st500:
	if ( ++p == pe )
		goto _test_eof500;
case 500:
	switch( (*p) ) {
		case 83: goto st501;
		case 115: goto st501;
	}
	goto tr1183;
st501:
	if ( ++p == pe )
		goto _test_eof501;
case 501:
	switch( (*p) ) {
		case 65: goto st502;
		case 97: goto st502;
	}
	goto tr1183;
st502:
	if ( ++p == pe )
		goto _test_eof502;
case 502:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1661;
		case 32: goto tr1661;
		case 40: goto tr1662;
		case 41: goto tr1663;
		case 2058: goto tr1664;
		case 2107: goto tr1665;
		case 2314: goto tr1666;
		case 2363: goto tr1666;
		case 2570: goto tr1667;
		case 2619: goto tr1668;
	}
	goto tr1183;
st503:
	if ( ++p == pe )
		goto _test_eof503;
case 503:
	switch( (*p) ) {
		case 84: goto st504;
		case 116: goto st504;
	}
	goto tr1183;
st504:
	if ( ++p == pe )
		goto _test_eof504;
case 504:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1670;
		case 32: goto tr1670;
		case 40: goto tr1671;
		case 41: goto tr1672;
		case 2058: goto tr1673;
		case 2107: goto tr1674;
		case 2314: goto tr1675;
		case 2363: goto tr1675;
		case 2570: goto tr1676;
		case 2619: goto tr1677;
	}
	goto tr1183;
st505:
	if ( ++p == pe )
		goto _test_eof505;
case 505:
	switch( (*p) ) {
		case 80: goto st506;
		case 112: goto st506;
	}
	goto tr1183;
st506:
	if ( ++p == pe )
		goto _test_eof506;
case 506:
	switch( (*p) ) {
		case 69: goto st507;
		case 101: goto st507;
	}
	goto tr1183;
st507:
	if ( ++p == pe )
		goto _test_eof507;
case 507:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1680;
	goto tr1183;
tr1680:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st508;
tr1684:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st508;
st508:
	if ( ++p == pe )
		goto _test_eof508;
case 508:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1681;
		case 32: goto tr1681;
		case 40: goto tr1682;
		case 41: goto tr1683;
		case 2058: goto tr1685;
		case 2107: goto tr1686;
		case 2314: goto tr1687;
		case 2363: goto tr1687;
		case 2570: goto tr1688;
		case 2619: goto tr1689;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1684;
	goto tr1183;
st509:
	if ( ++p == pe )
		goto _test_eof509;
case 509:
	switch( (*p) ) {
		case 82: goto st510;
		case 114: goto st510;
	}
	goto tr1183;
st510:
	if ( ++p == pe )
		goto _test_eof510;
case 510:
	switch( (*p) ) {
		case 73: goto st511;
		case 105: goto st511;
	}
	goto tr1183;
st511:
	if ( ++p == pe )
		goto _test_eof511;
case 511:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1692;
		case 32: goto tr1692;
		case 40: goto tr1693;
		case 41: goto tr1694;
		case 2058: goto tr1695;
		case 2107: goto tr1696;
		case 2314: goto tr1697;
		case 2363: goto tr1697;
		case 2570: goto tr1698;
		case 2619: goto tr1699;
	}
	goto tr1183;
st512:
	if ( ++p == pe )
		goto _test_eof512;
case 512:
	switch( (*p) ) {
		case 79: goto st513;
		case 111: goto st513;
	}
	goto tr1183;
st513:
	if ( ++p == pe )
		goto _test_eof513;
case 513:
	switch( (*p) ) {
		case 78: goto st514;
		case 110: goto st514;
	}
	goto tr1183;
st514:
	if ( ++p == pe )
		goto _test_eof514;
case 514:
	switch( (*p) ) {
		case 69: goto st515;
		case 101: goto st515;
	}
	goto tr1183;
st515:
	if ( ++p == pe )
		goto _test_eof515;
case 515:
	switch( (*p) ) {
		case 77: goto st516;
		case 109: goto st516;
	}
	goto tr1183;
st516:
	if ( ++p == pe )
		goto _test_eof516;
case 516:
	switch( (*p) ) {
		case 68: goto st517;
		case 100: goto st517;
	}
	goto tr1183;
st517:
	if ( ++p == pe )
		goto _test_eof517;
case 517:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1705;
		case 32: goto tr1705;
		case 40: goto tr1706;
		case 41: goto tr1707;
		case 2058: goto tr1708;
		case 2107: goto tr1709;
		case 2314: goto tr1710;
		case 2363: goto tr1710;
		case 2570: goto tr1711;
		case 2619: goto tr1712;
	}
	goto tr1183;
st518:
	if ( ++p == pe )
		goto _test_eof518;
case 518:
	switch( (*p) ) {
		case 83: goto st519;
		case 115: goto st519;
	}
	goto tr1183;
st519:
	if ( ++p == pe )
		goto _test_eof519;
case 519:
	switch( (*p) ) {
		case 68: goto st520;
		case 100: goto st520;
	}
	goto tr1183;
st520:
	if ( ++p == pe )
		goto _test_eof520;
case 520:
	switch( (*p) ) {
		case 66: goto st521;
		case 98: goto st521;
	}
	goto tr1183;
st521:
	if ( ++p == pe )
		goto _test_eof521;
case 521:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1716;
		case 32: goto tr1716;
		case 40: goto tr1717;
		case 41: goto tr1718;
		case 2058: goto tr1719;
		case 2107: goto tr1720;
		case 2314: goto tr1721;
		case 2363: goto tr1721;
		case 2570: goto tr1722;
		case 2619: goto tr1723;
	}
	goto tr1183;
st522:
	if ( ++p == pe )
		goto _test_eof522;
case 522:
	switch( (*p) ) {
		case 76: goto st523;
		case 108: goto st523;
	}
	goto tr1183;
st523:
	if ( ++p == pe )
		goto _test_eof523;
case 523:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1725;
		case 32: goto tr1725;
		case 40: goto tr1726;
		case 41: goto tr1727;
		case 2058: goto tr1728;
		case 2107: goto tr1729;
		case 2314: goto tr1730;
		case 2363: goto tr1730;
		case 2570: goto tr1731;
		case 2619: goto tr1732;
	}
	goto tr1183;
st524:
	if ( ++p == pe )
		goto _test_eof524;
case 524:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st525;
		case 32: goto st525;
		case 40: goto tr1735;
		case 41: goto tr1736;
		case 1034: goto tr1737;
		case 1083: goto tr1738;
	}
	goto tr1733;
tr1735:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st525;
tr1736:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st525;
tr1737:
	{
		s->line_counter++;
	}
	goto st525;
tr1768:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 525; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st525;
st525:
	if ( ++p == pe )
		goto _test_eof525;
case 525:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st525;
		case 32: goto st525;
		case 40: goto tr1735;
		case 41: goto tr1736;
		case 1034: goto tr1737;
		case 1083: goto tr1738;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1740;
	goto tr1739;
tr1740:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st526;
tr1744:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st526;
st526:
	if ( ++p == pe )
		goto _test_eof526;
case 526:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1741;
		case 32: goto tr1741;
		case 40: goto tr1742;
		case 41: goto tr1743;
		case 1034: goto tr1745;
		case 1083: goto tr1746;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1744;
	goto tr1739;
tr1749:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st527;
tr1750:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st527;
tr1752:
	{
		s->line_counter++;
	}
	goto st527;
tr1766:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 527; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st527;
tr1741:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st527;
tr1742:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st527;
tr1743:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st527;
tr1745:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st527;
st527:
	if ( ++p == pe )
		goto _test_eof527;
case 527:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st527;
		case 32: goto st527;
		case 40: goto tr1749;
		case 41: goto tr1750;
		case 1034: goto tr1752;
		case 1083: goto tr1753;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1751;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1751;
	} else
		goto tr1751;
	goto tr1747;
tr1751:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st528;
st528:
	if ( ++p == pe )
		goto _test_eof528;
case 528:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1754;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1754;
	} else
		goto tr1754;
	goto tr1747;
tr1756:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st529;
tr1757:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st529;
tr1758:
	{
		s->line_counter++;
	}
	goto st529;
tr1764:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 529; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st529;
tr1754:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st529;
st529:
	if ( ++p == pe )
		goto _test_eof529;
case 529:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st529;
		case 32: goto st529;
		case 40: goto tr1756;
		case 41: goto tr1757;
		case 2058: goto tr1758;
		case 2107: goto tr1759;
		case 2314: goto tr1760;
		case 2363: goto tr1760;
		case 2570: goto tr1761;
		case 2619: goto tr1762;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1751;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1751;
	} else
		goto tr1751;
	goto tr1747;
tr1759:
	{
		s->buffer_length = 0;
	}
	goto st530;
tr1763:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st530;
st530:
	if ( ++p == pe )
		goto _test_eof530;
case 530:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1764;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1763;
	goto tr1747;
tr1760:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1251;
st1251:
	if ( ++p == pe )
		goto _test_eof1251;
case 1251:
	goto st0;
tr1761:
	{
		s->line_counter++;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1252;
st1252:
	if ( ++p == pe )
		goto _test_eof1252;
case 1252:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st529;
		case 32: goto st529;
		case 40: goto tr1756;
		case 41: goto tr1757;
		case 2058: goto tr1758;
		case 2107: goto tr1759;
		case 2314: goto tr1760;
		case 2363: goto tr1760;
		case 2570: goto tr1761;
		case 2619: goto tr1762;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1751;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1751;
	} else
		goto tr1751;
	goto tr1747;
tr1762:
	{
		s->buffer_length = 0;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1253;
st1253:
	if ( ++p == pe )
		goto _test_eof1253;
case 1253:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1764;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1763;
	goto tr1747;
tr1753:
	{
		s->buffer_length = 0;
	}
	goto st531;
tr1765:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st531;
tr1746:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st531;
st531:
	if ( ++p == pe )
		goto _test_eof531;
case 531:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1766;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1765;
	goto tr1733;
tr1738:
	{
		s->buffer_length = 0;
	}
	goto st532;
tr1767:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st532;
st532:
	if ( ++p == pe )
		goto _test_eof532;
case 532:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1768;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1767;
	goto tr1733;
st533:
	if ( ++p == pe )
		goto _test_eof533;
case 533:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st534;
		case 32: goto st534;
		case 40: goto tr1770;
		case 41: goto tr1771;
		case 1034: goto tr1772;
		case 1083: goto tr1773;
	}
	goto tr1733;
tr1770:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st534;
tr1771:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st534;
tr1772:
	{
		s->line_counter++;
	}
	goto st534;
tr1808:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 534; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st534;
st534:
	if ( ++p == pe )
		goto _test_eof534;
case 534:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st534;
		case 32: goto st534;
		case 40: goto tr1770;
		case 41: goto tr1771;
		case 48: goto tr1774;
		case 1034: goto tr1772;
		case 1083: goto tr1773;
	}
	if ( 49 <= _widec && _widec <= 57 )
		goto tr1775;
	goto tr1739;
tr1774:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st535;
st535:
	if ( ++p == pe )
		goto _test_eof535;
case 535:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1776;
		case 32: goto tr1776;
		case 40: goto tr1777;
		case 41: goto tr1778;
		case 778: goto tr1780;
		case 827: goto tr1780;
		case 1034: goto tr1781;
		case 1083: goto tr1782;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1779;
	goto tr1739;
tr1776:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1254;
tr1777:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1254;
tr1778:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1254;
tr1781:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1254;
st1254:
	if ( ++p == pe )
		goto _test_eof1254;
case 1254:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st536;
		case 32: goto st536;
		case 40: goto tr1784;
		case 41: goto tr1785;
		case 1034: goto tr1787;
		case 1083: goto tr1788;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1786;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1786;
	} else
		goto tr1786;
	goto tr1747;
tr1784:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st536;
tr1785:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st536;
tr1787:
	{
		s->line_counter++;
	}
	goto st536;
tr1801:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 536; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st536;
tr1802:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st536;
tr1803:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st536;
tr1804:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st536;
tr1805:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st536;
st536:
	if ( ++p == pe )
		goto _test_eof536;
case 536:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st536;
		case 32: goto st536;
		case 40: goto tr1784;
		case 41: goto tr1785;
		case 1034: goto tr1787;
		case 1083: goto tr1788;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1786;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1786;
	} else
		goto tr1786;
	goto tr1747;
tr1786:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st537;
st537:
	if ( ++p == pe )
		goto _test_eof537;
case 537:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1789;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1789;
	} else
		goto tr1789;
	goto tr1747;
tr1791:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st538;
tr1792:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st538;
tr1793:
	{
		s->line_counter++;
	}
	goto st538;
tr1799:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 538; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st538;
tr1789:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st538;
st538:
	if ( ++p == pe )
		goto _test_eof538;
case 538:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st538;
		case 32: goto st538;
		case 40: goto tr1791;
		case 41: goto tr1792;
		case 2058: goto tr1793;
		case 2107: goto tr1794;
		case 2314: goto tr1795;
		case 2363: goto tr1795;
		case 2570: goto tr1796;
		case 2619: goto tr1797;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1786;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1786;
	} else
		goto tr1786;
	goto tr1747;
tr1794:
	{
		s->buffer_length = 0;
	}
	goto st539;
tr1798:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st539;
st539:
	if ( ++p == pe )
		goto _test_eof539;
case 539:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1799;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1798;
	goto tr1747;
tr1780:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1255;
tr1795:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1255;
st1255:
	if ( ++p == pe )
		goto _test_eof1255;
case 1255:
	goto tr1733;
tr1796:
	{
		s->line_counter++;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1256;
st1256:
	if ( ++p == pe )
		goto _test_eof1256;
case 1256:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st538;
		case 32: goto st538;
		case 40: goto tr1791;
		case 41: goto tr1792;
		case 2058: goto tr1793;
		case 2107: goto tr1794;
		case 2314: goto tr1795;
		case 2363: goto tr1795;
		case 2570: goto tr1796;
		case 2619: goto tr1797;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1786;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1786;
	} else
		goto tr1786;
	goto tr1747;
tr1797:
	{
		s->buffer_length = 0;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1257;
st1257:
	if ( ++p == pe )
		goto _test_eof1257;
case 1257:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1799;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1798;
	goto tr1747;
tr1788:
	{
		s->buffer_length = 0;
	}
	goto st540;
tr1800:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st540;
tr1806:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st540;
st540:
	if ( ++p == pe )
		goto _test_eof540;
case 540:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1801;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1800;
	goto tr1733;
tr1775:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st541;
tr1779:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st541;
st541:
	if ( ++p == pe )
		goto _test_eof541;
case 541:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1802;
		case 32: goto tr1802;
		case 40: goto tr1803;
		case 41: goto tr1804;
		case 1034: goto tr1805;
		case 1083: goto tr1806;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1779;
	goto tr1739;
tr1782:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1258;
st1258:
	if ( ++p == pe )
		goto _test_eof1258;
case 1258:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1801;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1800;
	goto tr1733;
tr1773:
	{
		s->buffer_length = 0;
	}
	goto st542;
tr1807:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st542;
st542:
	if ( ++p == pe )
		goto _test_eof542;
case 542:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1808;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1807;
	goto tr1733;
st543:
	if ( ++p == pe )
		goto _test_eof543;
case 543:
	switch( (*p) ) {
		case 68: goto st545;
		case 69: goto st560;
		case 73: goto st599;
		case 80: goto st607;
		case 82: goto st620;
		case 100: goto st545;
		case 101: goto st560;
		case 105: goto st599;
		case 112: goto st607;
		case 114: goto st620;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1810;
	goto tr1809;
tr1810:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st544;
tr1817:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st544;
st544:
	if ( ++p == pe )
		goto _test_eof544;
case 544:
	switch( (*p) ) {
		case 32: goto tr1816;
		case 59: goto tr1816;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1816;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1817;
	} else
		goto tr1816;
	goto tr1809;
tr1816:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1820:
	{
		*(rdata_tail++) = 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1822:
	{
		*(rdata_tail++) = 3;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1834:
	{
		*(rdata_tail++) = 6;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1844:
	{
		*(rdata_tail++) = 12;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1858:
	{
		*(rdata_tail++) = 13;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1867:
	{
		*(rdata_tail++) = 14;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1874:
	{
		*(rdata_tail++) = 15;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1877:
	{
		*(rdata_tail++) = 16;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1885:
	{
		*(rdata_tail++) = 252;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1896:
	{
		*(rdata_tail++) = 253;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1899:
	{
		*(rdata_tail++) = 254;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1906:
	{
		*(rdata_tail++) = 1;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1912:
	{
		*(rdata_tail++) = 5;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1924:
	{
		*(rdata_tail++) = 7;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1927:
	{
		*(rdata_tail++) = 8;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
tr1930:
	{
		*(rdata_tail++) = 10;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1259;
st1259:
	if ( ++p == pe )
		goto _test_eof1259;
case 1259:
	goto st0;
st545:
	if ( ++p == pe )
		goto _test_eof545;
case 545:
	switch( (*p) ) {
		case 72: goto st546;
		case 83: goto st547;
		case 104: goto st546;
		case 115: goto st547;
	}
	goto tr1809;
st546:
	if ( ++p == pe )
		goto _test_eof546;
case 546:
	switch( (*p) ) {
		case 32: goto tr1820;
		case 59: goto tr1820;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1820;
	} else if ( (*p) >= 9 )
		goto tr1820;
	goto tr1809;
st547:
	if ( ++p == pe )
		goto _test_eof547;
case 547:
	switch( (*p) ) {
		case 65: goto st548;
		case 97: goto st548;
	}
	goto tr1809;
st548:
	if ( ++p == pe )
		goto _test_eof548;
case 548:
	switch( (*p) ) {
		case 32: goto tr1822;
		case 45: goto st549;
		case 59: goto tr1822;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1822;
	} else if ( (*p) >= 9 )
		goto tr1822;
	goto tr1809;
st549:
	if ( ++p == pe )
		goto _test_eof549;
case 549:
	switch( (*p) ) {
		case 78: goto st550;
		case 110: goto st550;
	}
	goto tr1809;
st550:
	if ( ++p == pe )
		goto _test_eof550;
case 550:
	switch( (*p) ) {
		case 83: goto st551;
		case 115: goto st551;
	}
	goto tr1809;
st551:
	if ( ++p == pe )
		goto _test_eof551;
case 551:
	switch( (*p) ) {
		case 69: goto st552;
		case 101: goto st552;
	}
	goto tr1809;
st552:
	if ( ++p == pe )
		goto _test_eof552;
case 552:
	switch( (*p) ) {
		case 67: goto st553;
		case 99: goto st553;
	}
	goto tr1809;
st553:
	if ( ++p == pe )
		goto _test_eof553;
case 553:
	if ( (*p) == 51 )
		goto st554;
	goto tr1809;
st554:
	if ( ++p == pe )
		goto _test_eof554;
case 554:
	if ( (*p) == 45 )
		goto st555;
	goto tr1809;
st555:
	if ( ++p == pe )
		goto _test_eof555;
case 555:
	switch( (*p) ) {
		case 83: goto st556;
		case 115: goto st556;
	}
	goto tr1809;
st556:
	if ( ++p == pe )
		goto _test_eof556;
case 556:
	switch( (*p) ) {
		case 72: goto st557;
		case 104: goto st557;
	}
	goto tr1809;
st557:
	if ( ++p == pe )
		goto _test_eof557;
case 557:
	switch( (*p) ) {
		case 65: goto st558;
		case 97: goto st558;
	}
	goto tr1809;
st558:
	if ( ++p == pe )
		goto _test_eof558;
case 558:
	if ( (*p) == 49 )
		goto st559;
	goto tr1809;
st559:
	if ( ++p == pe )
		goto _test_eof559;
case 559:
	switch( (*p) ) {
		case 32: goto tr1834;
		case 59: goto tr1834;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1834;
	} else if ( (*p) >= 9 )
		goto tr1834;
	goto tr1809;
st560:
	if ( ++p == pe )
		goto _test_eof560;
case 560:
	switch( (*p) ) {
		case 67: goto st561;
		case 68: goto st590;
		case 99: goto st561;
		case 100: goto st590;
	}
	goto tr1809;
st561:
	if ( ++p == pe )
		goto _test_eof561;
case 561:
	switch( (*p) ) {
		case 67: goto st562;
		case 68: goto st568;
		case 99: goto st562;
		case 100: goto st568;
	}
	goto tr1809;
st562:
	if ( ++p == pe )
		goto _test_eof562;
case 562:
	if ( (*p) == 45 )
		goto st563;
	goto tr1809;
st563:
	if ( ++p == pe )
		goto _test_eof563;
case 563:
	switch( (*p) ) {
		case 71: goto st564;
		case 103: goto st564;
	}
	goto tr1809;
st564:
	if ( ++p == pe )
		goto _test_eof564;
case 564:
	switch( (*p) ) {
		case 79: goto st565;
		case 111: goto st565;
	}
	goto tr1809;
st565:
	if ( ++p == pe )
		goto _test_eof565;
case 565:
	switch( (*p) ) {
		case 83: goto st566;
		case 115: goto st566;
	}
	goto tr1809;
st566:
	if ( ++p == pe )
		goto _test_eof566;
case 566:
	switch( (*p) ) {
		case 84: goto st567;
		case 116: goto st567;
	}
	goto tr1809;
st567:
	if ( ++p == pe )
		goto _test_eof567;
case 567:
	switch( (*p) ) {
		case 32: goto tr1844;
		case 59: goto tr1844;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1844;
	} else if ( (*p) >= 9 )
		goto tr1844;
	goto tr1809;
st568:
	if ( ++p == pe )
		goto _test_eof568;
case 568:
	switch( (*p) ) {
		case 83: goto st569;
		case 115: goto st569;
	}
	goto tr1809;
st569:
	if ( ++p == pe )
		goto _test_eof569;
case 569:
	switch( (*p) ) {
		case 65: goto st570;
		case 97: goto st570;
	}
	goto tr1809;
st570:
	if ( ++p == pe )
		goto _test_eof570;
case 570:
	switch( (*p) ) {
		case 80: goto st571;
		case 112: goto st571;
	}
	goto tr1809;
st571:
	if ( ++p == pe )
		goto _test_eof571;
case 571:
	switch( (*p) ) {
		case 50: goto st572;
		case 51: goto st581;
	}
	goto tr1809;
st572:
	if ( ++p == pe )
		goto _test_eof572;
case 572:
	if ( (*p) == 53 )
		goto st573;
	goto tr1809;
st573:
	if ( ++p == pe )
		goto _test_eof573;
case 573:
	if ( (*p) == 54 )
		goto st574;
	goto tr1809;
st574:
	if ( ++p == pe )
		goto _test_eof574;
case 574:
	switch( (*p) ) {
		case 83: goto st575;
		case 115: goto st575;
	}
	goto tr1809;
st575:
	if ( ++p == pe )
		goto _test_eof575;
case 575:
	switch( (*p) ) {
		case 72: goto st576;
		case 104: goto st576;
	}
	goto tr1809;
st576:
	if ( ++p == pe )
		goto _test_eof576;
case 576:
	switch( (*p) ) {
		case 65: goto st577;
		case 97: goto st577;
	}
	goto tr1809;
st577:
	if ( ++p == pe )
		goto _test_eof577;
case 577:
	if ( (*p) == 50 )
		goto st578;
	goto tr1809;
st578:
	if ( ++p == pe )
		goto _test_eof578;
case 578:
	if ( (*p) == 53 )
		goto st579;
	goto tr1809;
st579:
	if ( ++p == pe )
		goto _test_eof579;
case 579:
	if ( (*p) == 54 )
		goto st580;
	goto tr1809;
st580:
	if ( ++p == pe )
		goto _test_eof580;
case 580:
	switch( (*p) ) {
		case 32: goto tr1858;
		case 59: goto tr1858;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1858;
	} else if ( (*p) >= 9 )
		goto tr1858;
	goto tr1809;
st581:
	if ( ++p == pe )
		goto _test_eof581;
case 581:
	if ( (*p) == 56 )
		goto st582;
	goto tr1809;
st582:
	if ( ++p == pe )
		goto _test_eof582;
case 582:
	if ( (*p) == 52 )
		goto st583;
	goto tr1809;
st583:
	if ( ++p == pe )
		goto _test_eof583;
case 583:
	switch( (*p) ) {
		case 83: goto st584;
		case 115: goto st584;
	}
	goto tr1809;
st584:
	if ( ++p == pe )
		goto _test_eof584;
case 584:
	switch( (*p) ) {
		case 72: goto st585;
		case 104: goto st585;
	}
	goto tr1809;
st585:
	if ( ++p == pe )
		goto _test_eof585;
case 585:
	switch( (*p) ) {
		case 65: goto st586;
		case 97: goto st586;
	}
	goto tr1809;
st586:
	if ( ++p == pe )
		goto _test_eof586;
case 586:
	if ( (*p) == 51 )
		goto st587;
	goto tr1809;
st587:
	if ( ++p == pe )
		goto _test_eof587;
case 587:
	if ( (*p) == 56 )
		goto st588;
	goto tr1809;
st588:
	if ( ++p == pe )
		goto _test_eof588;
case 588:
	if ( (*p) == 52 )
		goto st589;
	goto tr1809;
st589:
	if ( ++p == pe )
		goto _test_eof589;
case 589:
	switch( (*p) ) {
		case 32: goto tr1867;
		case 59: goto tr1867;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1867;
	} else if ( (*p) >= 9 )
		goto tr1867;
	goto tr1809;
st590:
	if ( ++p == pe )
		goto _test_eof590;
case 590:
	switch( (*p) ) {
		case 50: goto st591;
		case 52: goto st596;
	}
	goto tr1809;
st591:
	if ( ++p == pe )
		goto _test_eof591;
case 591:
	if ( (*p) == 53 )
		goto st592;
	goto tr1809;
st592:
	if ( ++p == pe )
		goto _test_eof592;
case 592:
	if ( (*p) == 53 )
		goto st593;
	goto tr1809;
st593:
	if ( ++p == pe )
		goto _test_eof593;
case 593:
	if ( (*p) == 49 )
		goto st594;
	goto tr1809;
st594:
	if ( ++p == pe )
		goto _test_eof594;
case 594:
	if ( (*p) == 57 )
		goto st595;
	goto tr1809;
st595:
	if ( ++p == pe )
		goto _test_eof595;
case 595:
	switch( (*p) ) {
		case 32: goto tr1874;
		case 59: goto tr1874;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1874;
	} else if ( (*p) >= 9 )
		goto tr1874;
	goto tr1809;
st596:
	if ( ++p == pe )
		goto _test_eof596;
case 596:
	if ( (*p) == 52 )
		goto st597;
	goto tr1809;
st597:
	if ( ++p == pe )
		goto _test_eof597;
case 597:
	if ( (*p) == 56 )
		goto st598;
	goto tr1809;
st598:
	if ( ++p == pe )
		goto _test_eof598;
case 598:
	switch( (*p) ) {
		case 32: goto tr1877;
		case 59: goto tr1877;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1877;
	} else if ( (*p) >= 9 )
		goto tr1877;
	goto tr1809;
st599:
	if ( ++p == pe )
		goto _test_eof599;
case 599:
	switch( (*p) ) {
		case 78: goto st600;
		case 110: goto st600;
	}
	goto tr1809;
st600:
	if ( ++p == pe )
		goto _test_eof600;
case 600:
	switch( (*p) ) {
		case 68: goto st601;
		case 100: goto st601;
	}
	goto tr1809;
st601:
	if ( ++p == pe )
		goto _test_eof601;
case 601:
	switch( (*p) ) {
		case 73: goto st602;
		case 105: goto st602;
	}
	goto tr1809;
st602:
	if ( ++p == pe )
		goto _test_eof602;
case 602:
	switch( (*p) ) {
		case 82: goto st603;
		case 114: goto st603;
	}
	goto tr1809;
st603:
	if ( ++p == pe )
		goto _test_eof603;
case 603:
	switch( (*p) ) {
		case 69: goto st604;
		case 101: goto st604;
	}
	goto tr1809;
st604:
	if ( ++p == pe )
		goto _test_eof604;
case 604:
	switch( (*p) ) {
		case 67: goto st605;
		case 99: goto st605;
	}
	goto tr1809;
st605:
	if ( ++p == pe )
		goto _test_eof605;
case 605:
	switch( (*p) ) {
		case 84: goto st606;
		case 116: goto st606;
	}
	goto tr1809;
st606:
	if ( ++p == pe )
		goto _test_eof606;
case 606:
	switch( (*p) ) {
		case 32: goto tr1885;
		case 59: goto tr1885;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1885;
	} else if ( (*p) >= 9 )
		goto tr1885;
	goto tr1809;
st607:
	if ( ++p == pe )
		goto _test_eof607;
case 607:
	switch( (*p) ) {
		case 82: goto st608;
		case 114: goto st608;
	}
	goto tr1809;
st608:
	if ( ++p == pe )
		goto _test_eof608;
case 608:
	switch( (*p) ) {
		case 73: goto st609;
		case 105: goto st609;
	}
	goto tr1809;
st609:
	if ( ++p == pe )
		goto _test_eof609;
case 609:
	switch( (*p) ) {
		case 86: goto st610;
		case 118: goto st610;
	}
	goto tr1809;
st610:
	if ( ++p == pe )
		goto _test_eof610;
case 610:
	switch( (*p) ) {
		case 65: goto st611;
		case 97: goto st611;
	}
	goto tr1809;
st611:
	if ( ++p == pe )
		goto _test_eof611;
case 611:
	switch( (*p) ) {
		case 84: goto st612;
		case 116: goto st612;
	}
	goto tr1809;
st612:
	if ( ++p == pe )
		goto _test_eof612;
case 612:
	switch( (*p) ) {
		case 69: goto st613;
		case 101: goto st613;
	}
	goto tr1809;
st613:
	if ( ++p == pe )
		goto _test_eof613;
case 613:
	switch( (*p) ) {
		case 68: goto st614;
		case 79: goto st617;
		case 100: goto st614;
		case 111: goto st617;
	}
	goto tr1809;
st614:
	if ( ++p == pe )
		goto _test_eof614;
case 614:
	switch( (*p) ) {
		case 78: goto st615;
		case 110: goto st615;
	}
	goto tr1809;
st615:
	if ( ++p == pe )
		goto _test_eof615;
case 615:
	switch( (*p) ) {
		case 83: goto st616;
		case 115: goto st616;
	}
	goto tr1809;
st616:
	if ( ++p == pe )
		goto _test_eof616;
case 616:
	switch( (*p) ) {
		case 32: goto tr1896;
		case 59: goto tr1896;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1896;
	} else if ( (*p) >= 9 )
		goto tr1896;
	goto tr1809;
st617:
	if ( ++p == pe )
		goto _test_eof617;
case 617:
	switch( (*p) ) {
		case 73: goto st618;
		case 105: goto st618;
	}
	goto tr1809;
st618:
	if ( ++p == pe )
		goto _test_eof618;
case 618:
	switch( (*p) ) {
		case 68: goto st619;
		case 100: goto st619;
	}
	goto tr1809;
st619:
	if ( ++p == pe )
		goto _test_eof619;
case 619:
	switch( (*p) ) {
		case 32: goto tr1899;
		case 59: goto tr1899;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1899;
	} else if ( (*p) >= 9 )
		goto tr1899;
	goto tr1809;
st620:
	if ( ++p == pe )
		goto _test_eof620;
case 620:
	switch( (*p) ) {
		case 83: goto st621;
		case 115: goto st621;
	}
	goto tr1809;
st621:
	if ( ++p == pe )
		goto _test_eof621;
case 621:
	switch( (*p) ) {
		case 65: goto st622;
		case 97: goto st622;
	}
	goto tr1809;
st622:
	if ( ++p == pe )
		goto _test_eof622;
case 622:
	switch( (*p) ) {
		case 77: goto st623;
		case 83: goto st626;
		case 109: goto st623;
		case 115: goto st626;
	}
	goto tr1809;
st623:
	if ( ++p == pe )
		goto _test_eof623;
case 623:
	switch( (*p) ) {
		case 68: goto st624;
		case 100: goto st624;
	}
	goto tr1809;
st624:
	if ( ++p == pe )
		goto _test_eof624;
case 624:
	if ( (*p) == 53 )
		goto st625;
	goto tr1809;
st625:
	if ( ++p == pe )
		goto _test_eof625;
case 625:
	switch( (*p) ) {
		case 32: goto tr1906;
		case 59: goto tr1906;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1906;
	} else if ( (*p) >= 9 )
		goto tr1906;
	goto tr1809;
st626:
	if ( ++p == pe )
		goto _test_eof626;
case 626:
	switch( (*p) ) {
		case 72: goto st627;
		case 104: goto st627;
	}
	goto tr1809;
st627:
	if ( ++p == pe )
		goto _test_eof627;
case 627:
	switch( (*p) ) {
		case 65: goto st628;
		case 97: goto st628;
	}
	goto tr1809;
st628:
	if ( ++p == pe )
		goto _test_eof628;
case 628:
	switch( (*p) ) {
		case 49: goto st629;
		case 50: goto st641;
		case 53: goto st644;
	}
	goto tr1809;
st629:
	if ( ++p == pe )
		goto _test_eof629;
case 629:
	switch( (*p) ) {
		case 32: goto tr1912;
		case 45: goto st630;
		case 59: goto tr1912;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1912;
	} else if ( (*p) >= 9 )
		goto tr1912;
	goto tr1809;
st630:
	if ( ++p == pe )
		goto _test_eof630;
case 630:
	switch( (*p) ) {
		case 78: goto st631;
		case 110: goto st631;
	}
	goto tr1809;
st631:
	if ( ++p == pe )
		goto _test_eof631;
case 631:
	switch( (*p) ) {
		case 83: goto st632;
		case 115: goto st632;
	}
	goto tr1809;
st632:
	if ( ++p == pe )
		goto _test_eof632;
case 632:
	switch( (*p) ) {
		case 69: goto st633;
		case 101: goto st633;
	}
	goto tr1809;
st633:
	if ( ++p == pe )
		goto _test_eof633;
case 633:
	switch( (*p) ) {
		case 67: goto st634;
		case 99: goto st634;
	}
	goto tr1809;
st634:
	if ( ++p == pe )
		goto _test_eof634;
case 634:
	if ( (*p) == 51 )
		goto st635;
	goto tr1809;
st635:
	if ( ++p == pe )
		goto _test_eof635;
case 635:
	if ( (*p) == 45 )
		goto st636;
	goto tr1809;
st636:
	if ( ++p == pe )
		goto _test_eof636;
case 636:
	switch( (*p) ) {
		case 83: goto st637;
		case 115: goto st637;
	}
	goto tr1809;
st637:
	if ( ++p == pe )
		goto _test_eof637;
case 637:
	switch( (*p) ) {
		case 72: goto st638;
		case 104: goto st638;
	}
	goto tr1809;
st638:
	if ( ++p == pe )
		goto _test_eof638;
case 638:
	switch( (*p) ) {
		case 65: goto st639;
		case 97: goto st639;
	}
	goto tr1809;
st639:
	if ( ++p == pe )
		goto _test_eof639;
case 639:
	if ( (*p) == 49 )
		goto st640;
	goto tr1809;
st640:
	if ( ++p == pe )
		goto _test_eof640;
case 640:
	switch( (*p) ) {
		case 32: goto tr1924;
		case 59: goto tr1924;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1924;
	} else if ( (*p) >= 9 )
		goto tr1924;
	goto tr1809;
st641:
	if ( ++p == pe )
		goto _test_eof641;
case 641:
	if ( (*p) == 53 )
		goto st642;
	goto tr1809;
st642:
	if ( ++p == pe )
		goto _test_eof642;
case 642:
	if ( (*p) == 54 )
		goto st643;
	goto tr1809;
st643:
	if ( ++p == pe )
		goto _test_eof643;
case 643:
	switch( (*p) ) {
		case 32: goto tr1927;
		case 59: goto tr1927;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1927;
	} else if ( (*p) >= 9 )
		goto tr1927;
	goto tr1809;
st644:
	if ( ++p == pe )
		goto _test_eof644;
case 644:
	if ( (*p) == 49 )
		goto st645;
	goto tr1809;
st645:
	if ( ++p == pe )
		goto _test_eof645;
case 645:
	if ( (*p) == 50 )
		goto st646;
	goto tr1809;
st646:
	if ( ++p == pe )
		goto _test_eof646;
case 646:
	switch( (*p) ) {
		case 32: goto tr1930;
		case 59: goto tr1930;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1930;
	} else if ( (*p) >= 9 )
		goto tr1930;
	goto tr1809;
st647:
	if ( ++p == pe )
		goto _test_eof647;
case 647:
	switch( (*p) ) {
		case 65: goto st649;
		case 73: goto st655;
		case 79: goto st672;
		case 80: goto st675;
		case 83: goto st681;
		case 85: goto st685;
		case 97: goto st649;
		case 105: goto st655;
		case 111: goto st672;
		case 112: goto st675;
		case 115: goto st681;
		case 117: goto st685;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1932;
	goto tr1931;
tr1932:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st648;
tr1940:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st648;
st648:
	if ( ++p == pe )
		goto _test_eof648;
case 648:
	switch( (*p) ) {
		case 32: goto tr1939;
		case 59: goto tr1939;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1939;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1940;
	} else
		goto tr1939;
	goto tr1931;
tr1939:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1946:
	{
		*((uint16_t *)rdata_tail) = htons(7);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1955:
	{
		*((uint16_t *)rdata_tail) = htons(8);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1959:
	{
		*((uint16_t *)rdata_tail) = htons(6);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1962:
	{
		*((uint16_t *)rdata_tail) = htons(4);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1966:
	{
		*((uint16_t *)rdata_tail) = htons(5);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1969:
	{
		*((uint16_t *)rdata_tail) = htons(254);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1973:
	{
		*((uint16_t *)rdata_tail) = htons(3);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1976:
	{
		*((uint16_t *)rdata_tail) = htons(1);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1980:
	{
		*((uint16_t *)rdata_tail) = htons(2);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
tr1983:
	{
		*((uint16_t *)rdata_tail) = htons(253);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1260;
st1260:
	if ( ++p == pe )
		goto _test_eof1260;
case 1260:
	goto st0;
st649:
	if ( ++p == pe )
		goto _test_eof649;
case 649:
	switch( (*p) ) {
		case 67: goto st650;
		case 99: goto st650;
	}
	goto tr1931;
st650:
	if ( ++p == pe )
		goto _test_eof650;
case 650:
	switch( (*p) ) {
		case 80: goto st651;
		case 112: goto st651;
	}
	goto tr1931;
st651:
	if ( ++p == pe )
		goto _test_eof651;
case 651:
	switch( (*p) ) {
		case 75: goto st652;
		case 107: goto st652;
	}
	goto tr1931;
st652:
	if ( ++p == pe )
		goto _test_eof652;
case 652:
	switch( (*p) ) {
		case 73: goto st653;
		case 105: goto st653;
	}
	goto tr1931;
st653:
	if ( ++p == pe )
		goto _test_eof653;
case 653:
	switch( (*p) ) {
		case 88: goto st654;
		case 120: goto st654;
	}
	goto tr1931;
st654:
	if ( ++p == pe )
		goto _test_eof654;
case 654:
	switch( (*p) ) {
		case 32: goto tr1946;
		case 59: goto tr1946;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1946;
	} else if ( (*p) >= 9 )
		goto tr1946;
	goto tr1931;
st655:
	if ( ++p == pe )
		goto _test_eof655;
case 655:
	switch( (*p) ) {
		case 65: goto st656;
		case 80: goto st662;
		case 83: goto st668;
		case 97: goto st656;
		case 112: goto st662;
		case 115: goto st668;
	}
	goto tr1931;
st656:
	if ( ++p == pe )
		goto _test_eof656;
case 656:
	switch( (*p) ) {
		case 67: goto st657;
		case 99: goto st657;
	}
	goto tr1931;
st657:
	if ( ++p == pe )
		goto _test_eof657;
case 657:
	switch( (*p) ) {
		case 80: goto st658;
		case 112: goto st658;
	}
	goto tr1931;
st658:
	if ( ++p == pe )
		goto _test_eof658;
case 658:
	switch( (*p) ) {
		case 75: goto st659;
		case 107: goto st659;
	}
	goto tr1931;
st659:
	if ( ++p == pe )
		goto _test_eof659;
case 659:
	switch( (*p) ) {
		case 73: goto st660;
		case 105: goto st660;
	}
	goto tr1931;
st660:
	if ( ++p == pe )
		goto _test_eof660;
case 660:
	switch( (*p) ) {
		case 88: goto st661;
		case 120: goto st661;
	}
	goto tr1931;
st661:
	if ( ++p == pe )
		goto _test_eof661;
case 661:
	switch( (*p) ) {
		case 32: goto tr1955;
		case 59: goto tr1955;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1955;
	} else if ( (*p) >= 9 )
		goto tr1955;
	goto tr1931;
st662:
	if ( ++p == pe )
		goto _test_eof662;
case 662:
	switch( (*p) ) {
		case 71: goto st663;
		case 75: goto st665;
		case 103: goto st663;
		case 107: goto st665;
	}
	goto tr1931;
st663:
	if ( ++p == pe )
		goto _test_eof663;
case 663:
	switch( (*p) ) {
		case 80: goto st664;
		case 112: goto st664;
	}
	goto tr1931;
st664:
	if ( ++p == pe )
		goto _test_eof664;
case 664:
	switch( (*p) ) {
		case 32: goto tr1959;
		case 59: goto tr1959;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1959;
	} else if ( (*p) >= 9 )
		goto tr1959;
	goto tr1931;
st665:
	if ( ++p == pe )
		goto _test_eof665;
case 665:
	switch( (*p) ) {
		case 73: goto st666;
		case 105: goto st666;
	}
	goto tr1931;
st666:
	if ( ++p == pe )
		goto _test_eof666;
case 666:
	switch( (*p) ) {
		case 88: goto st667;
		case 120: goto st667;
	}
	goto tr1931;
st667:
	if ( ++p == pe )
		goto _test_eof667;
case 667:
	switch( (*p) ) {
		case 32: goto tr1962;
		case 59: goto tr1962;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1962;
	} else if ( (*p) >= 9 )
		goto tr1962;
	goto tr1931;
st668:
	if ( ++p == pe )
		goto _test_eof668;
case 668:
	switch( (*p) ) {
		case 80: goto st669;
		case 112: goto st669;
	}
	goto tr1931;
st669:
	if ( ++p == pe )
		goto _test_eof669;
case 669:
	switch( (*p) ) {
		case 75: goto st670;
		case 107: goto st670;
	}
	goto tr1931;
st670:
	if ( ++p == pe )
		goto _test_eof670;
case 670:
	switch( (*p) ) {
		case 73: goto st671;
		case 105: goto st671;
	}
	goto tr1931;
st671:
	if ( ++p == pe )
		goto _test_eof671;
case 671:
	switch( (*p) ) {
		case 32: goto tr1966;
		case 59: goto tr1966;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1966;
	} else if ( (*p) >= 9 )
		goto tr1966;
	goto tr1931;
st672:
	if ( ++p == pe )
		goto _test_eof672;
case 672:
	switch( (*p) ) {
		case 73: goto st673;
		case 105: goto st673;
	}
	goto tr1931;
st673:
	if ( ++p == pe )
		goto _test_eof673;
case 673:
	switch( (*p) ) {
		case 68: goto st674;
		case 100: goto st674;
	}
	goto tr1931;
st674:
	if ( ++p == pe )
		goto _test_eof674;
case 674:
	switch( (*p) ) {
		case 32: goto tr1969;
		case 59: goto tr1969;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1969;
	} else if ( (*p) >= 9 )
		goto tr1969;
	goto tr1931;
st675:
	if ( ++p == pe )
		goto _test_eof675;
case 675:
	switch( (*p) ) {
		case 71: goto st676;
		case 75: goto st678;
		case 103: goto st676;
		case 107: goto st678;
	}
	goto tr1931;
st676:
	if ( ++p == pe )
		goto _test_eof676;
case 676:
	switch( (*p) ) {
		case 80: goto st677;
		case 112: goto st677;
	}
	goto tr1931;
st677:
	if ( ++p == pe )
		goto _test_eof677;
case 677:
	switch( (*p) ) {
		case 32: goto tr1973;
		case 59: goto tr1973;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1973;
	} else if ( (*p) >= 9 )
		goto tr1973;
	goto tr1931;
st678:
	if ( ++p == pe )
		goto _test_eof678;
case 678:
	switch( (*p) ) {
		case 73: goto st679;
		case 105: goto st679;
	}
	goto tr1931;
st679:
	if ( ++p == pe )
		goto _test_eof679;
case 679:
	switch( (*p) ) {
		case 88: goto st680;
		case 120: goto st680;
	}
	goto tr1931;
st680:
	if ( ++p == pe )
		goto _test_eof680;
case 680:
	switch( (*p) ) {
		case 32: goto tr1976;
		case 59: goto tr1976;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1976;
	} else if ( (*p) >= 9 )
		goto tr1976;
	goto tr1931;
st681:
	if ( ++p == pe )
		goto _test_eof681;
case 681:
	switch( (*p) ) {
		case 80: goto st682;
		case 112: goto st682;
	}
	goto tr1931;
st682:
	if ( ++p == pe )
		goto _test_eof682;
case 682:
	switch( (*p) ) {
		case 75: goto st683;
		case 107: goto st683;
	}
	goto tr1931;
st683:
	if ( ++p == pe )
		goto _test_eof683;
case 683:
	switch( (*p) ) {
		case 73: goto st684;
		case 105: goto st684;
	}
	goto tr1931;
st684:
	if ( ++p == pe )
		goto _test_eof684;
case 684:
	switch( (*p) ) {
		case 32: goto tr1980;
		case 59: goto tr1980;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1980;
	} else if ( (*p) >= 9 )
		goto tr1980;
	goto tr1931;
st685:
	if ( ++p == pe )
		goto _test_eof685;
case 685:
	switch( (*p) ) {
		case 82: goto st686;
		case 114: goto st686;
	}
	goto tr1931;
st686:
	if ( ++p == pe )
		goto _test_eof686;
case 686:
	switch( (*p) ) {
		case 73: goto st687;
		case 105: goto st687;
	}
	goto tr1931;
st687:
	if ( ++p == pe )
		goto _test_eof687;
case 687:
	switch( (*p) ) {
		case 32: goto tr1983;
		case 59: goto tr1983;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1983;
	} else if ( (*p) >= 9 )
		goto tr1983;
	goto tr1931;
st688:
	if ( ++p == pe )
		goto _test_eof688;
case 688:
	if ( (*p) == 46 )
		goto tr1985;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1985;
	goto tr1984;
tr1985:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st689;
tr1987:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st689;
st689:
	if ( ++p == pe )
		goto _test_eof689;
case 689:
	switch( (*p) ) {
		case 32: goto tr1986;
		case 46: goto tr1987;
		case 59: goto tr1986;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1986;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1987;
	} else
		goto tr1986;
	goto tr1984;
tr1986:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1261;
st1261:
	if ( ++p == pe )
		goto _test_eof1261;
case 1261:
	goto st0;
st690:
	if ( ++p == pe )
		goto _test_eof690;
case 690:
	switch( (*p) ) {
		case 42: goto tr1988;
		case 92: goto tr1988;
		case 95: goto tr1988;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1988;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1988;
	} else
		goto tr1988;
	goto tr75;
tr1988:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 691;goto st301;} }
	goto st691;
st691:
	if ( ++p == pe )
		goto _test_eof691;
case 691:
	switch( (*p) ) {
		case 32: goto tr1989;
		case 59: goto tr1989;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1989;
	} else if ( (*p) >= 9 )
		goto tr1989;
	goto tr75;
tr1989:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1262;
st1262:
	if ( ++p == pe )
		goto _test_eof1262;
case 1262:
	goto st0;
st692:
	if ( ++p == pe )
		goto _test_eof692;
case 692:
	switch( (*p) ) {
		case 42: goto tr1990;
		case 92: goto tr1990;
		case 95: goto tr1990;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1990;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1990;
	} else
		goto tr1990;
	goto tr75;
tr1990:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 693;goto st301;} }
	goto st693;
st693:
	if ( ++p == pe )
		goto _test_eof693;
case 693:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1991;
		case 32: goto tr1991;
		case 40: goto tr1992;
		case 41: goto tr1993;
		case 1034: goto tr1994;
		case 1083: goto tr1995;
	}
	goto tr75;
tr1997:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st694;
tr1998:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st694;
tr2000:
	{
		s->line_counter++;
	}
	goto st694;
tr2144:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 694; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st694;
tr1991:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st694;
tr1992:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st694;
tr1993:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st694;
tr1994:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st694;
st694:
	if ( ++p == pe )
		goto _test_eof694;
case 694:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st694;
		case 32: goto st694;
		case 40: goto tr1997;
		case 41: goto tr1998;
		case 42: goto tr1999;
		case 92: goto tr1999;
		case 95: goto tr1999;
		case 1034: goto tr2000;
		case 1083: goto tr2001;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1999;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1999;
	} else
		goto tr1999;
	goto tr75;
tr1999:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 695;goto st301;} }
	goto st695;
st695:
	if ( ++p == pe )
		goto _test_eof695;
case 695:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2002;
		case 32: goto tr2002;
		case 40: goto tr2003;
		case 41: goto tr2004;
		case 1034: goto tr2005;
		case 1083: goto tr2006;
	}
	goto tr75;
tr2009:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st696;
tr2010:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st696;
tr2012:
	{
		s->line_counter++;
	}
	goto st696;
tr2142:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 696; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st696;
tr2002:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st696;
tr2003:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st696;
tr2004:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st696;
tr2005:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st696;
st696:
	if ( ++p == pe )
		goto _test_eof696;
case 696:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st696;
		case 32: goto st696;
		case 40: goto tr2009;
		case 41: goto tr2010;
		case 1034: goto tr2012;
		case 1083: goto tr2013;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2011;
	goto tr2007;
tr2011:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st697;
tr2017:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st697;
st697:
	if ( ++p == pe )
		goto _test_eof697;
case 697:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2014;
		case 32: goto tr2014;
		case 40: goto tr2015;
		case 41: goto tr2016;
		case 1034: goto tr2018;
		case 1083: goto tr2019;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2017;
	goto tr2007;
tr2021:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st698;
tr2022:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st698;
tr2024:
	{
		s->line_counter++;
	}
	goto st698;
tr2140:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 698; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st698;
tr2014:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st698;
tr2015:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st698;
tr2016:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st698;
tr2018:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st698;
st698:
	if ( ++p == pe )
		goto _test_eof698;
case 698:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st698;
		case 32: goto st698;
		case 40: goto tr2021;
		case 41: goto tr2022;
		case 1034: goto tr2024;
		case 1083: goto tr2025;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2023;
	goto tr2007;
tr2023:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st699;
tr2030:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st699;
st699:
	if ( ++p == pe )
		goto _test_eof699;
case 699:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2027;
		case 32: goto tr2027;
		case 40: goto tr2028;
		case 41: goto tr2029;
		case 68: goto tr2031;
		case 72: goto tr2032;
		case 77: goto tr2033;
		case 83: goto st718;
		case 87: goto tr2035;
		case 100: goto tr2031;
		case 104: goto tr2032;
		case 109: goto tr2033;
		case 115: goto st718;
		case 119: goto tr2035;
		case 1034: goto tr2036;
		case 1083: goto tr2037;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2030;
	goto tr2026;
tr2039:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st700;
tr2040:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st700;
tr2042:
	{
		s->line_counter++;
	}
	goto st700;
tr2125:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 700; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st700;
tr2027:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st700;
tr2028:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st700;
tr2029:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st700;
tr2036:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st700;
tr2133:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st700;
tr2134:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st700;
tr2135:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st700;
tr2137:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st700;
st700:
	if ( ++p == pe )
		goto _test_eof700;
case 700:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st700;
		case 32: goto st700;
		case 40: goto tr2039;
		case 41: goto tr2040;
		case 1034: goto tr2042;
		case 1083: goto tr2043;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2041;
	goto tr2007;
tr2041:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st701;
tr2047:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st701;
st701:
	if ( ++p == pe )
		goto _test_eof701;
case 701:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2044;
		case 32: goto tr2044;
		case 40: goto tr2045;
		case 41: goto tr2046;
		case 68: goto tr2048;
		case 72: goto tr2049;
		case 77: goto tr2050;
		case 83: goto st714;
		case 87: goto tr2052;
		case 100: goto tr2048;
		case 104: goto tr2049;
		case 109: goto tr2050;
		case 115: goto st714;
		case 119: goto tr2052;
		case 1034: goto tr2053;
		case 1083: goto tr2054;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2047;
	goto tr2026;
tr2056:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st702;
tr2057:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st702;
tr2059:
	{
		s->line_counter++;
	}
	goto st702;
tr2110:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 702; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st702;
tr2044:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st702;
tr2045:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st702;
tr2046:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st702;
tr2053:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st702;
tr2118:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st702;
tr2119:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st702;
tr2120:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st702;
tr2122:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st702;
st702:
	if ( ++p == pe )
		goto _test_eof702;
case 702:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st702;
		case 32: goto st702;
		case 40: goto tr2056;
		case 41: goto tr2057;
		case 1034: goto tr2059;
		case 1083: goto tr2060;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2058;
	goto tr2007;
tr2058:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st703;
tr2064:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st703;
st703:
	if ( ++p == pe )
		goto _test_eof703;
case 703:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2061;
		case 32: goto tr2061;
		case 40: goto tr2062;
		case 41: goto tr2063;
		case 68: goto tr2065;
		case 72: goto tr2066;
		case 77: goto tr2067;
		case 83: goto st710;
		case 87: goto tr2069;
		case 100: goto tr2065;
		case 104: goto tr2066;
		case 109: goto tr2067;
		case 115: goto st710;
		case 119: goto tr2069;
		case 1034: goto tr2070;
		case 1083: goto tr2071;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2064;
	goto tr2026;
tr2073:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st704;
tr2074:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st704;
tr2076:
	{
		s->line_counter++;
	}
	goto st704;
tr2095:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 704; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st704;
tr2061:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st704;
tr2062:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st704;
tr2063:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st704;
tr2070:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st704;
tr2103:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st704;
tr2104:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st704;
tr2105:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st704;
tr2107:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st704;
st704:
	if ( ++p == pe )
		goto _test_eof704;
case 704:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st704;
		case 32: goto st704;
		case 40: goto tr2073;
		case 41: goto tr2074;
		case 1034: goto tr2076;
		case 1083: goto tr2077;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2075;
	goto tr2007;
tr2075:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st705;
tr2079:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st705;
st705:
	if ( ++p == pe )
		goto _test_eof705;
case 705:
	switch( (*p) ) {
		case 32: goto tr2078;
		case 59: goto tr2078;
		case 68: goto tr2080;
		case 72: goto tr2081;
		case 77: goto tr2082;
		case 83: goto st706;
		case 87: goto tr2084;
		case 100: goto tr2080;
		case 104: goto tr2081;
		case 109: goto tr2082;
		case 115: goto st706;
		case 119: goto tr2084;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr2078;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2079;
	} else
		goto tr2078;
	goto tr2026;
tr2078:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1263;
tr2092:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1263;
st1263:
	if ( ++p == pe )
		goto _test_eof1263;
case 1263:
	goto st0;
tr2080:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st706;
tr2081:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st706;
tr2082:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st706;
tr2084:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st706;
st706:
	if ( ++p == pe )
		goto _test_eof706;
case 706:
	switch( (*p) ) {
		case 32: goto tr2078;
		case 59: goto tr2078;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr2078;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2085;
	} else
		goto tr2078;
	goto tr2026;
tr2086:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st707;
tr2085:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st707;
tr2093:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st707;
st707:
	if ( ++p == pe )
		goto _test_eof707;
case 707:
	switch( (*p) ) {
		case 68: goto tr2087;
		case 72: goto tr2088;
		case 77: goto tr2089;
		case 83: goto st708;
		case 87: goto tr2091;
		case 100: goto tr2087;
		case 104: goto tr2088;
		case 109: goto tr2089;
		case 115: goto st708;
		case 119: goto tr2091;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2086;
	goto tr2026;
tr2087:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st708;
tr2088:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st708;
tr2089:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st708;
tr2091:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st708;
st708:
	if ( ++p == pe )
		goto _test_eof708;
case 708:
	switch( (*p) ) {
		case 32: goto tr2092;
		case 59: goto tr2092;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr2092;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2093;
	} else
		goto tr2092;
	goto tr2026;
tr2077:
	{
		s->buffer_length = 0;
	}
	goto st709;
tr2094:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st709;
tr2071:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st709;
tr2108:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st709;
st709:
	if ( ++p == pe )
		goto _test_eof709;
case 709:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2095;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2094;
	goto tr75;
tr2065:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st710;
tr2066:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st710;
tr2067:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st710;
tr2069:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st710;
st710:
	if ( ++p == pe )
		goto _test_eof710;
case 710:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2061;
		case 32: goto tr2061;
		case 40: goto tr2062;
		case 41: goto tr2063;
		case 1034: goto tr2070;
		case 1083: goto tr2071;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2096;
	goto tr2026;
tr2097:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st711;
tr2096:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st711;
tr2106:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st711;
st711:
	if ( ++p == pe )
		goto _test_eof711;
case 711:
	switch( (*p) ) {
		case 68: goto tr2098;
		case 72: goto tr2099;
		case 77: goto tr2100;
		case 83: goto st712;
		case 87: goto tr2102;
		case 100: goto tr2098;
		case 104: goto tr2099;
		case 109: goto tr2100;
		case 115: goto st712;
		case 119: goto tr2102;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2097;
	goto tr2026;
tr2098:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st712;
tr2099:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st712;
tr2100:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st712;
tr2102:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st712;
st712:
	if ( ++p == pe )
		goto _test_eof712;
case 712:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2103;
		case 32: goto tr2103;
		case 40: goto tr2104;
		case 41: goto tr2105;
		case 1034: goto tr2107;
		case 1083: goto tr2108;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2106;
	goto tr2026;
tr2060:
	{
		s->buffer_length = 0;
	}
	goto st713;
tr2109:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st713;
tr2054:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st713;
tr2123:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st713;
st713:
	if ( ++p == pe )
		goto _test_eof713;
case 713:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2110;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2109;
	goto tr75;
tr2048:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st714;
tr2049:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st714;
tr2050:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st714;
tr2052:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st714;
st714:
	if ( ++p == pe )
		goto _test_eof714;
case 714:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2044;
		case 32: goto tr2044;
		case 40: goto tr2045;
		case 41: goto tr2046;
		case 1034: goto tr2053;
		case 1083: goto tr2054;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2111;
	goto tr2026;
tr2112:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st715;
tr2111:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st715;
tr2121:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st715;
st715:
	if ( ++p == pe )
		goto _test_eof715;
case 715:
	switch( (*p) ) {
		case 68: goto tr2113;
		case 72: goto tr2114;
		case 77: goto tr2115;
		case 83: goto st716;
		case 87: goto tr2117;
		case 100: goto tr2113;
		case 104: goto tr2114;
		case 109: goto tr2115;
		case 115: goto st716;
		case 119: goto tr2117;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2112;
	goto tr2026;
tr2113:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st716;
tr2114:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st716;
tr2115:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st716;
tr2117:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st716;
st716:
	if ( ++p == pe )
		goto _test_eof716;
case 716:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2118;
		case 32: goto tr2118;
		case 40: goto tr2119;
		case 41: goto tr2120;
		case 1034: goto tr2122;
		case 1083: goto tr2123;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2121;
	goto tr2026;
tr2043:
	{
		s->buffer_length = 0;
	}
	goto st717;
tr2124:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st717;
tr2037:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st717;
tr2138:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st717;
st717:
	if ( ++p == pe )
		goto _test_eof717;
case 717:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2125;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2124;
	goto tr75;
tr2031:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st718;
tr2032:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st718;
tr2033:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st718;
tr2035:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st718;
st718:
	if ( ++p == pe )
		goto _test_eof718;
case 718:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2027;
		case 32: goto tr2027;
		case 40: goto tr2028;
		case 41: goto tr2029;
		case 1034: goto tr2036;
		case 1083: goto tr2037;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2126;
	goto tr2026;
tr2127:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st719;
tr2126:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st719;
tr2136:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st719;
st719:
	if ( ++p == pe )
		goto _test_eof719;
case 719:
	switch( (*p) ) {
		case 68: goto tr2128;
		case 72: goto tr2129;
		case 77: goto tr2130;
		case 83: goto st720;
		case 87: goto tr2132;
		case 100: goto tr2128;
		case 104: goto tr2129;
		case 109: goto tr2130;
		case 115: goto st720;
		case 119: goto tr2132;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2127;
	goto tr2026;
tr2128:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st720;
tr2129:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st720;
tr2130:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st720;
tr2132:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st299;}
	              }
	            }
	goto st720;
st720:
	if ( ++p == pe )
		goto _test_eof720;
case 720:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2133;
		case 32: goto tr2133;
		case 40: goto tr2134;
		case 41: goto tr2135;
		case 1034: goto tr2137;
		case 1083: goto tr2138;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2136;
	goto tr2026;
tr2025:
	{
		s->buffer_length = 0;
	}
	goto st721;
tr2139:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st721;
tr2019:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st721;
st721:
	if ( ++p == pe )
		goto _test_eof721;
case 721:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2140;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2139;
	goto tr75;
tr2013:
	{
		s->buffer_length = 0;
	}
	goto st722;
tr2141:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st722;
tr2006:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st722;
st722:
	if ( ++p == pe )
		goto _test_eof722;
case 722:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2142;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2141;
	goto tr75;
tr2001:
	{
		s->buffer_length = 0;
	}
	goto st723;
tr2143:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st723;
tr1995:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st723;
st723:
	if ( ++p == pe )
		goto _test_eof723;
case 723:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2144;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2143;
	goto tr75;
st724:
	if ( ++p == pe )
		goto _test_eof724;
case 724:
	switch( (*p) ) {
		case 32: goto tr75;
		case 59: goto tr75;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr75;
	} else if ( (*p) >= 9 )
		goto tr75;
	goto tr2145;
tr2145:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 725;goto st310;} }
	goto st725;
st725:
	if ( ++p == pe )
		goto _test_eof725;
case 725:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2146;
		case 32: goto tr2146;
		case 40: goto tr2147;
		case 41: goto tr2148;
		case 1034: goto tr2149;
		case 1083: goto tr2150;
	}
	goto tr75;
tr2153:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st726;
tr2154:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st726;
tr2155:
	{
		s->line_counter++;
	}
	goto st726;
tr2159:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 726; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st726;
tr2146:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st726;
tr2147:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st726;
tr2148:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st726;
tr2149:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st726;
st726:
	if ( ++p == pe )
		goto _test_eof726;
case 726:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st726;
		case 32: goto st726;
		case 40: goto tr2153;
		case 41: goto tr2154;
		case 1034: goto tr2155;
		case 1083: goto tr2156;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2151;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2151;
	} else
		goto tr2151;
	goto tr75;
tr2151:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 727;goto st310;} }
	goto st727;
st727:
	if ( ++p == pe )
		goto _test_eof727;
case 727:
	switch( (*p) ) {
		case 32: goto tr2157;
		case 59: goto tr2157;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2157;
	} else if ( (*p) >= 9 )
		goto tr2157;
	goto tr75;
tr2157:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1264;
st1264:
	if ( ++p == pe )
		goto _test_eof1264;
case 1264:
	goto st0;
tr2156:
	{
		s->buffer_length = 0;
	}
	goto st728;
tr2158:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st728;
tr2150:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st728;
st728:
	if ( ++p == pe )
		goto _test_eof728;
case 728:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2159;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2158;
	goto tr75;
st729:
	if ( ++p == pe )
		goto _test_eof729;
case 729:
	switch( (*p) ) {
		case 42: goto tr2160;
		case 92: goto tr2160;
		case 95: goto tr2160;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr2160;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2160;
	} else
		goto tr2160;
	goto tr75;
tr2160:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 730;goto st301;} }
	goto st730;
st730:
	if ( ++p == pe )
		goto _test_eof730;
case 730:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2161;
		case 32: goto tr2161;
		case 40: goto tr2162;
		case 41: goto tr2163;
		case 1034: goto tr2164;
		case 1083: goto tr2165;
	}
	goto tr75;
tr2167:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st731;
tr2168:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st731;
tr2170:
	{
		s->line_counter++;
	}
	goto st731;
tr2174:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 731; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st731;
tr2161:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st731;
tr2162:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st731;
tr2163:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st731;
tr2164:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st731;
st731:
	if ( ++p == pe )
		goto _test_eof731;
case 731:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st731;
		case 32: goto st731;
		case 40: goto tr2167;
		case 41: goto tr2168;
		case 42: goto tr2169;
		case 92: goto tr2169;
		case 95: goto tr2169;
		case 1034: goto tr2170;
		case 1083: goto tr2171;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2169;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2169;
	} else
		goto tr2169;
	goto tr75;
tr2169:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 732;goto st301;} }
	goto st732;
st732:
	if ( ++p == pe )
		goto _test_eof732;
case 732:
	switch( (*p) ) {
		case 32: goto tr2172;
		case 59: goto tr2172;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2172;
	} else if ( (*p) >= 9 )
		goto tr2172;
	goto tr75;
tr2172:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1265;
st1265:
	if ( ++p == pe )
		goto _test_eof1265;
case 1265:
	goto st0;
tr2171:
	{
		s->buffer_length = 0;
	}
	goto st733;
tr2173:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st733;
tr2165:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st733;
st733:
	if ( ++p == pe )
		goto _test_eof733;
case 733:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2174;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2173;
	goto tr75;
st734:
	if ( ++p == pe )
		goto _test_eof734;
case 734:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2175;
	goto tr2007;
tr2175:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st735;
tr2179:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st735;
st735:
	if ( ++p == pe )
		goto _test_eof735;
case 735:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2176;
		case 32: goto tr2176;
		case 40: goto tr2177;
		case 41: goto tr2178;
		case 1034: goto tr2180;
		case 1083: goto tr2181;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2179;
	goto tr2007;
tr2183:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st736;
tr2184:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st736;
tr2186:
	{
		s->line_counter++;
	}
	goto st736;
tr2190:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 736; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st736;
tr2176:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st736;
tr2177:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st736;
tr2178:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st736;
tr2180:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st736;
st736:
	if ( ++p == pe )
		goto _test_eof736;
case 736:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st736;
		case 32: goto st736;
		case 40: goto tr2183;
		case 41: goto tr2184;
		case 42: goto tr2185;
		case 92: goto tr2185;
		case 95: goto tr2185;
		case 1034: goto tr2186;
		case 1083: goto tr2187;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2185;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2185;
	} else
		goto tr2185;
	goto tr75;
tr2185:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 737;goto st301;} }
	goto st737;
st737:
	if ( ++p == pe )
		goto _test_eof737;
case 737:
	switch( (*p) ) {
		case 32: goto tr2188;
		case 59: goto tr2188;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2188;
	} else if ( (*p) >= 9 )
		goto tr2188;
	goto tr75;
tr2188:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1266;
st1266:
	if ( ++p == pe )
		goto _test_eof1266;
case 1266:
	goto st0;
tr2187:
	{
		s->buffer_length = 0;
	}
	goto st738;
tr2189:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st738;
tr2181:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st738;
st738:
	if ( ++p == pe )
		goto _test_eof738;
case 738:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2190;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2189;
	goto tr75;
st739:
	if ( ++p == pe )
		goto _test_eof739;
case 739:
	switch( (*p) ) {
		case 32: goto tr2192;
		case 59: goto tr2192;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2192;
	} else if ( (*p) >= 9 )
		goto tr2192;
	goto tr2191;
tr2201:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 740;goto st310;} }
	goto st740;
tr2191:
	{
		s->long_string = true;
	}
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 740;goto st310;} }
	goto st740;
st740:
	if ( ++p == pe )
		goto _test_eof740;
case 740:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2193;
		case 32: goto tr2193;
		case 40: goto tr2194;
		case 41: goto tr2195;
		case 2058: goto tr2196;
		case 2107: goto tr2197;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2199;
		case 2619: goto tr2200;
	}
	goto tr2192;
tr2203:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st741;
tr2204:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st741;
tr2205:
	{
		s->line_counter++;
	}
	goto st741;
tr2211:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 741; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st741;
tr2193:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st741;
tr2194:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st741;
tr2195:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st741;
tr2196:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st741;
st741:
	if ( ++p == pe )
		goto _test_eof741;
case 741:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st741;
		case 32: goto st741;
		case 40: goto tr2203;
		case 41: goto tr2204;
		case 2058: goto tr2205;
		case 2107: goto tr2206;
		case 2314: goto tr2207;
		case 2363: goto tr2207;
		case 2570: goto tr2208;
		case 2619: goto tr2209;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2201;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2201;
	} else
		goto tr2201;
	goto tr2192;
tr2206:
	{
		s->buffer_length = 0;
	}
	goto st742;
tr2210:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st742;
tr2197:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st742;
st742:
	if ( ++p == pe )
		goto _test_eof742;
case 742:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2211;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2210;
	goto tr2192;
tr2198:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1267;
tr2207:
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1267;
st1267:
	if ( ++p == pe )
		goto _test_eof1267;
case 1267:
	goto st0;
tr2199:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1268;
tr2208:
	{
		s->line_counter++;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1268;
st1268:
	if ( ++p == pe )
		goto _test_eof1268;
case 1268:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st741;
		case 32: goto st741;
		case 40: goto tr2203;
		case 41: goto tr2204;
		case 2058: goto tr2205;
		case 2107: goto tr2206;
		case 2314: goto tr2207;
		case 2363: goto tr2207;
		case 2570: goto tr2208;
		case 2619: goto tr2209;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2201;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2201;
	} else
		goto tr2201;
	goto tr2192;
tr2200:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1269;
tr2209:
	{
		s->buffer_length = 0;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1269;
st1269:
	if ( ++p == pe )
		goto _test_eof1269;
case 1269:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2211;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2210;
	goto tr2192;
st743:
	if ( ++p == pe )
		goto _test_eof743;
case 743:
	if ( (*p) == 46 )
		goto tr2212;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr2212;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2212;
	} else
		goto tr2212;
	goto tr1984;
tr2212:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st744;
tr2214:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st744;
st744:
	if ( ++p == pe )
		goto _test_eof744;
case 744:
	switch( (*p) ) {
		case 32: goto tr2213;
		case 46: goto tr2214;
		case 59: goto tr2213;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr2213;
		} else if ( (*p) >= 9 )
			goto tr2213;
	} else if ( (*p) > 58 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr2214;
		} else if ( (*p) >= 65 )
			goto tr2214;
	} else
		goto tr2214;
	goto tr1984;
tr2213:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1270;
st1270:
	if ( ++p == pe )
		goto _test_eof1270;
case 1270:
	goto st0;
st745:
	if ( ++p == pe )
		goto _test_eof745;
case 745:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2216;
	goto tr2215;
tr2220:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st746;
tr2216:
	{
		memset(&(s->loc), 0, sizeof(s->loc));
		// Defaults.
		s->loc.siz = 100;
		s->loc.vp  = 1000;
		s->loc.hp  = 1000000;
		s->loc.lat_sign  = 1;
		s->loc.long_sign = 1;
		s->loc.alt_sign  = 1;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st746;
st746:
	if ( ++p == pe )
		goto _test_eof746;
case 746:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2217;
		case 32: goto tr2217;
		case 40: goto tr2218;
		case 41: goto tr2219;
		case 1034: goto tr2221;
		case 1083: goto tr2222;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2220;
	goto tr2215;
tr2224:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st747;
tr2225:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st747;
tr2229:
	{
		s->line_counter++;
	}
	goto st747;
tr2472:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 747; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st747;
tr2217:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st747;
tr2218:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st747;
tr2219:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st747;
tr2221:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st747;
st747:
	if ( ++p == pe )
		goto _test_eof747;
case 747:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st747;
		case 32: goto st747;
		case 40: goto tr2224;
		case 41: goto tr2225;
		case 78: goto st752;
		case 83: goto st794;
		case 1034: goto tr2229;
		case 1083: goto tr2230;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2226;
	goto tr2215;
tr2226:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st748;
tr2234:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st748;
st748:
	if ( ++p == pe )
		goto _test_eof748;
case 748:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2231;
		case 32: goto tr2231;
		case 40: goto tr2232;
		case 41: goto tr2233;
		case 1034: goto tr2235;
		case 1083: goto tr2236;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2234;
	goto tr2215;
tr2238:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st749;
tr2239:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st749;
tr2241:
	{
		s->line_counter++;
	}
	goto st749;
tr2470:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 749; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st749;
tr2231:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st749;
tr2232:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st749;
tr2233:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st749;
tr2235:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st749;
st749:
	if ( ++p == pe )
		goto _test_eof749;
case 749:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st749;
		case 32: goto st749;
		case 40: goto tr2238;
		case 41: goto tr2239;
		case 78: goto st752;
		case 83: goto st794;
		case 1034: goto tr2241;
		case 1083: goto tr2242;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2240;
	goto tr2215;
tr2247:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st750;
tr2240:
	{
		s->decimals = 3;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st750;
st750:
	if ( ++p == pe )
		goto _test_eof750;
case 750:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2243;
		case 32: goto tr2243;
		case 40: goto tr2244;
		case 41: goto tr2245;
		case 46: goto st796;
		case 1034: goto tr2248;
		case 1083: goto tr2249;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2247;
	goto tr2215;
tr2251:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st751;
tr2252:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st751;
tr2253:
	{
		s->line_counter++;
	}
	goto st751;
tr2461:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 751; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st751;
tr2243:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st751;
tr2244:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st751;
tr2245:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st751;
tr2248:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st751;
tr2462:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st751;
tr2463:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st751;
tr2464:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st751;
tr2466:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st751;
st751:
	if ( ++p == pe )
		goto _test_eof751;
case 751:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st751;
		case 32: goto st751;
		case 40: goto tr2251;
		case 41: goto tr2252;
		case 78: goto st752;
		case 83: goto st794;
		case 1034: goto tr2253;
		case 1083: goto tr2254;
	}
	goto tr2215;
st752:
	if ( ++p == pe )
		goto _test_eof752;
case 752:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st753;
		case 32: goto st753;
		case 40: goto tr2256;
		case 41: goto tr2257;
		case 1034: goto tr2258;
		case 1083: goto tr2259;
	}
	goto tr2215;
tr2256:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st753;
tr2257:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st753;
tr2258:
	{
		s->line_counter++;
	}
	goto st753;
tr2454:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 753; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st753;
tr2455:
	{
		s->loc.lat_sign = -1;
	}
	goto st753;
tr2456:
	{
		s->loc.lat_sign = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st753;
tr2457:
	{
		s->loc.lat_sign = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st753;
tr2458:
	{
		s->loc.lat_sign = -1;
	}
	{
		s->line_counter++;
	}
	goto st753;
st753:
	if ( ++p == pe )
		goto _test_eof753;
case 753:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st753;
		case 32: goto st753;
		case 40: goto tr2256;
		case 41: goto tr2257;
		case 1034: goto tr2258;
		case 1083: goto tr2259;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2260;
	goto tr2215;
tr2260:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st754;
tr2264:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st754;
st754:
	if ( ++p == pe )
		goto _test_eof754;
case 754:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2261;
		case 32: goto tr2261;
		case 40: goto tr2262;
		case 41: goto tr2263;
		case 1034: goto tr2265;
		case 1083: goto tr2266;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2264;
	goto tr2215;
tr2268:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st755;
tr2269:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st755;
tr2273:
	{
		s->line_counter++;
	}
	goto st755;
tr2452:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 755; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st755;
tr2261:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st755;
tr2262:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st755;
tr2263:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st755;
tr2265:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st755;
st755:
	if ( ++p == pe )
		goto _test_eof755;
case 755:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st755;
		case 32: goto st755;
		case 40: goto tr2268;
		case 41: goto tr2269;
		case 69: goto st760;
		case 87: goto st787;
		case 1034: goto tr2273;
		case 1083: goto tr2274;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2270;
	goto tr2215;
tr2270:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st756;
tr2278:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st756;
st756:
	if ( ++p == pe )
		goto _test_eof756;
case 756:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2275;
		case 32: goto tr2275;
		case 40: goto tr2276;
		case 41: goto tr2277;
		case 1034: goto tr2279;
		case 1083: goto tr2280;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2278;
	goto tr2215;
tr2282:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st757;
tr2283:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st757;
tr2285:
	{
		s->line_counter++;
	}
	goto st757;
tr2450:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 757; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st757;
tr2275:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st757;
tr2276:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st757;
tr2277:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st757;
tr2279:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st757;
st757:
	if ( ++p == pe )
		goto _test_eof757;
case 757:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st757;
		case 32: goto st757;
		case 40: goto tr2282;
		case 41: goto tr2283;
		case 69: goto st760;
		case 87: goto st787;
		case 1034: goto tr2285;
		case 1083: goto tr2286;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2284;
	goto tr2215;
tr2291:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st758;
tr2284:
	{
		s->decimals = 3;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st758;
st758:
	if ( ++p == pe )
		goto _test_eof758;
case 758:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2287;
		case 32: goto tr2287;
		case 40: goto tr2288;
		case 41: goto tr2289;
		case 46: goto st789;
		case 1034: goto tr2292;
		case 1083: goto tr2293;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2291;
	goto tr2215;
tr2295:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st759;
tr2296:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st759;
tr2297:
	{
		s->line_counter++;
	}
	goto st759;
tr2441:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 759; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st759;
tr2287:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st759;
tr2288:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st759;
tr2289:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st759;
tr2292:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st759;
tr2442:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st759;
tr2443:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st759;
tr2444:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st759;
tr2446:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st759;
st759:
	if ( ++p == pe )
		goto _test_eof759;
case 759:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st759;
		case 32: goto st759;
		case 40: goto tr2295;
		case 41: goto tr2296;
		case 69: goto st760;
		case 87: goto st787;
		case 1034: goto tr2297;
		case 1083: goto tr2298;
	}
	goto tr2215;
st760:
	if ( ++p == pe )
		goto _test_eof760;
case 760:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st761;
		case 32: goto st761;
		case 40: goto tr2300;
		case 41: goto tr2301;
		case 1034: goto tr2302;
		case 1083: goto tr2303;
	}
	goto tr2215;
tr2300:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st761;
tr2301:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st761;
tr2302:
	{
		s->line_counter++;
	}
	goto st761;
tr2434:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 761; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st761;
tr2435:
	{
		s->loc.long_sign = -1;
	}
	goto st761;
tr2436:
	{
		s->loc.long_sign = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st761;
tr2437:
	{
		s->loc.long_sign = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st761;
tr2438:
	{
		s->loc.long_sign = -1;
	}
	{
		s->line_counter++;
	}
	goto st761;
st761:
	if ( ++p == pe )
		goto _test_eof761;
case 761:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st761;
		case 32: goto st761;
		case 40: goto tr2300;
		case 41: goto tr2301;
		case 45: goto st762;
		case 1034: goto tr2302;
		case 1083: goto tr2303;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2305;
	goto tr2215;
st762:
	if ( ++p == pe )
		goto _test_eof762;
case 762:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2306;
	goto tr2215;
tr2311:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st763;
tr2305:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st763;
tr2306:
	{
		s->loc.alt_sign = -1;
	}
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st763;
st763:
	if ( ++p == pe )
		goto _test_eof763;
case 763:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2307;
		case 32: goto tr2307;
		case 40: goto tr2308;
		case 41: goto tr2309;
		case 46: goto st783;
		case 109: goto tr2312;
		case 2058: goto tr2313;
		case 2107: goto tr2314;
		case 2314: goto tr2315;
		case 2363: goto tr2315;
		case 2570: goto tr2316;
		case 2619: goto tr2317;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2311;
	goto tr2215;
tr2319:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st764;
tr2320:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st764;
tr2322:
	{
		s->line_counter++;
	}
	goto st764;
tr2421:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 764; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st764;
tr2307:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st764;
tr2308:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st764;
tr2309:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st764;
tr2313:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st764;
tr2422:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st764;
tr2423:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st764;
tr2424:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st764;
tr2427:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st764;
st764:
	if ( ++p == pe )
		goto _test_eof764;
case 764:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st764;
		case 32: goto st764;
		case 40: goto tr2319;
		case 41: goto tr2320;
		case 2058: goto tr2322;
		case 2107: goto tr2323;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2325;
		case 2619: goto tr2326;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2321;
	goto tr2215;
tr2331:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st765;
tr2321:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st765;
st765:
	if ( ++p == pe )
		goto _test_eof765;
case 765:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2327;
		case 32: goto tr2327;
		case 40: goto tr2328;
		case 41: goto tr2329;
		case 46: goto st779;
		case 109: goto tr2332;
		case 2058: goto tr2333;
		case 2107: goto tr2334;
		case 2314: goto tr2335;
		case 2363: goto tr2335;
		case 2570: goto tr2336;
		case 2619: goto tr2337;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2331;
	goto tr2215;
tr2339:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st766;
tr2340:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st766;
tr2342:
	{
		s->line_counter++;
	}
	goto st766;
tr2408:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 766; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st766;
tr2327:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st766;
tr2328:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st766;
tr2329:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st766;
tr2333:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st766;
tr2409:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st766;
tr2410:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st766;
tr2411:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st766;
tr2414:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st766;
st766:
	if ( ++p == pe )
		goto _test_eof766;
case 766:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st766;
		case 32: goto st766;
		case 40: goto tr2339;
		case 41: goto tr2340;
		case 2058: goto tr2342;
		case 2107: goto tr2343;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2344;
		case 2619: goto tr2345;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2341;
	goto tr2215;
tr2350:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st767;
tr2341:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st767;
st767:
	if ( ++p == pe )
		goto _test_eof767;
case 767:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2346;
		case 32: goto tr2346;
		case 40: goto tr2347;
		case 41: goto tr2348;
		case 46: goto st775;
		case 109: goto tr2351;
		case 2058: goto tr2352;
		case 2107: goto tr2353;
		case 2314: goto tr2354;
		case 2363: goto tr2354;
		case 2570: goto tr2355;
		case 2619: goto tr2356;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2350;
	goto tr2215;
tr2358:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st768;
tr2359:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st768;
tr2361:
	{
		s->line_counter++;
	}
	goto st768;
tr2395:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 768; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st768;
tr2346:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st768;
tr2347:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st768;
tr2348:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st768;
tr2352:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st768;
tr2396:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st768;
tr2397:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st768;
tr2398:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st768;
tr2401:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st768;
st768:
	if ( ++p == pe )
		goto _test_eof768;
case 768:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st768;
		case 32: goto st768;
		case 40: goto tr2358;
		case 41: goto tr2359;
		case 2058: goto tr2361;
		case 2107: goto tr2362;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2363;
		case 2619: goto tr2364;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2360;
	goto tr2215;
tr2369:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st769;
tr2360:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st769;
st769:
	if ( ++p == pe )
		goto _test_eof769;
case 769:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2365;
		case 32: goto tr2365;
		case 40: goto tr2366;
		case 41: goto tr2367;
		case 46: goto st772;
		case 109: goto tr2365;
		case 2058: goto tr2370;
		case 2107: goto tr2371;
		case 2314: goto tr2372;
		case 2363: goto tr2372;
		case 2570: goto tr2373;
		case 2619: goto tr2374;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2369;
	goto tr2215;
tr2376:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st770;
tr2377:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st770;
tr2378:
	{
		s->line_counter++;
	}
	goto st770;
tr2383:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 770; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st770;
tr2365:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st770;
tr2366:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st770;
tr2367:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st770;
tr2370:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st770;
tr2384:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st770;
tr2385:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st770;
tr2386:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st770;
tr2388:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st770;
st770:
	if ( ++p == pe )
		goto _test_eof770;
case 770:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st770;
		case 32: goto st770;
		case 40: goto tr2376;
		case 41: goto tr2377;
		case 2058: goto tr2378;
		case 2107: goto tr2379;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2380;
		case 2619: goto tr2381;
	}
	goto tr2215;
tr2379:
	{
		s->buffer_length = 0;
	}
	goto st771;
tr2382:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st771;
tr2371:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st771;
tr2389:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st771;
st771:
	if ( ++p == pe )
		goto _test_eof771;
case 771:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2383;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2382;
	goto tr2215;
tr2315:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2324:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2335:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2354:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2372:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2390:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2403:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2416:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
tr2429:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1271;
st1271:
	if ( ++p == pe )
		goto _test_eof1271;
case 1271:
	goto st0;
tr2380:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1272;
tr2373:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1272;
tr2391:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1272;
st1272:
	if ( ++p == pe )
		goto _test_eof1272;
case 1272:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st770;
		case 32: goto st770;
		case 40: goto tr2376;
		case 41: goto tr2377;
		case 2058: goto tr2378;
		case 2107: goto tr2379;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2380;
		case 2619: goto tr2381;
	}
	goto tr2215;
tr2381:
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1273;
tr2374:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1273;
tr2392:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1273;
st1273:
	if ( ++p == pe )
		goto _test_eof1273;
case 1273:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2383;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2382;
	goto tr2215;
st772:
	if ( ++p == pe )
		goto _test_eof772;
case 772:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2384;
		case 32: goto tr2384;
		case 40: goto tr2385;
		case 41: goto tr2386;
		case 109: goto tr2384;
		case 2058: goto tr2388;
		case 2107: goto tr2389;
		case 2314: goto tr2390;
		case 2363: goto tr2390;
		case 2570: goto tr2391;
		case 2619: goto tr2392;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2387;
	goto tr2215;
tr2387:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st773;
tr2393:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st773;
st773:
	if ( ++p == pe )
		goto _test_eof773;
case 773:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2365;
		case 32: goto tr2365;
		case 40: goto tr2366;
		case 41: goto tr2367;
		case 109: goto tr2365;
		case 2058: goto tr2370;
		case 2107: goto tr2371;
		case 2314: goto tr2372;
		case 2363: goto tr2372;
		case 2570: goto tr2373;
		case 2619: goto tr2374;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2393;
	goto tr2215;
tr2362:
	{
		s->buffer_length = 0;
	}
	goto st774;
tr2394:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st774;
tr2353:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st774;
tr2402:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st774;
st774:
	if ( ++p == pe )
		goto _test_eof774;
case 774:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2395;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2394;
	goto tr2215;
tr2363:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1274;
tr2355:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1274;
tr2404:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1274;
st1274:
	if ( ++p == pe )
		goto _test_eof1274;
case 1274:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st768;
		case 32: goto st768;
		case 40: goto tr2358;
		case 41: goto tr2359;
		case 2058: goto tr2361;
		case 2107: goto tr2362;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2363;
		case 2619: goto tr2364;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2360;
	goto tr2215;
tr2364:
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1275;
tr2356:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1275;
tr2405:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1275;
st1275:
	if ( ++p == pe )
		goto _test_eof1275;
case 1275:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2395;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2394;
	goto tr2215;
st775:
	if ( ++p == pe )
		goto _test_eof775;
case 775:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2396;
		case 32: goto tr2396;
		case 40: goto tr2397;
		case 41: goto tr2398;
		case 109: goto tr2400;
		case 2058: goto tr2401;
		case 2107: goto tr2402;
		case 2314: goto tr2403;
		case 2363: goto tr2403;
		case 2570: goto tr2404;
		case 2619: goto tr2405;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2399;
	goto tr2215;
tr2399:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st776;
tr2406:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st776;
st776:
	if ( ++p == pe )
		goto _test_eof776;
case 776:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2346;
		case 32: goto tr2346;
		case 40: goto tr2347;
		case 41: goto tr2348;
		case 109: goto tr2351;
		case 2058: goto tr2352;
		case 2107: goto tr2353;
		case 2314: goto tr2354;
		case 2363: goto tr2354;
		case 2570: goto tr2355;
		case 2619: goto tr2356;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2406;
	goto tr2215;
tr2351:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st777;
tr2400:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st777;
st777:
	if ( ++p == pe )
		goto _test_eof777;
case 777:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st768;
		case 32: goto st768;
		case 40: goto tr2358;
		case 41: goto tr2359;
		case 2058: goto tr2361;
		case 2107: goto tr2362;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2363;
		case 2619: goto tr2364;
	}
	goto tr2215;
tr2343:
	{
		s->buffer_length = 0;
	}
	goto st778;
tr2407:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st778;
tr2334:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st778;
tr2415:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st778;
st778:
	if ( ++p == pe )
		goto _test_eof778;
case 778:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2408;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2407;
	goto tr2215;
tr2344:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1276;
tr2336:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1276;
tr2417:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1276;
st1276:
	if ( ++p == pe )
		goto _test_eof1276;
case 1276:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st766;
		case 32: goto st766;
		case 40: goto tr2339;
		case 41: goto tr2340;
		case 2058: goto tr2342;
		case 2107: goto tr2343;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2344;
		case 2619: goto tr2345;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2341;
	goto tr2215;
tr2345:
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1277;
tr2337:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1277;
tr2418:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1277;
st1277:
	if ( ++p == pe )
		goto _test_eof1277;
case 1277:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2408;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2407;
	goto tr2215;
st779:
	if ( ++p == pe )
		goto _test_eof779;
case 779:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2409;
		case 32: goto tr2409;
		case 40: goto tr2410;
		case 41: goto tr2411;
		case 109: goto tr2413;
		case 2058: goto tr2414;
		case 2107: goto tr2415;
		case 2314: goto tr2416;
		case 2363: goto tr2416;
		case 2570: goto tr2417;
		case 2619: goto tr2418;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2412;
	goto tr2215;
tr2412:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st780;
tr2419:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st780;
st780:
	if ( ++p == pe )
		goto _test_eof780;
case 780:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2327;
		case 32: goto tr2327;
		case 40: goto tr2328;
		case 41: goto tr2329;
		case 109: goto tr2332;
		case 2058: goto tr2333;
		case 2107: goto tr2334;
		case 2314: goto tr2335;
		case 2363: goto tr2335;
		case 2570: goto tr2336;
		case 2619: goto tr2337;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2419;
	goto tr2215;
tr2332:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st781;
tr2413:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st781;
st781:
	if ( ++p == pe )
		goto _test_eof781;
case 781:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st766;
		case 32: goto st766;
		case 40: goto tr2339;
		case 41: goto tr2340;
		case 2058: goto tr2342;
		case 2107: goto tr2343;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2344;
		case 2619: goto tr2345;
	}
	goto tr2215;
tr2323:
	{
		s->buffer_length = 0;
	}
	goto st782;
tr2420:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st782;
tr2314:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st782;
tr2428:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st782;
st782:
	if ( ++p == pe )
		goto _test_eof782;
case 782:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2421;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2420;
	goto tr2215;
tr2316:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1278;
tr2325:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1278;
tr2430:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1278;
st1278:
	if ( ++p == pe )
		goto _test_eof1278;
case 1278:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st764;
		case 32: goto st764;
		case 40: goto tr2319;
		case 41: goto tr2320;
		case 2058: goto tr2322;
		case 2107: goto tr2323;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2325;
		case 2619: goto tr2326;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2321;
	goto tr2215;
tr2317:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1279;
tr2326:
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1279;
tr2431:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1279;
st1279:
	if ( ++p == pe )
		goto _test_eof1279;
case 1279:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2421;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2420;
	goto tr2215;
st783:
	if ( ++p == pe )
		goto _test_eof783;
case 783:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2422;
		case 32: goto tr2422;
		case 40: goto tr2423;
		case 41: goto tr2424;
		case 109: goto tr2426;
		case 2058: goto tr2427;
		case 2107: goto tr2428;
		case 2314: goto tr2429;
		case 2363: goto tr2429;
		case 2570: goto tr2430;
		case 2619: goto tr2431;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2425;
	goto tr2215;
tr2425:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st784;
tr2432:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st784;
st784:
	if ( ++p == pe )
		goto _test_eof784;
case 784:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2307;
		case 32: goto tr2307;
		case 40: goto tr2308;
		case 41: goto tr2309;
		case 109: goto tr2312;
		case 2058: goto tr2313;
		case 2107: goto tr2314;
		case 2314: goto tr2315;
		case 2363: goto tr2315;
		case 2570: goto tr2316;
		case 2619: goto tr2317;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2432;
	goto tr2215;
tr2312:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st785;
tr2426:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	goto st785;
st785:
	if ( ++p == pe )
		goto _test_eof785;
case 785:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st764;
		case 32: goto st764;
		case 40: goto tr2319;
		case 41: goto tr2320;
		case 2058: goto tr2322;
		case 2107: goto tr2323;
		case 2314: goto tr2324;
		case 2363: goto tr2324;
		case 2570: goto tr2325;
		case 2619: goto tr2326;
	}
	goto tr2215;
tr2303:
	{
		s->buffer_length = 0;
	}
	goto st786;
tr2433:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st786;
tr2439:
	{
		s->loc.long_sign = -1;
	}
	{
		s->buffer_length = 0;
	}
	goto st786;
st786:
	if ( ++p == pe )
		goto _test_eof786;
case 786:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2434;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2433;
	goto tr2215;
st787:
	if ( ++p == pe )
		goto _test_eof787;
case 787:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2435;
		case 32: goto tr2435;
		case 40: goto tr2436;
		case 41: goto tr2437;
		case 1034: goto tr2438;
		case 1083: goto tr2439;
	}
	goto tr2215;
tr2298:
	{
		s->buffer_length = 0;
	}
	goto st788;
tr2440:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st788;
tr2293:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st788;
tr2447:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st788;
st788:
	if ( ++p == pe )
		goto _test_eof788;
case 788:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2441;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2440;
	goto tr2215;
st789:
	if ( ++p == pe )
		goto _test_eof789;
case 789:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2442;
		case 32: goto tr2442;
		case 40: goto tr2443;
		case 41: goto tr2444;
		case 1034: goto tr2446;
		case 1083: goto tr2447;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2445;
	goto tr2215;
tr2445:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st790;
tr2448:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st790;
st790:
	if ( ++p == pe )
		goto _test_eof790;
case 790:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2287;
		case 32: goto tr2287;
		case 40: goto tr2288;
		case 41: goto tr2289;
		case 1034: goto tr2292;
		case 1083: goto tr2293;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2448;
	goto tr2215;
tr2286:
	{
		s->buffer_length = 0;
	}
	goto st791;
tr2449:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st791;
tr2280:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st791;
st791:
	if ( ++p == pe )
		goto _test_eof791;
case 791:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2450;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2449;
	goto tr2215;
tr2274:
	{
		s->buffer_length = 0;
	}
	goto st792;
tr2451:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st792;
tr2266:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st792;
st792:
	if ( ++p == pe )
		goto _test_eof792;
case 792:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2452;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2451;
	goto tr2215;
tr2259:
	{
		s->buffer_length = 0;
	}
	goto st793;
tr2453:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st793;
tr2459:
	{
		s->loc.lat_sign = -1;
	}
	{
		s->buffer_length = 0;
	}
	goto st793;
st793:
	if ( ++p == pe )
		goto _test_eof793;
case 793:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2454;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2453;
	goto tr2215;
st794:
	if ( ++p == pe )
		goto _test_eof794;
case 794:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2455;
		case 32: goto tr2455;
		case 40: goto tr2456;
		case 41: goto tr2457;
		case 1034: goto tr2458;
		case 1083: goto tr2459;
	}
	goto tr2215;
tr2254:
	{
		s->buffer_length = 0;
	}
	goto st795;
tr2460:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st795;
tr2249:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st795;
tr2467:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st795;
st795:
	if ( ++p == pe )
		goto _test_eof795;
case 795:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2461;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2460;
	goto tr2215;
st796:
	if ( ++p == pe )
		goto _test_eof796;
case 796:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2462;
		case 32: goto tr2462;
		case 40: goto tr2463;
		case 41: goto tr2464;
		case 1034: goto tr2466;
		case 1083: goto tr2467;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2465;
	goto tr2215;
tr2465:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st797;
tr2468:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st797;
st797:
	if ( ++p == pe )
		goto _test_eof797;
case 797:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2243;
		case 32: goto tr2243;
		case 40: goto tr2244;
		case 41: goto tr2245;
		case 1034: goto tr2248;
		case 1083: goto tr2249;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2468;
	goto tr2215;
tr2242:
	{
		s->buffer_length = 0;
	}
	goto st798;
tr2469:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st798;
tr2236:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st798;
st798:
	if ( ++p == pe )
		goto _test_eof798;
case 798:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2470;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2469;
	goto tr2215;
tr2230:
	{
		s->buffer_length = 0;
	}
	goto st799;
tr2471:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st799;
tr2222:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st799;
st799:
	if ( ++p == pe )
		goto _test_eof799;
case 799:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2472;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2471;
	goto tr2215;
st800:
	if ( ++p == pe )
		goto _test_eof800;
case 800:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2473;
	goto tr2007;
tr2473:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st801;
tr2477:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st801;
st801:
	if ( ++p == pe )
		goto _test_eof801;
case 801:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2474;
		case 32: goto tr2474;
		case 40: goto tr2475;
		case 41: goto tr2476;
		case 1034: goto tr2478;
		case 1083: goto tr2479;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2477;
	goto tr2007;
tr2481:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st802;
tr2482:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st802;
tr2484:
	{
		s->line_counter++;
	}
	goto st802;
tr2516:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 802; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st802;
tr2474:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st802;
tr2475:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st802;
tr2476:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st802;
tr2478:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st802;
st802:
	if ( ++p == pe )
		goto _test_eof802;
case 802:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st802;
		case 32: goto st802;
		case 40: goto tr2481;
		case 41: goto tr2482;
		case 1034: goto tr2484;
		case 1083: goto tr2485;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2483;
	goto tr2007;
tr2483:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st803;
tr2489:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st803;
st803:
	if ( ++p == pe )
		goto _test_eof803;
case 803:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2486;
		case 32: goto tr2486;
		case 40: goto tr2487;
		case 41: goto tr2488;
		case 1034: goto tr2490;
		case 1083: goto tr2491;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2489;
	goto tr2007;
tr2493:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st804;
tr2494:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st804;
tr2496:
	{
		s->line_counter++;
	}
	goto st804;
tr2514:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 804; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st804;
tr2486:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st804;
tr2487:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st804;
tr2488:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st804;
tr2490:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st804;
st804:
	if ( ++p == pe )
		goto _test_eof804;
case 804:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st804;
		case 32: goto st804;
		case 40: goto tr2493;
		case 41: goto tr2494;
		case 1034: goto tr2496;
		case 1083: goto tr2497;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2495;
	goto tr2007;
tr2495:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st805;
tr2501:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st805;
st805:
	if ( ++p == pe )
		goto _test_eof805;
case 805:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2498;
		case 32: goto tr2498;
		case 40: goto tr2499;
		case 41: goto tr2500;
		case 1034: goto tr2502;
		case 1083: goto tr2503;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2501;
	goto tr2007;
tr2505:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st806;
tr2506:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st806;
tr2508:
	{
		s->line_counter++;
	}
	goto st806;
tr2512:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 806; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st806;
tr2498:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st806;
tr2499:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st806;
tr2500:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st806;
tr2502:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st806;
st806:
	if ( ++p == pe )
		goto _test_eof806;
case 806:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st806;
		case 32: goto st806;
		case 40: goto tr2505;
		case 41: goto tr2506;
		case 42: goto tr2507;
		case 92: goto tr2507;
		case 95: goto tr2507;
		case 1034: goto tr2508;
		case 1083: goto tr2509;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2507;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2507;
	} else
		goto tr2507;
	goto tr75;
tr2507:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 807;goto st301;} }
	goto st807;
st807:
	if ( ++p == pe )
		goto _test_eof807;
case 807:
	switch( (*p) ) {
		case 32: goto tr2510;
		case 59: goto tr2510;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2510;
	} else if ( (*p) >= 9 )
		goto tr2510;
	goto tr75;
tr2510:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1280;
st1280:
	if ( ++p == pe )
		goto _test_eof1280;
case 1280:
	goto st0;
tr2509:
	{
		s->buffer_length = 0;
	}
	goto st808;
tr2511:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st808;
tr2503:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st808;
st808:
	if ( ++p == pe )
		goto _test_eof808;
case 808:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2512;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2511;
	goto tr75;
tr2497:
	{
		s->buffer_length = 0;
	}
	goto st809;
tr2513:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st809;
tr2491:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st809;
st809:
	if ( ++p == pe )
		goto _test_eof809;
case 809:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2514;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2513;
	goto tr75;
tr2485:
	{
		s->buffer_length = 0;
	}
	goto st810;
tr2515:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st810;
tr2479:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st810;
st810:
	if ( ++p == pe )
		goto _test_eof810;
case 810:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2516;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2515;
	goto tr75;
st811:
	if ( ++p == pe )
		goto _test_eof811;
case 811:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2517;
	goto tr2007;
tr2517:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st812;
tr2521:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st812;
st812:
	if ( ++p == pe )
		goto _test_eof812;
case 812:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2518;
		case 32: goto tr2518;
		case 40: goto tr2519;
		case 41: goto tr2520;
		case 1034: goto tr2522;
		case 1083: goto tr2523;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2521;
	goto tr2007;
tr2525:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st813;
tr2526:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st813;
tr2528:
	{
		s->line_counter++;
	}
	goto st813;
tr2585:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 813; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st813;
tr2518:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st813;
tr2519:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st813;
tr2520:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st813;
tr2522:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st813;
st813:
	if ( ++p == pe )
		goto _test_eof813;
case 813:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st813;
		case 32: goto st813;
		case 40: goto tr2525;
		case 41: goto tr2526;
		case 1034: goto tr2528;
		case 1083: goto tr2529;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2527;
	goto tr2007;
tr2527:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st814;
tr2533:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st814;
st814:
	if ( ++p == pe )
		goto _test_eof814;
case 814:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2530;
		case 32: goto tr2530;
		case 40: goto tr2531;
		case 41: goto tr2532;
		case 1034: goto tr2534;
		case 1083: goto tr2535;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2533;
	goto tr2007;
tr2538:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st815;
tr2539:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st815;
tr2540:
	{
		s->line_counter++;
	}
	goto st815;
tr2583:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 815; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st815;
tr2530:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st815;
tr2531:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st815;
tr2532:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st815;
tr2534:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st815;
st815:
	if ( ++p == pe )
		goto _test_eof815;
case 815:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st815;
		case 32: goto st815;
		case 40: goto tr2538;
		case 41: goto tr2539;
		case 1034: goto tr2540;
		case 1083: goto tr2541;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2536;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2536;
	} else
		goto tr2536;
	goto tr75;
tr2536:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 816;goto st310;} }
	goto st816;
st816:
	if ( ++p == pe )
		goto _test_eof816;
case 816:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2542;
		case 32: goto tr2542;
		case 40: goto tr2543;
		case 41: goto tr2544;
		case 1034: goto tr2545;
		case 1083: goto tr2546;
	}
	goto tr75;
tr2549:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st817;
tr2550:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st817;
tr2551:
	{
		s->line_counter++;
	}
	goto st817;
tr2581:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 817; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st817;
tr2542:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st817;
tr2543:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st817;
tr2544:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st817;
tr2545:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st817;
st817:
	if ( ++p == pe )
		goto _test_eof817;
case 817:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st817;
		case 32: goto st817;
		case 40: goto tr2549;
		case 41: goto tr2550;
		case 1034: goto tr2551;
		case 1083: goto tr2552;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2547;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2547;
	} else
		goto tr2547;
	goto tr75;
tr2547:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 818;goto st310;} }
	goto st818;
st818:
	if ( ++p == pe )
		goto _test_eof818;
case 818:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2553;
		case 32: goto tr2553;
		case 40: goto tr2554;
		case 41: goto tr2555;
		case 1034: goto tr2556;
		case 1083: goto tr2557;
	}
	goto tr75;
tr2560:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st819;
tr2561:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st819;
tr2562:
	{
		s->line_counter++;
	}
	goto st819;
tr2579:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 819; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st819;
tr2553:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st819;
tr2554:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st819;
tr2555:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st819;
tr2556:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st819;
st819:
	if ( ++p == pe )
		goto _test_eof819;
case 819:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st819;
		case 32: goto st819;
		case 40: goto tr2560;
		case 41: goto tr2561;
		case 1034: goto tr2562;
		case 1083: goto tr2563;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2558;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2558;
	} else
		goto tr2558;
	goto tr75;
tr2558:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 820;goto st310;} }
	goto st820;
st820:
	if ( ++p == pe )
		goto _test_eof820;
case 820:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2564;
		case 32: goto tr2564;
		case 40: goto tr2565;
		case 41: goto tr2566;
		case 1034: goto tr2567;
		case 1083: goto tr2568;
	}
	goto tr75;
tr2570:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st821;
tr2571:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st821;
tr2573:
	{
		s->line_counter++;
	}
	goto st821;
tr2577:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 821; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st821;
tr2564:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st821;
tr2565:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st821;
tr2566:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st821;
tr2567:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st821;
st821:
	if ( ++p == pe )
		goto _test_eof821;
case 821:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st821;
		case 32: goto st821;
		case 40: goto tr2570;
		case 41: goto tr2571;
		case 42: goto tr2572;
		case 92: goto tr2572;
		case 95: goto tr2572;
		case 1034: goto tr2573;
		case 1083: goto tr2574;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2572;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2572;
	} else
		goto tr2572;
	goto tr75;
tr2572:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 822;goto st301;} }
	goto st822;
st822:
	if ( ++p == pe )
		goto _test_eof822;
case 822:
	switch( (*p) ) {
		case 32: goto tr2575;
		case 59: goto tr2575;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2575;
	} else if ( (*p) >= 9 )
		goto tr2575;
	goto tr75;
tr2575:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1281;
st1281:
	if ( ++p == pe )
		goto _test_eof1281;
case 1281:
	goto st0;
tr2574:
	{
		s->buffer_length = 0;
	}
	goto st823;
tr2576:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st823;
tr2568:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st823;
st823:
	if ( ++p == pe )
		goto _test_eof823;
case 823:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2577;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2576;
	goto tr75;
tr2563:
	{
		s->buffer_length = 0;
	}
	goto st824;
tr2578:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st824;
tr2557:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st824;
st824:
	if ( ++p == pe )
		goto _test_eof824;
case 824:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2579;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2578;
	goto tr75;
tr2552:
	{
		s->buffer_length = 0;
	}
	goto st825;
tr2580:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st825;
tr2546:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st825;
st825:
	if ( ++p == pe )
		goto _test_eof825;
case 825:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2581;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2580;
	goto tr75;
tr2541:
	{
		s->buffer_length = 0;
	}
	goto st826;
tr2582:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st826;
tr2535:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st826;
st826:
	if ( ++p == pe )
		goto _test_eof826;
case 826:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2583;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2582;
	goto tr75;
tr2529:
	{
		s->buffer_length = 0;
	}
	goto st827;
tr2584:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st827;
tr2523:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st827;
st827:
	if ( ++p == pe )
		goto _test_eof827;
case 827:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2585;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2584;
	goto tr75;
st828:
	if ( ++p == pe )
		goto _test_eof828;
case 828:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2586;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2586;
	} else
		goto tr2586;
	goto tr75;
tr2586:
	{ p--; {stack[top++] = 829;goto st647;} }
	goto st829;
st829:
	if ( ++p == pe )
		goto _test_eof829;
case 829:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st830;
		case 32: goto st830;
		case 40: goto tr2588;
		case 41: goto tr2589;
		case 1034: goto tr2590;
		case 1083: goto tr2591;
	}
	goto tr75;
tr2588:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st830;
tr2589:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st830;
tr2590:
	{
		s->line_counter++;
	}
	goto st830;
tr2617:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 830; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st830;
st830:
	if ( ++p == pe )
		goto _test_eof830;
case 830:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st830;
		case 32: goto st830;
		case 40: goto tr2588;
		case 41: goto tr2589;
		case 1034: goto tr2590;
		case 1083: goto tr2591;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2592;
	goto tr2007;
tr2592:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st831;
tr2596:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st831;
st831:
	if ( ++p == pe )
		goto _test_eof831;
case 831:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2593;
		case 32: goto tr2593;
		case 40: goto tr2594;
		case 41: goto tr2595;
		case 1034: goto tr2597;
		case 1083: goto tr2598;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2596;
	goto tr2007;
tr2600:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st832;
tr2601:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st832;
tr2603:
	{
		s->line_counter++;
	}
	goto st832;
tr2615:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 832; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st832;
tr2593:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st832;
tr2594:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st832;
tr2595:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st832;
tr2597:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st832;
st832:
	if ( ++p == pe )
		goto _test_eof832;
case 832:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st832;
		case 32: goto st832;
		case 40: goto tr2600;
		case 41: goto tr2601;
		case 1034: goto tr2603;
		case 1083: goto tr2604;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2602;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2602;
	} else
		goto tr2602;
	goto tr75;
tr2602:
	{ p--; {stack[top++] = 833;goto st543;} }
	goto st833;
st833:
	if ( ++p == pe )
		goto _test_eof833;
case 833:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st834;
		case 32: goto st834;
		case 40: goto tr2606;
		case 41: goto tr2607;
		case 1034: goto tr2608;
		case 1083: goto tr2609;
	}
	goto tr75;
tr2606:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st834;
tr2607:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st834;
tr2608:
	{
		s->line_counter++;
	}
	goto st834;
tr2613:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 834; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st834;
st834:
	if ( ++p == pe )
		goto _test_eof834;
case 834:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st834;
		case 32: goto st834;
		case 40: goto tr2606;
		case 41: goto tr2607;
		case 43: goto tr2610;
		case 1034: goto tr2608;
		case 1083: goto tr2609;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2610;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2610;
	} else
		goto tr2610;
	goto tr75;
tr2610:
	{ p--; {stack[top++] = 835;goto st360;} }
	goto st835;
st835:
	if ( ++p == pe )
		goto _test_eof835;
case 835:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2611;
		case 1595: goto tr2611;
	}
	goto tr75;
tr2611:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1282;
st1282:
	if ( ++p == pe )
		goto _test_eof1282;
case 1282:
	goto st0;
tr2609:
	{
		s->buffer_length = 0;
	}
	goto st836;
tr2612:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st836;
st836:
	if ( ++p == pe )
		goto _test_eof836;
case 836:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2613;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2612;
	goto tr75;
tr2604:
	{
		s->buffer_length = 0;
	}
	goto st837;
tr2614:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st837;
tr2598:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st837;
st837:
	if ( ++p == pe )
		goto _test_eof837;
case 837:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2615;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2614;
	goto tr75;
tr2591:
	{
		s->buffer_length = 0;
	}
	goto st838;
tr2616:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st838;
st838:
	if ( ++p == pe )
		goto _test_eof838;
case 838:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2617;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2616;
	goto tr75;
tr2621:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st839;
tr2622:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st839;
tr2625:
	{
		s->line_counter++;
	}
	goto st839;
tr2648:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 839; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st839;
tr2638:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st839;
tr2639:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st839;
tr2640:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st839;
tr2642:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->line_counter++;
	}
	goto st839;
st839:
	if ( ++p == pe )
		goto _test_eof839;
case 839:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st839;
		case 32: goto st839;
		case 33: goto tr2620;
		case 40: goto tr2621;
		case 41: goto tr2622;
		case 49: goto tr2623;
		case 50: goto tr2624;
		case 2058: goto tr2625;
		case 2107: goto tr2626;
		case 2314: goto tr2627;
		case 2363: goto tr2627;
		case 2570: goto tr2628;
		case 2619: goto tr2629;
	}
	goto tr2618;
tr2620:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.excl_flag = 128; // dec 128  = bin 10000000.
	}
	goto st840;
st840:
	if ( ++p == pe )
		goto _test_eof840;
case 840:
	switch( (*p) ) {
		case 49: goto tr2630;
		case 50: goto tr2631;
	}
	goto tr2618;
tr2623:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.addr_family = 1;
	}
	goto st841;
tr2630:
	{
		s->apl.addr_family = 1;
	}
	goto st841;
st841:
	if ( ++p == pe )
		goto _test_eof841;
case 841:
	if ( (*p) == 58 )
		goto st842;
	goto tr2618;
st842:
	if ( ++p == pe )
		goto _test_eof842;
case 842:
	if ( (*p) == 46 )
		goto tr2634;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2634;
	goto tr2633;
tr2634:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st843;
tr2635:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st843;
st843:
	if ( ++p == pe )
		goto _test_eof843;
case 843:
	if ( (*p) == 47 )
		goto tr2636;
	if ( 46 <= (*p) && (*p) <= 57 )
		goto tr2635;
	goto tr2633;
tr2636:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	goto st844;
tr2652:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	goto st844;
st844:
	if ( ++p == pe )
		goto _test_eof844;
case 844:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2637;
	goto tr2618;
tr2637:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st845;
tr2641:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st845;
st845:
	if ( ++p == pe )
		goto _test_eof845;
case 845:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2638;
		case 32: goto tr2638;
		case 40: goto tr2639;
		case 41: goto tr2640;
		case 2058: goto tr2642;
		case 2107: goto tr2643;
		case 2314: goto tr2644;
		case 2363: goto tr2644;
		case 2570: goto tr2645;
		case 2619: goto tr2646;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2641;
	goto tr2618;
tr2626:
	{
		s->buffer_length = 0;
	}
	goto st846;
tr2647:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st846;
tr2643:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->buffer_length = 0;
	}
	goto st846;
st846:
	if ( ++p == pe )
		goto _test_eof846;
case 846:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2648;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2647;
	goto tr75;
tr2627:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1283;
tr2644:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1283;
st1283:
	if ( ++p == pe )
		goto _test_eof1283;
case 1283:
	goto st0;
tr2628:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1284;
tr2645:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1284;
st1284:
	if ( ++p == pe )
		goto _test_eof1284;
case 1284:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st839;
		case 32: goto st839;
		case 33: goto tr2620;
		case 40: goto tr2621;
		case 41: goto tr2622;
		case 49: goto tr2623;
		case 50: goto tr2624;
		case 2058: goto tr2625;
		case 2107: goto tr2626;
		case 2314: goto tr2627;
		case 2363: goto tr2627;
		case 2570: goto tr2628;
		case 2619: goto tr2629;
	}
	goto tr2618;
tr2624:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.addr_family = 2;
	}
	goto st847;
tr2631:
	{
		s->apl.addr_family = 2;
	}
	goto st847;
st847:
	if ( ++p == pe )
		goto _test_eof847;
case 847:
	if ( (*p) == 58 )
		goto st848;
	goto tr2618;
st848:
	if ( ++p == pe )
		goto _test_eof848;
case 848:
	if ( (*p) == 46 )
		goto tr2650;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr2650;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2650;
	} else
		goto tr2650;
	goto tr2633;
tr2650:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st849;
tr2651:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st849;
st849:
	if ( ++p == pe )
		goto _test_eof849;
case 849:
	if ( (*p) == 47 )
		goto tr2652;
	if ( (*p) < 65 ) {
		if ( 46 <= (*p) && (*p) <= 58 )
			goto tr2651;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2651;
	} else
		goto tr2651;
	goto tr2633;
tr2629:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1285;
tr2646:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
	}
	{
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = ZS_INET4_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		case 2:
			len = ZS_INET6_ADDR_LENGTH;
			memcpy(s->buffer, s->addr, len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st299;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Check for rdata overflow.
		if (rdata_tail + 4 + len > rdata_stop) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bits.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1285;
st1285:
	if ( ++p == pe )
		goto _test_eof1285;
case 1285:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2648;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2647;
	goto tr75;
st850:
	if ( ++p == pe )
		goto _test_eof850;
case 850:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2653;
	goto tr2007;
tr2653:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st851;
tr2657:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st851;
st851:
	if ( ++p == pe )
		goto _test_eof851;
case 851:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2654;
		case 32: goto tr2654;
		case 40: goto tr2655;
		case 41: goto tr2656;
		case 1034: goto tr2658;
		case 1083: goto tr2659;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2657;
	goto tr2007;
tr2661:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st852;
tr2662:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st852;
tr2664:
	{
		s->line_counter++;
	}
	goto st852;
tr2701:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 852; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st852;
tr2654:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st852;
tr2655:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st852;
tr2656:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st852;
tr2658:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st852;
st852:
	if ( ++p == pe )
		goto _test_eof852;
case 852:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st852;
		case 32: goto st852;
		case 40: goto tr2661;
		case 41: goto tr2662;
		case 1034: goto tr2664;
		case 1083: goto tr2665;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2663;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2663;
	} else
		goto tr2663;
	goto tr75;
tr2663:
	{ p--; {stack[top++] = 853;goto st543;} }
	goto st853;
st853:
	if ( ++p == pe )
		goto _test_eof853;
case 853:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st854;
		case 32: goto st854;
		case 40: goto tr2667;
		case 41: goto tr2668;
		case 1034: goto tr2669;
		case 1083: goto tr2670;
	}
	goto tr75;
tr2667:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st854;
tr2668:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st854;
tr2669:
	{
		s->line_counter++;
	}
	goto st854;
tr2699:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 854; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st854;
st854:
	if ( ++p == pe )
		goto _test_eof854;
case 854:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st854;
		case 32: goto st854;
		case 40: goto tr2667;
		case 41: goto tr2668;
		case 1034: goto tr2669;
		case 1083: goto tr2670;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2671;
	goto tr2007;
tr2671:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st855;
tr2675:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st855;
st855:
	if ( ++p == pe )
		goto _test_eof855;
case 855:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2672;
		case 32: goto tr2672;
		case 40: goto tr2673;
		case 41: goto tr2674;
		case 1034: goto tr2676;
		case 1083: goto tr2677;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2675;
	goto tr2007;
tr2680:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st856;
tr2681:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st856;
tr2683:
	{
		s->line_counter++;
	}
	goto st856;
tr2697:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 856; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st856;
tr2672:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st856;
tr2673:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st856;
tr2674:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st856;
tr2676:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st856;
st856:
	if ( ++p == pe )
		goto _test_eof856;
case 856:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st856;
		case 32: goto st856;
		case 40: goto tr2680;
		case 41: goto tr2681;
		case 1034: goto tr2683;
		case 1083: goto tr2684;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2682;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2682;
	} else
		goto tr2682;
	goto tr2678;
tr2682:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st857;
st857:
	if ( ++p == pe )
		goto _test_eof857;
case 857:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2685;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2685;
	} else
		goto tr2685;
	goto tr2678;
tr2687:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st858;
tr2688:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st858;
tr2689:
	{
		s->line_counter++;
	}
	goto st858;
tr2695:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 858; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st858;
tr2685:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st858;
st858:
	if ( ++p == pe )
		goto _test_eof858;
case 858:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st858;
		case 32: goto st858;
		case 40: goto tr2687;
		case 41: goto tr2688;
		case 2058: goto tr2689;
		case 2107: goto tr2690;
		case 2314: goto tr2691;
		case 2363: goto tr2691;
		case 2570: goto tr2692;
		case 2619: goto tr2693;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2682;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2682;
	} else
		goto tr2682;
	goto tr2678;
tr2690:
	{
		s->buffer_length = 0;
	}
	goto st859;
tr2694:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st859;
st859:
	if ( ++p == pe )
		goto _test_eof859;
case 859:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2695;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2694;
	goto tr2678;
tr2691:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1286;
st1286:
	if ( ++p == pe )
		goto _test_eof1286;
case 1286:
	goto st0;
tr2692:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1287;
st1287:
	if ( ++p == pe )
		goto _test_eof1287;
case 1287:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st858;
		case 32: goto st858;
		case 40: goto tr2687;
		case 41: goto tr2688;
		case 2058: goto tr2689;
		case 2107: goto tr2690;
		case 2314: goto tr2691;
		case 2363: goto tr2691;
		case 2570: goto tr2692;
		case 2619: goto tr2693;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2682;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2682;
	} else
		goto tr2682;
	goto tr2678;
tr2693:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1288;
st1288:
	if ( ++p == pe )
		goto _test_eof1288;
case 1288:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2695;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2694;
	goto tr2678;
tr2684:
	{
		s->buffer_length = 0;
	}
	goto st860;
tr2696:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st860;
tr2677:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st860;
st860:
	if ( ++p == pe )
		goto _test_eof860;
case 860:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2697;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2696;
	goto tr75;
tr2670:
	{
		s->buffer_length = 0;
	}
	goto st861;
tr2698:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st861;
st861:
	if ( ++p == pe )
		goto _test_eof861;
case 861:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2699;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2698;
	goto tr75;
tr2665:
	{
		s->buffer_length = 0;
	}
	goto st862;
tr2700:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st862;
tr2659:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st862;
st862:
	if ( ++p == pe )
		goto _test_eof862;
case 862:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2701;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2700;
	goto tr75;
st863:
	if ( ++p == pe )
		goto _test_eof863;
case 863:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2702;
	goto tr2007;
tr2702:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st864;
tr2706:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st864;
st864:
	if ( ++p == pe )
		goto _test_eof864;
case 864:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2703;
		case 32: goto tr2703;
		case 40: goto tr2704;
		case 41: goto tr2705;
		case 1034: goto tr2707;
		case 1083: goto tr2708;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2706;
	goto tr2007;
tr2710:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st865;
tr2711:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st865;
tr2713:
	{
		s->line_counter++;
	}
	goto st865;
tr2741:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 865; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st865;
tr2703:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st865;
tr2704:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st865;
tr2705:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st865;
tr2707:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st865;
st865:
	if ( ++p == pe )
		goto _test_eof865;
case 865:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st865;
		case 32: goto st865;
		case 40: goto tr2710;
		case 41: goto tr2711;
		case 1034: goto tr2713;
		case 1083: goto tr2714;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2712;
	goto tr2007;
tr2712:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st866;
tr2718:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st866;
st866:
	if ( ++p == pe )
		goto _test_eof866;
case 866:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2715;
		case 32: goto tr2715;
		case 40: goto tr2716;
		case 41: goto tr2717;
		case 1034: goto tr2719;
		case 1083: goto tr2720;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2718;
	goto tr2007;
tr2722:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st867;
tr2723:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st867;
tr2725:
	{
		s->line_counter++;
	}
	goto st867;
tr2739:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 867; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st867;
tr2715:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st867;
tr2716:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st867;
tr2717:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st867;
tr2719:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st867;
st867:
	if ( ++p == pe )
		goto _test_eof867;
case 867:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st867;
		case 32: goto st867;
		case 40: goto tr2722;
		case 41: goto tr2723;
		case 1034: goto tr2725;
		case 1083: goto tr2726;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2724;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2724;
	} else
		goto tr2724;
	goto tr2678;
tr2724:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st868;
st868:
	if ( ++p == pe )
		goto _test_eof868;
case 868:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2727;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2727;
	} else
		goto tr2727;
	goto tr2678;
tr2729:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st869;
tr2730:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st869;
tr2731:
	{
		s->line_counter++;
	}
	goto st869;
tr2737:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 869; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st869;
tr2727:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st869;
st869:
	if ( ++p == pe )
		goto _test_eof869;
case 869:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st869;
		case 32: goto st869;
		case 40: goto tr2729;
		case 41: goto tr2730;
		case 2058: goto tr2731;
		case 2107: goto tr2732;
		case 2314: goto tr2733;
		case 2363: goto tr2733;
		case 2570: goto tr2734;
		case 2619: goto tr2735;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2724;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2724;
	} else
		goto tr2724;
	goto tr2678;
tr2732:
	{
		s->buffer_length = 0;
	}
	goto st870;
tr2736:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st870;
st870:
	if ( ++p == pe )
		goto _test_eof870;
case 870:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2737;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2736;
	goto tr2678;
tr2733:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1289;
st1289:
	if ( ++p == pe )
		goto _test_eof1289;
case 1289:
	goto st0;
tr2734:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1290;
st1290:
	if ( ++p == pe )
		goto _test_eof1290;
case 1290:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st869;
		case 32: goto st869;
		case 40: goto tr2729;
		case 41: goto tr2730;
		case 2058: goto tr2731;
		case 2107: goto tr2732;
		case 2314: goto tr2733;
		case 2363: goto tr2733;
		case 2570: goto tr2734;
		case 2619: goto tr2735;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2724;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2724;
	} else
		goto tr2724;
	goto tr2678;
tr2735:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1291;
st1291:
	if ( ++p == pe )
		goto _test_eof1291;
case 1291:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2737;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2736;
	goto tr2678;
tr2726:
	{
		s->buffer_length = 0;
	}
	goto st871;
tr2738:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st871;
tr2720:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st871;
st871:
	if ( ++p == pe )
		goto _test_eof871;
case 871:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2739;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2738;
	goto tr75;
tr2714:
	{
		s->buffer_length = 0;
	}
	goto st872;
tr2740:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st872;
tr2708:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st872;
st872:
	if ( ++p == pe )
		goto _test_eof872;
case 872:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2741;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2740;
	goto tr75;
st873:
	if ( ++p == pe )
		goto _test_eof873;
case 873:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2742;
	goto tr2007;
tr2742:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st874;
tr2746:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st874;
st874:
	if ( ++p == pe )
		goto _test_eof874;
case 874:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2743;
		case 32: goto tr2743;
		case 40: goto tr2744;
		case 41: goto tr2745;
		case 1034: goto tr2747;
		case 1083: goto tr2748;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2746;
	goto tr2007;
tr2751:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st875;
tr2752:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st875;
tr2757:
	{
		s->line_counter++;
	}
	goto st875;
tr2950:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 875; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st875;
tr2743:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st875;
tr2744:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st875;
tr2745:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st875;
tr2747:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st875;
st875:
	if ( ++p == pe )
		goto _test_eof875;
case 875:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st875;
		case 32: goto st875;
		case 40: goto tr2751;
		case 41: goto tr2752;
		case 48: goto tr2753;
		case 49: goto tr2754;
		case 50: goto tr2755;
		case 51: goto tr2756;
		case 1034: goto tr2757;
		case 1083: goto tr2758;
	}
	goto tr2749;
tr2753:
	{
		*(rdata_tail++) = 0;
	}
	goto st876;
st876:
	if ( ++p == pe )
		goto _test_eof876;
case 876:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st877;
		case 32: goto st877;
		case 40: goto tr2760;
		case 41: goto tr2761;
		case 1034: goto tr2762;
		case 1083: goto tr2763;
	}
	goto tr2749;
tr2760:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st877;
tr2761:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st877;
tr2762:
	{
		s->line_counter++;
	}
	goto st877;
tr2812:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 877; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st877;
st877:
	if ( ++p == pe )
		goto _test_eof877;
case 877:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st877;
		case 32: goto st877;
		case 40: goto tr2760;
		case 41: goto tr2761;
		case 1034: goto tr2762;
		case 1083: goto tr2763;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2765;
	goto tr2764;
tr2765:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st878;
tr2769:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st878;
st878:
	if ( ++p == pe )
		goto _test_eof878;
case 878:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2766;
		case 32: goto tr2766;
		case 40: goto tr2767;
		case 41: goto tr2768;
		case 1034: goto tr2770;
		case 1083: goto tr2771;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2769;
	goto tr2764;
tr2773:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st879;
tr2774:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st879;
tr2776:
	{
		s->line_counter++;
	}
	goto st879;
tr2810:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 879; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st879;
tr2766:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st879;
tr2767:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st879;
tr2768:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st879;
tr2770:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st879;
st879:
	if ( ++p == pe )
		goto _test_eof879;
case 879:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st879;
		case 32: goto st879;
		case 40: goto tr2773;
		case 41: goto tr2774;
		case 46: goto st880;
		case 1034: goto tr2776;
		case 1083: goto tr2777;
	}
	goto tr2749;
st880:
	if ( ++p == pe )
		goto _test_eof880;
case 880:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto st881;
		case 6176: goto st881;
		case 6184: goto tr2780;
		case 6185: goto tr2781;
		case 6409: goto st884;
		case 6432: goto st884;
		case 6440: goto tr2783;
		case 6441: goto tr2784;
		case 6665: goto st886;
		case 6688: goto st886;
		case 6696: goto tr2786;
		case 6697: goto tr2787;
		case 9482: goto tr2788;
		case 9531: goto tr2788;
		case 9738: goto tr2788;
		case 9787: goto tr2788;
		case 10250: goto tr2789;
		case 10299: goto tr2790;
		case 10506: goto tr2788;
		case 10555: goto tr2788;
		case 10762: goto tr2791;
		case 10811: goto tr2792;
		case 11274: goto tr2793;
		case 11323: goto tr2794;
		case 11530: goto tr2788;
		case 11579: goto tr2788;
		case 11786: goto tr2795;
		case 11835: goto tr2796;
		case 12298: goto tr2797;
		case 12347: goto tr2798;
		case 12554: goto tr2788;
		case 12603: goto tr2788;
		case 12810: goto tr2799;
		case 12859: goto tr2800;
	}
	goto tr2778;
tr2780:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st881;
tr2781:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st881;
tr2789:
	{
		s->line_counter++;
	}
	goto st881;
tr2804:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 881; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st881;
tr2923:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st881;
tr2924:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st881;
tr2925:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st881;
tr2933:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st881;
tr2834:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	goto st881;
tr2835:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st881;
tr2836:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st881;
tr2844:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st881;
tr2879:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	goto st881;
tr2880:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st881;
tr2881:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st881;
tr2889:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st881;
st881:
	if ( ++p == pe )
		goto _test_eof881;
case 881:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3200 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2688 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st881;
		case 3104: goto st881;
		case 3112: goto tr2780;
		case 3113: goto tr2781;
		case 3115: goto tr2802;
		case 4106: goto tr2789;
		case 4155: goto tr2790;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2802;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2802;
	} else
		goto tr2802;
	goto tr2801;
tr2802:
	{ p--; {stack[top++] = 882;goto st360;} }
	goto st882;
st882:
	if ( ++p == pe )
		goto _test_eof882;
case 882:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2788;
		case 1595: goto tr2788;
	}
	goto tr2801;
tr2788:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1292;
tr2843:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1292;
tr2932:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1292;
tr2888:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1292;
st1292:
	if ( ++p == pe )
		goto _test_eof1292;
case 1292:
	goto st0;
tr2790:
	{
		s->buffer_length = 0;
	}
	goto st883;
tr2803:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st883;
tr2934:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st883;
tr2845:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	goto st883;
tr2890:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	goto st883;
st883:
	if ( ++p == pe )
		goto _test_eof883;
case 883:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3200 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2804;
	if ( 3968 <= _widec && _widec <= 4223 )
		goto tr2803;
	goto tr2801;
tr2783:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st884;
tr2784:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st884;
tr2793:
	{
		s->line_counter++;
	}
	goto st884;
tr2806:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 884; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st884;
tr2926:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st884;
tr2927:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st884;
tr2928:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st884;
tr2937:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st884;
tr2837:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	goto st884;
tr2838:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st884;
tr2839:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st884;
tr2848:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st884;
tr2882:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	goto st884;
tr2883:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st884;
tr2884:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st884;
tr2893:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st884;
st884:
	if ( ++p == pe )
		goto _test_eof884;
case 884:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4224 + ((*p) - -128));
				if (
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4224 + ((*p) - -128));
		if (
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st884;
		case 4640: goto st884;
		case 4648: goto tr2783;
		case 4649: goto tr2784;
		case 13578: goto tr2788;
		case 13627: goto tr2788;
		case 13834: goto tr2788;
		case 13883: goto tr2788;
		case 14346: goto tr2793;
		case 14395: goto tr2794;
		case 14602: goto tr2788;
		case 14651: goto tr2788;
		case 14858: goto tr2795;
		case 14907: goto tr2796;
	}
	goto tr2801;
tr2794:
	{
		s->buffer_length = 0;
	}
	goto st885;
tr2805:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st885;
tr2938:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st885;
tr2849:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	goto st885;
tr2894:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	goto st885;
st885:
	if ( ++p == pe )
		goto _test_eof885;
case 885:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4736 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2806;
	if ( 5504 <= _widec && _widec <= 5759 )
		goto tr2805;
	goto tr2801;
tr2795:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1293;
tr2850:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1293;
tr2895:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1293;
tr2939:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1293;
st1293:
	if ( ++p == pe )
		goto _test_eof1293;
case 1293:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4224 + ((*p) - -128));
				if (
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4224 + ((*p) - -128));
		if (
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st884;
		case 4640: goto st884;
		case 4648: goto tr2783;
		case 4649: goto tr2784;
		case 13578: goto tr2788;
		case 13627: goto tr2788;
		case 13834: goto tr2788;
		case 13883: goto tr2788;
		case 14346: goto tr2793;
		case 14395: goto tr2794;
		case 14602: goto tr2788;
		case 14651: goto tr2788;
		case 14858: goto tr2795;
		case 14907: goto tr2796;
	}
	goto tr2801;
tr2796:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1294;
tr2851:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1294;
tr2896:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1294;
tr2940:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1294;
st1294:
	if ( ++p == pe )
		goto _test_eof1294;
case 1294:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4736 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2806;
	if ( 5504 <= _widec && _widec <= 5759 )
		goto tr2805;
	goto tr2801;
tr2786:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st886;
tr2787:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st886;
tr2797:
	{
		s->line_counter++;
	}
	goto st886;
tr2808:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 886; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st886;
tr2929:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st886;
tr2930:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st886;
tr2931:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st886;
tr2941:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st886;
tr2840:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	goto st886;
tr2841:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st886;
tr2842:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st886;
tr2852:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st886;
tr2885:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	goto st886;
tr2886:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st886;
tr2887:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st886;
tr2897:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st886;
st886:
	if ( ++p == pe )
		goto _test_eof886;
case 886:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8832 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 !s->multiline  ) _widec += 512;
					if (
 s->number64 != 0  ) _widec += 1024;
					if (
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5760 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
					if (
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8832 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
			if (
 s->number64 != 0  ) _widec += 1024;
			if (
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2802;
		case 6153: goto st881;
		case 6176: goto st881;
		case 6184: goto tr2780;
		case 6185: goto tr2781;
		case 6409: goto st884;
		case 6432: goto st884;
		case 6440: goto tr2783;
		case 6441: goto tr2784;
		case 6665: goto st886;
		case 6688: goto st886;
		case 6696: goto tr2786;
		case 6697: goto tr2787;
		case 9482: goto tr2788;
		case 9531: goto tr2788;
		case 9738: goto tr2788;
		case 9787: goto tr2788;
		case 10250: goto tr2789;
		case 10299: goto tr2790;
		case 10506: goto tr2788;
		case 10555: goto tr2788;
		case 10762: goto tr2791;
		case 10811: goto tr2792;
		case 11274: goto tr2793;
		case 11323: goto tr2794;
		case 11530: goto tr2788;
		case 11579: goto tr2788;
		case 11786: goto tr2795;
		case 11835: goto tr2796;
		case 12298: goto tr2797;
		case 12347: goto tr2798;
		case 12554: goto tr2788;
		case 12603: goto tr2788;
		case 12810: goto tr2799;
		case 12859: goto tr2800;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2802;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2802;
	} else
		goto tr2802;
	goto tr2801;
tr2791:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1295;
tr2846:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1295;
tr2891:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1295;
tr2935:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1295;
st1295:
	if ( ++p == pe )
		goto _test_eof1295;
case 1295:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3200 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2688 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st881;
		case 3104: goto st881;
		case 3112: goto tr2780;
		case 3113: goto tr2781;
		case 3115: goto tr2802;
		case 4106: goto tr2789;
		case 4155: goto tr2790;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2802;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2802;
	} else
		goto tr2802;
	goto tr2801;
tr2792:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1296;
tr2847:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1296;
tr2892:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1296;
tr2936:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1296;
st1296:
	if ( ++p == pe )
		goto _test_eof1296;
case 1296:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3200 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2804;
	if ( 3968 <= _widec && _widec <= 4223 )
		goto tr2803;
	goto tr2801;
tr2798:
	{
		s->buffer_length = 0;
	}
	goto st887;
tr2807:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st887;
tr2942:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st887;
tr2853:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	goto st887;
tr2898:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	goto st887;
st887:
	if ( ++p == pe )
		goto _test_eof887;
case 887:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6784 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
		if (
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2804;
		case 8202: goto tr2806;
		case 8714: goto tr2808;
	}
	if ( _widec < 8064 ) {
		if ( 7552 <= _widec && _widec <= 7807 )
			goto tr2803;
	} else if ( _widec > 8319 ) {
		if ( 8576 <= _widec && _widec <= 8831 )
			goto tr2807;
	} else
		goto tr2805;
	goto tr2801;
tr2799:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1297;
tr2854:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1297;
tr2899:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1297;
tr2943:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1297;
st1297:
	if ( ++p == pe )
		goto _test_eof1297;
case 1297:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8832 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 !s->multiline  ) _widec += 512;
					if (
 s->number64 != 0  ) _widec += 1024;
					if (
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5760 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
					if (
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8832 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
			if (
 s->number64 != 0  ) _widec += 1024;
			if (
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2802;
		case 6153: goto st881;
		case 6176: goto st881;
		case 6184: goto tr2780;
		case 6185: goto tr2781;
		case 6409: goto st884;
		case 6432: goto st884;
		case 6440: goto tr2783;
		case 6441: goto tr2784;
		case 6665: goto st886;
		case 6688: goto st886;
		case 6696: goto tr2786;
		case 6697: goto tr2787;
		case 9482: goto tr2788;
		case 9531: goto tr2788;
		case 9738: goto tr2788;
		case 9787: goto tr2788;
		case 10250: goto tr2789;
		case 10299: goto tr2790;
		case 10506: goto tr2788;
		case 10555: goto tr2788;
		case 10762: goto tr2791;
		case 10811: goto tr2792;
		case 11274: goto tr2793;
		case 11323: goto tr2794;
		case 11530: goto tr2788;
		case 11579: goto tr2788;
		case 11786: goto tr2795;
		case 11835: goto tr2796;
		case 12298: goto tr2797;
		case 12347: goto tr2798;
		case 12554: goto tr2788;
		case 12603: goto tr2788;
		case 12810: goto tr2799;
		case 12859: goto tr2800;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2802;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2802;
	} else
		goto tr2802;
	goto tr2801;
tr2800:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1298;
tr2855:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1298;
tr2900:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET6_ADDR_LENGTH);
		rdata_tail += ZS_INET6_ADDR_LENGTH;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1298;
tr2944:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1298;
st1298:
	if ( ++p == pe )
		goto _test_eof1298;
case 1298:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6784 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
		if (
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2804;
		case 8202: goto tr2806;
		case 8714: goto tr2808;
	}
	if ( _widec < 8064 ) {
		if ( 7552 <= _widec && _widec <= 7807 )
			goto tr2803;
	} else if ( _widec > 8319 ) {
		if ( 8576 <= _widec && _widec <= 8831 )
			goto tr2807;
	} else
		goto tr2805;
	goto tr2801;
tr2777:
	{
		s->buffer_length = 0;
	}
	goto st888;
tr2809:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st888;
tr2771:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st888;
st888:
	if ( ++p == pe )
		goto _test_eof888;
case 888:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2810;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2809;
	goto tr2749;
tr2763:
	{
		s->buffer_length = 0;
	}
	goto st889;
tr2811:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st889;
st889:
	if ( ++p == pe )
		goto _test_eof889;
case 889:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2812;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2811;
	goto tr2749;
tr2754:
	{
		*(rdata_tail++) = 1;
	}
	goto st890;
st890:
	if ( ++p == pe )
		goto _test_eof890;
case 890:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st891;
		case 32: goto st891;
		case 40: goto tr2814;
		case 41: goto tr2815;
		case 1034: goto tr2816;
		case 1083: goto tr2817;
	}
	goto tr2749;
tr2814:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st891;
tr2815:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st891;
tr2816:
	{
		s->line_counter++;
	}
	goto st891;
tr2859:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 891; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st891;
st891:
	if ( ++p == pe )
		goto _test_eof891;
case 891:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st891;
		case 32: goto st891;
		case 40: goto tr2814;
		case 41: goto tr2815;
		case 1034: goto tr2816;
		case 1083: goto tr2817;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2818;
	goto tr2764;
tr2818:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st892;
tr2822:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st892;
st892:
	if ( ++p == pe )
		goto _test_eof892;
case 892:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2819;
		case 32: goto tr2819;
		case 40: goto tr2820;
		case 41: goto tr2821;
		case 1034: goto tr2823;
		case 1083: goto tr2824;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2822;
	goto tr2764;
tr2827:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st893;
tr2828:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st893;
tr2830:
	{
		s->line_counter++;
	}
	goto st893;
tr2857:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 893; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st893;
tr2819:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st893;
tr2820:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st893;
tr2821:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st893;
tr2823:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st893;
st893:
	if ( ++p == pe )
		goto _test_eof893;
case 893:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st893;
		case 32: goto st893;
		case 40: goto tr2827;
		case 41: goto tr2828;
		case 46: goto tr2829;
		case 1034: goto tr2830;
		case 1083: goto tr2831;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2829;
	goto tr2825;
tr2829:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st894;
tr2833:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st894;
st894:
	if ( ++p == pe )
		goto _test_eof894;
case 894:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2833;
		case 6153: goto tr2834;
		case 6176: goto tr2834;
		case 6184: goto tr2835;
		case 6185: goto tr2836;
		case 6409: goto tr2837;
		case 6432: goto tr2837;
		case 6440: goto tr2838;
		case 6441: goto tr2839;
		case 6665: goto tr2840;
		case 6688: goto tr2840;
		case 6696: goto tr2841;
		case 6697: goto tr2842;
		case 9482: goto tr2843;
		case 9531: goto tr2843;
		case 9738: goto tr2843;
		case 9787: goto tr2843;
		case 10250: goto tr2844;
		case 10299: goto tr2845;
		case 10506: goto tr2843;
		case 10555: goto tr2843;
		case 10762: goto tr2846;
		case 10811: goto tr2847;
		case 11274: goto tr2848;
		case 11323: goto tr2849;
		case 11530: goto tr2843;
		case 11579: goto tr2843;
		case 11786: goto tr2850;
		case 11835: goto tr2851;
		case 12298: goto tr2852;
		case 12347: goto tr2853;
		case 12554: goto tr2843;
		case 12603: goto tr2843;
		case 12810: goto tr2854;
		case 12859: goto tr2855;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2833;
	goto tr2832;
tr2831:
	{
		s->buffer_length = 0;
	}
	goto st895;
tr2856:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st895;
tr2824:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st895;
st895:
	if ( ++p == pe )
		goto _test_eof895;
case 895:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2857;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2856;
	goto tr2749;
tr2817:
	{
		s->buffer_length = 0;
	}
	goto st896;
tr2858:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st896;
st896:
	if ( ++p == pe )
		goto _test_eof896;
case 896:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2859;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2858;
	goto tr2749;
tr2755:
	{
		*(rdata_tail++) = 2;
	}
	goto st897;
st897:
	if ( ++p == pe )
		goto _test_eof897;
case 897:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st898;
		case 32: goto st898;
		case 40: goto tr2861;
		case 41: goto tr2862;
		case 1034: goto tr2863;
		case 1083: goto tr2864;
	}
	goto tr2749;
tr2861:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st898;
tr2862:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st898;
tr2863:
	{
		s->line_counter++;
	}
	goto st898;
tr2904:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 898; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st898;
st898:
	if ( ++p == pe )
		goto _test_eof898;
case 898:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st898;
		case 32: goto st898;
		case 40: goto tr2861;
		case 41: goto tr2862;
		case 1034: goto tr2863;
		case 1083: goto tr2864;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2865;
	goto tr2764;
tr2865:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st899;
tr2869:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st899;
st899:
	if ( ++p == pe )
		goto _test_eof899;
case 899:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2866;
		case 32: goto tr2866;
		case 40: goto tr2867;
		case 41: goto tr2868;
		case 1034: goto tr2870;
		case 1083: goto tr2871;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2869;
	goto tr2764;
tr2873:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st900;
tr2874:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st900;
tr2876:
	{
		s->line_counter++;
	}
	goto st900;
tr2902:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 900; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st900;
tr2866:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st900;
tr2867:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st900;
tr2868:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st900;
tr2870:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st900;
st900:
	if ( ++p == pe )
		goto _test_eof900;
case 900:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st900;
		case 32: goto st900;
		case 40: goto tr2873;
		case 41: goto tr2874;
		case 46: goto tr2875;
		case 1034: goto tr2876;
		case 1083: goto tr2877;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2875;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2875;
	} else
		goto tr2875;
	goto tr2825;
tr2875:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st901;
tr2878:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st901;
st901:
	if ( ++p == pe )
		goto _test_eof901;
case 901:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2878;
		case 6153: goto tr2879;
		case 6176: goto tr2879;
		case 6184: goto tr2880;
		case 6185: goto tr2881;
		case 6409: goto tr2882;
		case 6432: goto tr2882;
		case 6440: goto tr2883;
		case 6441: goto tr2884;
		case 6665: goto tr2885;
		case 6688: goto tr2885;
		case 6696: goto tr2886;
		case 6697: goto tr2887;
		case 9482: goto tr2888;
		case 9531: goto tr2888;
		case 9738: goto tr2888;
		case 9787: goto tr2888;
		case 10250: goto tr2889;
		case 10299: goto tr2890;
		case 10506: goto tr2888;
		case 10555: goto tr2888;
		case 10762: goto tr2891;
		case 10811: goto tr2892;
		case 11274: goto tr2893;
		case 11323: goto tr2894;
		case 11530: goto tr2888;
		case 11579: goto tr2888;
		case 11786: goto tr2895;
		case 11835: goto tr2896;
		case 12298: goto tr2897;
		case 12347: goto tr2898;
		case 12554: goto tr2888;
		case 12603: goto tr2888;
		case 12810: goto tr2899;
		case 12859: goto tr2900;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2878;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2878;
	} else
		goto tr2878;
	goto tr2832;
tr2877:
	{
		s->buffer_length = 0;
	}
	goto st902;
tr2901:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st902;
tr2871:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st902;
st902:
	if ( ++p == pe )
		goto _test_eof902;
case 902:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2902;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2901;
	goto tr2749;
tr2864:
	{
		s->buffer_length = 0;
	}
	goto st903;
tr2903:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st903;
st903:
	if ( ++p == pe )
		goto _test_eof903;
case 903:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2904;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2903;
	goto tr2749;
tr2756:
	{
		*(rdata_tail++) = 3;
	}
	goto st904;
st904:
	if ( ++p == pe )
		goto _test_eof904;
case 904:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st905;
		case 32: goto st905;
		case 40: goto tr2906;
		case 41: goto tr2907;
		case 1034: goto tr2908;
		case 1083: goto tr2909;
	}
	goto tr2749;
tr2906:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st905;
tr2907:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st905;
tr2908:
	{
		s->line_counter++;
	}
	goto st905;
tr2948:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 905; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st905;
st905:
	if ( ++p == pe )
		goto _test_eof905;
case 905:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st905;
		case 32: goto st905;
		case 40: goto tr2906;
		case 41: goto tr2907;
		case 1034: goto tr2908;
		case 1083: goto tr2909;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2910;
	goto tr2764;
tr2910:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st906;
tr2914:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st906;
st906:
	if ( ++p == pe )
		goto _test_eof906;
case 906:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2911;
		case 32: goto tr2911;
		case 40: goto tr2912;
		case 41: goto tr2913;
		case 1034: goto tr2915;
		case 1083: goto tr2916;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2914;
	goto tr2764;
tr2918:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st907;
tr2919:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st907;
tr2921:
	{
		s->line_counter++;
	}
	goto st907;
tr2946:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 907; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st907;
tr2911:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st907;
tr2912:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st907;
tr2913:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st907;
tr2915:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st907;
st907:
	if ( ++p == pe )
		goto _test_eof907;
case 907:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st907;
		case 32: goto st907;
		case 40: goto tr2918;
		case 41: goto tr2919;
		case 42: goto tr2920;
		case 92: goto tr2920;
		case 95: goto tr2920;
		case 1034: goto tr2921;
		case 1083: goto tr2922;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2920;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2920;
	} else
		goto tr2920;
	goto tr2749;
tr2920:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 908;goto st301;} }
	goto st908;
st908:
	if ( ++p == pe )
		goto _test_eof908;
case 908:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto tr2923;
		case 6176: goto tr2923;
		case 6184: goto tr2924;
		case 6185: goto tr2925;
		case 6409: goto tr2926;
		case 6432: goto tr2926;
		case 6440: goto tr2927;
		case 6441: goto tr2928;
		case 6665: goto tr2929;
		case 6688: goto tr2929;
		case 6696: goto tr2930;
		case 6697: goto tr2931;
		case 9482: goto tr2932;
		case 9531: goto tr2932;
		case 9738: goto tr2932;
		case 9787: goto tr2932;
		case 10250: goto tr2933;
		case 10299: goto tr2934;
		case 10506: goto tr2932;
		case 10555: goto tr2932;
		case 10762: goto tr2935;
		case 10811: goto tr2936;
		case 11274: goto tr2937;
		case 11323: goto tr2938;
		case 11530: goto tr2932;
		case 11579: goto tr2932;
		case 11786: goto tr2939;
		case 11835: goto tr2940;
		case 12298: goto tr2941;
		case 12347: goto tr2942;
		case 12554: goto tr2932;
		case 12603: goto tr2932;
		case 12810: goto tr2943;
		case 12859: goto tr2944;
	}
	goto tr2778;
tr2922:
	{
		s->buffer_length = 0;
	}
	goto st909;
tr2945:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st909;
tr2916:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st909;
st909:
	if ( ++p == pe )
		goto _test_eof909;
case 909:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2946;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2945;
	goto tr2749;
tr2909:
	{
		s->buffer_length = 0;
	}
	goto st910;
tr2947:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st910;
st910:
	if ( ++p == pe )
		goto _test_eof910;
case 910:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2948;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2947;
	goto tr2749;
tr2758:
	{
		s->buffer_length = 0;
	}
	goto st911;
tr2949:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st911;
tr2748:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st911;
st911:
	if ( ++p == pe )
		goto _test_eof911;
case 911:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2950;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2949;
	goto tr75;
st912:
	if ( ++p == pe )
		goto _test_eof912;
case 912:
	switch( (*p) ) {
		case 65: goto st913;
		case 67: goto st947;
		case 68: goto st968;
		case 69: goto st982;
		case 72: goto st989;
		case 73: goto st994;
		case 75: goto st1002;
		case 76: goto st1006;
		case 77: goto st1014;
		case 78: goto st1020;
		case 79: goto st1036;
		case 80: goto st1046;
		case 82: goto st1049;
		case 83: goto st1056;
		case 84: goto st1072;
		case 85: goto st1082;
		case 90: goto st1085;
		case 97: goto st913;
		case 99: goto st947;
		case 100: goto st968;
		case 101: goto st982;
		case 104: goto st989;
		case 105: goto st994;
		case 107: goto st1002;
		case 108: goto st1006;
		case 109: goto st1014;
		case 110: goto st1020;
		case 111: goto st1036;
		case 112: goto st1046;
		case 114: goto st1049;
		case 115: goto st1056;
		case 116: goto st1072;
		case 117: goto st1082;
		case 122: goto st1085;
	}
	goto tr2951;
st913:
	if ( ++p == pe )
		goto _test_eof913;
case 913:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2969;
		case 32: goto tr2969;
		case 40: goto tr2970;
		case 41: goto tr2971;
		case 65: goto st938;
		case 70: goto st941;
		case 80: goto st945;
		case 97: goto st938;
		case 102: goto st941;
		case 112: goto st945;
		case 1034: goto tr2975;
		case 1083: goto tr2976;
	}
	goto tr2951;
tr2978:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr2979:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr2981:
	{
		s->line_counter++;
	}
	goto st914;
tr3077:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 914; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st914;
tr3421:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st914;
tr3422:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3423:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3425:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st914;
tr2969:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st914;
tr2970:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr2971:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr2975:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3080:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st914;
tr3081:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3082:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3083:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3088:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st914;
tr3089:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3090:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3091:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3094:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st914;
tr3095:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3096:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3097:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3105:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	goto st914;
tr3106:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3107:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3108:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3116:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	goto st914;
tr3117:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3118:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3119:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3121:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	goto st914;
tr3122:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3123:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3124:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3128:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st914;
tr3129:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3130:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3131:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3136:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st914;
tr3137:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3138:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3139:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3144:
	{ type_num(KNOT_RRTYPE_CSYNC, &rdata_tail); }
	goto st914;
tr3145:
	{ type_num(KNOT_RRTYPE_CSYNC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3146:
	{ type_num(KNOT_RRTYPE_CSYNC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3147:
	{ type_num(KNOT_RRTYPE_CSYNC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3155:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st914;
tr3156:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3157:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3158:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3164:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st914;
tr3165:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3166:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3167:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3172:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st914;
tr3173:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3174:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3175:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3177:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st914;
tr3178:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3179:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3180:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3187:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st914;
tr3188:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3189:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3190:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3193:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st914;
tr3194:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3195:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3196:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3202:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st914;
tr3203:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3204:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3205:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3214:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st914;
tr3215:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3216:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3217:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3222:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st914;
tr3223:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3224:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3225:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3227:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st914;
tr3228:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3229:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3230:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3237:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st914;
tr3238:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3239:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3240:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3243:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st914;
tr3244:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3245:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3246:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3249:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st914;
tr3250:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3251:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3252:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3254:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st914;
tr3255:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3256:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3257:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3264:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st914;
tr3265:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3266:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3267:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3269:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st914;
tr3270:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3271:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3272:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3280:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st914;
tr3281:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3282:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3283:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3286:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st914;
tr3287:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3288:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3289:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3291:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st914;
tr3292:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3293:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3295:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3298:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st914;
tr3299:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3300:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3302:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3304:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st914;
tr3305:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3306:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3308:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3314:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st914;
tr3315:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3316:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3317:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3328:
	{ type_num(KNOT_RRTYPE_OPENPGPKEY, &rdata_tail); }
	goto st914;
tr3329:
	{ type_num(KNOT_RRTYPE_OPENPGPKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3330:
	{ type_num(KNOT_RRTYPE_OPENPGPKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3331:
	{ type_num(KNOT_RRTYPE_OPENPGPKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3335:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st914;
tr3336:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3337:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3338:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3343:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st914;
tr3344:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3345:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3346:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3351:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st914;
tr3352:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3353:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3354:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3356:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st914;
tr3357:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3358:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3359:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3370:
	{ type_num(KNOT_RRTYPE_SMIMEA, &rdata_tail); }
	goto st914;
tr3371:
	{ type_num(KNOT_RRTYPE_SMIMEA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3372:
	{ type_num(KNOT_RRTYPE_SMIMEA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3373:
	{ type_num(KNOT_RRTYPE_SMIMEA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3376:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st914;
tr3377:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3378:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3379:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3382:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st914;
tr3383:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3384:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3385:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3388:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st914;
tr3389:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3390:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3391:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3396:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st914;
tr3397:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3398:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3399:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3406:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st914;
tr3407:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3408:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3409:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3412:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st914;
tr3413:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3414:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3415:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3429:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	goto st914;
tr3430:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3431:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3432:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
tr3439:
	{ type_num(KNOT_RRTYPE_ZONEMD, &rdata_tail); }
	goto st914;
tr3440:
	{ type_num(KNOT_RRTYPE_ZONEMD, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st914;
tr3441:
	{ type_num(KNOT_RRTYPE_ZONEMD, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st914;
tr3442:
	{ type_num(KNOT_RRTYPE_ZONEMD, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st914;
st914:
	if ( ++p == pe )
		goto _test_eof914;
case 914:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st914;
		case 32: goto st914;
		case 40: goto tr2978;
		case 41: goto tr2979;
		case 1034: goto tr2981;
		case 1083: goto tr2982;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2980;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2980;
	} else
		goto tr2980;
	goto tr75;
tr2980:
	{ p--; {stack[top++] = 915;goto st543;} }
	goto st915;
st915:
	if ( ++p == pe )
		goto _test_eof915;
case 915:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st916;
		case 32: goto st916;
		case 40: goto tr2984;
		case 41: goto tr2985;
		case 1034: goto tr2986;
		case 1083: goto tr2987;
	}
	goto tr75;
tr2984:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st916;
tr2985:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st916;
tr2986:
	{
		s->line_counter++;
	}
	goto st916;
tr3075:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 916; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st916;
st916:
	if ( ++p == pe )
		goto _test_eof916;
case 916:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st916;
		case 32: goto st916;
		case 40: goto tr2984;
		case 41: goto tr2985;
		case 1034: goto tr2986;
		case 1083: goto tr2987;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2988;
	goto tr2007;
tr2988:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st917;
tr2992:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st917;
st917:
	if ( ++p == pe )
		goto _test_eof917;
case 917:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2989;
		case 32: goto tr2989;
		case 40: goto tr2990;
		case 41: goto tr2991;
		case 1034: goto tr2993;
		case 1083: goto tr2994;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2992;
	goto tr2007;
tr2996:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st918;
tr2997:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st918;
tr2999:
	{
		s->line_counter++;
	}
	goto st918;
tr3073:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 918; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st918;
tr2989:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st918;
tr2990:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st918;
tr2991:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st918;
tr2993:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st918;
st918:
	if ( ++p == pe )
		goto _test_eof918;
case 918:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st918;
		case 32: goto st918;
		case 40: goto tr2996;
		case 41: goto tr2997;
		case 1034: goto tr2999;
		case 1083: goto tr3000;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2998;
	goto tr2007;
tr2998:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st919;
tr3004:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st919;
st919:
	if ( ++p == pe )
		goto _test_eof919;
case 919:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3001;
		case 32: goto tr3001;
		case 40: goto tr3002;
		case 41: goto tr3003;
		case 1034: goto tr3005;
		case 1083: goto tr3006;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3004;
	goto tr2007;
tr3009:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st920;
tr3010:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st920;
tr3012:
	{
		s->line_counter++;
	}
	goto st920;
tr3071:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 920; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st920;
tr3001:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st920;
tr3002:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st920;
tr3003:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st920;
tr3005:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st920;
st920:
	if ( ++p == pe )
		goto _test_eof920;
case 920:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st920;
		case 32: goto st920;
		case 40: goto tr3009;
		case 41: goto tr3010;
		case 1034: goto tr3012;
		case 1083: goto tr3013;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3011;
	goto tr3007;
tr3011:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st921;
tr3017:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st921;
st921:
	if ( ++p == pe )
		goto _test_eof921;
case 921:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3014;
		case 32: goto tr3014;
		case 40: goto tr3015;
		case 41: goto tr3016;
		case 1034: goto tr3018;
		case 1083: goto tr3019;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3017;
	goto tr3007;
tr3021:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st922;
tr3022:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st922;
tr3024:
	{
		s->line_counter++;
	}
	goto st922;
tr3069:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 922; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st922;
tr3014:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	goto st922;
tr3015:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st922;
tr3016:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st922;
tr3018:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st922;
st922:
	if ( ++p == pe )
		goto _test_eof922;
case 922:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st922;
		case 32: goto st922;
		case 40: goto tr3021;
		case 41: goto tr3022;
		case 1034: goto tr3024;
		case 1083: goto tr3025;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3023;
	goto tr3007;
tr3023:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st923;
tr3029:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st923;
st923:
	if ( ++p == pe )
		goto _test_eof923;
case 923:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3026;
		case 32: goto tr3026;
		case 40: goto tr3027;
		case 41: goto tr3028;
		case 1034: goto tr3030;
		case 1083: goto tr3031;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3029;
	goto tr3007;
tr3033:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st924;
tr3034:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st924;
tr3036:
	{
		s->line_counter++;
	}
	goto st924;
tr3067:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 924; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st924;
tr3026:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	goto st924;
tr3027:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st924;
tr3028:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st924;
tr3030:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st924;
st924:
	if ( ++p == pe )
		goto _test_eof924;
case 924:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st924;
		case 32: goto st924;
		case 40: goto tr3033;
		case 41: goto tr3034;
		case 1034: goto tr3036;
		case 1083: goto tr3037;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3035;
	goto tr2007;
tr3035:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st925;
tr3041:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st925;
st925:
	if ( ++p == pe )
		goto _test_eof925;
case 925:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3038;
		case 32: goto tr3038;
		case 40: goto tr3039;
		case 41: goto tr3040;
		case 1034: goto tr3042;
		case 1083: goto tr3043;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3041;
	goto tr2007;
tr3045:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st926;
tr3046:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st926;
tr3048:
	{
		s->line_counter++;
	}
	goto st926;
tr3065:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 926; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st926;
tr3038:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st926;
tr3039:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st926;
tr3040:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st926;
tr3042:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st926;
st926:
	if ( ++p == pe )
		goto _test_eof926;
case 926:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st926;
		case 32: goto st926;
		case 40: goto tr3045;
		case 41: goto tr3046;
		case 42: goto tr3047;
		case 92: goto tr3047;
		case 95: goto tr3047;
		case 1034: goto tr3048;
		case 1083: goto tr3049;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3047;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3047;
	} else
		goto tr3047;
	goto tr75;
tr3047:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 927;goto st301;} }
	goto st927;
st927:
	if ( ++p == pe )
		goto _test_eof927;
case 927:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3050;
		case 32: goto tr3050;
		case 40: goto tr3051;
		case 41: goto tr3052;
		case 1034: goto tr3053;
		case 1083: goto tr3054;
	}
	goto tr75;
tr3056:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st928;
tr3057:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st928;
tr3059:
	{
		s->line_counter++;
	}
	goto st928;
tr3063:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 928; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st928;
tr3050:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st928;
tr3051:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st928;
tr3052:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st928;
tr3053:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st928;
st928:
	if ( ++p == pe )
		goto _test_eof928;
case 928:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st928;
		case 32: goto st928;
		case 40: goto tr3056;
		case 41: goto tr3057;
		case 43: goto tr3058;
		case 1034: goto tr3059;
		case 1083: goto tr3060;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr3058;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3058;
	} else
		goto tr3058;
	goto tr75;
tr3058:
	{ p--; {stack[top++] = 929;goto st360;} }
	goto st929;
st929:
	if ( ++p == pe )
		goto _test_eof929;
case 929:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr3061;
		case 1595: goto tr3061;
	}
	goto tr75;
tr3061:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1299;
st1299:
	if ( ++p == pe )
		goto _test_eof1299;
case 1299:
	goto st0;
tr3060:
	{
		s->buffer_length = 0;
	}
	goto st930;
tr3062:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st930;
tr3054:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->buffer_length = 0;
	}
	goto st930;
st930:
	if ( ++p == pe )
		goto _test_eof930;
case 930:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3063;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3062;
	goto tr75;
tr3049:
	{
		s->buffer_length = 0;
	}
	goto st931;
tr3064:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st931;
tr3043:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st931;
st931:
	if ( ++p == pe )
		goto _test_eof931;
case 931:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3065;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3064;
	goto tr75;
tr3037:
	{
		s->buffer_length = 0;
	}
	goto st932;
tr3066:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st932;
tr3031:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st932;
st932:
	if ( ++p == pe )
		goto _test_eof932;
case 932:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3067;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3066;
	goto tr75;
tr3025:
	{
		s->buffer_length = 0;
	}
	goto st933;
tr3068:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st933;
tr3019:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			uint32_t timestamp;
			int ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st299;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st299;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st299;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st933;
st933:
	if ( ++p == pe )
		goto _test_eof933;
case 933:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3069;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3068;
	goto tr75;
tr3013:
	{
		s->buffer_length = 0;
	}
	goto st934;
tr3070:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st934;
tr3006:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st934;
st934:
	if ( ++p == pe )
		goto _test_eof934;
case 934:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3071;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3070;
	goto tr75;
tr3000:
	{
		s->buffer_length = 0;
	}
	goto st935;
tr3072:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st935;
tr2994:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st935;
st935:
	if ( ++p == pe )
		goto _test_eof935;
case 935:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3073;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3072;
	goto tr75;
tr2987:
	{
		s->buffer_length = 0;
	}
	goto st936;
tr3074:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st936;
st936:
	if ( ++p == pe )
		goto _test_eof936;
case 936:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3075;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3074;
	goto tr75;
tr2982:
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3076:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st937;
tr3426:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st937;
tr2976:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3084:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3092:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3098:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3109:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3120:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3125:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3132:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3140:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3148:
	{ type_num(KNOT_RRTYPE_CSYNC, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3159:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3168:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3176:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3181:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3191:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3197:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3206:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3218:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3226:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3231:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3241:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3247:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3253:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3258:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3268:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3273:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3284:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3290:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3296:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3303:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3309:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3318:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3332:
	{ type_num(KNOT_RRTYPE_OPENPGPKEY, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3339:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3347:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3355:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3360:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3374:
	{ type_num(KNOT_RRTYPE_SMIMEA, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3380:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3386:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3392:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3400:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3410:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3416:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3433:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
tr3443:
	{ type_num(KNOT_RRTYPE_ZONEMD, &rdata_tail); }
	{
		s->buffer_length = 0;
	}
	goto st937;
st937:
	if ( ++p == pe )
		goto _test_eof937;
case 937:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3077;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3076;
	goto tr75;
st938:
	if ( ++p == pe )
		goto _test_eof938;
case 938:
	switch( (*p) ) {
		case 65: goto st939;
		case 97: goto st939;
	}
	goto tr2951;
st939:
	if ( ++p == pe )
		goto _test_eof939;
case 939:
	switch( (*p) ) {
		case 65: goto st940;
		case 97: goto st940;
	}
	goto tr2951;
st940:
	if ( ++p == pe )
		goto _test_eof940;
case 940:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3080;
		case 32: goto tr3080;
		case 40: goto tr3081;
		case 41: goto tr3082;
		case 1034: goto tr3083;
		case 1083: goto tr3084;
	}
	goto tr2951;
st941:
	if ( ++p == pe )
		goto _test_eof941;
case 941:
	switch( (*p) ) {
		case 83: goto st942;
		case 115: goto st942;
	}
	goto tr2951;
st942:
	if ( ++p == pe )
		goto _test_eof942;
case 942:
	switch( (*p) ) {
		case 68: goto st943;
		case 100: goto st943;
	}
	goto tr2951;
st943:
	if ( ++p == pe )
		goto _test_eof943;
case 943:
	switch( (*p) ) {
		case 66: goto st944;
		case 98: goto st944;
	}
	goto tr2951;
st944:
	if ( ++p == pe )
		goto _test_eof944;
case 944:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3088;
		case 32: goto tr3088;
		case 40: goto tr3089;
		case 41: goto tr3090;
		case 1034: goto tr3091;
		case 1083: goto tr3092;
	}
	goto tr2951;
st945:
	if ( ++p == pe )
		goto _test_eof945;
case 945:
	switch( (*p) ) {
		case 76: goto st946;
		case 108: goto st946;
	}
	goto tr2951;
st946:
	if ( ++p == pe )
		goto _test_eof946;
case 946:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3094;
		case 32: goto tr3094;
		case 40: goto tr3095;
		case 41: goto tr3096;
		case 1034: goto tr3097;
		case 1083: goto tr3098;
	}
	goto tr2951;
st947:
	if ( ++p == pe )
		goto _test_eof947;
case 947:
	switch( (*p) ) {
		case 65: goto st948;
		case 68: goto st950;
		case 69: goto st957;
		case 78: goto st960;
		case 83: goto st964;
		case 97: goto st948;
		case 100: goto st950;
		case 101: goto st957;
		case 110: goto st960;
		case 115: goto st964;
	}
	goto tr2951;
st948:
	if ( ++p == pe )
		goto _test_eof948;
case 948:
	switch( (*p) ) {
		case 65: goto st949;
		case 97: goto st949;
	}
	goto tr2951;
st949:
	if ( ++p == pe )
		goto _test_eof949;
case 949:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3105;
		case 32: goto tr3105;
		case 40: goto tr3106;
		case 41: goto tr3107;
		case 1034: goto tr3108;
		case 1083: goto tr3109;
	}
	goto tr2951;
st950:
	if ( ++p == pe )
		goto _test_eof950;
case 950:
	switch( (*p) ) {
		case 78: goto st951;
		case 83: goto st956;
		case 110: goto st951;
		case 115: goto st956;
	}
	goto tr2951;
st951:
	if ( ++p == pe )
		goto _test_eof951;
case 951:
	switch( (*p) ) {
		case 83: goto st952;
		case 115: goto st952;
	}
	goto tr2951;
st952:
	if ( ++p == pe )
		goto _test_eof952;
case 952:
	switch( (*p) ) {
		case 75: goto st953;
		case 107: goto st953;
	}
	goto tr2951;
st953:
	if ( ++p == pe )
		goto _test_eof953;
case 953:
	switch( (*p) ) {
		case 69: goto st954;
		case 101: goto st954;
	}
	goto tr2951;
st954:
	if ( ++p == pe )
		goto _test_eof954;
case 954:
	switch( (*p) ) {
		case 89: goto st955;
		case 121: goto st955;
	}
	goto tr2951;
st955:
	if ( ++p == pe )
		goto _test_eof955;
case 955:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3116;
		case 32: goto tr3116;
		case 40: goto tr3117;
		case 41: goto tr3118;
		case 1034: goto tr3119;
		case 1083: goto tr3120;
	}
	goto tr2951;
st956:
	if ( ++p == pe )
		goto _test_eof956;
case 956:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3121;
		case 32: goto tr3121;
		case 40: goto tr3122;
		case 41: goto tr3123;
		case 1034: goto tr3124;
		case 1083: goto tr3125;
	}
	goto tr2951;
st957:
	if ( ++p == pe )
		goto _test_eof957;
case 957:
	switch( (*p) ) {
		case 82: goto st958;
		case 114: goto st958;
	}
	goto tr2951;
st958:
	if ( ++p == pe )
		goto _test_eof958;
case 958:
	switch( (*p) ) {
		case 84: goto st959;
		case 116: goto st959;
	}
	goto tr2951;
st959:
	if ( ++p == pe )
		goto _test_eof959;
case 959:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3128;
		case 32: goto tr3128;
		case 40: goto tr3129;
		case 41: goto tr3130;
		case 1034: goto tr3131;
		case 1083: goto tr3132;
	}
	goto tr2951;
st960:
	if ( ++p == pe )
		goto _test_eof960;
case 960:
	switch( (*p) ) {
		case 65: goto st961;
		case 97: goto st961;
	}
	goto tr2951;
st961:
	if ( ++p == pe )
		goto _test_eof961;
case 961:
	switch( (*p) ) {
		case 77: goto st962;
		case 109: goto st962;
	}
	goto tr2951;
st962:
	if ( ++p == pe )
		goto _test_eof962;
case 962:
	switch( (*p) ) {
		case 69: goto st963;
		case 101: goto st963;
	}
	goto tr2951;
st963:
	if ( ++p == pe )
		goto _test_eof963;
case 963:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3136;
		case 32: goto tr3136;
		case 40: goto tr3137;
		case 41: goto tr3138;
		case 1034: goto tr3139;
		case 1083: goto tr3140;
	}
	goto tr2951;
st964:
	if ( ++p == pe )
		goto _test_eof964;
case 964:
	switch( (*p) ) {
		case 89: goto st965;
		case 121: goto st965;
	}
	goto tr2951;
st965:
	if ( ++p == pe )
		goto _test_eof965;
case 965:
	switch( (*p) ) {
		case 78: goto st966;
		case 110: goto st966;
	}
	goto tr2951;
st966:
	if ( ++p == pe )
		goto _test_eof966;
case 966:
	switch( (*p) ) {
		case 67: goto st967;
		case 99: goto st967;
	}
	goto tr2951;
st967:
	if ( ++p == pe )
		goto _test_eof967;
case 967:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3144;
		case 32: goto tr3144;
		case 40: goto tr3145;
		case 41: goto tr3146;
		case 1034: goto tr3147;
		case 1083: goto tr3148;
	}
	goto tr2951;
st968:
	if ( ++p == pe )
		goto _test_eof968;
case 968:
	switch( (*p) ) {
		case 72: goto st969;
		case 78: goto st973;
		case 83: goto st981;
		case 104: goto st969;
		case 110: goto st973;
		case 115: goto st981;
	}
	goto tr2951;
st969:
	if ( ++p == pe )
		goto _test_eof969;
case 969:
	switch( (*p) ) {
		case 67: goto st970;
		case 99: goto st970;
	}
	goto tr2951;
st970:
	if ( ++p == pe )
		goto _test_eof970;
case 970:
	switch( (*p) ) {
		case 73: goto st971;
		case 105: goto st971;
	}
	goto tr2951;
st971:
	if ( ++p == pe )
		goto _test_eof971;
case 971:
	switch( (*p) ) {
		case 68: goto st972;
		case 100: goto st972;
	}
	goto tr2951;
st972:
	if ( ++p == pe )
		goto _test_eof972;
case 972:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3155;
		case 32: goto tr3155;
		case 40: goto tr3156;
		case 41: goto tr3157;
		case 1034: goto tr3158;
		case 1083: goto tr3159;
	}
	goto tr2951;
st973:
	if ( ++p == pe )
		goto _test_eof973;
case 973:
	switch( (*p) ) {
		case 65: goto st974;
		case 83: goto st977;
		case 97: goto st974;
		case 115: goto st977;
	}
	goto tr2951;
st974:
	if ( ++p == pe )
		goto _test_eof974;
case 974:
	switch( (*p) ) {
		case 77: goto st975;
		case 109: goto st975;
	}
	goto tr2951;
st975:
	if ( ++p == pe )
		goto _test_eof975;
case 975:
	switch( (*p) ) {
		case 69: goto st976;
		case 101: goto st976;
	}
	goto tr2951;
st976:
	if ( ++p == pe )
		goto _test_eof976;
case 976:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3164;
		case 32: goto tr3164;
		case 40: goto tr3165;
		case 41: goto tr3166;
		case 1034: goto tr3167;
		case 1083: goto tr3168;
	}
	goto tr2951;
st977:
	if ( ++p == pe )
		goto _test_eof977;
case 977:
	switch( (*p) ) {
		case 75: goto st978;
		case 107: goto st978;
	}
	goto tr2951;
st978:
	if ( ++p == pe )
		goto _test_eof978;
case 978:
	switch( (*p) ) {
		case 69: goto st979;
		case 101: goto st979;
	}
	goto tr2951;
st979:
	if ( ++p == pe )
		goto _test_eof979;
case 979:
	switch( (*p) ) {
		case 89: goto st980;
		case 121: goto st980;
	}
	goto tr2951;
st980:
	if ( ++p == pe )
		goto _test_eof980;
case 980:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3172;
		case 32: goto tr3172;
		case 40: goto tr3173;
		case 41: goto tr3174;
		case 1034: goto tr3175;
		case 1083: goto tr3176;
	}
	goto tr2951;
st981:
	if ( ++p == pe )
		goto _test_eof981;
case 981:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3177;
		case 32: goto tr3177;
		case 40: goto tr3178;
		case 41: goto tr3179;
		case 1034: goto tr3180;
		case 1083: goto tr3181;
	}
	goto tr2951;
st982:
	if ( ++p == pe )
		goto _test_eof982;
case 982:
	switch( (*p) ) {
		case 85: goto st983;
		case 117: goto st983;
	}
	goto tr2951;
st983:
	if ( ++p == pe )
		goto _test_eof983;
case 983:
	switch( (*p) ) {
		case 73: goto st984;
		case 105: goto st984;
	}
	goto tr2951;
st984:
	if ( ++p == pe )
		goto _test_eof984;
case 984:
	switch( (*p) ) {
		case 52: goto st985;
		case 54: goto st987;
	}
	goto tr2951;
st985:
	if ( ++p == pe )
		goto _test_eof985;
case 985:
	if ( (*p) == 56 )
		goto st986;
	goto tr2951;
st986:
	if ( ++p == pe )
		goto _test_eof986;
case 986:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3187;
		case 32: goto tr3187;
		case 40: goto tr3188;
		case 41: goto tr3189;
		case 1034: goto tr3190;
		case 1083: goto tr3191;
	}
	goto tr2951;
st987:
	if ( ++p == pe )
		goto _test_eof987;
case 987:
	if ( (*p) == 52 )
		goto st988;
	goto tr2951;
st988:
	if ( ++p == pe )
		goto _test_eof988;
case 988:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3193;
		case 32: goto tr3193;
		case 40: goto tr3194;
		case 41: goto tr3195;
		case 1034: goto tr3196;
		case 1083: goto tr3197;
	}
	goto tr2951;
st989:
	if ( ++p == pe )
		goto _test_eof989;
case 989:
	switch( (*p) ) {
		case 73: goto st990;
		case 105: goto st990;
	}
	goto tr2951;
st990:
	if ( ++p == pe )
		goto _test_eof990;
case 990:
	switch( (*p) ) {
		case 78: goto st991;
		case 110: goto st991;
	}
	goto tr2951;
st991:
	if ( ++p == pe )
		goto _test_eof991;
case 991:
	switch( (*p) ) {
		case 70: goto st992;
		case 102: goto st992;
	}
	goto tr2951;
st992:
	if ( ++p == pe )
		goto _test_eof992;
case 992:
	switch( (*p) ) {
		case 79: goto st993;
		case 111: goto st993;
	}
	goto tr2951;
st993:
	if ( ++p == pe )
		goto _test_eof993;
case 993:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3202;
		case 32: goto tr3202;
		case 40: goto tr3203;
		case 41: goto tr3204;
		case 1034: goto tr3205;
		case 1083: goto tr3206;
	}
	goto tr2951;
st994:
	if ( ++p == pe )
		goto _test_eof994;
case 994:
	switch( (*p) ) {
		case 80: goto st995;
		case 112: goto st995;
	}
	goto tr2951;
st995:
	if ( ++p == pe )
		goto _test_eof995;
case 995:
	switch( (*p) ) {
		case 83: goto st996;
		case 115: goto st996;
	}
	goto tr2951;
st996:
	if ( ++p == pe )
		goto _test_eof996;
case 996:
	switch( (*p) ) {
		case 69: goto st997;
		case 101: goto st997;
	}
	goto tr2951;
st997:
	if ( ++p == pe )
		goto _test_eof997;
case 997:
	switch( (*p) ) {
		case 67: goto st998;
		case 99: goto st998;
	}
	goto tr2951;
st998:
	if ( ++p == pe )
		goto _test_eof998;
case 998:
	switch( (*p) ) {
		case 75: goto st999;
		case 107: goto st999;
	}
	goto tr2951;
st999:
	if ( ++p == pe )
		goto _test_eof999;
case 999:
	switch( (*p) ) {
		case 69: goto st1000;
		case 101: goto st1000;
	}
	goto tr2951;
st1000:
	if ( ++p == pe )
		goto _test_eof1000;
case 1000:
	switch( (*p) ) {
		case 89: goto st1001;
		case 121: goto st1001;
	}
	goto tr2951;
st1001:
	if ( ++p == pe )
		goto _test_eof1001;
case 1001:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3214;
		case 32: goto tr3214;
		case 40: goto tr3215;
		case 41: goto tr3216;
		case 1034: goto tr3217;
		case 1083: goto tr3218;
	}
	goto tr2951;
st1002:
	if ( ++p == pe )
		goto _test_eof1002;
case 1002:
	switch( (*p) ) {
		case 69: goto st1003;
		case 88: goto st1005;
		case 101: goto st1003;
		case 120: goto st1005;
	}
	goto tr2951;
st1003:
	if ( ++p == pe )
		goto _test_eof1003;
case 1003:
	switch( (*p) ) {
		case 89: goto st1004;
		case 121: goto st1004;
	}
	goto tr2951;
st1004:
	if ( ++p == pe )
		goto _test_eof1004;
case 1004:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3222;
		case 32: goto tr3222;
		case 40: goto tr3223;
		case 41: goto tr3224;
		case 1034: goto tr3225;
		case 1083: goto tr3226;
	}
	goto tr2951;
st1005:
	if ( ++p == pe )
		goto _test_eof1005;
case 1005:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3227;
		case 32: goto tr3227;
		case 40: goto tr3228;
		case 41: goto tr3229;
		case 1034: goto tr3230;
		case 1083: goto tr3231;
	}
	goto tr2951;
st1006:
	if ( ++p == pe )
		goto _test_eof1006;
case 1006:
	switch( (*p) ) {
		case 51: goto st1007;
		case 54: goto st1009;
		case 79: goto st1011;
		case 80: goto st1013;
		case 111: goto st1011;
		case 112: goto st1013;
	}
	goto tr2951;
st1007:
	if ( ++p == pe )
		goto _test_eof1007;
case 1007:
	if ( (*p) == 50 )
		goto st1008;
	goto tr2951;
st1008:
	if ( ++p == pe )
		goto _test_eof1008;
case 1008:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3237;
		case 32: goto tr3237;
		case 40: goto tr3238;
		case 41: goto tr3239;
		case 1034: goto tr3240;
		case 1083: goto tr3241;
	}
	goto tr2951;
st1009:
	if ( ++p == pe )
		goto _test_eof1009;
case 1009:
	if ( (*p) == 52 )
		goto st1010;
	goto tr2951;
st1010:
	if ( ++p == pe )
		goto _test_eof1010;
case 1010:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3243;
		case 32: goto tr3243;
		case 40: goto tr3244;
		case 41: goto tr3245;
		case 1034: goto tr3246;
		case 1083: goto tr3247;
	}
	goto tr2951;
st1011:
	if ( ++p == pe )
		goto _test_eof1011;
case 1011:
	switch( (*p) ) {
		case 67: goto st1012;
		case 99: goto st1012;
	}
	goto tr2951;
st1012:
	if ( ++p == pe )
		goto _test_eof1012;
case 1012:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3249;
		case 32: goto tr3249;
		case 40: goto tr3250;
		case 41: goto tr3251;
		case 1034: goto tr3252;
		case 1083: goto tr3253;
	}
	goto tr2951;
st1013:
	if ( ++p == pe )
		goto _test_eof1013;
case 1013:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3254;
		case 32: goto tr3254;
		case 40: goto tr3255;
		case 41: goto tr3256;
		case 1034: goto tr3257;
		case 1083: goto tr3258;
	}
	goto tr2951;
st1014:
	if ( ++p == pe )
		goto _test_eof1014;
case 1014:
	switch( (*p) ) {
		case 73: goto st1015;
		case 88: goto st1019;
		case 105: goto st1015;
		case 120: goto st1019;
	}
	goto tr2951;
st1015:
	if ( ++p == pe )
		goto _test_eof1015;
case 1015:
	switch( (*p) ) {
		case 78: goto st1016;
		case 110: goto st1016;
	}
	goto tr2951;
st1016:
	if ( ++p == pe )
		goto _test_eof1016;
case 1016:
	switch( (*p) ) {
		case 70: goto st1017;
		case 102: goto st1017;
	}
	goto tr2951;
st1017:
	if ( ++p == pe )
		goto _test_eof1017;
case 1017:
	switch( (*p) ) {
		case 79: goto st1018;
		case 111: goto st1018;
	}
	goto tr2951;
st1018:
	if ( ++p == pe )
		goto _test_eof1018;
case 1018:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3264;
		case 32: goto tr3264;
		case 40: goto tr3265;
		case 41: goto tr3266;
		case 1034: goto tr3267;
		case 1083: goto tr3268;
	}
	goto tr2951;
st1019:
	if ( ++p == pe )
		goto _test_eof1019;
case 1019:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3269;
		case 32: goto tr3269;
		case 40: goto tr3270;
		case 41: goto tr3271;
		case 1034: goto tr3272;
		case 1083: goto tr3273;
	}
	goto tr2951;
st1020:
	if ( ++p == pe )
		goto _test_eof1020;
case 1020:
	switch( (*p) ) {
		case 65: goto st1021;
		case 73: goto st1025;
		case 83: goto st1027;
		case 97: goto st1021;
		case 105: goto st1025;
		case 115: goto st1027;
	}
	goto tr2951;
st1021:
	if ( ++p == pe )
		goto _test_eof1021;
case 1021:
	switch( (*p) ) {
		case 80: goto st1022;
		case 112: goto st1022;
	}
	goto tr2951;
st1022:
	if ( ++p == pe )
		goto _test_eof1022;
case 1022:
	switch( (*p) ) {
		case 84: goto st1023;
		case 116: goto st1023;
	}
	goto tr2951;
st1023:
	if ( ++p == pe )
		goto _test_eof1023;
case 1023:
	switch( (*p) ) {
		case 82: goto st1024;
		case 114: goto st1024;
	}
	goto tr2951;
st1024:
	if ( ++p == pe )
		goto _test_eof1024;
case 1024:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3280;
		case 32: goto tr3280;
		case 40: goto tr3281;
		case 41: goto tr3282;
		case 1034: goto tr3283;
		case 1083: goto tr3284;
	}
	goto tr2951;
st1025:
	if ( ++p == pe )
		goto _test_eof1025;
case 1025:
	switch( (*p) ) {
		case 68: goto st1026;
		case 100: goto st1026;
	}
	goto tr2951;
st1026:
	if ( ++p == pe )
		goto _test_eof1026;
case 1026:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3286;
		case 32: goto tr3286;
		case 40: goto tr3287;
		case 41: goto tr3288;
		case 1034: goto tr3289;
		case 1083: goto tr3290;
	}
	goto tr2951;
st1027:
	if ( ++p == pe )
		goto _test_eof1027;
case 1027:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3291;
		case 32: goto tr3291;
		case 40: goto tr3292;
		case 41: goto tr3293;
		case 69: goto st1028;
		case 101: goto st1028;
		case 1034: goto tr3295;
		case 1083: goto tr3296;
	}
	goto tr2951;
st1028:
	if ( ++p == pe )
		goto _test_eof1028;
case 1028:
	switch( (*p) ) {
		case 67: goto st1029;
		case 99: goto st1029;
	}
	goto tr2951;
st1029:
	if ( ++p == pe )
		goto _test_eof1029;
case 1029:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3298;
		case 32: goto tr3298;
		case 40: goto tr3299;
		case 41: goto tr3300;
		case 51: goto st1030;
		case 1034: goto tr3302;
		case 1083: goto tr3303;
	}
	goto tr2951;
st1030:
	if ( ++p == pe )
		goto _test_eof1030;
case 1030:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3304;
		case 32: goto tr3304;
		case 40: goto tr3305;
		case 41: goto tr3306;
		case 80: goto st1031;
		case 112: goto st1031;
		case 1034: goto tr3308;
		case 1083: goto tr3309;
	}
	goto tr2951;
st1031:
	if ( ++p == pe )
		goto _test_eof1031;
case 1031:
	switch( (*p) ) {
		case 65: goto st1032;
		case 97: goto st1032;
	}
	goto tr2951;
st1032:
	if ( ++p == pe )
		goto _test_eof1032;
case 1032:
	switch( (*p) ) {
		case 82: goto st1033;
		case 114: goto st1033;
	}
	goto tr2951;
st1033:
	if ( ++p == pe )
		goto _test_eof1033;
case 1033:
	switch( (*p) ) {
		case 65: goto st1034;
		case 97: goto st1034;
	}
	goto tr2951;
st1034:
	if ( ++p == pe )
		goto _test_eof1034;
case 1034:
	switch( (*p) ) {
		case 77: goto st1035;
		case 109: goto st1035;
	}
	goto tr2951;
st1035:
	if ( ++p == pe )
		goto _test_eof1035;
case 1035:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3314;
		case 32: goto tr3314;
		case 40: goto tr3315;
		case 41: goto tr3316;
		case 1034: goto tr3317;
		case 1083: goto tr3318;
	}
	goto tr2951;
st1036:
	if ( ++p == pe )
		goto _test_eof1036;
case 1036:
	switch( (*p) ) {
		case 80: goto st1037;
		case 112: goto st1037;
	}
	goto tr2951;
st1037:
	if ( ++p == pe )
		goto _test_eof1037;
case 1037:
	switch( (*p) ) {
		case 69: goto st1038;
		case 101: goto st1038;
	}
	goto tr2951;
st1038:
	if ( ++p == pe )
		goto _test_eof1038;
case 1038:
	switch( (*p) ) {
		case 78: goto st1039;
		case 110: goto st1039;
	}
	goto tr2951;
st1039:
	if ( ++p == pe )
		goto _test_eof1039;
case 1039:
	switch( (*p) ) {
		case 80: goto st1040;
		case 112: goto st1040;
	}
	goto tr2951;
st1040:
	if ( ++p == pe )
		goto _test_eof1040;
case 1040:
	switch( (*p) ) {
		case 71: goto st1041;
		case 103: goto st1041;
	}
	goto tr2951;
st1041:
	if ( ++p == pe )
		goto _test_eof1041;
case 1041:
	switch( (*p) ) {
		case 80: goto st1042;
		case 112: goto st1042;
	}
	goto tr2951;
st1042:
	if ( ++p == pe )
		goto _test_eof1042;
case 1042:
	switch( (*p) ) {
		case 75: goto st1043;
		case 107: goto st1043;
	}
	goto tr2951;
st1043:
	if ( ++p == pe )
		goto _test_eof1043;
case 1043:
	switch( (*p) ) {
		case 69: goto st1044;
		case 101: goto st1044;
	}
	goto tr2951;
st1044:
	if ( ++p == pe )
		goto _test_eof1044;
case 1044:
	switch( (*p) ) {
		case 89: goto st1045;
		case 121: goto st1045;
	}
	goto tr2951;
st1045:
	if ( ++p == pe )
		goto _test_eof1045;
case 1045:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3328;
		case 32: goto tr3328;
		case 40: goto tr3329;
		case 41: goto tr3330;
		case 1034: goto tr3331;
		case 1083: goto tr3332;
	}
	goto tr2951;
st1046:
	if ( ++p == pe )
		goto _test_eof1046;
case 1046:
	switch( (*p) ) {
		case 84: goto st1047;
		case 116: goto st1047;
	}
	goto tr2951;
st1047:
	if ( ++p == pe )
		goto _test_eof1047;
case 1047:
	switch( (*p) ) {
		case 82: goto st1048;
		case 114: goto st1048;
	}
	goto tr2951;
st1048:
	if ( ++p == pe )
		goto _test_eof1048;
case 1048:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3335;
		case 32: goto tr3335;
		case 40: goto tr3336;
		case 41: goto tr3337;
		case 1034: goto tr3338;
		case 1083: goto tr3339;
	}
	goto tr2951;
st1049:
	if ( ++p == pe )
		goto _test_eof1049;
case 1049:
	switch( (*p) ) {
		case 80: goto st1050;
		case 82: goto st1051;
		case 84: goto st1055;
		case 112: goto st1050;
		case 114: goto st1051;
		case 116: goto st1055;
	}
	goto tr2951;
st1050:
	if ( ++p == pe )
		goto _test_eof1050;
case 1050:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3343;
		case 32: goto tr3343;
		case 40: goto tr3344;
		case 41: goto tr3345;
		case 1034: goto tr3346;
		case 1083: goto tr3347;
	}
	goto tr2951;
st1051:
	if ( ++p == pe )
		goto _test_eof1051;
case 1051:
	switch( (*p) ) {
		case 83: goto st1052;
		case 115: goto st1052;
	}
	goto tr2951;
st1052:
	if ( ++p == pe )
		goto _test_eof1052;
case 1052:
	switch( (*p) ) {
		case 73: goto st1053;
		case 105: goto st1053;
	}
	goto tr2951;
st1053:
	if ( ++p == pe )
		goto _test_eof1053;
case 1053:
	switch( (*p) ) {
		case 71: goto st1054;
		case 103: goto st1054;
	}
	goto tr2951;
st1054:
	if ( ++p == pe )
		goto _test_eof1054;
case 1054:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3351;
		case 32: goto tr3351;
		case 40: goto tr3352;
		case 41: goto tr3353;
		case 1034: goto tr3354;
		case 1083: goto tr3355;
	}
	goto tr2951;
st1055:
	if ( ++p == pe )
		goto _test_eof1055;
case 1055:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3356;
		case 32: goto tr3356;
		case 40: goto tr3357;
		case 41: goto tr3358;
		case 1034: goto tr3359;
		case 1083: goto tr3360;
	}
	goto tr2951;
st1056:
	if ( ++p == pe )
		goto _test_eof1056;
case 1056:
	switch( (*p) ) {
		case 77: goto st1057;
		case 79: goto st1062;
		case 80: goto st1064;
		case 82: goto st1066;
		case 83: goto st1068;
		case 109: goto st1057;
		case 111: goto st1062;
		case 112: goto st1064;
		case 114: goto st1066;
		case 115: goto st1068;
	}
	goto tr2951;
st1057:
	if ( ++p == pe )
		goto _test_eof1057;
case 1057:
	switch( (*p) ) {
		case 73: goto st1058;
		case 105: goto st1058;
	}
	goto tr2951;
st1058:
	if ( ++p == pe )
		goto _test_eof1058;
case 1058:
	switch( (*p) ) {
		case 77: goto st1059;
		case 109: goto st1059;
	}
	goto tr2951;
st1059:
	if ( ++p == pe )
		goto _test_eof1059;
case 1059:
	switch( (*p) ) {
		case 69: goto st1060;
		case 101: goto st1060;
	}
	goto tr2951;
st1060:
	if ( ++p == pe )
		goto _test_eof1060;
case 1060:
	switch( (*p) ) {
		case 65: goto st1061;
		case 97: goto st1061;
	}
	goto tr2951;
st1061:
	if ( ++p == pe )
		goto _test_eof1061;
case 1061:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3370;
		case 32: goto tr3370;
		case 40: goto tr3371;
		case 41: goto tr3372;
		case 1034: goto tr3373;
		case 1083: goto tr3374;
	}
	goto tr2951;
st1062:
	if ( ++p == pe )
		goto _test_eof1062;
case 1062:
	switch( (*p) ) {
		case 65: goto st1063;
		case 97: goto st1063;
	}
	goto tr2951;
st1063:
	if ( ++p == pe )
		goto _test_eof1063;
case 1063:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3376;
		case 32: goto tr3376;
		case 40: goto tr3377;
		case 41: goto tr3378;
		case 1034: goto tr3379;
		case 1083: goto tr3380;
	}
	goto tr2951;
st1064:
	if ( ++p == pe )
		goto _test_eof1064;
case 1064:
	switch( (*p) ) {
		case 70: goto st1065;
		case 102: goto st1065;
	}
	goto tr2951;
st1065:
	if ( ++p == pe )
		goto _test_eof1065;
case 1065:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3382;
		case 32: goto tr3382;
		case 40: goto tr3383;
		case 41: goto tr3384;
		case 1034: goto tr3385;
		case 1083: goto tr3386;
	}
	goto tr2951;
st1066:
	if ( ++p == pe )
		goto _test_eof1066;
case 1066:
	switch( (*p) ) {
		case 86: goto st1067;
		case 118: goto st1067;
	}
	goto tr2951;
st1067:
	if ( ++p == pe )
		goto _test_eof1067;
case 1067:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3388;
		case 32: goto tr3388;
		case 40: goto tr3389;
		case 41: goto tr3390;
		case 1034: goto tr3391;
		case 1083: goto tr3392;
	}
	goto tr2951;
st1068:
	if ( ++p == pe )
		goto _test_eof1068;
case 1068:
	switch( (*p) ) {
		case 72: goto st1069;
		case 104: goto st1069;
	}
	goto tr2951;
st1069:
	if ( ++p == pe )
		goto _test_eof1069;
case 1069:
	switch( (*p) ) {
		case 70: goto st1070;
		case 102: goto st1070;
	}
	goto tr2951;
st1070:
	if ( ++p == pe )
		goto _test_eof1070;
case 1070:
	switch( (*p) ) {
		case 80: goto st1071;
		case 112: goto st1071;
	}
	goto tr2951;
st1071:
	if ( ++p == pe )
		goto _test_eof1071;
case 1071:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3396;
		case 32: goto tr3396;
		case 40: goto tr3397;
		case 41: goto tr3398;
		case 1034: goto tr3399;
		case 1083: goto tr3400;
	}
	goto tr2951;
st1072:
	if ( ++p == pe )
		goto _test_eof1072;
case 1072:
	switch( (*p) ) {
		case 76: goto st1073;
		case 88: goto st1076;
		case 89: goto st1078;
		case 108: goto st1073;
		case 120: goto st1076;
		case 121: goto st1078;
	}
	goto tr2951;
st1073:
	if ( ++p == pe )
		goto _test_eof1073;
case 1073:
	switch( (*p) ) {
		case 83: goto st1074;
		case 115: goto st1074;
	}
	goto tr2951;
st1074:
	if ( ++p == pe )
		goto _test_eof1074;
case 1074:
	switch( (*p) ) {
		case 65: goto st1075;
		case 97: goto st1075;
	}
	goto tr2951;
st1075:
	if ( ++p == pe )
		goto _test_eof1075;
case 1075:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3406;
		case 32: goto tr3406;
		case 40: goto tr3407;
		case 41: goto tr3408;
		case 1034: goto tr3409;
		case 1083: goto tr3410;
	}
	goto tr2951;
st1076:
	if ( ++p == pe )
		goto _test_eof1076;
case 1076:
	switch( (*p) ) {
		case 84: goto st1077;
		case 116: goto st1077;
	}
	goto tr2951;
st1077:
	if ( ++p == pe )
		goto _test_eof1077;
case 1077:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3412;
		case 32: goto tr3412;
		case 40: goto tr3413;
		case 41: goto tr3414;
		case 1034: goto tr3415;
		case 1083: goto tr3416;
	}
	goto tr2951;
st1078:
	if ( ++p == pe )
		goto _test_eof1078;
case 1078:
	switch( (*p) ) {
		case 80: goto st1079;
		case 112: goto st1079;
	}
	goto tr2951;
st1079:
	if ( ++p == pe )
		goto _test_eof1079;
case 1079:
	switch( (*p) ) {
		case 69: goto st1080;
		case 101: goto st1080;
	}
	goto tr2951;
st1080:
	if ( ++p == pe )
		goto _test_eof1080;
case 1080:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3420;
	goto tr3419;
tr3420:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1081;
tr3424:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1081;
st1081:
	if ( ++p == pe )
		goto _test_eof1081;
case 1081:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3421;
		case 32: goto tr3421;
		case 40: goto tr3422;
		case 41: goto tr3423;
		case 1034: goto tr3425;
		case 1083: goto tr3426;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3424;
	goto tr3419;
st1082:
	if ( ++p == pe )
		goto _test_eof1082;
case 1082:
	switch( (*p) ) {
		case 82: goto st1083;
		case 114: goto st1083;
	}
	goto tr2951;
st1083:
	if ( ++p == pe )
		goto _test_eof1083;
case 1083:
	switch( (*p) ) {
		case 73: goto st1084;
		case 105: goto st1084;
	}
	goto tr2951;
st1084:
	if ( ++p == pe )
		goto _test_eof1084;
case 1084:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3429;
		case 32: goto tr3429;
		case 40: goto tr3430;
		case 41: goto tr3431;
		case 1034: goto tr3432;
		case 1083: goto tr3433;
	}
	goto tr2951;
st1085:
	if ( ++p == pe )
		goto _test_eof1085;
case 1085:
	switch( (*p) ) {
		case 79: goto st1086;
		case 111: goto st1086;
	}
	goto tr2951;
st1086:
	if ( ++p == pe )
		goto _test_eof1086;
case 1086:
	switch( (*p) ) {
		case 78: goto st1087;
		case 110: goto st1087;
	}
	goto tr2951;
st1087:
	if ( ++p == pe )
		goto _test_eof1087;
case 1087:
	switch( (*p) ) {
		case 69: goto st1088;
		case 101: goto st1088;
	}
	goto tr2951;
st1088:
	if ( ++p == pe )
		goto _test_eof1088;
case 1088:
	switch( (*p) ) {
		case 77: goto st1089;
		case 109: goto st1089;
	}
	goto tr2951;
st1089:
	if ( ++p == pe )
		goto _test_eof1089;
case 1089:
	switch( (*p) ) {
		case 68: goto st1090;
		case 100: goto st1090;
	}
	goto tr2951;
st1090:
	if ( ++p == pe )
		goto _test_eof1090;
case 1090:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3439;
		case 32: goto tr3439;
		case 40: goto tr3440;
		case 41: goto tr3441;
		case 1034: goto tr3442;
		case 1083: goto tr3443;
	}
	goto tr2951;
st1091:
	if ( ++p == pe )
		goto _test_eof1091;
case 1091:
	switch( (*p) ) {
		case 42: goto tr3444;
		case 92: goto tr3444;
		case 95: goto tr3444;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr3444;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr3444;
	} else
		goto tr3444;
	goto tr75;
tr3444:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 1092;goto st301;} }
	goto st1092;
st1092:
	if ( ++p == pe )
		goto _test_eof1092;
case 1092:
	switch( (*p) ) {
		case 32: goto tr3445;
		case 59: goto tr3445;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3445;
	} else if ( (*p) >= 9 )
		goto tr3445;
	goto tr75;
tr3445:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{ p--; {stack[top++] = 1093;goto st367;} }
	goto st1093;
st1093:
	if ( ++p == pe )
		goto _test_eof1093;
case 1093:
	switch( (*p) ) {
		case 32: goto tr3446;
		case 59: goto tr3446;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3446;
	} else if ( (*p) >= 9 )
		goto tr3446;
	goto tr75;
tr3446:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1300;
st1300:
	if ( ++p == pe )
		goto _test_eof1300;
case 1300:
	goto st0;
st1094:
	if ( ++p == pe )
		goto _test_eof1094;
case 1094:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3447;
	goto tr2007;
tr3447:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1095;
tr3451:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1095;
st1095:
	if ( ++p == pe )
		goto _test_eof1095;
case 1095:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3448;
		case 32: goto tr3448;
		case 40: goto tr3449;
		case 41: goto tr3450;
		case 1034: goto tr3452;
		case 1083: goto tr3453;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3451;
	goto tr2007;
tr3455:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1096;
tr3456:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1096;
tr3458:
	{
		s->line_counter++;
	}
	goto st1096;
tr3484:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1096; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1096;
tr3448:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1096;
tr3449:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1096;
tr3450:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1096;
tr3452:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1096;
st1096:
	if ( ++p == pe )
		goto _test_eof1096;
case 1096:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1096;
		case 32: goto st1096;
		case 40: goto tr3455;
		case 41: goto tr3456;
		case 1034: goto tr3458;
		case 1083: goto tr3459;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3457;
	goto tr2007;
tr3457:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1097;
tr3463:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1097;
st1097:
	if ( ++p == pe )
		goto _test_eof1097;
case 1097:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3460;
		case 32: goto tr3460;
		case 40: goto tr3461;
		case 41: goto tr3462;
		case 1034: goto tr3464;
		case 1083: goto tr3465;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3463;
	goto tr2007;
tr3467:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1098;
tr3468:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1098;
tr3470:
	{
		s->line_counter++;
	}
	goto st1098;
tr3482:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1098; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1098;
tr3460:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1098;
tr3461:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1098;
tr3462:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1098;
tr3464:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1098;
st1098:
	if ( ++p == pe )
		goto _test_eof1098;
case 1098:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1098;
		case 32: goto st1098;
		case 40: goto tr3467;
		case 41: goto tr3468;
		case 1034: goto tr3470;
		case 1083: goto tr3471;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3469;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3469;
	} else
		goto tr3469;
	goto tr75;
tr3469:
	{ p--; {stack[top++] = 1099;goto st543;} }
	goto st1099;
st1099:
	if ( ++p == pe )
		goto _test_eof1099;
case 1099:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1100;
		case 32: goto st1100;
		case 40: goto tr3473;
		case 41: goto tr3474;
		case 1034: goto tr3475;
		case 1083: goto tr3476;
	}
	goto tr75;
tr3473:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1100;
tr3474:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1100;
tr3475:
	{
		s->line_counter++;
	}
	goto st1100;
tr3480:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1100; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1100;
st1100:
	if ( ++p == pe )
		goto _test_eof1100;
case 1100:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1100;
		case 32: goto st1100;
		case 40: goto tr3473;
		case 41: goto tr3474;
		case 43: goto tr3477;
		case 1034: goto tr3475;
		case 1083: goto tr3476;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr3477;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3477;
	} else
		goto tr3477;
	goto tr75;
tr3477:
	{ p--; {stack[top++] = 1101;goto st360;} }
	goto st1101;
st1101:
	if ( ++p == pe )
		goto _test_eof1101;
case 1101:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr3478;
		case 1595: goto tr3478;
	}
	goto tr75;
tr3478:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1301;
st1301:
	if ( ++p == pe )
		goto _test_eof1301;
case 1301:
	goto st0;
tr3476:
	{
		s->buffer_length = 0;
	}
	goto st1102;
tr3479:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1102;
st1102:
	if ( ++p == pe )
		goto _test_eof1102;
case 1102:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3480;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3479;
	goto tr75;
tr3471:
	{
		s->buffer_length = 0;
	}
	goto st1103;
tr3481:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1103;
tr3465:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1103;
st1103:
	if ( ++p == pe )
		goto _test_eof1103;
case 1103:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3482;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3481;
	goto tr75;
tr3459:
	{
		s->buffer_length = 0;
	}
	goto st1104;
tr3483:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1104;
tr3453:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1104;
st1104:
	if ( ++p == pe )
		goto _test_eof1104;
case 1104:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3484;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3483;
	goto tr75;
st1105:
	if ( ++p == pe )
		goto _test_eof1105;
case 1105:
	if ( (*p) == 43 )
		goto tr3485;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr3485;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr3485;
	} else
		goto tr3485;
	goto tr75;
tr3485:
	{ p--; {stack[top++] = 1106;goto st360;} }
	goto st1106;
st1106:
	if ( ++p == pe )
		goto _test_eof1106;
case 1106:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr3486;
		case 1595: goto tr3486;
	}
	goto tr75;
tr3486:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1302;
st1302:
	if ( ++p == pe )
		goto _test_eof1302;
case 1302:
	goto st0;
st1107:
	if ( ++p == pe )
		goto _test_eof1107;
case 1107:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3487;
	goto tr2007;
tr3487:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1108;
tr3491:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1108;
st1108:
	if ( ++p == pe )
		goto _test_eof1108;
case 1108:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3488;
		case 32: goto tr3488;
		case 40: goto tr3489;
		case 41: goto tr3490;
		case 1034: goto tr3492;
		case 1083: goto tr3493;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3491;
	goto tr2007;
tr3495:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1109;
tr3496:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1109;
tr3498:
	{
		s->line_counter++;
	}
	goto st1109;
tr3562:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1109; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1109;
tr3488:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1109;
tr3489:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1109;
tr3490:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1109;
tr3492:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1109;
st1109:
	if ( ++p == pe )
		goto _test_eof1109;
case 1109:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1109;
		case 32: goto st1109;
		case 40: goto tr3495;
		case 41: goto tr3496;
		case 1034: goto tr3498;
		case 1083: goto tr3499;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3497;
	goto tr2007;
tr3497:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1110;
tr3503:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1110;
st1110:
	if ( ++p == pe )
		goto _test_eof1110;
case 1110:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3500;
		case 32: goto tr3500;
		case 40: goto tr3501;
		case 41: goto tr3502;
		case 1034: goto tr3504;
		case 1083: goto tr3505;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3503;
	goto tr2007;
tr3507:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1111;
tr3508:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1111;
tr3510:
	{
		s->line_counter++;
	}
	goto st1111;
tr3560:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1111; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1111;
tr3500:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1111;
tr3501:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1111;
tr3502:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1111;
tr3504:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1111;
st1111:
	if ( ++p == pe )
		goto _test_eof1111;
case 1111:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1111;
		case 32: goto st1111;
		case 40: goto tr3507;
		case 41: goto tr3508;
		case 1034: goto tr3510;
		case 1083: goto tr3511;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3509;
	goto tr2007;
tr3509:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1112;
tr3515:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1112;
st1112:
	if ( ++p == pe )
		goto _test_eof1112;
case 1112:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3512;
		case 32: goto tr3512;
		case 40: goto tr3513;
		case 41: goto tr3514;
		case 1034: goto tr3516;
		case 1083: goto tr3517;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3515;
	goto tr2007;
tr3519:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1113;
tr3520:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1113;
tr3523:
	{
		s->line_counter++;
	}
	goto st1113;
tr3558:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1113; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1113;
tr3512:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1113;
tr3513:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1113;
tr3514:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1113;
tr3516:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1113;
st1113:
	if ( ++p == pe )
		goto _test_eof1113;
case 1113:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1113;
		case 32: goto st1113;
		case 40: goto tr3519;
		case 41: goto tr3520;
		case 45: goto tr3521;
		case 1034: goto tr3523;
		case 1083: goto tr3524;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3522;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3522;
	} else
		goto tr3522;
	goto tr2678;
tr3521:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1114;
st1114:
	if ( ++p == pe )
		goto _test_eof1114;
case 1114:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3525;
		case 32: goto tr3525;
		case 40: goto tr3526;
		case 41: goto tr3527;
		case 1034: goto tr3528;
		case 1083: goto tr3529;
	}
	goto tr2678;
tr3532:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1115;
tr3533:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1115;
tr3535:
	{
		s->line_counter++;
	}
	goto st1115;
tr3554:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1115; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1115;
tr3525:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1115;
tr3526:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1115;
tr3527:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1115;
tr3528:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1115;
st1115:
	if ( ++p == pe )
		goto _test_eof1115;
case 1115:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1115;
		case 32: goto st1115;
		case 40: goto tr3532;
		case 41: goto tr3533;
		case 1034: goto tr3535;
		case 1083: goto tr3536;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3534;
	} else if ( _widec > 86 ) {
		if ( 97 <= _widec && _widec <= 118 )
			goto tr3534;
	} else
		goto tr3534;
	goto tr3530;
tr3534:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1116;
tr3549:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1116;
st1116:
	if ( ++p == pe )
		goto _test_eof1116;
case 1116:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3537;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3537;
	} else
		goto tr3537;
	goto tr3530;
tr3537:
	{
		*(rdata_tail++) += second_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1117;
st1117:
	if ( ++p == pe )
		goto _test_eof1117;
case 1117:
	if ( (*p) == 61 )
		goto st1128;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3538;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3538;
	} else
		goto tr3538;
	goto tr3530;
tr3538:
	{
		*rdata_tail += third_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1118;
st1118:
	if ( ++p == pe )
		goto _test_eof1118;
case 1118:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3540;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3540;
	} else
		goto tr3540;
	goto tr3530;
tr3540:
	{
		*(rdata_tail++) += fourth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fourth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1119;
st1119:
	if ( ++p == pe )
		goto _test_eof1119;
case 1119:
	if ( (*p) == 61 )
		goto st1127;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3541;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3541;
	} else
		goto tr3541;
	goto tr3530;
tr3541:
	{
		*(rdata_tail++) += fifth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fifth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1120;
st1120:
	if ( ++p == pe )
		goto _test_eof1120;
case 1120:
	if ( (*p) == 61 )
		goto st1125;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3543;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3543;
	} else
		goto tr3543;
	goto tr3530;
tr3543:
	{
		*rdata_tail += sixth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1121;
st1121:
	if ( ++p == pe )
		goto _test_eof1121;
case 1121:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3545;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3545;
	} else
		goto tr3545;
	goto tr3530;
tr3545:
	{
		*(rdata_tail++) += seventh_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = seventh_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1122;
st1122:
	if ( ++p == pe )
		goto _test_eof1122;
case 1122:
	if ( (*p) == 61 )
		goto st1123;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3546;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3546;
	} else
		goto tr3546;
	goto tr3530;
tr3546:
	{
		*(rdata_tail++) += eighth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1123;
st1123:
	if ( ++p == pe )
		goto _test_eof1123;
case 1123:
	switch( (*p) ) {
		case 32: goto tr3548;
		case 59: goto tr3548;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3548;
		} else if ( (*p) >= 9 )
			goto tr3548;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 86 ) {
			if ( 97 <= (*p) && (*p) <= 118 )
				goto tr3549;
		} else if ( (*p) >= 65 )
			goto tr3549;
	} else
		goto tr3549;
	goto tr3530;
tr3548:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 1124;goto st367;} }
	goto st1124;
st1124:
	if ( ++p == pe )
		goto _test_eof1124;
case 1124:
	switch( (*p) ) {
		case 32: goto tr3550;
		case 59: goto tr3550;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3550;
	} else if ( (*p) >= 9 )
		goto tr3550;
	goto tr75;
tr3550:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1303;
st1303:
	if ( ++p == pe )
		goto _test_eof1303;
case 1303:
	goto st0;
st1125:
	if ( ++p == pe )
		goto _test_eof1125;
case 1125:
	if ( (*p) == 61 )
		goto st1126;
	goto tr3530;
st1126:
	if ( ++p == pe )
		goto _test_eof1126;
case 1126:
	if ( (*p) == 61 )
		goto st1123;
	goto tr3530;
st1127:
	if ( ++p == pe )
		goto _test_eof1127;
case 1127:
	if ( (*p) == 61 )
		goto st1125;
	goto tr3530;
st1128:
	if ( ++p == pe )
		goto _test_eof1128;
case 1128:
	if ( (*p) == 61 )
		goto st1129;
	goto tr3530;
st1129:
	if ( ++p == pe )
		goto _test_eof1129;
case 1129:
	if ( (*p) == 61 )
		goto st1127;
	goto tr3530;
tr3536:
	{
		s->buffer_length = 0;
	}
	goto st1130;
tr3553:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1130;
tr3529:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1130;
st1130:
	if ( ++p == pe )
		goto _test_eof1130;
case 1130:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3554;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3553;
	goto tr75;
tr3556:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1131;
tr3522:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1131;
st1131:
	if ( ++p == pe )
		goto _test_eof1131;
case 1131:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3555;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3555;
	} else
		goto tr3555;
	goto tr2678;
tr3555:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1132;
st1132:
	if ( ++p == pe )
		goto _test_eof1132;
case 1132:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3525;
		case 32: goto tr3525;
		case 40: goto tr3526;
		case 41: goto tr3527;
		case 1034: goto tr3528;
		case 1083: goto tr3529;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3556;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3556;
	} else
		goto tr3556;
	goto tr2678;
tr3524:
	{
		s->buffer_length = 0;
	}
	goto st1133;
tr3557:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1133;
tr3517:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1133;
st1133:
	if ( ++p == pe )
		goto _test_eof1133;
case 1133:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3558;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3557;
	goto tr75;
tr3511:
	{
		s->buffer_length = 0;
	}
	goto st1134;
tr3559:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1134;
tr3505:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1134;
st1134:
	if ( ++p == pe )
		goto _test_eof1134;
case 1134:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3560;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3559;
	goto tr75;
tr3499:
	{
		s->buffer_length = 0;
	}
	goto st1135;
tr3561:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1135;
tr3493:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1135;
st1135:
	if ( ++p == pe )
		goto _test_eof1135;
case 1135:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3562;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3561;
	goto tr75;
st1136:
	if ( ++p == pe )
		goto _test_eof1136;
case 1136:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3563;
	goto tr2007;
tr3563:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1137;
tr3567:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1137;
st1137:
	if ( ++p == pe )
		goto _test_eof1137;
case 1137:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3564;
		case 32: goto tr3564;
		case 40: goto tr3565;
		case 41: goto tr3566;
		case 1034: goto tr3568;
		case 1083: goto tr3569;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3567;
	goto tr2007;
tr3571:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1138;
tr3572:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1138;
tr3574:
	{
		s->line_counter++;
	}
	goto st1138;
tr3609:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1138; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1138;
tr3564:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1138;
tr3565:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1138;
tr3566:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1138;
tr3568:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1138;
st1138:
	if ( ++p == pe )
		goto _test_eof1138;
case 1138:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1138;
		case 32: goto st1138;
		case 40: goto tr3571;
		case 41: goto tr3572;
		case 1034: goto tr3574;
		case 1083: goto tr3575;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3573;
	goto tr2007;
tr3573:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1139;
tr3579:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1139;
st1139:
	if ( ++p == pe )
		goto _test_eof1139;
case 1139:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3576;
		case 32: goto tr3576;
		case 40: goto tr3577;
		case 41: goto tr3578;
		case 1034: goto tr3580;
		case 1083: goto tr3581;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3579;
	goto tr2007;
tr3583:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1140;
tr3584:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1140;
tr3586:
	{
		s->line_counter++;
	}
	goto st1140;
tr3607:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1140; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1140;
tr3576:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1140;
tr3577:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1140;
tr3578:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1140;
tr3580:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1140;
st1140:
	if ( ++p == pe )
		goto _test_eof1140;
case 1140:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1140;
		case 32: goto st1140;
		case 40: goto tr3583;
		case 41: goto tr3584;
		case 1034: goto tr3586;
		case 1083: goto tr3587;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3585;
	goto tr2007;
tr3585:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1141;
tr3591:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1141;
st1141:
	if ( ++p == pe )
		goto _test_eof1141;
case 1141:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3588;
		case 32: goto tr3588;
		case 40: goto tr3589;
		case 41: goto tr3590;
		case 1034: goto tr3592;
		case 1083: goto tr3593;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3591;
	goto tr2007;
tr3595:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1142;
tr3596:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1142;
tr3599:
	{
		s->line_counter++;
	}
	goto st1142;
tr3605:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1142; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1142;
tr3588:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1142;
tr3589:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1142;
tr3590:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1142;
tr3592:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1142;
st1142:
	if ( ++p == pe )
		goto _test_eof1142;
case 1142:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1142;
		case 32: goto st1142;
		case 40: goto tr3595;
		case 41: goto tr3596;
		case 45: goto tr3597;
		case 1034: goto tr3599;
		case 1083: goto tr3600;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3598;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3598;
	} else
		goto tr3598;
	goto tr2678;
tr3597:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1143;
st1143:
	if ( ++p == pe )
		goto _test_eof1143;
case 1143:
	switch( (*p) ) {
		case 32: goto tr3601;
		case 59: goto tr3601;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3601;
	} else if ( (*p) >= 9 )
		goto tr3601;
	goto tr2678;
tr3601:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1304;
st1304:
	if ( ++p == pe )
		goto _test_eof1304;
case 1304:
	goto st0;
tr3603:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1144;
tr3598:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1144;
st1144:
	if ( ++p == pe )
		goto _test_eof1144;
case 1144:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3602;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3602;
	} else
		goto tr3602;
	goto tr2678;
tr3602:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1145;
st1145:
	if ( ++p == pe )
		goto _test_eof1145;
case 1145:
	switch( (*p) ) {
		case 32: goto tr3601;
		case 59: goto tr3601;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3601;
		} else if ( (*p) >= 9 )
			goto tr3601;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr3603;
		} else if ( (*p) >= 65 )
			goto tr3603;
	} else
		goto tr3603;
	goto tr2678;
tr3600:
	{
		s->buffer_length = 0;
	}
	goto st1146;
tr3604:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1146;
tr3593:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1146;
st1146:
	if ( ++p == pe )
		goto _test_eof1146;
case 1146:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3605;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3604;
	goto tr75;
tr3587:
	{
		s->buffer_length = 0;
	}
	goto st1147;
tr3606:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1147;
tr3581:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1147;
st1147:
	if ( ++p == pe )
		goto _test_eof1147;
case 1147:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3607;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3606;
	goto tr75;
tr3575:
	{
		s->buffer_length = 0;
	}
	goto st1148;
tr3608:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1148;
tr3569:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1148;
st1148:
	if ( ++p == pe )
		goto _test_eof1148;
case 1148:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3609;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3608;
	goto tr75;
st1149:
	if ( ++p == pe )
		goto _test_eof1149;
case 1149:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3610;
	goto tr2007;
tr3610:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1150;
tr3614:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1150;
st1150:
	if ( ++p == pe )
		goto _test_eof1150;
case 1150:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3611;
		case 32: goto tr3611;
		case 40: goto tr3612;
		case 41: goto tr3613;
		case 1034: goto tr3615;
		case 1083: goto tr3616;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3614;
	goto tr2007;
tr3618:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1151;
tr3619:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1151;
tr3621:
	{
		s->line_counter++;
	}
	goto st1151;
tr3663:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1151; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1151;
tr3611:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1151;
tr3612:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1151;
tr3613:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1151;
tr3615:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1151;
st1151:
	if ( ++p == pe )
		goto _test_eof1151;
case 1151:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1151;
		case 32: goto st1151;
		case 40: goto tr3618;
		case 41: goto tr3619;
		case 1034: goto tr3621;
		case 1083: goto tr3622;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3620;
	goto tr2007;
tr3620:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1152;
tr3626:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1152;
st1152:
	if ( ++p == pe )
		goto _test_eof1152;
case 1152:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3623;
		case 32: goto tr3623;
		case 40: goto tr3624;
		case 41: goto tr3625;
		case 1034: goto tr3627;
		case 1083: goto tr3628;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3626;
	goto tr2007;
tr3630:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1153;
tr3631:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1153;
tr3633:
	{
		s->line_counter++;
	}
	goto st1153;
tr3661:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1153; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1153;
tr3623:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1153;
tr3624:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1153;
tr3625:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1153;
tr3627:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1153;
st1153:
	if ( ++p == pe )
		goto _test_eof1153;
case 1153:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1153;
		case 32: goto st1153;
		case 40: goto tr3630;
		case 41: goto tr3631;
		case 1034: goto tr3633;
		case 1083: goto tr3634;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3632;
	goto tr2007;
tr3632:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1154;
tr3638:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1154;
st1154:
	if ( ++p == pe )
		goto _test_eof1154;
case 1154:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3635;
		case 32: goto tr3635;
		case 40: goto tr3636;
		case 41: goto tr3637;
		case 1034: goto tr3639;
		case 1083: goto tr3640;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3638;
	goto tr2007;
tr3642:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1155;
tr3643:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1155;
tr3645:
	{
		s->line_counter++;
	}
	goto st1155;
tr3659:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1155; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1155;
tr3635:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1155;
tr3636:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1155;
tr3637:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1155;
tr3639:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1155;
st1155:
	if ( ++p == pe )
		goto _test_eof1155;
case 1155:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1155;
		case 32: goto st1155;
		case 40: goto tr3642;
		case 41: goto tr3643;
		case 1034: goto tr3645;
		case 1083: goto tr3646;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3644;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3644;
	} else
		goto tr3644;
	goto tr2678;
tr3644:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1156;
st1156:
	if ( ++p == pe )
		goto _test_eof1156;
case 1156:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3647;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3647;
	} else
		goto tr3647;
	goto tr2678;
tr3649:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1157;
tr3650:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1157;
tr3651:
	{
		s->line_counter++;
	}
	goto st1157;
tr3657:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1157; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1157;
tr3647:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1157;
st1157:
	if ( ++p == pe )
		goto _test_eof1157;
case 1157:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1157;
		case 32: goto st1157;
		case 40: goto tr3649;
		case 41: goto tr3650;
		case 2058: goto tr3651;
		case 2107: goto tr3652;
		case 2314: goto tr3653;
		case 2363: goto tr3653;
		case 2570: goto tr3654;
		case 2619: goto tr3655;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3644;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3644;
	} else
		goto tr3644;
	goto tr2678;
tr3652:
	{
		s->buffer_length = 0;
	}
	goto st1158;
tr3656:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1158;
st1158:
	if ( ++p == pe )
		goto _test_eof1158;
case 1158:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3657;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3656;
	goto tr2678;
tr3653:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1305;
st1305:
	if ( ++p == pe )
		goto _test_eof1305;
case 1305:
	goto st0;
tr3654:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1306;
st1306:
	if ( ++p == pe )
		goto _test_eof1306;
case 1306:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1157;
		case 32: goto st1157;
		case 40: goto tr3649;
		case 41: goto tr3650;
		case 2058: goto tr3651;
		case 2107: goto tr3652;
		case 2314: goto tr3653;
		case 2363: goto tr3653;
		case 2570: goto tr3654;
		case 2619: goto tr3655;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3644;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3644;
	} else
		goto tr3644;
	goto tr2678;
tr3655:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1307;
st1307:
	if ( ++p == pe )
		goto _test_eof1307;
case 1307:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3657;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3656;
	goto tr2678;
tr3646:
	{
		s->buffer_length = 0;
	}
	goto st1159;
tr3658:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1159;
tr3640:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1159;
st1159:
	if ( ++p == pe )
		goto _test_eof1159;
case 1159:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3659;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3658;
	goto tr75;
tr3634:
	{
		s->buffer_length = 0;
	}
	goto st1160;
tr3660:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1160;
tr3628:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1160;
st1160:
	if ( ++p == pe )
		goto _test_eof1160;
case 1160:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3661;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3660;
	goto tr75;
tr3622:
	{
		s->buffer_length = 0;
	}
	goto st1161;
tr3662:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1161;
tr3616:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1161;
st1161:
	if ( ++p == pe )
		goto _test_eof1161;
case 1161:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3663;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3662;
	goto tr75;
st1162:
	if ( ++p == pe )
		goto _test_eof1162;
case 1162:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3664;
	goto tr2007;
tr3664:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1163;
tr3668:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1163;
st1163:
	if ( ++p == pe )
		goto _test_eof1163;
case 1163:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3665;
		case 32: goto tr3665;
		case 40: goto tr3666;
		case 41: goto tr3667;
		case 1034: goto tr3669;
		case 1083: goto tr3670;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3668;
	goto tr2007;
tr3672:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1164;
tr3673:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1164;
tr3675:
	{
		s->line_counter++;
	}
	goto st1164;
tr3681:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1164; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1164;
tr3665:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1164;
tr3666:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1164;
tr3667:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1164;
tr3669:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1164;
st1164:
	if ( ++p == pe )
		goto _test_eof1164;
case 1164:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1164;
		case 32: goto st1164;
		case 40: goto tr3672;
		case 41: goto tr3673;
		case 1034: goto tr3675;
		case 1083: goto tr3676;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3674;
	goto tr2007;
tr3674:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1165;
tr3678:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1165;
st1165:
	if ( ++p == pe )
		goto _test_eof1165;
case 1165:
	switch( (*p) ) {
		case 32: goto tr3677;
		case 59: goto tr3677;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr3677;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3678;
	} else
		goto tr3677;
	goto tr2007;
tr3677:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 1166;goto st367;} }
	goto st1166;
st1166:
	if ( ++p == pe )
		goto _test_eof1166;
case 1166:
	switch( (*p) ) {
		case 32: goto tr3679;
		case 59: goto tr3679;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3679;
	} else if ( (*p) >= 9 )
		goto tr3679;
	goto tr75;
tr3679:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1308;
st1308:
	if ( ++p == pe )
		goto _test_eof1308;
case 1308:
	goto st0;
tr3676:
	{
		s->buffer_length = 0;
	}
	goto st1167;
tr3680:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1167;
tr3670:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1167;
st1167:
	if ( ++p == pe )
		goto _test_eof1167;
case 1167:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3681;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3680;
	goto tr75;
st1168:
	if ( ++p == pe )
		goto _test_eof1168;
case 1168:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3682;
	goto tr2007;
tr3682:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1169;
tr3686:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1169;
st1169:
	if ( ++p == pe )
		goto _test_eof1169;
case 1169:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3683;
		case 32: goto tr3683;
		case 40: goto tr3684;
		case 41: goto tr3685;
		case 1034: goto tr3687;
		case 1083: goto tr3688;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3686;
	goto tr2007;
tr3690:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1170;
tr3691:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1170;
tr3693:
	{
		s->line_counter++;
	}
	goto st1170;
tr3735:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1170; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1170;
tr3683:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1170;
tr3684:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1170;
tr3685:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1170;
tr3687:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1170;
st1170:
	if ( ++p == pe )
		goto _test_eof1170;
case 1170:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1170;
		case 32: goto st1170;
		case 40: goto tr3690;
		case 41: goto tr3691;
		case 1034: goto tr3693;
		case 1083: goto tr3694;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3692;
	goto tr2007;
tr3692:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1171;
tr3698:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1171;
st1171:
	if ( ++p == pe )
		goto _test_eof1171;
case 1171:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3695;
		case 32: goto tr3695;
		case 40: goto tr3696;
		case 41: goto tr3697;
		case 1034: goto tr3699;
		case 1083: goto tr3700;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3698;
	goto tr2007;
tr3702:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1172;
tr3703:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1172;
tr3705:
	{
		s->line_counter++;
	}
	goto st1172;
tr3733:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1172; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1172;
tr3695:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1172;
tr3696:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1172;
tr3697:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1172;
tr3699:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1172;
st1172:
	if ( ++p == pe )
		goto _test_eof1172;
case 1172:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1172;
		case 32: goto st1172;
		case 40: goto tr3702;
		case 41: goto tr3703;
		case 1034: goto tr3705;
		case 1083: goto tr3706;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3704;
	goto tr2007;
tr3704:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1173;
tr3710:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1173;
st1173:
	if ( ++p == pe )
		goto _test_eof1173;
case 1173:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3707;
		case 32: goto tr3707;
		case 40: goto tr3708;
		case 41: goto tr3709;
		case 1034: goto tr3711;
		case 1083: goto tr3712;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3710;
	goto tr2007;
tr3714:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1174;
tr3715:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1174;
tr3717:
	{
		s->line_counter++;
	}
	goto st1174;
tr3731:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1174; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1174;
tr3707:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1174;
tr3708:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1174;
tr3709:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1174;
tr3711:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1174;
st1174:
	if ( ++p == pe )
		goto _test_eof1174;
case 1174:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1174;
		case 32: goto st1174;
		case 40: goto tr3714;
		case 41: goto tr3715;
		case 1034: goto tr3717;
		case 1083: goto tr3718;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3716;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3716;
	} else
		goto tr3716;
	goto tr2678;
tr3716:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1175;
st1175:
	if ( ++p == pe )
		goto _test_eof1175;
case 1175:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3719;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3719;
	} else
		goto tr3719;
	goto tr2678;
tr3721:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1176;
tr3722:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1176;
tr3723:
	{
		s->line_counter++;
	}
	goto st1176;
tr3729:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1176; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1176;
tr3719:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1176;
st1176:
	if ( ++p == pe )
		goto _test_eof1176;
case 1176:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1176;
		case 32: goto st1176;
		case 40: goto tr3721;
		case 41: goto tr3722;
		case 2058: goto tr3723;
		case 2107: goto tr3724;
		case 2314: goto tr3725;
		case 2363: goto tr3725;
		case 2570: goto tr3726;
		case 2619: goto tr3727;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3716;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3716;
	} else
		goto tr3716;
	goto tr2678;
tr3724:
	{
		s->buffer_length = 0;
	}
	goto st1177;
tr3728:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1177;
st1177:
	if ( ++p == pe )
		goto _test_eof1177;
case 1177:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3729;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3728;
	goto tr2678;
tr3725:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1309;
st1309:
	if ( ++p == pe )
		goto _test_eof1309;
case 1309:
	goto st0;
tr3726:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1310;
st1310:
	if ( ++p == pe )
		goto _test_eof1310;
case 1310:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1176;
		case 32: goto st1176;
		case 40: goto tr3721;
		case 41: goto tr3722;
		case 2058: goto tr3723;
		case 2107: goto tr3724;
		case 2314: goto tr3725;
		case 2363: goto tr3725;
		case 2570: goto tr3726;
		case 2619: goto tr3727;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3716;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3716;
	} else
		goto tr3716;
	goto tr2678;
tr3727:
	{
		s->buffer_length = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1311;
st1311:
	if ( ++p == pe )
		goto _test_eof1311;
case 1311:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3729;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3728;
	goto tr2678;
tr3718:
	{
		s->buffer_length = 0;
	}
	goto st1178;
tr3730:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1178;
tr3712:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1178;
st1178:
	if ( ++p == pe )
		goto _test_eof1178;
case 1178:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3731;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3730;
	goto tr75;
tr3706:
	{
		s->buffer_length = 0;
	}
	goto st1179;
tr3732:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1179;
tr3700:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1179;
st1179:
	if ( ++p == pe )
		goto _test_eof1179;
case 1179:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3733;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3732;
	goto tr75;
tr3694:
	{
		s->buffer_length = 0;
	}
	goto st1180;
tr3734:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1180;
tr3688:
	{
		if (s->number64 <= UINT32_MAX) {
			uint32_t num32 = htonl((uint32_t)s->number64);
			memcpy(rdata_tail, &num32, 4);
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1180;
st1180:
	if ( ++p == pe )
		goto _test_eof1180;
case 1180:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3735;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3734;
	goto tr75;
st1181:
	if ( ++p == pe )
		goto _test_eof1181;
case 1181:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3736;
	goto tr2007;
tr3736:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1182;
tr3740:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1182;
st1182:
	if ( ++p == pe )
		goto _test_eof1182;
case 1182:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3737;
		case 32: goto tr3737;
		case 40: goto tr3738;
		case 41: goto tr3739;
		case 1034: goto tr3741;
		case 1083: goto tr3742;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3740;
	goto tr2007;
tr3744:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1183;
tr3745:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1183;
tr3747:
	{
		s->line_counter++;
	}
	goto st1183;
tr3752:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1183; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1183;
tr3737:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1183;
tr3738:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1183;
tr3739:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1183;
tr3741:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1183;
st1183:
	if ( ++p == pe )
		goto _test_eof1183;
case 1183:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1183;
		case 32: goto st1183;
		case 40: goto tr3744;
		case 41: goto tr3745;
		case 46: goto tr3746;
		case 1034: goto tr3747;
		case 1083: goto tr3748;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3746;
	goto tr1984;
tr3746:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1184;
tr3750:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1184;
st1184:
	if ( ++p == pe )
		goto _test_eof1184;
case 1184:
	switch( (*p) ) {
		case 32: goto tr3749;
		case 46: goto tr3750;
		case 59: goto tr3749;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr3749;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3750;
	} else
		goto tr3749;
	goto tr1984;
tr3749:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, s->addr) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st299;}
		}
	}
	{
		memcpy(rdata_tail, s->addr, ZS_INET4_ADDR_LENGTH);
		rdata_tail += ZS_INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1312;
st1312:
	if ( ++p == pe )
		goto _test_eof1312;
case 1312:
	goto st0;
tr3748:
	{
		s->buffer_length = 0;
	}
	goto st1185;
tr3751:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1185;
tr3742:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1185;
st1185:
	if ( ++p == pe )
		goto _test_eof1185;
case 1185:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3752;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3751;
	goto tr75;
st1186:
	if ( ++p == pe )
		goto _test_eof1186;
case 1186:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3753;
	goto tr2007;
tr3753:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1187;
tr3757:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1187;
st1187:
	if ( ++p == pe )
		goto _test_eof1187;
case 1187:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3754;
		case 32: goto tr3754;
		case 40: goto tr3755;
		case 41: goto tr3756;
		case 1034: goto tr3758;
		case 1083: goto tr3759;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3757;
	goto tr2007;
tr3761:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1188;
tr3762:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1188;
tr3764:
	{
		s->line_counter++;
	}
	goto st1188;
tr3778:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1188; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1188;
tr3754:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1188;
tr3755:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1188;
tr3756:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1188;
tr3758:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1188;
st1188:
	if ( ++p == pe )
		goto _test_eof1188;
case 1188:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1188;
		case 32: goto st1188;
		case 40: goto tr3761;
		case 41: goto tr3762;
		case 1034: goto tr3764;
		case 1083: goto tr3765;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3763;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3763;
	} else
		goto tr3763;
	goto tr2678;
tr3763:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1189;
st1189:
	if ( ++p == pe )
		goto _test_eof1189;
case 1189:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3766;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3766;
	} else
		goto tr3766;
	goto tr2678;
tr3766:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1190;
st1190:
	if ( ++p == pe )
		goto _test_eof1190;
case 1190:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3767;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3767;
	} else
		goto tr3767;
	goto tr2678;
tr3767:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1191;
st1191:
	if ( ++p == pe )
		goto _test_eof1191;
case 1191:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3768;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3768;
	} else
		goto tr3768;
	goto tr2678;
tr3768:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1192;
st1192:
	if ( ++p == pe )
		goto _test_eof1192;
case 1192:
	if ( (*p) == 58 )
		goto tr3770;
	goto tr3769;
tr3770:
	{
		s->item_length++;
	}
	goto st1193;
st1193:
	if ( ++p == pe )
		goto _test_eof1193;
case 1193:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3771;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3771;
	} else
		goto tr3771;
	goto tr2678;
tr3771:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1194;
st1194:
	if ( ++p == pe )
		goto _test_eof1194;
case 1194:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3772;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3772;
	} else
		goto tr3772;
	goto tr2678;
tr3772:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1195;
st1195:
	if ( ++p == pe )
		goto _test_eof1195;
case 1195:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3773;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3773;
	} else
		goto tr3773;
	goto tr2678;
tr3773:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1196;
st1196:
	if ( ++p == pe )
		goto _test_eof1196;
case 1196:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3774;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3774;
	} else
		goto tr3774;
	goto tr2678;
tr3774:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1197;
st1197:
	if ( ++p == pe )
		goto _test_eof1197;
case 1197:
	switch( (*p) ) {
		case 32: goto tr3776;
		case 58: goto tr3770;
		case 59: goto tr3776;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3776;
	} else if ( (*p) >= 9 )
		goto tr3776;
	goto tr3775;
tr3776:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 4) {
			WARN(ZS_BAD_L64_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1313;
st1313:
	if ( ++p == pe )
		goto _test_eof1313;
case 1313:
	goto st0;
tr3765:
	{
		s->buffer_length = 0;
	}
	goto st1198;
tr3777:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1198;
tr3759:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1198;
st1198:
	if ( ++p == pe )
		goto _test_eof1198;
case 1198:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3778;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3777;
	goto tr75;
st1199:
	if ( ++p == pe )
		goto _test_eof1199;
case 1199:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3779;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3779;
	} else
		goto tr3779;
	goto tr2678;
tr3779:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1200;
st1200:
	if ( ++p == pe )
		goto _test_eof1200;
case 1200:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3780;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3780;
	} else
		goto tr3780;
	goto tr2678;
tr3780:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1201;
st1201:
	if ( ++p == pe )
		goto _test_eof1201;
case 1201:
	if ( (*p) == 45 )
		goto tr3782;
	goto tr3781;
tr3782:
	{
		s->item_length++;
	}
	goto st1202;
st1202:
	if ( ++p == pe )
		goto _test_eof1202;
case 1202:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3783;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3783;
	} else
		goto tr3783;
	goto tr2678;
tr3783:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1203;
st1203:
	if ( ++p == pe )
		goto _test_eof1203;
case 1203:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3784;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3784;
	} else
		goto tr3784;
	goto tr2678;
tr3784:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1204;
st1204:
	if ( ++p == pe )
		goto _test_eof1204;
case 1204:
	switch( (*p) ) {
		case 32: goto tr3785;
		case 45: goto tr3782;
		case 59: goto tr3785;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3785;
	} else if ( (*p) >= 9 )
		goto tr3785;
	goto tr3781;
tr3785:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 6) {
			WARN(ZS_BAD_EUI_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1314;
st1314:
	if ( ++p == pe )
		goto _test_eof1314;
case 1314:
	goto st0;
st1205:
	if ( ++p == pe )
		goto _test_eof1205;
case 1205:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3786;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3786;
	} else
		goto tr3786;
	goto tr2678;
tr3786:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1206;
st1206:
	if ( ++p == pe )
		goto _test_eof1206;
case 1206:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3787;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3787;
	} else
		goto tr3787;
	goto tr2678;
tr3787:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1207;
st1207:
	if ( ++p == pe )
		goto _test_eof1207;
case 1207:
	if ( (*p) == 45 )
		goto tr3788;
	goto tr3781;
tr3788:
	{
		s->item_length++;
	}
	goto st1208;
st1208:
	if ( ++p == pe )
		goto _test_eof1208;
case 1208:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3789;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3789;
	} else
		goto tr3789;
	goto tr2678;
tr3789:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1209;
st1209:
	if ( ++p == pe )
		goto _test_eof1209;
case 1209:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3790;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3790;
	} else
		goto tr3790;
	goto tr2678;
tr3790:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1210;
st1210:
	if ( ++p == pe )
		goto _test_eof1210;
case 1210:
	switch( (*p) ) {
		case 32: goto tr3791;
		case 45: goto tr3788;
		case 59: goto tr3791;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3791;
	} else if ( (*p) >= 9 )
		goto tr3791;
	goto tr3781;
tr3791:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 8) {
			WARN(ZS_BAD_EUI_LENGTH);
			p--; {goto st299;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1315;
st1315:
	if ( ++p == pe )
		goto _test_eof1315;
case 1315:
	goto st0;
st1211:
	if ( ++p == pe )
		goto _test_eof1211;
case 1211:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3792;
	goto tr2007;
tr3792:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1212;
tr3796:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1212;
st1212:
	if ( ++p == pe )
		goto _test_eof1212;
case 1212:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3793;
		case 32: goto tr3793;
		case 40: goto tr3794;
		case 41: goto tr3795;
		case 1034: goto tr3797;
		case 1083: goto tr3798;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3796;
	goto tr2007;
tr3800:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1213;
tr3801:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1213;
tr3803:
	{
		s->line_counter++;
	}
	goto st1213;
tr3821:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1213; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1213;
tr3793:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1213;
tr3794:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1213;
tr3795:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1213;
tr3797:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1213;
st1213:
	if ( ++p == pe )
		goto _test_eof1213;
case 1213:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1213;
		case 32: goto st1213;
		case 40: goto tr3800;
		case 41: goto tr3801;
		case 1034: goto tr3803;
		case 1083: goto tr3804;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3802;
	goto tr2007;
tr3802:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1214;
tr3808:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1214;
st1214:
	if ( ++p == pe )
		goto _test_eof1214;
case 1214:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3805;
		case 32: goto tr3805;
		case 40: goto tr3806;
		case 41: goto tr3807;
		case 1034: goto tr3809;
		case 1083: goto tr3810;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3808;
	goto tr2007;
tr3813:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1215;
tr3814:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1215;
tr3815:
	{
		s->line_counter++;
	}
	goto st1215;
tr3819:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1215; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1215;
tr3805:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1215;
tr3806:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1215;
tr3807:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1215;
tr3809:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1215;
st1215:
	if ( ++p == pe )
		goto _test_eof1215;
case 1215:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1215;
		case 32: goto st1215;
		case 40: goto tr3813;
		case 41: goto tr3814;
		case 1034: goto tr3815;
		case 1083: goto tr3816;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr3811;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr3811;
	} else
		goto tr3811;
	goto tr75;
tr3811:
	{ p--; {stack[top++] = 1216;goto st310;} }
	goto st1216;
st1216:
	if ( ++p == pe )
		goto _test_eof1216;
case 1216:
	switch( (*p) ) {
		case 32: goto tr3817;
		case 59: goto tr3817;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3817;
	} else if ( (*p) >= 9 )
		goto tr3817;
	goto tr75;
tr3817:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1316;
st1316:
	if ( ++p == pe )
		goto _test_eof1316;
case 1316:
	goto st0;
tr3816:
	{
		s->buffer_length = 0;
	}
	goto st1217;
tr3818:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1217;
tr3810:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1217;
st1217:
	if ( ++p == pe )
		goto _test_eof1217;
case 1217:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3819;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3818;
	goto tr75;
tr3804:
	{
		s->buffer_length = 0;
	}
	goto st1218;
tr3820:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1218;
tr3798:
	{
		if (s->number64 <= UINT16_MAX) {
			uint16_t num16 = htons((uint16_t)s->number64);
			memcpy(rdata_tail, &num16, 2);
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1218;
st1218:
	if ( ++p == pe )
		goto _test_eof1218;
case 1218:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3821;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3820;
	goto tr75;
st1219:
	if ( ++p == pe )
		goto _test_eof1219;
case 1219:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3822;
	goto tr2007;
tr3822:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1220;
tr3826:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1220;
st1220:
	if ( ++p == pe )
		goto _test_eof1220;
case 1220:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3823;
		case 32: goto tr3823;
		case 40: goto tr3824;
		case 41: goto tr3825;
		case 1034: goto tr3827;
		case 1083: goto tr3828;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3826;
	goto tr2007;
tr3831:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1221;
tr3832:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1221;
tr3833:
	{
		s->line_counter++;
	}
	goto st1221;
tr3850:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1221; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1221;
tr3823:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1221;
tr3824:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1221;
tr3825:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1221;
tr3827:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1221;
st1221:
	if ( ++p == pe )
		goto _test_eof1221;
case 1221:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1221;
		case 32: goto st1221;
		case 40: goto tr3831;
		case 41: goto tr3832;
		case 1034: goto tr3833;
		case 1083: goto tr3834;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr3829;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr3829;
	} else
		goto tr3829;
	goto tr75;
tr3829:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{ p--; {stack[top++] = 1222;goto st310;} }
	goto st1222;
st1222:
	if ( ++p == pe )
		goto _test_eof1222;
case 1222:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3835;
		case 32: goto tr3835;
		case 40: goto tr3836;
		case 41: goto tr3837;
		case 1034: goto tr3838;
		case 1083: goto tr3839;
	}
	goto tr75;
tr3842:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1223;
tr3843:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1223;
tr3844:
	{
		s->line_counter++;
	}
	goto st1223;
tr3848:
	{
		s->buffer[s->buffer_length++] = 0;

		// Execute the comment callback.
		if (s->process.automatic && s->process.comment != NULL) {
			s->process.comment(s);

			// Stop if required from the callback.
			if (s->state == ZS_STATE_STOP) {
				{p++; cs = 1223; goto _out;}
			}
		}
	}
	{
		s->line_counter++;
	}
	goto st1223;
tr3835:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	goto st1223;
tr3836:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = true;
	}
	goto st1223;
tr3837:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st299;}
		}
		s->multiline = false;
	}
	goto st1223;
tr3838:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1223;
st1223:
	if ( ++p == pe )
		goto _test_eof1223;
case 1223:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1223;
		case 32: goto st1223;
		case 40: goto tr3842;
		case 41: goto tr3843;
		case 1034: goto tr3844;
		case 1083: goto tr3845;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr3840;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr3840;
	} else
		goto tr3840;
	goto tr75;
tr3840:
	{ p--; {stack[top++] = 1224;goto st310;} }
	goto st1224;
st1224:
	if ( ++p == pe )
		goto _test_eof1224;
case 1224:
	switch( (*p) ) {
		case 32: goto tr3846;
		case 59: goto tr3846;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3846;
	} else if ( (*p) >= 9 )
		goto tr3846;
	goto tr75;
tr3846:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1317;
st1317:
	if ( ++p == pe )
		goto _test_eof1317;
case 1317:
	goto st0;
tr3845:
	{
		s->buffer_length = 0;
	}
	goto st1225;
tr3847:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1225;
tr3839:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1225;
st1225:
	if ( ++p == pe )
		goto _test_eof1225;
case 1225:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3848;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3847;
	goto tr75;
tr3834:
	{
		s->buffer_length = 0;
	}
	goto st1226;
tr3849:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1226;
tr3828:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st299;}
		}
	}
	{
		s->buffer_length = 0;
	}
	goto st1226;
st1226:
	if ( ++p == pe )
		goto _test_eof1226;
case 1226:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3850;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3849;
	goto tr75;
	}
	_test_eof1227: cs = 1227; goto _test_eof;
	_test_eof1: cs = 1; goto _test_eof;
	_test_eof2: cs = 2; goto _test_eof;
	_test_eof3: cs = 3; goto _test_eof;
	_test_eof4: cs = 4; goto _test_eof;
	_test_eof5: cs = 5; goto _test_eof;
	_test_eof6: cs = 6; goto _test_eof;
	_test_eof7: cs = 7; goto _test_eof;
	_test_eof8: cs = 8; goto _test_eof;
	_test_eof9: cs = 9; goto _test_eof;
	_test_eof10: cs = 10; goto _test_eof;
	_test_eof11: cs = 11; goto _test_eof;
	_test_eof12: cs = 12; goto _test_eof;
	_test_eof13: cs = 13; goto _test_eof;
	_test_eof1228: cs = 1228; goto _test_eof;
	_test_eof14: cs = 14; goto _test_eof;
	_test_eof15: cs = 15; goto _test_eof;
	_test_eof16: cs = 16; goto _test_eof;
	_test_eof17: cs = 17; goto _test_eof;
	_test_eof18: cs = 18; goto _test_eof;
	_test_eof19: cs = 19; goto _test_eof;
	_test_eof20: cs = 20; goto _test_eof;
	_test_eof21: cs = 21; goto _test_eof;
	_test_eof22: cs = 22; goto _test_eof;
	_test_eof23: cs = 23; goto _test_eof;
	_test_eof24: cs = 24; goto _test_eof;
	_test_eof25: cs = 25; goto _test_eof;
	_test_eof26: cs = 26; goto _test_eof;
	_test_eof27: cs = 27; goto _test_eof;
	_test_eof28: cs = 28; goto _test_eof;
	_test_eof29: cs = 29; goto _test_eof;
	_test_eof30: cs = 30; goto _test_eof;
	_test_eof31: cs = 31; goto _test_eof;
	_test_eof32: cs = 32; goto _test_eof;
	_test_eof33: cs = 33; goto _test_eof;
	_test_eof34: cs = 34; goto _test_eof;
	_test_eof35: cs = 35; goto _test_eof;
	_test_eof36: cs = 36; goto _test_eof;
	_test_eof37: cs = 37; goto _test_eof;
	_test_eof38: cs = 38; goto _test_eof;
	_test_eof39: cs = 39; goto _test_eof;
	_test_eof40: cs = 40; goto _test_eof;
	_test_eof41: cs = 41; goto _test_eof;
	_test_eof42: cs = 42; goto _test_eof;
	_test_eof43: cs = 43; goto _test_eof;
	_test_eof44: cs = 44; goto _test_eof;
	_test_eof45: cs = 45; goto _test_eof;
	_test_eof46: cs = 46; goto _test_eof;
	_test_eof47: cs = 47; goto _test_eof;
	_test_eof48: cs = 48; goto _test_eof;
	_test_eof49: cs = 49; goto _test_eof;
	_test_eof50: cs = 50; goto _test_eof;
	_test_eof51: cs = 51; goto _test_eof;
	_test_eof52: cs = 52; goto _test_eof;
	_test_eof53: cs = 53; goto _test_eof;
	_test_eof54: cs = 54; goto _test_eof;
	_test_eof55: cs = 55; goto _test_eof;
	_test_eof56: cs = 56; goto _test_eof;
	_test_eof57: cs = 57; goto _test_eof;
	_test_eof58: cs = 58; goto _test_eof;
	_test_eof59: cs = 59; goto _test_eof;
	_test_eof60: cs = 60; goto _test_eof;
	_test_eof61: cs = 61; goto _test_eof;
	_test_eof62: cs = 62; goto _test_eof;
	_test_eof63: cs = 63; goto _test_eof;
	_test_eof64: cs = 64; goto _test_eof;
	_test_eof65: cs = 65; goto _test_eof;
	_test_eof66: cs = 66; goto _test_eof;
	_test_eof67: cs = 67; goto _test_eof;
	_test_eof68: cs = 68; goto _test_eof;
	_test_eof69: cs = 69; goto _test_eof;
	_test_eof70: cs = 70; goto _test_eof;
	_test_eof71: cs = 71; goto _test_eof;
	_test_eof72: cs = 72; goto _test_eof;
	_test_eof73: cs = 73; goto _test_eof;
	_test_eof74: cs = 74; goto _test_eof;
	_test_eof75: cs = 75; goto _test_eof;
	_test_eof76: cs = 76; goto _test_eof;
	_test_eof77: cs = 77; goto _test_eof;
	_test_eof78: cs = 78; goto _test_eof;
	_test_eof79: cs = 79; goto _test_eof;
	_test_eof80: cs = 80; goto _test_eof;
	_test_eof81: cs = 81; goto _test_eof;
	_test_eof82: cs = 82; goto _test_eof;
	_test_eof83: cs = 83; goto _test_eof;
	_test_eof84: cs = 84; goto _test_eof;
	_test_eof85: cs = 85; goto _test_eof;
	_test_eof86: cs = 86; goto _test_eof;
	_test_eof87: cs = 87; goto _test_eof;
	_test_eof88: cs = 88; goto _test_eof;
	_test_eof89: cs = 89; goto _test_eof;
	_test_eof90: cs = 90; goto _test_eof;
	_test_eof91: cs = 91; goto _test_eof;
	_test_eof92: cs = 92; goto _test_eof;
	_test_eof93: cs = 93; goto _test_eof;
	_test_eof94: cs = 94; goto _test_eof;
	_test_eof95: cs = 95; goto _test_eof;
	_test_eof96: cs = 96; goto _test_eof;
	_test_eof97: cs = 97; goto _test_eof;
	_test_eof98: cs = 98; goto _test_eof;
	_test_eof99: cs = 99; goto _test_eof;
	_test_eof100: cs = 100; goto _test_eof;
	_test_eof101: cs = 101; goto _test_eof;
	_test_eof102: cs = 102; goto _test_eof;
	_test_eof103: cs = 103; goto _test_eof;
	_test_eof104: cs = 104; goto _test_eof;
	_test_eof105: cs = 105; goto _test_eof;
	_test_eof106: cs = 106; goto _test_eof;
	_test_eof107: cs = 107; goto _test_eof;
	_test_eof108: cs = 108; goto _test_eof;
	_test_eof109: cs = 109; goto _test_eof;
	_test_eof110: cs = 110; goto _test_eof;
	_test_eof111: cs = 111; goto _test_eof;
	_test_eof112: cs = 112; goto _test_eof;
	_test_eof113: cs = 113; goto _test_eof;
	_test_eof114: cs = 114; goto _test_eof;
	_test_eof115: cs = 115; goto _test_eof;
	_test_eof116: cs = 116; goto _test_eof;
	_test_eof117: cs = 117; goto _test_eof;
	_test_eof118: cs = 118; goto _test_eof;
	_test_eof119: cs = 119; goto _test_eof;
	_test_eof120: cs = 120; goto _test_eof;
	_test_eof121: cs = 121; goto _test_eof;
	_test_eof122: cs = 122; goto _test_eof;
	_test_eof123: cs = 123; goto _test_eof;
	_test_eof124: cs = 124; goto _test_eof;
	_test_eof125: cs = 125; goto _test_eof;
	_test_eof126: cs = 126; goto _test_eof;
	_test_eof127: cs = 127; goto _test_eof;
	_test_eof128: cs = 128; goto _test_eof;
	_test_eof129: cs = 129; goto _test_eof;
	_test_eof130: cs = 130; goto _test_eof;
	_test_eof131: cs = 131; goto _test_eof;
	_test_eof132: cs = 132; goto _test_eof;
	_test_eof133: cs = 133; goto _test_eof;
	_test_eof134: cs = 134; goto _test_eof;
	_test_eof135: cs = 135; goto _test_eof;
	_test_eof136: cs = 136; goto _test_eof;
	_test_eof137: cs = 137; goto _test_eof;
	_test_eof138: cs = 138; goto _test_eof;
	_test_eof139: cs = 139; goto _test_eof;
	_test_eof140: cs = 140; goto _test_eof;
	_test_eof141: cs = 141; goto _test_eof;
	_test_eof142: cs = 142; goto _test_eof;
	_test_eof143: cs = 143; goto _test_eof;
	_test_eof144: cs = 144; goto _test_eof;
	_test_eof145: cs = 145; goto _test_eof;
	_test_eof146: cs = 146; goto _test_eof;
	_test_eof147: cs = 147; goto _test_eof;
	_test_eof148: cs = 148; goto _test_eof;
	_test_eof149: cs = 149; goto _test_eof;
	_test_eof150: cs = 150; goto _test_eof;
	_test_eof151: cs = 151; goto _test_eof;
	_test_eof152: cs = 152; goto _test_eof;
	_test_eof153: cs = 153; goto _test_eof;
	_test_eof154: cs = 154; goto _test_eof;
	_test_eof155: cs = 155; goto _test_eof;
	_test_eof156: cs = 156; goto _test_eof;
	_test_eof157: cs = 157; goto _test_eof;
	_test_eof158: cs = 158; goto _test_eof;
	_test_eof159: cs = 159; goto _test_eof;
	_test_eof160: cs = 160; goto _test_eof;
	_test_eof161: cs = 161; goto _test_eof;
	_test_eof162: cs = 162; goto _test_eof;
	_test_eof163: cs = 163; goto _test_eof;
	_test_eof164: cs = 164; goto _test_eof;
	_test_eof165: cs = 165; goto _test_eof;
	_test_eof166: cs = 166; goto _test_eof;
	_test_eof1229: cs = 1229; goto _test_eof;
	_test_eof167: cs = 167; goto _test_eof;
	_test_eof168: cs = 168; goto _test_eof;
	_test_eof169: cs = 169; goto _test_eof;
	_test_eof170: cs = 170; goto _test_eof;
	_test_eof171: cs = 171; goto _test_eof;
	_test_eof172: cs = 172; goto _test_eof;
	_test_eof173: cs = 173; goto _test_eof;
	_test_eof174: cs = 174; goto _test_eof;
	_test_eof175: cs = 175; goto _test_eof;
	_test_eof176: cs = 176; goto _test_eof;
	_test_eof1230: cs = 1230; goto _test_eof;
	_test_eof177: cs = 177; goto _test_eof;
	_test_eof178: cs = 178; goto _test_eof;
	_test_eof179: cs = 179; goto _test_eof;
	_test_eof180: cs = 180; goto _test_eof;
	_test_eof181: cs = 181; goto _test_eof;
	_test_eof182: cs = 182; goto _test_eof;
	_test_eof183: cs = 183; goto _test_eof;
	_test_eof184: cs = 184; goto _test_eof;
	_test_eof1231: cs = 1231; goto _test_eof;
	_test_eof185: cs = 185; goto _test_eof;
	_test_eof186: cs = 186; goto _test_eof;
	_test_eof187: cs = 187; goto _test_eof;
	_test_eof1232: cs = 1232; goto _test_eof;
	_test_eof188: cs = 188; goto _test_eof;
	_test_eof189: cs = 189; goto _test_eof;
	_test_eof190: cs = 190; goto _test_eof;
	_test_eof191: cs = 191; goto _test_eof;
	_test_eof192: cs = 192; goto _test_eof;
	_test_eof193: cs = 193; goto _test_eof;
	_test_eof194: cs = 194; goto _test_eof;
	_test_eof195: cs = 195; goto _test_eof;
	_test_eof196: cs = 196; goto _test_eof;
	_test_eof197: cs = 197; goto _test_eof;
	_test_eof198: cs = 198; goto _test_eof;
	_test_eof1233: cs = 1233; goto _test_eof;
	_test_eof199: cs = 199; goto _test_eof;
	_test_eof200: cs = 200; goto _test_eof;
	_test_eof201: cs = 201; goto _test_eof;
	_test_eof202: cs = 202; goto _test_eof;
	_test_eof1234: cs = 1234; goto _test_eof;
	_test_eof203: cs = 203; goto _test_eof;
	_test_eof204: cs = 204; goto _test_eof;
	_test_eof205: cs = 205; goto _test_eof;
	_test_eof206: cs = 206; goto _test_eof;
	_test_eof207: cs = 207; goto _test_eof;
	_test_eof208: cs = 208; goto _test_eof;
	_test_eof209: cs = 209; goto _test_eof;
	_test_eof210: cs = 210; goto _test_eof;
	_test_eof211: cs = 211; goto _test_eof;
	_test_eof212: cs = 212; goto _test_eof;
	_test_eof213: cs = 213; goto _test_eof;
	_test_eof214: cs = 214; goto _test_eof;
	_test_eof215: cs = 215; goto _test_eof;
	_test_eof216: cs = 216; goto _test_eof;
	_test_eof217: cs = 217; goto _test_eof;
	_test_eof218: cs = 218; goto _test_eof;
	_test_eof219: cs = 219; goto _test_eof;
	_test_eof220: cs = 220; goto _test_eof;
	_test_eof1235: cs = 1235; goto _test_eof;
	_test_eof221: cs = 221; goto _test_eof;
	_test_eof222: cs = 222; goto _test_eof;
	_test_eof223: cs = 223; goto _test_eof;
	_test_eof224: cs = 224; goto _test_eof;
	_test_eof225: cs = 225; goto _test_eof;
	_test_eof226: cs = 226; goto _test_eof;
	_test_eof227: cs = 227; goto _test_eof;
	_test_eof228: cs = 228; goto _test_eof;
	_test_eof229: cs = 229; goto _test_eof;
	_test_eof230: cs = 230; goto _test_eof;
	_test_eof231: cs = 231; goto _test_eof;
	_test_eof232: cs = 232; goto _test_eof;
	_test_eof233: cs = 233; goto _test_eof;
	_test_eof234: cs = 234; goto _test_eof;
	_test_eof235: cs = 235; goto _test_eof;
	_test_eof236: cs = 236; goto _test_eof;
	_test_eof1236: cs = 1236; goto _test_eof;
	_test_eof237: cs = 237; goto _test_eof;
	_test_eof238: cs = 238; goto _test_eof;
	_test_eof239: cs = 239; goto _test_eof;
	_test_eof240: cs = 240; goto _test_eof;
	_test_eof241: cs = 241; goto _test_eof;
	_test_eof242: cs = 242; goto _test_eof;
	_test_eof243: cs = 243; goto _test_eof;
	_test_eof244: cs = 244; goto _test_eof;
	_test_eof245: cs = 245; goto _test_eof;
	_test_eof246: cs = 246; goto _test_eof;
	_test_eof247: cs = 247; goto _test_eof;
	_test_eof248: cs = 248; goto _test_eof;
	_test_eof249: cs = 249; goto _test_eof;
	_test_eof250: cs = 250; goto _test_eof;
	_test_eof251: cs = 251; goto _test_eof;
	_test_eof252: cs = 252; goto _test_eof;
	_test_eof253: cs = 253; goto _test_eof;
	_test_eof254: cs = 254; goto _test_eof;
	_test_eof255: cs = 255; goto _test_eof;
	_test_eof256: cs = 256; goto _test_eof;
	_test_eof257: cs = 257; goto _test_eof;
	_test_eof258: cs = 258; goto _test_eof;
	_test_eof259: cs = 259; goto _test_eof;
	_test_eof260: cs = 260; goto _test_eof;
	_test_eof261: cs = 261; goto _test_eof;
	_test_eof262: cs = 262; goto _test_eof;
	_test_eof263: cs = 263; goto _test_eof;
	_test_eof264: cs = 264; goto _test_eof;
	_test_eof265: cs = 265; goto _test_eof;
	_test_eof266: cs = 266; goto _test_eof;
	_test_eof267: cs = 267; goto _test_eof;
	_test_eof268: cs = 268; goto _test_eof;
	_test_eof269: cs = 269; goto _test_eof;
	_test_eof270: cs = 270; goto _test_eof;
	_test_eof271: cs = 271; goto _test_eof;
	_test_eof272: cs = 272; goto _test_eof;
	_test_eof273: cs = 273; goto _test_eof;
	_test_eof274: cs = 274; goto _test_eof;
	_test_eof275: cs = 275; goto _test_eof;
	_test_eof276: cs = 276; goto _test_eof;
	_test_eof277: cs = 277; goto _test_eof;
	_test_eof278: cs = 278; goto _test_eof;
	_test_eof279: cs = 279; goto _test_eof;
	_test_eof280: cs = 280; goto _test_eof;
	_test_eof281: cs = 281; goto _test_eof;
	_test_eof282: cs = 282; goto _test_eof;
	_test_eof283: cs = 283; goto _test_eof;
	_test_eof284: cs = 284; goto _test_eof;
	_test_eof285: cs = 285; goto _test_eof;
	_test_eof286: cs = 286; goto _test_eof;
	_test_eof287: cs = 287; goto _test_eof;
	_test_eof288: cs = 288; goto _test_eof;
	_test_eof289: cs = 289; goto _test_eof;
	_test_eof290: cs = 290; goto _test_eof;
	_test_eof291: cs = 291; goto _test_eof;
	_test_eof292: cs = 292; goto _test_eof;
	_test_eof293: cs = 293; goto _test_eof;
	_test_eof294: cs = 294; goto _test_eof;
	_test_eof295: cs = 295; goto _test_eof;
	_test_eof296: cs = 296; goto _test_eof;
	_test_eof297: cs = 297; goto _test_eof;
	_test_eof298: cs = 298; goto _test_eof;
	_test_eof299: cs = 299; goto _test_eof;
	_test_eof300: cs = 300; goto _test_eof;
	_test_eof1237: cs = 1237; goto _test_eof;
	_test_eof301: cs = 301; goto _test_eof;
	_test_eof302: cs = 302; goto _test_eof;
	_test_eof1238: cs = 1238; goto _test_eof;
	_test_eof303: cs = 303; goto _test_eof;
	_test_eof304: cs = 304; goto _test_eof;
	_test_eof305: cs = 305; goto _test_eof;
	_test_eof306: cs = 306; goto _test_eof;
	_test_eof307: cs = 307; goto _test_eof;
	_test_eof308: cs = 308; goto _test_eof;
	_test_eof309: cs = 309; goto _test_eof;
	_test_eof310: cs = 310; goto _test_eof;
	_test_eof311: cs = 311; goto _test_eof;
	_test_eof1239: cs = 1239; goto _test_eof;
	_test_eof1240: cs = 1240; goto _test_eof;
	_test_eof312: cs = 312; goto _test_eof;
	_test_eof313: cs = 313; goto _test_eof;
	_test_eof314: cs = 314; goto _test_eof;
	_test_eof315: cs = 315; goto _test_eof;
	_test_eof316: cs = 316; goto _test_eof;
	_test_eof317: cs = 317; goto _test_eof;
	_test_eof318: cs = 318; goto _test_eof;
	_test_eof319: cs = 319; goto _test_eof;
	_test_eof320: cs = 320; goto _test_eof;
	_test_eof321: cs = 321; goto _test_eof;
	_test_eof322: cs = 322; goto _test_eof;
	_test_eof323: cs = 323; goto _test_eof;
	_test_eof324: cs = 324; goto _test_eof;
	_test_eof325: cs = 325; goto _test_eof;
	_test_eof1241: cs = 1241; goto _test_eof;
	_test_eof326: cs = 326; goto _test_eof;
	_test_eof327: cs = 327; goto _test_eof;
	_test_eof328: cs = 328; goto _test_eof;
	_test_eof329: cs = 329; goto _test_eof;
	_test_eof330: cs = 330; goto _test_eof;
	_test_eof331: cs = 331; goto _test_eof;
	_test_eof332: cs = 332; goto _test_eof;
	_test_eof333: cs = 333; goto _test_eof;
	_test_eof334: cs = 334; goto _test_eof;
	_test_eof335: cs = 335; goto _test_eof;
	_test_eof1242: cs = 1242; goto _test_eof;
	_test_eof336: cs = 336; goto _test_eof;
	_test_eof337: cs = 337; goto _test_eof;
	_test_eof338: cs = 338; goto _test_eof;
	_test_eof339: cs = 339; goto _test_eof;
	_test_eof340: cs = 340; goto _test_eof;
	_test_eof341: cs = 341; goto _test_eof;
	_test_eof342: cs = 342; goto _test_eof;
	_test_eof343: cs = 343; goto _test_eof;
	_test_eof344: cs = 344; goto _test_eof;
	_test_eof345: cs = 345; goto _test_eof;
	_test_eof346: cs = 346; goto _test_eof;
	_test_eof347: cs = 347; goto _test_eof;
	_test_eof348: cs = 348; goto _test_eof;
	_test_eof349: cs = 349; goto _test_eof;
	_test_eof1243: cs = 1243; goto _test_eof;
	_test_eof350: cs = 350; goto _test_eof;
	_test_eof351: cs = 351; goto _test_eof;
	_test_eof352: cs = 352; goto _test_eof;
	_test_eof353: cs = 353; goto _test_eof;
	_test_eof354: cs = 354; goto _test_eof;
	_test_eof355: cs = 355; goto _test_eof;
	_test_eof356: cs = 356; goto _test_eof;
	_test_eof1244: cs = 1244; goto _test_eof;
	_test_eof357: cs = 357; goto _test_eof;
	_test_eof358: cs = 358; goto _test_eof;
	_test_eof359: cs = 359; goto _test_eof;
	_test_eof360: cs = 360; goto _test_eof;
	_test_eof361: cs = 361; goto _test_eof;
	_test_eof362: cs = 362; goto _test_eof;
	_test_eof363: cs = 363; goto _test_eof;
	_test_eof364: cs = 364; goto _test_eof;
	_test_eof365: cs = 365; goto _test_eof;
	_test_eof1245: cs = 1245; goto _test_eof;
	_test_eof1246: cs = 1246; goto _test_eof;
	_test_eof1247: cs = 1247; goto _test_eof;
	_test_eof366: cs = 366; goto _test_eof;
	_test_eof367: cs = 367; goto _test_eof;
	_test_eof368: cs = 368; goto _test_eof;
	_test_eof369: cs = 369; goto _test_eof;
	_test_eof370: cs = 370; goto _test_eof;
	_test_eof371: cs = 371; goto _test_eof;
	_test_eof372: cs = 372; goto _test_eof;
	_test_eof373: cs = 373; goto _test_eof;
	_test_eof1248: cs = 1248; goto _test_eof;
	_test_eof1249: cs = 1249; goto _test_eof;
	_test_eof374: cs = 374; goto _test_eof;
	_test_eof375: cs = 375; goto _test_eof;
	_test_eof376: cs = 376; goto _test_eof;
	_test_eof1250: cs = 1250; goto _test_eof;
	_test_eof377: cs = 377; goto _test_eof;
	_test_eof378: cs = 378; goto _test_eof;
	_test_eof379: cs = 379; goto _test_eof;
	_test_eof380: cs = 380; goto _test_eof;
	_test_eof381: cs = 381; goto _test_eof;
	_test_eof382: cs = 382; goto _test_eof;
	_test_eof383: cs = 383; goto _test_eof;
	_test_eof384: cs = 384; goto _test_eof;
	_test_eof385: cs = 385; goto _test_eof;
	_test_eof386: cs = 386; goto _test_eof;
	_test_eof387: cs = 387; goto _test_eof;
	_test_eof388: cs = 388; goto _test_eof;
	_test_eof389: cs = 389; goto _test_eof;
	_test_eof390: cs = 390; goto _test_eof;
	_test_eof391: cs = 391; goto _test_eof;
	_test_eof392: cs = 392; goto _test_eof;
	_test_eof393: cs = 393; goto _test_eof;
	_test_eof394: cs = 394; goto _test_eof;
	_test_eof395: cs = 395; goto _test_eof;
	_test_eof396: cs = 396; goto _test_eof;
	_test_eof397: cs = 397; goto _test_eof;
	_test_eof398: cs = 398; goto _test_eof;
	_test_eof399: cs = 399; goto _test_eof;
	_test_eof400: cs = 400; goto _test_eof;
	_test_eof401: cs = 401; goto _test_eof;
	_test_eof402: cs = 402; goto _test_eof;
	_test_eof403: cs = 403; goto _test_eof;
	_test_eof404: cs = 404; goto _test_eof;
	_test_eof405: cs = 405; goto _test_eof;
	_test_eof406: cs = 406; goto _test_eof;
	_test_eof407: cs = 407; goto _test_eof;
	_test_eof408: cs = 408; goto _test_eof;
	_test_eof409: cs = 409; goto _test_eof;
	_test_eof410: cs = 410; goto _test_eof;
	_test_eof411: cs = 411; goto _test_eof;
	_test_eof412: cs = 412; goto _test_eof;
	_test_eof413: cs = 413; goto _test_eof;
	_test_eof414: cs = 414; goto _test_eof;
	_test_eof415: cs = 415; goto _test_eof;
	_test_eof416: cs = 416; goto _test_eof;
	_test_eof417: cs = 417; goto _test_eof;
	_test_eof418: cs = 418; goto _test_eof;
	_test_eof419: cs = 419; goto _test_eof;
	_test_eof420: cs = 420; goto _test_eof;
	_test_eof421: cs = 421; goto _test_eof;
	_test_eof422: cs = 422; goto _test_eof;
	_test_eof423: cs = 423; goto _test_eof;
	_test_eof424: cs = 424; goto _test_eof;
	_test_eof425: cs = 425; goto _test_eof;
	_test_eof426: cs = 426; goto _test_eof;
	_test_eof427: cs = 427; goto _test_eof;
	_test_eof428: cs = 428; goto _test_eof;
	_test_eof429: cs = 429; goto _test_eof;
	_test_eof430: cs = 430; goto _test_eof;
	_test_eof431: cs = 431; goto _test_eof;
	_test_eof432: cs = 432; goto _test_eof;
	_test_eof433: cs = 433; goto _test_eof;
	_test_eof434: cs = 434; goto _test_eof;
	_test_eof435: cs = 435; goto _test_eof;
	_test_eof436: cs = 436; goto _test_eof;
	_test_eof437: cs = 437; goto _test_eof;
	_test_eof438: cs = 438; goto _test_eof;
	_test_eof439: cs = 439; goto _test_eof;
	_test_eof440: cs = 440; goto _test_eof;
	_test_eof441: cs = 441; goto _test_eof;
	_test_eof442: cs = 442; goto _test_eof;
	_test_eof443: cs = 443; goto _test_eof;
	_test_eof444: cs = 444; goto _test_eof;
	_test_eof445: cs = 445; goto _test_eof;
	_test_eof446: cs = 446; goto _test_eof;
	_test_eof447: cs = 447; goto _test_eof;
	_test_eof448: cs = 448; goto _test_eof;
	_test_eof449: cs = 449; goto _test_eof;
	_test_eof450: cs = 450; goto _test_eof;
	_test_eof451: cs = 451; goto _test_eof;
	_test_eof452: cs = 452; goto _test_eof;
	_test_eof453: cs = 453; goto _test_eof;
	_test_eof454: cs = 454; goto _test_eof;
	_test_eof455: cs = 455; goto _test_eof;
	_test_eof456: cs = 456; goto _test_eof;
	_test_eof457: cs = 457; goto _test_eof;
	_test_eof458: cs = 458; goto _test_eof;
	_test_eof459: cs = 459; goto _test_eof;
	_test_eof460: cs = 460; goto _test_eof;
	_test_eof461: cs = 461; goto _test_eof;
	_test_eof462: cs = 462; goto _test_eof;
	_test_eof463: cs = 463; goto _test_eof;
	_test_eof464: cs = 464; goto _test_eof;
	_test_eof465: cs = 465; goto _test_eof;
	_test_eof466: cs = 466; goto _test_eof;
	_test_eof467: cs = 467; goto _test_eof;
	_test_eof468: cs = 468; goto _test_eof;
	_test_eof469: cs = 469; goto _test_eof;
	_test_eof470: cs = 470; goto _test_eof;
	_test_eof471: cs = 471; goto _test_eof;
	_test_eof472: cs = 472; goto _test_eof;
	_test_eof473: cs = 473; goto _test_eof;
	_test_eof474: cs = 474; goto _test_eof;
	_test_eof475: cs = 475; goto _test_eof;
	_test_eof476: cs = 476; goto _test_eof;
	_test_eof477: cs = 477; goto _test_eof;
	_test_eof478: cs = 478; goto _test_eof;
	_test_eof479: cs = 479; goto _test_eof;
	_test_eof480: cs = 480; goto _test_eof;
	_test_eof481: cs = 481; goto _test_eof;
	_test_eof482: cs = 482; goto _test_eof;
	_test_eof483: cs = 483; goto _test_eof;
	_test_eof484: cs = 484; goto _test_eof;
	_test_eof485: cs = 485; goto _test_eof;
	_test_eof486: cs = 486; goto _test_eof;
	_test_eof487: cs = 487; goto _test_eof;
	_test_eof488: cs = 488; goto _test_eof;
	_test_eof489: cs = 489; goto _test_eof;
	_test_eof490: cs = 490; goto _test_eof;
	_test_eof491: cs = 491; goto _test_eof;
	_test_eof492: cs = 492; goto _test_eof;
	_test_eof493: cs = 493; goto _test_eof;
	_test_eof494: cs = 494; goto _test_eof;
	_test_eof495: cs = 495; goto _test_eof;
	_test_eof496: cs = 496; goto _test_eof;
	_test_eof497: cs = 497; goto _test_eof;
	_test_eof498: cs = 498; goto _test_eof;
	_test_eof499: cs = 499; goto _test_eof;
	_test_eof500: cs = 500; goto _test_eof;
	_test_eof501: cs = 501; goto _test_eof;
	_test_eof502: cs = 502; goto _test_eof;
	_test_eof503: cs = 503; goto _test_eof;
	_test_eof504: cs = 504; goto _test_eof;
	_test_eof505: cs = 505; goto _test_eof;
	_test_eof506: cs = 506; goto _test_eof;
	_test_eof507: cs = 507; goto _test_eof;
	_test_eof508: cs = 508; goto _test_eof;
	_test_eof509: cs = 509; goto _test_eof;
	_test_eof510: cs = 510; goto _test_eof;
	_test_eof511: cs = 511; goto _test_eof;
	_test_eof512: cs = 512; goto _test_eof;
	_test_eof513: cs = 513; goto _test_eof;
	_test_eof514: cs = 514; goto _test_eof;
	_test_eof515: cs = 515; goto _test_eof;
	_test_eof516: cs = 516; goto _test_eof;
	_test_eof517: cs = 517; goto _test_eof;
	_test_eof518: cs = 518; goto _test_eof;
	_test_eof519: cs = 519; goto _test_eof;
	_test_eof520: cs = 520; goto _test_eof;
	_test_eof521: cs = 521; goto _test_eof;
	_test_eof522: cs = 522; goto _test_eof;
	_test_eof523: cs = 523; goto _test_eof;
	_test_eof524: cs = 524; goto _test_eof;
	_test_eof525: cs = 525; goto _test_eof;
	_test_eof526: cs = 526; goto _test_eof;
	_test_eof527: cs = 527; goto _test_eof;
	_test_eof528: cs = 528; goto _test_eof;
	_test_eof529: cs = 529; goto _test_eof;
	_test_eof530: cs = 530; goto _test_eof;
	_test_eof1251: cs = 1251; goto _test_eof;
	_test_eof1252: cs = 1252; goto _test_eof;
	_test_eof1253: cs = 1253; goto _test_eof;
	_test_eof531: cs = 531; goto _test_eof;
	_test_eof532: cs = 532; goto _test_eof;
	_test_eof533: cs = 533; goto _test_eof;
	_test_eof534: cs = 534; goto _test_eof;
	_test_eof535: cs = 535; goto _test_eof;
	_test_eof1254: cs = 1254; goto _test_eof;
	_test_eof536: cs = 536; goto _test_eof;
	_test_eof537: cs = 537; goto _test_eof;
	_test_eof538: cs = 538; goto _test_eof;
	_test_eof539: cs = 539; goto _test_eof;
	_test_eof1255: cs = 1255; goto _test_eof;
	_test_eof1256: cs = 1256; goto _test_eof;
	_test_eof1257: cs = 1257; goto _test_eof;
	_test_eof540: cs = 540; goto _test_eof;
	_test_eof541: cs = 541; goto _test_eof;
	_test_eof1258: cs = 1258; goto _test_eof;
	_test_eof542: cs = 542; goto _test_eof;
	_test_eof543: cs = 543; goto _test_eof;
	_test_eof544: cs = 544; goto _test_eof;
	_test_eof1259: cs = 1259; goto _test_eof;
	_test_eof545: cs = 545; goto _test_eof;
	_test_eof546: cs = 546; goto _test_eof;
	_test_eof547: cs = 547; goto _test_eof;
	_test_eof548: cs = 548; goto _test_eof;
	_test_eof549: cs = 549; goto _test_eof;
	_test_eof550: cs = 550; goto _test_eof;
	_test_eof551: cs = 551; goto _test_eof;
	_test_eof552: cs = 552; goto _test_eof;
	_test_eof553: cs = 553; goto _test_eof;
	_test_eof554: cs = 554; goto _test_eof;
	_test_eof555: cs = 555; goto _test_eof;
	_test_eof556: cs = 556; goto _test_eof;
	_test_eof557: cs = 557; goto _test_eof;
	_test_eof558: cs = 558; goto _test_eof;
	_test_eof559: cs = 559; goto _test_eof;
	_test_eof560: cs = 560; goto _test_eof;
	_test_eof561: cs = 561; goto _test_eof;
	_test_eof562: cs = 562; goto _test_eof;
	_test_eof563: cs = 563; goto _test_eof;
	_test_eof564: cs = 564; goto _test_eof;
	_test_eof565: cs = 565; goto _test_eof;
	_test_eof566: cs = 566; goto _test_eof;
	_test_eof567: cs = 567; goto _test_eof;
	_test_eof568: cs = 568; goto _test_eof;
	_test_eof569: cs = 569; goto _test_eof;
	_test_eof570: cs = 570; goto _test_eof;
	_test_eof571: cs = 571; goto _test_eof;
	_test_eof572: cs = 572; goto _test_eof;
	_test_eof573: cs = 573; goto _test_eof;
	_test_eof574: cs = 574; goto _test_eof;
	_test_eof575: cs = 575; goto _test_eof;
	_test_eof576: cs = 576; goto _test_eof;
	_test_eof577: cs = 577; goto _test_eof;
	_test_eof578: cs = 578; goto _test_eof;
	_test_eof579: cs = 579; goto _test_eof;
	_test_eof580: cs = 580; goto _test_eof;
	_test_eof581: cs = 581; goto _test_eof;
	_test_eof582: cs = 582; goto _test_eof;
	_test_eof583: cs = 583; goto _test_eof;
	_test_eof584: cs = 584; goto _test_eof;
	_test_eof585: cs = 585; goto _test_eof;
	_test_eof586: cs = 586; goto _test_eof;
	_test_eof587: cs = 587; goto _test_eof;
	_test_eof588: cs = 588; goto _test_eof;
	_test_eof589: cs = 589; goto _test_eof;
	_test_eof590: cs = 590; goto _test_eof;
	_test_eof591: cs = 591; goto _test_eof;
	_test_eof592: cs = 592; goto _test_eof;
	_test_eof593: cs = 593; goto _test_eof;
	_test_eof594: cs = 594; goto _test_eof;
	_test_eof595: cs = 595; goto _test_eof;
	_test_eof596: cs = 596; goto _test_eof;
	_test_eof597: cs = 597; goto _test_eof;
	_test_eof598: cs = 598; goto _test_eof;
	_test_eof599: cs = 599; goto _test_eof;
	_test_eof600: cs = 600; goto _test_eof;
	_test_eof601: cs = 601; goto _test_eof;
	_test_eof602: cs = 602; goto _test_eof;
	_test_eof603: cs = 603; goto _test_eof;
	_test_eof604: cs = 604; goto _test_eof;
	_test_eof605: cs = 605; goto _test_eof;
	_test_eof606: cs = 606; goto _test_eof;
	_test_eof607: cs = 607; goto _test_eof;
	_test_eof608: cs = 608; goto _test_eof;
	_test_eof609: cs = 609; goto _test_eof;
	_test_eof610: cs = 610; goto _test_eof;
	_test_eof611: cs = 611; goto _test_eof;
	_test_eof612: cs = 612; goto _test_eof;
	_test_eof613: cs = 613; goto _test_eof;
	_test_eof614: cs = 614; goto _test_eof;
	_test_eof615: cs = 615; goto _test_eof;
	_test_eof616: cs = 616; goto _test_eof;
	_test_eof617: cs = 617; goto _test_eof;
	_test_eof618: cs = 618; goto _test_eof;
	_test_eof619: cs = 619; goto _test_eof;
	_test_eof620: cs = 620; goto _test_eof;
	_test_eof621: cs = 621; goto _test_eof;
	_test_eof622: cs = 622; goto _test_eof;
	_test_eof623: cs = 623; goto _test_eof;
	_test_eof624: cs = 624; goto _test_eof;
	_test_eof625: cs = 625; goto _test_eof;
	_test_eof626: cs = 626; goto _test_eof;
	_test_eof627: cs = 627; goto _test_eof;
	_test_eof628: cs = 628; goto _test_eof;
	_test_eof629: cs = 629; goto _test_eof;
	_test_eof630: cs = 630; goto _test_eof;
	_test_eof631: cs = 631; goto _test_eof;
	_test_eof632: cs = 632; goto _test_eof;
	_test_eof633: cs = 633; goto _test_eof;
	_test_eof634: cs = 634; goto _test_eof;
	_test_eof635: cs = 635; goto _test_eof;
	_test_eof636: cs = 636; goto _test_eof;
	_test_eof637: cs = 637; goto _test_eof;
	_test_eof638: cs = 638; goto _test_eof;
	_test_eof639: cs = 639; goto _test_eof;
	_test_eof640: cs = 640; goto _test_eof;
	_test_eof641: cs = 641; goto _test_eof;
	_test_eof642: cs = 642; goto _test_eof;
	_test_eof643: cs = 643; goto _test_eof;
	_test_eof644: cs = 644; goto _test_eof;
	_test_eof645: cs = 645; goto _test_eof;
	_test_eof646: cs = 646; goto _test_eof;
	_test_eof647: cs = 647; goto _test_eof;
	_test_eof648: cs = 648; goto _test_eof;
	_test_eof1260: cs = 1260; goto _test_eof;
	_test_eof649: cs = 649; goto _test_eof;
	_test_eof650: cs = 650; goto _test_eof;
	_test_eof651: cs = 651; goto _test_eof;
	_test_eof652: cs = 652; goto _test_eof;
	_test_eof653: cs = 653; goto _test_eof;
	_test_eof654: cs = 654; goto _test_eof;
	_test_eof655: cs = 655; goto _test_eof;
	_test_eof656: cs = 656; goto _test_eof;
	_test_eof657: cs = 657; goto _test_eof;
	_test_eof658: cs = 658; goto _test_eof;
	_test_eof659: cs = 659; goto _test_eof;
	_test_eof660: cs = 660; goto _test_eof;
	_test_eof661: cs = 661; goto _test_eof;
	_test_eof662: cs = 662; goto _test_eof;
	_test_eof663: cs = 663; goto _test_eof;
	_test_eof664: cs = 664; goto _test_eof;
	_test_eof665: cs = 665; goto _test_eof;
	_test_eof666: cs = 666; goto _test_eof;
	_test_eof667: cs = 667; goto _test_eof;
	_test_eof668: cs = 668; goto _test_eof;
	_test_eof669: cs = 669; goto _test_eof;
	_test_eof670: cs = 670; goto _test_eof;
	_test_eof671: cs = 671; goto _test_eof;
	_test_eof672: cs = 672; goto _test_eof;
	_test_eof673: cs = 673; goto _test_eof;
	_test_eof674: cs = 674; goto _test_eof;
	_test_eof675: cs = 675; goto _test_eof;
	_test_eof676: cs = 676; goto _test_eof;
	_test_eof677: cs = 677; goto _test_eof;
	_test_eof678: cs = 678; goto _test_eof;
	_test_eof679: cs = 679; goto _test_eof;
	_test_eof680: cs = 680; goto _test_eof;
	_test_eof681: cs = 681; goto _test_eof;
	_test_eof682: cs = 682; goto _test_eof;
	_test_eof683: cs = 683; goto _test_eof;
	_test_eof684: cs = 684; goto _test_eof;
	_test_eof685: cs = 685; goto _test_eof;
	_test_eof686: cs = 686; goto _test_eof;
	_test_eof687: cs = 687; goto _test_eof;
	_test_eof688: cs = 688; goto _test_eof;
	_test_eof689: cs = 689; goto _test_eof;
	_test_eof1261: cs = 1261; goto _test_eof;
	_test_eof690: cs = 690; goto _test_eof;
	_test_eof691: cs = 691; goto _test_eof;
	_test_eof1262: cs = 1262; goto _test_eof;
	_test_eof692: cs = 692; goto _test_eof;
	_test_eof693: cs = 693; goto _test_eof;
	_test_eof694: cs = 694; goto _test_eof;
	_test_eof695: cs = 695; goto _test_eof;
	_test_eof696: cs = 696; goto _test_eof;
	_test_eof697: cs = 697; goto _test_eof;
	_test_eof698: cs = 698; goto _test_eof;
	_test_eof699: cs = 699; goto _test_eof;
	_test_eof700: cs = 700; goto _test_eof;
	_test_eof701: cs = 701; goto _test_eof;
	_test_eof702: cs = 702; goto _test_eof;
	_test_eof703: cs = 703; goto _test_eof;
	_test_eof704: cs = 704; goto _test_eof;
	_test_eof705: cs = 705; goto _test_eof;
	_test_eof1263: cs = 1263; goto _test_eof;
	_test_eof706: cs = 706; goto _test_eof;
	_test_eof707: cs = 707; goto _test_eof;
	_test_eof708: cs = 708; goto _test_eof;
	_test_eof709: cs = 709; goto _test_eof;
	_test_eof710: cs = 710; goto _test_eof;
	_test_eof711: cs = 711; goto _test_eof;
	_test_eof712: cs = 712; goto _test_eof;
	_test_eof713: cs = 713; goto _test_eof;
	_test_eof714: cs = 714; goto _test_eof;
	_test_eof715: cs = 715; goto _test_eof;
	_test_eof716: cs = 716; goto _test_eof;
	_test_eof717: cs = 717; goto _test_eof;
	_test_eof718: cs = 718; goto _test_eof;
	_test_eof719: cs = 719; goto _test_eof;
	_test_eof720: cs = 720; goto _test_eof;
	_test_eof721: cs = 721; goto _test_eof;
	_test_eof722: cs = 722; goto _test_eof;
	_test_eof723: cs = 723; goto _test_eof;
	_test_eof724: cs = 724; goto _test_eof;
	_test_eof725: cs = 725; goto _test_eof;
	_test_eof726: cs = 726; goto _test_eof;
	_test_eof727: cs = 727; goto _test_eof;
	_test_eof1264: cs = 1264; goto _test_eof;
	_test_eof728: cs = 728; goto _test_eof;
	_test_eof729: cs = 729; goto _test_eof;
	_test_eof730: cs = 730; goto _test_eof;
	_test_eof731: cs = 731; goto _test_eof;
	_test_eof732: cs = 732; goto _test_eof;
	_test_eof1265: cs = 1265; goto _test_eof;
	_test_eof733: cs = 733; goto _test_eof;
	_test_eof734: cs = 734; goto _test_eof;
	_test_eof735: cs = 735; goto _test_eof;
	_test_eof736: cs = 736; goto _test_eof;
	_test_eof737: cs = 737; goto _test_eof;
	_test_eof1266: cs = 1266; goto _test_eof;
	_test_eof738: cs = 738; goto _test_eof;
	_test_eof739: cs = 739; goto _test_eof;
	_test_eof740: cs = 740; goto _test_eof;
	_test_eof741: cs = 741; goto _test_eof;
	_test_eof742: cs = 742; goto _test_eof;
	_test_eof1267: cs = 1267; goto _test_eof;
	_test_eof1268: cs = 1268; goto _test_eof;
	_test_eof1269: cs = 1269; goto _test_eof;
	_test_eof743: cs = 743; goto _test_eof;
	_test_eof744: cs = 744; goto _test_eof;
	_test_eof1270: cs = 1270; goto _test_eof;
	_test_eof745: cs = 745; goto _test_eof;
	_test_eof746: cs = 746; goto _test_eof;
	_test_eof747: cs = 747; goto _test_eof;
	_test_eof748: cs = 748; goto _test_eof;
	_test_eof749: cs = 749; goto _test_eof;
	_test_eof750: cs = 750; goto _test_eof;
	_test_eof751: cs = 751; goto _test_eof;
	_test_eof752: cs = 752; goto _test_eof;
	_test_eof753: cs = 753; goto _test_eof;
	_test_eof754: cs = 754; goto _test_eof;
	_test_eof755: cs = 755; goto _test_eof;
	_test_eof756: cs = 756; goto _test_eof;
	_test_eof757: cs = 757; goto _test_eof;
	_test_eof758: cs = 758; goto _test_eof;
	_test_eof759: cs = 759; goto _test_eof;
	_test_eof760: cs = 760; goto _test_eof;
	_test_eof761: cs = 761; goto _test_eof;
	_test_eof762: cs = 762; goto _test_eof;
	_test_eof763: cs = 763; goto _test_eof;
	_test_eof764: cs = 764; goto _test_eof;
	_test_eof765: cs = 765; goto _test_eof;
	_test_eof766: cs = 766; goto _test_eof;
	_test_eof767: cs = 767; goto _test_eof;
	_test_eof768: cs = 768; goto _test_eof;
	_test_eof769: cs = 769; goto _test_eof;
	_test_eof770: cs = 770; goto _test_eof;
	_test_eof771: cs = 771; goto _test_eof;
	_test_eof1271: cs = 1271; goto _test_eof;
	_test_eof1272: cs = 1272; goto _test_eof;
	_test_eof1273: cs = 1273; goto _test_eof;
	_test_eof772: cs = 772; goto _test_eof;
	_test_eof773: cs = 773; goto _test_eof;
	_test_eof774: cs = 774; goto _test_eof;
	_test_eof1274: cs = 1274; goto _test_eof;
	_test_eof1275: cs = 1275; goto _test_eof;
	_test_eof775: cs = 775; goto _test_eof;
	_test_eof776: cs = 776; goto _test_eof;
	_test_eof777: cs = 777; goto _test_eof;
	_test_eof778: cs = 778; goto _test_eof;
	_test_eof1276: cs = 1276; goto _test_eof;
	_test_eof1277: cs = 1277; goto _test_eof;
	_test_eof779: cs = 779; goto _test_eof;
	_test_eof780: cs = 780; goto _test_eof;
	_test_eof781: cs = 781; goto _test_eof;
	_test_eof782: cs = 782; goto _test_eof;
	_test_eof1278: cs = 1278; goto _test_eof;
	_test_eof1279: cs = 1279; goto _test_eof;
	_test_eof783: cs = 783; goto _test_eof;
	_test_eof784: cs = 784; goto _test_eof;
	_test_eof785: cs = 785; goto _test_eof;
	_test_eof786: cs = 786; goto _test_eof;
	_test_eof787: cs = 787; goto _test_eof;
	_test_eof788: cs = 788; goto _test_eof;
	_test_eof789: cs = 789; goto _test_eof;
	_test_eof790: cs = 790; goto _test_eof;
	_test_eof791: cs = 791; goto _test_eof;
	_test_eof792: cs = 792; goto _test_eof;
	_test_eof793: cs = 793; goto _test_eof;
	_test_eof794: cs = 794; goto _test_eof;
	_test_eof795: cs = 795; goto _test_eof;
	_test_eof796: cs = 796; goto _test_eof;
	_test_eof797: cs = 797; goto _test_eof;
	_test_eof798: cs = 798; goto _test_eof;
	_test_eof799: cs = 799; goto _test_eof;
	_test_eof800: cs = 800; goto _test_eof;
	_test_eof801: cs = 801; goto _test_eof;
	_test_eof802: cs = 802; goto _test_eof;
	_test_eof803: cs = 803; goto _test_eof;
	_test_eof804: cs = 804; goto _test_eof;
	_test_eof805: cs = 805; goto _test_eof;
	_test_eof806: cs = 806; goto _test_eof;
	_test_eof807: cs = 807; goto _test_eof;
	_test_eof1280: cs = 1280; goto _test_eof;
	_test_eof808: cs = 808; goto _test_eof;
	_test_eof809: cs = 809; goto _test_eof;
	_test_eof810: cs = 810; goto _test_eof;
	_test_eof811: cs = 811; goto _test_eof;
	_test_eof812: cs = 812; goto _test_eof;
	_test_eof813: cs = 813; goto _test_eof;
	_test_eof814: cs = 814; goto _test_eof;
	_test_eof815: cs = 815; goto _test_eof;
	_test_eof816: cs = 816; goto _test_eof;
	_test_eof817: cs = 817; goto _test_eof;
	_test_eof818: cs = 818; goto _test_eof;
	_test_eof819: cs = 819; goto _test_eof;
	_test_eof820: cs = 820; goto _test_eof;
	_test_eof821: cs = 821; goto _test_eof;
	_test_eof822: cs = 822; goto _test_eof;
	_test_eof1281: cs = 1281; goto _test_eof;
	_test_eof823: cs = 823; goto _test_eof;
	_test_eof824: cs = 824; goto _test_eof;
	_test_eof825: cs = 825; goto _test_eof;
	_test_eof826: cs = 826; goto _test_eof;
	_test_eof827: cs = 827; goto _test_eof;
	_test_eof828: cs = 828; goto _test_eof;
	_test_eof829: cs = 829; goto _test_eof;
	_test_eof830: cs = 830; goto _test_eof;
	_test_eof831: cs = 831; goto _test_eof;
	_test_eof832: cs = 832; goto _test_eof;
	_test_eof833: cs = 833; goto _test_eof;
	_test_eof834: cs = 834; goto _test_eof;
	_test_eof835: cs = 835; goto _test_eof;
	_test_eof1282: cs = 1282; goto _test_eof;
	_test_eof836: cs = 836; goto _test_eof;
	_test_eof837: cs = 837; goto _test_eof;
	_test_eof838: cs = 838; goto _test_eof;
	_test_eof839: cs = 839; goto _test_eof;
	_test_eof840: cs = 840; goto _test_eof;
	_test_eof841: cs = 841; goto _test_eof;
	_test_eof842: cs = 842; goto _test_eof;
	_test_eof843: cs = 843; goto _test_eof;
	_test_eof844: cs = 844; goto _test_eof;
	_test_eof845: cs = 845; goto _test_eof;
	_test_eof846: cs = 846; goto _test_eof;
	_test_eof1283: cs = 1283; goto _test_eof;
	_test_eof1284: cs = 1284; goto _test_eof;
	_test_eof847: cs = 847; goto _test_eof;
	_test_eof848: cs = 848; goto _test_eof;
	_test_eof849: cs = 849; goto _test_eof;
	_test_eof1285: cs = 1285; goto _test_eof;
	_test_eof850: cs = 850; goto _test_eof;
	_test_eof851: cs = 851; goto _test_eof;
	_test_eof852: cs = 852; goto _test_eof;
	_test_eof853: cs = 853; goto _test_eof;
	_test_eof854: cs = 854; goto _test_eof;
	_test_eof855: cs = 855; goto _test_eof;
	_test_eof856: cs = 856; goto _test_eof;
	_test_eof857: cs = 857; goto _test_eof;
	_test_eof858: cs = 858; goto _test_eof;
	_test_eof859: cs = 859; goto _test_eof;
	_test_eof1286: cs = 1286; goto _test_eof;
	_test_eof1287: cs = 1287; goto _test_eof;
	_test_eof1288: cs = 1288; goto _test_eof;
	_test_eof860: cs = 860; goto _test_eof;
	_test_eof861: cs = 861; goto _test_eof;
	_test_eof862: cs = 862; goto _test_eof;
	_test_eof863: cs = 863; goto _test_eof;
	_test_eof864: cs = 864; goto _test_eof;
	_test_eof865: cs = 865; goto _test_eof;
	_test_eof866: cs = 866; goto _test_eof;
	_test_eof867: cs = 867; goto _test_eof;
	_test_eof868: cs = 868; goto _test_eof;
	_test_eof869: cs = 869; goto _test_eof;
	_test_eof870: cs = 870; goto _test_eof;
	_test_eof1289: cs = 1289; goto _test_eof;
	_test_eof1290: cs = 1290; goto _test_eof;
	_test_eof1291: cs = 1291; goto _test_eof;
	_test_eof871: cs = 871; goto _test_eof;
	_test_eof872: cs = 872; goto _test_eof;
	_test_eof873: cs = 873; goto _test_eof;
	_test_eof874: cs = 874; goto _test_eof;
	_test_eof875: cs = 875; goto _test_eof;
	_test_eof876: cs = 876; goto _test_eof;
	_test_eof877: cs = 877; goto _test_eof;
	_test_eof878: cs = 878; goto _test_eof;
	_test_eof879: cs = 879; goto _test_eof;
	_test_eof880: cs = 880; goto _test_eof;
	_test_eof881: cs = 881; goto _test_eof;
	_test_eof882: cs = 882; goto _test_eof;
	_test_eof1292: cs = 1292; goto _test_eof;
	_test_eof883: cs = 883; goto _test_eof;
	_test_eof884: cs = 884; goto _test_eof;
	_test_eof885: cs = 885; goto _test_eof;
	_test_eof1293: cs = 1293; goto _test_eof;
	_test_eof1294: cs = 1294; goto _test_eof;
	_test_eof886: cs = 886; goto _test_eof;
	_test_eof1295: cs = 1295; goto _test_eof;
	_test_eof1296: cs = 1296; goto _test_eof;
	_test_eof887: cs = 887; goto _test_eof;
	_test_eof1297: cs = 1297; goto _test_eof;
	_test_eof1298: cs = 1298; goto _test_eof;
	_test_eof888: cs = 888; goto _test_eof;
	_test_eof889: cs = 889; goto _test_eof;
	_test_eof890: cs = 890; goto _test_eof;
	_test_eof891: cs = 891; goto _test_eof;
	_test_eof892: cs = 892; goto _test_eof;
	_test_eof893: cs = 893; goto _test_eof;
	_test_eof894: cs = 894; goto _test_eof;
	_test_eof895: cs = 895; goto _test_eof;
	_test_eof896: cs = 896; goto _test_eof;
	_test_eof897: cs = 897; goto _test_eof;
	_test_eof898: cs = 898; goto _test_eof;
	_test_eof899: cs = 899; goto _test_eof;
	_test_eof900: cs = 900; goto _test_eof;
	_test_eof901: cs = 901; goto _test_eof;
	_test_eof902: cs = 902; goto _test_eof;
	_test_eof903: cs = 903; goto _test_eof;
	_test_eof904: cs = 904; goto _test_eof;
	_test_eof905: cs = 905; goto _test_eof;
	_test_eof906: cs = 906; goto _test_eof;
	_test_eof907: cs = 907; goto _test_eof;
	_test_eof908: cs = 908; goto _test_eof;
	_test_eof909: cs = 909; goto _test_eof;
	_test_eof910: cs = 910; goto _test_eof;
	_test_eof911: cs = 911; goto _test_eof;
	_test_eof912: cs = 912; goto _test_eof;
	_test_eof913: cs = 913; goto _test_eof;
	_test_eof914: cs = 914; goto _test_eof;
	_test_eof915: cs = 915; goto _test_eof;
	_test_eof916: cs = 916; goto _test_eof;
	_test_eof917: cs = 917; goto _test_eof;
	_test_eof918: cs = 918; goto _test_eof;
	_test_eof919: cs = 919; goto _test_eof;
	_test_eof920: cs = 920; goto _test_eof;
	_test_eof921: cs = 921; goto _test_eof;
	_test_eof922: cs = 922; goto _test_eof;
	_test_eof923: cs = 923; goto _test_eof;
	_test_eof924: cs = 924; goto _test_eof;
	_test_eof925: cs = 925; goto _test_eof;
	_test_eof926: cs = 926; goto _test_eof;
	_test_eof927: cs = 927; goto _test_eof;
	_test_eof928: cs = 928; goto _test_eof;
	_test_eof929: cs = 929; goto _test_eof;
	_test_eof1299: cs = 1299; goto _test_eof;
	_test_eof930: cs = 930; goto _test_eof;
	_test_eof931: cs = 931; goto _test_eof;
	_test_eof932: cs = 932; goto _test_eof;
	_test_eof933: cs = 933; goto _test_eof;
	_test_eof934: cs = 934; goto _test_eof;
	_test_eof935: cs = 935; goto _test_eof;
	_test_eof936: cs = 936; goto _test_eof;
	_test_eof937: cs = 937; goto _test_eof;
	_test_eof938: cs = 938; goto _test_eof;
	_test_eof939: cs = 939; goto _test_eof;
	_test_eof940: cs = 940; goto _test_eof;
	_test_eof941: cs = 941; goto _test_eof;
	_test_eof942: cs = 942; goto _test_eof;
	_test_eof943: cs = 943; goto _test_eof;
	_test_eof944: cs = 944; goto _test_eof;
	_test_eof945: cs = 945; goto _test_eof;
	_test_eof946: cs = 946; goto _test_eof;
	_test_eof947: cs = 947; goto _test_eof;
	_test_eof948: cs = 948; goto _test_eof;
	_test_eof949: cs = 949; goto _test_eof;
	_test_eof950: cs = 950; goto _test_eof;
	_test_eof951: cs = 951; goto _test_eof;
	_test_eof952: cs = 952; goto _test_eof;
	_test_eof953: cs = 953; goto _test_eof;
	_test_eof954: cs = 954; goto _test_eof;
	_test_eof955: cs = 955; goto _test_eof;
	_test_eof956: cs = 956; goto _test_eof;
	_test_eof957: cs = 957; goto _test_eof;
	_test_eof958: cs = 958; goto _test_eof;
	_test_eof959: cs = 959; goto _test_eof;
	_test_eof960: cs = 960; goto _test_eof;
	_test_eof961: cs = 961; goto _test_eof;
	_test_eof962: cs = 962; goto _test_eof;
	_test_eof963: cs = 963; goto _test_eof;
	_test_eof964: cs = 964; goto _test_eof;
	_test_eof965: cs = 965; goto _test_eof;
	_test_eof966: cs = 966; goto _test_eof;
	_test_eof967: cs = 967; goto _test_eof;
	_test_eof968: cs = 968; goto _test_eof;
	_test_eof969: cs = 969; goto _test_eof;
	_test_eof970: cs = 970; goto _test_eof;
	_test_eof971: cs = 971; goto _test_eof;
	_test_eof972: cs = 972; goto _test_eof;
	_test_eof973: cs = 973; goto _test_eof;
	_test_eof974: cs = 974; goto _test_eof;
	_test_eof975: cs = 975; goto _test_eof;
	_test_eof976: cs = 976; goto _test_eof;
	_test_eof977: cs = 977; goto _test_eof;
	_test_eof978: cs = 978; goto _test_eof;
	_test_eof979: cs = 979; goto _test_eof;
	_test_eof980: cs = 980; goto _test_eof;
	_test_eof981: cs = 981; goto _test_eof;
	_test_eof982: cs = 982; goto _test_eof;
	_test_eof983: cs = 983; goto _test_eof;
	_test_eof984: cs = 984; goto _test_eof;
	_test_eof985: cs = 985; goto _test_eof;
	_test_eof986: cs = 986; goto _test_eof;
	_test_eof987: cs = 987; goto _test_eof;
	_test_eof988: cs = 988; goto _test_eof;
	_test_eof989: cs = 989; goto _test_eof;
	_test_eof990: cs = 990; goto _test_eof;
	_test_eof991: cs = 991; goto _test_eof;
	_test_eof992: cs = 992; goto _test_eof;
	_test_eof993: cs = 993; goto _test_eof;
	_test_eof994: cs = 994; goto _test_eof;
	_test_eof995: cs = 995; goto _test_eof;
	_test_eof996: cs = 996; goto _test_eof;
	_test_eof997: cs = 997; goto _test_eof;
	_test_eof998: cs = 998; goto _test_eof;
	_test_eof999: cs = 999; goto _test_eof;
	_test_eof1000: cs = 1000; goto _test_eof;
	_test_eof1001: cs = 1001; goto _test_eof;
	_test_eof1002: cs = 1002; goto _test_eof;
	_test_eof1003: cs = 1003; goto _test_eof;
	_test_eof1004: cs = 1004; goto _test_eof;
	_test_eof1005: cs = 1005; goto _test_eof;
	_test_eof1006: cs = 1006; goto _test_eof;
	_test_eof1007: cs = 1007; goto _test_eof;
	_test_eof1008: cs = 1008; goto _test_eof;
	_test_eof1009: cs = 1009; goto _test_eof;
	_test_eof1010: cs = 1010; goto _test_eof;
	_test_eof1011: cs = 1011; goto _test_eof;
	_test_eof1012: cs = 1012; goto _test_eof;
	_test_eof1013: cs = 1013; goto _test_eof;
	_test_eof1014: cs = 1014; goto _test_eof;
	_test_eof1015: cs = 1015; goto _test_eof;
	_test_eof1016: cs = 1016; goto _test_eof;
	_test_eof1017: cs = 1017; goto _test_eof;
	_test_eof1018: cs = 1018; goto _test_eof;
	_test_eof1019: cs = 1019; goto _test_eof;
	_test_eof1020: cs = 1020; goto _test_eof;
	_test_eof1021: cs = 1021; goto _test_eof;
	_test_eof1022: cs = 1022; goto _test_eof;
	_test_eof1023: cs = 1023; goto _test_eof;
	_test_eof1024: cs = 1024; goto _test_eof;
	_test_eof1025: cs = 1025; goto _test_eof;
	_test_eof1026: cs = 1026; goto _test_eof;
	_test_eof1027: cs = 1027; goto _test_eof;
	_test_eof1028: cs = 1028; goto _test_eof;
	_test_eof1029: cs = 1029; goto _test_eof;
	_test_eof1030: cs = 1030; goto _test_eof;
	_test_eof1031: cs = 1031; goto _test_eof;
	_test_eof1032: cs = 1032; goto _test_eof;
	_test_eof1033: cs = 1033; goto _test_eof;
	_test_eof1034: cs = 1034; goto _test_eof;
	_test_eof1035: cs = 1035; goto _test_eof;
	_test_eof1036: cs = 1036; goto _test_eof;
	_test_eof1037: cs = 1037; goto _test_eof;
	_test_eof1038: cs = 1038; goto _test_eof;
	_test_eof1039: cs = 1039; goto _test_eof;
	_test_eof1040: cs = 1040; goto _test_eof;
	_test_eof1041: cs = 1041; goto _test_eof;
	_test_eof1042: cs = 1042; goto _test_eof;
	_test_eof1043: cs = 1043; goto _test_eof;
	_test_eof1044: cs = 1044; goto _test_eof;
	_test_eof1045: cs = 1045; goto _test_eof;
	_test_eof1046: cs = 1046; goto _test_eof;
	_test_eof1047: cs = 1047; goto _test_eof;
	_test_eof1048: cs = 1048; goto _test_eof;
	_test_eof1049: cs = 1049; goto _test_eof;
	_test_eof1050: cs = 1050; goto _test_eof;
	_test_eof1051: cs = 1051; goto _test_eof;
	_test_eof1052: cs = 1052; goto _test_eof;
	_test_eof1053: cs = 1053; goto _test_eof;
	_test_eof1054: cs = 1054; goto _test_eof;
	_test_eof1055: cs = 1055; goto _test_eof;
	_test_eof1056: cs = 1056; goto _test_eof;
	_test_eof1057: cs = 1057; goto _test_eof;
	_test_eof1058: cs = 1058; goto _test_eof;
	_test_eof1059: cs = 1059; goto _test_eof;
	_test_eof1060: cs = 1060; goto _test_eof;
	_test_eof1061: cs = 1061; goto _test_eof;
	_test_eof1062: cs = 1062; goto _test_eof;
	_test_eof1063: cs = 1063; goto _test_eof;
	_test_eof1064: cs = 1064; goto _test_eof;
	_test_eof1065: cs = 1065; goto _test_eof;
	_test_eof1066: cs = 1066; goto _test_eof;
	_test_eof1067: cs = 1067; goto _test_eof;
	_test_eof1068: cs = 1068; goto _test_eof;
	_test_eof1069: cs = 1069; goto _test_eof;
	_test_eof1070: cs = 1070; goto _test_eof;
	_test_eof1071: cs = 1071; goto _test_eof;
	_test_eof1072: cs = 1072; goto _test_eof;
	_test_eof1073: cs = 1073; goto _test_eof;
	_test_eof1074: cs = 1074; goto _test_eof;
	_test_eof1075: cs = 1075; goto _test_eof;
	_test_eof1076: cs = 1076; goto _test_eof;
	_test_eof1077: cs = 1077; goto _test_eof;
	_test_eof1078: cs = 1078; goto _test_eof;
	_test_eof1079: cs = 1079; goto _test_eof;
	_test_eof1080: cs = 1080; goto _test_eof;
	_test_eof1081: cs = 1081; goto _test_eof;
	_test_eof1082: cs = 1082; goto _test_eof;
	_test_eof1083: cs = 1083; goto _test_eof;
	_test_eof1084: cs = 1084; goto _test_eof;
	_test_eof1085: cs = 1085; goto _test_eof;
	_test_eof1086: cs = 1086; goto _test_eof;
	_test_eof1087: cs = 1087; goto _test_eof;
	_test_eof1088: cs = 1088; goto _test_eof;
	_test_eof1089: cs = 1089; goto _test_eof;
	_test_eof1090: cs = 1090; goto _test_eof;
	_test_eof1091: cs = 1091; goto _test_eof;
	_test_eof1092: cs = 1092; goto _test_eof;
	_test_eof1093: cs = 1093; goto _test_eof;
	_test_eof1300: cs = 1300; goto _test_eof;
	_test_eof1094: cs = 1094; goto _test_eof;
	_test_eof1095: cs = 1095; goto _test_eof;
	_test_eof1096: cs = 1096; goto _test_eof;
	_test_eof1097: cs = 1097; goto _test_eof;
	_test_eof1098: cs = 1098; goto _test_eof;
	_test_eof1099: cs = 1099; goto _test_eof;
	_test_eof1100: cs = 1100; goto _test_eof;
	_test_eof1101: cs = 1101; goto _test_eof;
	_test_eof1301: cs = 1301; goto _test_eof;
	_test_eof1102: cs = 1102; goto _test_eof;
	_test_eof1103: cs = 1103; goto _test_eof;
	_test_eof1104: cs = 1104; goto _test_eof;
	_test_eof1105: cs = 1105; goto _test_eof;
	_test_eof1106: cs = 1106; goto _test_eof;
	_test_eof1302: cs = 1302; goto _test_eof;
	_test_eof1107: cs = 1107; goto _test_eof;
	_test_eof1108: cs = 1108; goto _test_eof;
	_test_eof1109: cs = 1109; goto _test_eof;
	_test_eof1110: cs = 1110; goto _test_eof;
	_test_eof1111: cs = 1111; goto _test_eof;
	_test_eof1112: cs = 1112; goto _test_eof;
	_test_eof1113: cs = 1113; goto _test_eof;
	_test_eof1114: cs = 1114; goto _test_eof;
	_test_eof1115: cs = 1115; goto _test_eof;
	_test_eof1116: cs = 1116; goto _test_eof;
	_test_eof1117: cs = 1117; goto _test_eof;
	_test_eof1118: cs = 1118; goto _test_eof;
	_test_eof1119: cs = 1119; goto _test_eof;
	_test_eof1120: cs = 1120; goto _test_eof;
	_test_eof1121: cs = 1121; goto _test_eof;
	_test_eof1122: cs = 1122; goto _test_eof;
	_test_eof1123: cs = 1123; goto _test_eof;
	_test_eof1124: cs = 1124; goto _test_eof;
	_test_eof1303: cs = 1303; goto _test_eof;
	_test_eof1125: cs = 1125; goto _test_eof;
	_test_eof1126: cs = 1126; goto _test_eof;
	_test_eof1127: cs = 1127; goto _test_eof;
	_test_eof1128: cs = 1128; goto _test_eof;
	_test_eof1129: cs = 1129; goto _test_eof;
	_test_eof1130: cs = 1130; goto _test_eof;
	_test_eof1131: cs = 1131; goto _test_eof;
	_test_eof1132: cs = 1132; goto _test_eof;
	_test_eof1133: cs = 1133; goto _test_eof;
	_test_eof1134: cs = 1134; goto _test_eof;
	_test_eof1135: cs = 1135; goto _test_eof;
	_test_eof1136: cs = 1136; goto _test_eof;
	_test_eof1137: cs = 1137; goto _test_eof;
	_test_eof1138: cs = 1138; goto _test_eof;
	_test_eof1139: cs = 1139; goto _test_eof;
	_test_eof1140: cs = 1140; goto _test_eof;
	_test_eof1141: cs = 1141; goto _test_eof;
	_test_eof1142: cs = 1142; goto _test_eof;
	_test_eof1143: cs = 1143; goto _test_eof;
	_test_eof1304: cs = 1304; goto _test_eof;
	_test_eof1144: cs = 1144; goto _test_eof;
	_test_eof1145: cs = 1145; goto _test_eof;
	_test_eof1146: cs = 1146; goto _test_eof;
	_test_eof1147: cs = 1147; goto _test_eof;
	_test_eof1148: cs = 1148; goto _test_eof;
	_test_eof1149: cs = 1149; goto _test_eof;
	_test_eof1150: cs = 1150; goto _test_eof;
	_test_eof1151: cs = 1151; goto _test_eof;
	_test_eof1152: cs = 1152; goto _test_eof;
	_test_eof1153: cs = 1153; goto _test_eof;
	_test_eof1154: cs = 1154; goto _test_eof;
	_test_eof1155: cs = 1155; goto _test_eof;
	_test_eof1156: cs = 1156; goto _test_eof;
	_test_eof1157: cs = 1157; goto _test_eof;
	_test_eof1158: cs = 1158; goto _test_eof;
	_test_eof1305: cs = 1305; goto _test_eof;
	_test_eof1306: cs = 1306; goto _test_eof;
	_test_eof1307: cs = 1307; goto _test_eof;
	_test_eof1159: cs = 1159; goto _test_eof;
	_test_eof1160: cs = 1160; goto _test_eof;
	_test_eof1161: cs = 1161; goto _test_eof;
	_test_eof1162: cs = 1162; goto _test_eof;
	_test_eof1163: cs = 1163; goto _test_eof;
	_test_eof1164: cs = 1164; goto _test_eof;
	_test_eof1165: cs = 1165; goto _test_eof;
	_test_eof1166: cs = 1166; goto _test_eof;
	_test_eof1308: cs = 1308; goto _test_eof;
	_test_eof1167: cs = 1167; goto _test_eof;
	_test_eof1168: cs = 1168; goto _test_eof;
	_test_eof1169: cs = 1169; goto _test_eof;
	_test_eof1170: cs = 1170; goto _test_eof;
	_test_eof1171: cs = 1171; goto _test_eof;
	_test_eof1172: cs = 1172; goto _test_eof;
	_test_eof1173: cs = 1173; goto _test_eof;
	_test_eof1174: cs = 1174; goto _test_eof;
	_test_eof1175: cs = 1175; goto _test_eof;
	_test_eof1176: cs = 1176; goto _test_eof;
	_test_eof1177: cs = 1177; goto _test_eof;
	_test_eof1309: cs = 1309; goto _test_eof;
	_test_eof1310: cs = 1310; goto _test_eof;
	_test_eof1311: cs = 1311; goto _test_eof;
	_test_eof1178: cs = 1178; goto _test_eof;
	_test_eof1179: cs = 1179; goto _test_eof;
	_test_eof1180: cs = 1180; goto _test_eof;
	_test_eof1181: cs = 1181; goto _test_eof;
	_test_eof1182: cs = 1182; goto _test_eof;
	_test_eof1183: cs = 1183; goto _test_eof;
	_test_eof1184: cs = 1184; goto _test_eof;
	_test_eof1312: cs = 1312; goto _test_eof;
	_test_eof1185: cs = 1185; goto _test_eof;
	_test_eof1186: cs = 1186; goto _test_eof;
	_test_eof1187: cs = 1187; goto _test_eof;
	_test_eof1188: cs = 1188; goto _test_eof;
	_test_eof1189: cs = 1189; goto _test_eof;
	_test_eof1190: cs = 1190; goto _test_eof;
	_test_eof1191: cs = 1191; goto _test_eof;
	_test_eof1192: cs = 1192; goto _test_eof;
	_test_eof1193: cs = 1193; goto _test_eof;
	_test_eof1194: cs = 1194; goto _test_eof;
	_test_eof1195: cs = 1195; goto _test_eof;
	_test_eof1196: cs = 1196; goto _test_eof;
	_test_eof1197: cs = 1197; goto _test_eof;
	_test_eof1313: cs = 1313; goto _test_eof;
	_test_eof1198: cs = 1198; goto _test_eof;
	_test_eof1199: cs = 1199; goto _test_eof;
	_test_eof1200: cs = 1200; goto _test_eof;
	_test_eof1201: cs = 1201; goto _test_eof;
	_test_eof1202: cs = 1202; goto _test_eof;
	_test_eof1203: cs = 1203; goto _test_eof;
	_test_eof1204: cs = 1204; goto _test_eof;
	_test_eof1314: cs = 1314; goto _test_eof;
	_test_eof1205: cs = 1205; goto _test_eof;
	_test_eof1206: cs = 1206; goto _test_eof;
	_test_eof1207: cs = 1207; goto _test_eof;
	_test_eof1208: cs = 1208; goto _test_eof;
	_test_eof1209: cs = 1209; goto _test_eof;
	_test_eof1210: cs = 1210; goto _test_eof;
	_test_eof1315: cs = 1315; goto _test_eof;
	_test_eof1211: cs = 1211; goto _test_eof;
	_test_eof1212: cs = 1212; goto _test_eof;
	_test_eof1213: cs = 1213; goto _test_eof;
	_test_eof1214: cs = 1214; goto _test_eof;
	_test_eof1215: cs = 1215; goto _test_eof;
	_test_eof1216: cs = 1216; goto _test_eof;
	_test_eof1316: cs = 1316; goto _test_eof;
	_test_eof1217: cs = 1217; goto _test_eof;
	_test_eof1218: cs = 1218; goto _test_eof;
	_test_eof1219: cs = 1219; goto _test_eof;
	_test_eof1220: cs = 1220; goto _test_eof;
	_test_eof1221: cs = 1221; goto _test_eof;
	_test_eof1222: cs = 1222; goto _test_eof;
	_test_eof1223: cs = 1223; goto _test_eof;
	_test_eof1224: cs = 1224; goto _test_eof;
	_test_eof1317: cs = 1317; goto _test_eof;
	_test_eof1225: cs = 1225; goto _test_eof;
	_test_eof1226: cs = 1226; goto _test_eof;

	_test_eof: {}
	if ( p == eof )
	{
	switch ( cs ) {
	case 7:
	case 8:
	case 176:
	case 200:
	case 202:
	case 256:
	case 270:
	case 295:
	case 349:
	case 350:
	case 356:
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 301:
	case 302:
	case 303:
	case 308:
	case 309:
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 185:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 317:
	{
		WARN(ZS_BAD_TEXT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 322:
	case 330:
	{
		ERR(ZS_BAD_TTL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 331:
	case 332:
	case 333:
	case 342:
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 344:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 347:
	case 357:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1231:
	{
		NOERR;
	}
	break;
	case 177:
	case 178:
	case 179:
	case 180:
	case 181:
	case 182:
	case 183:
	case 184:
	case 222:
	case 223:
	case 224:
	case 225:
	case 226:
	case 227:
	case 228:
	case 229:
	case 230:
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 360:
	case 361:
	case 362:
	case 363:
	case 364:
	case 365:
	case 366:
	{
		WARN(ZS_BAD_BASE64_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 367:
	case 368:
	case 369:
	case 370:
	case 371:
	case 372:
	case 373:
	case 374:
	case 375:
	case 376:
	case 377:
	case 378:
	case 379:
	case 380:
	case 381:
	case 382:
	case 383:
	case 384:
	case 385:
	case 386:
	case 387:
	case 388:
	case 389:
	case 390:
	case 391:
	case 392:
	case 393:
	case 394:
	case 395:
	case 396:
	case 397:
	case 398:
	case 399:
	case 400:
	case 401:
	case 402:
	case 403:
	case 404:
	case 405:
	case 406:
	case 407:
	case 408:
	case 409:
	case 410:
	case 411:
	case 412:
	case 413:
	case 414:
	case 415:
	case 416:
	case 417:
	case 418:
	case 419:
	case 420:
	case 421:
	case 422:
	case 423:
	case 424:
	case 425:
	case 426:
	case 427:
	case 428:
	case 429:
	case 430:
	case 431:
	case 432:
	case 433:
	case 434:
	case 435:
	case 436:
	case 437:
	case 438:
	case 439:
	case 440:
	case 441:
	case 442:
	case 443:
	case 444:
	case 445:
	case 446:
	case 447:
	case 448:
	case 449:
	case 450:
	case 451:
	case 452:
	case 453:
	case 454:
	case 455:
	case 456:
	case 457:
	case 458:
	case 459:
	case 460:
	case 461:
	case 462:
	case 463:
	case 464:
	case 465:
	case 466:
	case 467:
	case 468:
	case 469:
	case 470:
	case 471:
	case 472:
	case 473:
	case 474:
	case 475:
	case 476:
	case 477:
	case 478:
	case 479:
	case 480:
	case 481:
	case 482:
	case 483:
	case 484:
	case 485:
	case 486:
	case 487:
	case 488:
	case 489:
	case 490:
	case 491:
	case 492:
	case 493:
	case 494:
	case 495:
	case 496:
	case 497:
	case 498:
	case 499:
	case 500:
	case 501:
	case 502:
	case 503:
	case 504:
	case 505:
	case 506:
	case 507:
	case 508:
	case 509:
	case 510:
	case 511:
	case 512:
	case 513:
	case 514:
	case 515:
	case 516:
	case 517:
	case 518:
	case 519:
	case 520:
	case 521:
	case 522:
	case 523:
	{
		WARN(ZS_BAD_BITMAP);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 524:
	case 531:
	case 532:
	case 533:
	case 540:
	case 542:
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 543:
	case 544:
	case 545:
	case 546:
	case 547:
	case 548:
	case 549:
	case 550:
	case 551:
	case 552:
	case 553:
	case 554:
	case 555:
	case 556:
	case 557:
	case 558:
	case 559:
	case 560:
	case 561:
	case 562:
	case 563:
	case 564:
	case 565:
	case 566:
	case 567:
	case 568:
	case 569:
	case 570:
	case 571:
	case 572:
	case 573:
	case 574:
	case 575:
	case 576:
	case 577:
	case 578:
	case 579:
	case 580:
	case 581:
	case 582:
	case 583:
	case 584:
	case 585:
	case 586:
	case 587:
	case 588:
	case 589:
	case 590:
	case 591:
	case 592:
	case 593:
	case 594:
	case 595:
	case 596:
	case 597:
	case 598:
	case 599:
	case 600:
	case 601:
	case 602:
	case 603:
	case 604:
	case 605:
	case 606:
	case 607:
	case 608:
	case 609:
	case 610:
	case 611:
	case 612:
	case 613:
	case 614:
	case 615:
	case 616:
	case 617:
	case 618:
	case 619:
	case 620:
	case 621:
	case 622:
	case 623:
	case 624:
	case 625:
	case 626:
	case 627:
	case 628:
	case 629:
	case 630:
	case 631:
	case 632:
	case 633:
	case 634:
	case 635:
	case 636:
	case 637:
	case 638:
	case 639:
	case 640:
	case 641:
	case 642:
	case 643:
	case 644:
	case 645:
	case 646:
	{
		WARN(ZS_BAD_ALGORITHM);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 647:
	case 648:
	case 649:
	case 650:
	case 651:
	case 652:
	case 653:
	case 654:
	case 655:
	case 656:
	case 657:
	case 658:
	case 659:
	case 660:
	case 661:
	case 662:
	case 663:
	case 664:
	case 665:
	case 666:
	case 667:
	case 668:
	case 669:
	case 670:
	case 671:
	case 672:
	case 673:
	case 674:
	case 675:
	case 676:
	case 677:
	case 678:
	case 679:
	case 680:
	case 681:
	case 682:
	case 683:
	case 684:
	case 685:
	case 686:
	case 687:
	{
		WARN(ZS_BAD_CERT_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 5:
	case 9:
	case 10:
	case 275:
	case 690:
	case 691:
	case 692:
	case 693:
	case 694:
	case 695:
	case 709:
	case 713:
	case 717:
	case 721:
	case 722:
	case 723:
	case 724:
	case 725:
	case 726:
	case 727:
	case 728:
	case 729:
	case 730:
	case 731:
	case 732:
	case 733:
	case 736:
	case 737:
	case 738:
	case 806:
	case 807:
	case 808:
	case 809:
	case 810:
	case 815:
	case 816:
	case 817:
	case 818:
	case 819:
	case 820:
	case 821:
	case 822:
	case 823:
	case 824:
	case 825:
	case 826:
	case 827:
	case 828:
	case 829:
	case 832:
	case 833:
	case 834:
	case 835:
	case 836:
	case 837:
	case 838:
	case 846:
	case 852:
	case 853:
	case 860:
	case 861:
	case 862:
	case 871:
	case 872:
	case 911:
	case 914:
	case 915:
	case 926:
	case 927:
	case 928:
	case 929:
	case 930:
	case 931:
	case 932:
	case 933:
	case 934:
	case 935:
	case 936:
	case 937:
	case 1091:
	case 1092:
	case 1093:
	case 1098:
	case 1099:
	case 1100:
	case 1101:
	case 1102:
	case 1103:
	case 1104:
	case 1105:
	case 1106:
	case 1124:
	case 1130:
	case 1133:
	case 1134:
	case 1135:
	case 1146:
	case 1147:
	case 1148:
	case 1159:
	case 1160:
	case 1161:
	case 1166:
	case 1167:
	case 1178:
	case 1179:
	case 1180:
	case 1185:
	case 1198:
	case 1215:
	case 1216:
	case 1217:
	case 1218:
	case 1221:
	case 1222:
	case 1223:
	case 1224:
	case 1225:
	case 1226:
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 3:
	case 17:
	case 18:
	case 22:
	case 23:
	case 24:
	case 25:
	case 26:
	case 29:
	case 30:
	case 32:
	case 33:
	case 34:
	case 36:
	case 37:
	case 38:
	case 40:
	case 41:
	case 42:
	case 43:
	case 45:
	case 46:
	case 47:
	case 49:
	case 50:
	case 51:
	case 54:
	case 55:
	case 56:
	case 57:
	case 59:
	case 61:
	case 62:
	case 63:
	case 64:
	case 66:
	case 67:
	case 68:
	case 69:
	case 70:
	case 71:
	case 72:
	case 73:
	case 74:
	case 75:
	case 77:
	case 78:
	case 81:
	case 82:
	case 84:
	case 86:
	case 89:
	case 90:
	case 91:
	case 92:
	case 95:
	case 96:
	case 97:
	case 98:
	case 100:
	case 103:
	case 106:
	case 107:
	case 108:
	case 109:
	case 111:
	case 112:
	case 113:
	case 114:
	case 115:
	case 116:
	case 117:
	case 118:
	case 119:
	case 121:
	case 122:
	case 124:
	case 126:
	case 127:
	case 128:
	case 131:
	case 132:
	case 133:
	case 134:
	case 135:
	case 137:
	case 139:
	case 141:
	case 143:
	case 144:
	case 145:
	case 147:
	case 148:
	case 149:
	case 151:
	case 153:
	case 154:
	case 157:
	case 158:
	case 160:
	case 161:
	case 162:
	case 163:
	case 164:
	case 166:
	case 168:
	case 169:
	case 175:
	case 190:
	case 194:
	case 259:
	case 260:
	case 262:
	case 263:
	case 264:
	case 266:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 325:
	case 326:
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 334:
	case 335:
	case 336:
	case 341:
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 304:
	case 305:
	case 306:
	case 307:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 337:
	case 338:
	case 339:
	case 340:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 351:
	case 352:
	case 353:
	case 354:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 253:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 279:
	case 280:
	case 281:
	case 282:
	case 283:
	case 284:
	case 285:
	case 286:
	case 287:
	case 288:
	case 289:
	case 290:
	case 291:
	case 292:
	case 293:
	case 294:
	case 297:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 323:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 525:
	case 526:
	case 534:
	case 535:
	case 541:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 696:
	case 697:
	case 698:
	case 700:
	case 702:
	case 704:
	case 734:
	case 735:
	case 800:
	case 801:
	case 802:
	case 803:
	case 804:
	case 805:
	case 811:
	case 812:
	case 813:
	case 814:
	case 830:
	case 831:
	case 850:
	case 851:
	case 854:
	case 855:
	case 863:
	case 864:
	case 865:
	case 866:
	case 873:
	case 874:
	case 916:
	case 917:
	case 918:
	case 919:
	case 924:
	case 925:
	case 1094:
	case 1095:
	case 1096:
	case 1097:
	case 1107:
	case 1108:
	case 1109:
	case 1110:
	case 1111:
	case 1112:
	case 1136:
	case 1137:
	case 1138:
	case 1139:
	case 1140:
	case 1141:
	case 1149:
	case 1150:
	case 1151:
	case 1152:
	case 1153:
	case 1154:
	case 1162:
	case 1163:
	case 1164:
	case 1165:
	case 1168:
	case 1169:
	case 1170:
	case 1171:
	case 1172:
	case 1173:
	case 1181:
	case 1182:
	case 1186:
	case 1187:
	case 1211:
	case 1212:
	case 1213:
	case 1214:
	case 1219:
	case 1220:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 155:
	case 170:
	case 186:
	case 189:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 920:
	case 921:
	case 922:
	case 923:
	{
		WARN(ZS_BAD_TIMESTAMP_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 310:
	case 311:
	case 316:
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 739:
	case 740:
	case 741:
	case 742:
	{
		s->long_string = false;
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 345:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 346:
	case 348:
	case 355:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 688:
	case 689:
	case 743:
	case 744:
	case 1183:
	case 1184:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 839:
	case 840:
	case 841:
	case 844:
	case 845:
	case 847:
	{
		WARN(ZS_BAD_APL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 856:
	case 857:
	case 858:
	case 859:
	case 867:
	case 868:
	case 869:
	case 870:
	case 1113:
	case 1114:
	case 1131:
	case 1132:
	case 1142:
	case 1143:
	case 1144:
	case 1145:
	case 1155:
	case 1156:
	case 1157:
	case 1158:
	case 1174:
	case 1175:
	case 1176:
	case 1177:
	case 1188:
	case 1189:
	case 1190:
	case 1191:
	case 1193:
	case 1194:
	case 1195:
	case 1196:
	case 1199:
	case 1200:
	case 1202:
	case 1203:
	case 1205:
	case 1206:
	case 1208:
	case 1209:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1115:
	case 1116:
	case 1117:
	case 1118:
	case 1119:
	case 1120:
	case 1121:
	case 1122:
	case 1123:
	case 1125:
	case 1126:
	case 1127:
	case 1128:
	case 1129:
	{
		WARN(ZS_BAD_BASE32HEX_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 875:
	case 876:
	case 879:
	case 888:
	case 889:
	case 890:
	case 895:
	case 896:
	case 897:
	case 902:
	case 903:
	case 904:
	case 907:
	case 909:
	case 910:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 881:
	case 882:
	case 883:
	case 884:
	case 885:
	case 886:
	case 887:
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 912:
	case 913:
	case 938:
	case 939:
	case 940:
	case 941:
	case 942:
	case 943:
	case 944:
	case 945:
	case 946:
	case 947:
	case 948:
	case 949:
	case 950:
	case 951:
	case 952:
	case 953:
	case 954:
	case 955:
	case 956:
	case 957:
	case 958:
	case 959:
	case 960:
	case 961:
	case 962:
	case 963:
	case 964:
	case 965:
	case 966:
	case 967:
	case 968:
	case 969:
	case 970:
	case 971:
	case 972:
	case 973:
	case 974:
	case 975:
	case 976:
	case 977:
	case 978:
	case 979:
	case 980:
	case 981:
	case 982:
	case 983:
	case 984:
	case 985:
	case 986:
	case 987:
	case 988:
	case 989:
	case 990:
	case 991:
	case 992:
	case 993:
	case 994:
	case 995:
	case 996:
	case 997:
	case 998:
	case 999:
	case 1000:
	case 1001:
	case 1002:
	case 1003:
	case 1004:
	case 1005:
	case 1006:
	case 1007:
	case 1008:
	case 1009:
	case 1010:
	case 1011:
	case 1012:
	case 1013:
	case 1014:
	case 1015:
	case 1016:
	case 1017:
	case 1018:
	case 1019:
	case 1020:
	case 1021:
	case 1022:
	case 1023:
	case 1024:
	case 1025:
	case 1026:
	case 1027:
	case 1028:
	case 1029:
	case 1030:
	case 1031:
	case 1032:
	case 1033:
	case 1034:
	case 1035:
	case 1036:
	case 1037:
	case 1038:
	case 1039:
	case 1040:
	case 1041:
	case 1042:
	case 1043:
	case 1044:
	case 1045:
	case 1046:
	case 1047:
	case 1048:
	case 1049:
	case 1050:
	case 1051:
	case 1052:
	case 1053:
	case 1054:
	case 1055:
	case 1056:
	case 1057:
	case 1058:
	case 1059:
	case 1060:
	case 1061:
	case 1062:
	case 1063:
	case 1064:
	case 1065:
	case 1066:
	case 1067:
	case 1068:
	case 1069:
	case 1070:
	case 1071:
	case 1072:
	case 1073:
	case 1074:
	case 1075:
	case 1076:
	case 1077:
	case 1078:
	case 1079:
	case 1082:
	case 1083:
	case 1084:
	case 1085:
	case 1086:
	case 1087:
	case 1088:
	case 1089:
	case 1090:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 745:
	case 746:
	case 747:
	case 748:
	case 749:
	case 750:
	case 751:
	case 752:
	case 753:
	case 754:
	case 755:
	case 756:
	case 757:
	case 758:
	case 759:
	case 760:
	case 761:
	case 762:
	case 763:
	case 764:
	case 765:
	case 766:
	case 767:
	case 768:
	case 769:
	case 770:
	case 771:
	case 772:
	case 773:
	case 774:
	case 775:
	case 776:
	case 777:
	case 778:
	case 779:
	case 780:
	case 781:
	case 782:
	case 783:
	case 784:
	case 785:
	case 786:
	case 787:
	case 788:
	case 789:
	case 790:
	case 791:
	case 792:
	case 793:
	case 794:
	case 795:
	case 796:
	case 797:
	case 798:
	case 799:
	{
		WARN(ZS_BAD_LOC_DATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 6:
	case 11:
	case 12:
	case 13:
	case 20:
	case 21:
	case 198:
	case 201:
	case 220:
	case 255:
	case 257:
	case 258:
	case 268:
	case 269:
	case 271:
	case 272:
	case 273:
	case 274:
	case 277:
	case 278:
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 16:
	case 234:
	case 296:
	case 298:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 4:
	case 19:
	case 27:
	case 28:
	case 31:
	case 35:
	case 39:
	case 44:
	case 48:
	case 52:
	case 53:
	case 58:
	case 60:
	case 65:
	case 76:
	case 79:
	case 80:
	case 83:
	case 85:
	case 87:
	case 88:
	case 93:
	case 94:
	case 99:
	case 101:
	case 102:
	case 104:
	case 105:
	case 110:
	case 120:
	case 123:
	case 125:
	case 129:
	case 130:
	case 136:
	case 138:
	case 140:
	case 142:
	case 146:
	case 150:
	case 152:
	case 159:
	case 165:
	case 261:
	case 265:
	case 267:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 231:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 195:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 877:
	case 878:
	case 891:
	case 892:
	case 898:
	case 899:
	case 905:
	case 906:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1080:
	case 1081:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1:
	case 167:
	case 199:
	case 233:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 156:
	case 196:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 328:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 699:
	case 701:
	case 703:
	case 705:
	case 706:
	case 707:
	case 708:
	case 710:
	case 711:
	case 712:
	case 714:
	case 715:
	case 716:
	case 718:
	case 719:
	case 720:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 2:
	case 171:
	case 172:
	case 173:
	case 174:
	case 191:
	case 192:
	case 193:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 312:
	case 313:
	case 314:
	case 315:
	case 318:
	case 319:
	case 320:
	case 321:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 842:
	case 843:
	case 848:
	case 849:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_APL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 893:
	case 900:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 527:
	case 528:
	case 529:
	case 530:
	case 536:
	case 537:
	case 538:
	case 539:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1192:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 880:
	case 908:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1201:
	case 1204:
	case 1207:
	case 1210:
	{
		WARN(ZS_BAD_CHAR_DASH);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 1197:
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 221:
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 197:
	case 204:
	case 205:
	case 206:
	case 207:
	case 208:
	case 209:
	case 210:
	case 211:
	case 212:
	case 213:
	case 214:
	case 215:
	case 216:
	case 217:
	case 218:
	case 219:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 188:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 235:
	case 237:
	case 238:
	case 239:
	case 240:
	case 241:
	case 242:
	case 243:
	case 244:
	case 245:
	case 246:
	case 247:
	case 248:
	case 249:
	case 250:
	case 251:
	case 252:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 14:
	case 203:
	case 236:
	case 254:
	case 276:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 324:
	case 327:
	case 329:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 894:
	case 901:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 15:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	case 232:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {       if ( p == pe )
               goto _test_eof299;
goto st299;}
	}
	break;
	}
	}

	_out: {}
	}


	// Check if the scanner state machine is in an uncovered state.
	bool extra_error = false;
	if (cs == 0) {
		ERR(ZS_UNCOVERED_STATE);
		extra_error = true;
	// Check for an unclosed multiline record.
	} else if (s->input.eof && s->multiline) {
		ERR(ZS_UNCLOSED_MULTILINE);
		extra_error = true;
	}

	// Treat the extra error.
	if (extra_error) {
		s->error.counter++;
		s->state = ZS_STATE_ERROR;

		// Copy the error context just for the part of the current line.
		s->buffer_length = 0;
		while (p < pe && *p != '\n' && s->buffer_length < 50) {
			s->buffer[s->buffer_length++] = *p++;
		}
		s->buffer[s->buffer_length++] = 0;

		// Execute the error callback.
		if (s->process.automatic && s->process.error != NULL) {
			s->process.error(s);
		}

		return;
	}

	// Storing scanner states.
	s->cs  = cs;
	s->top = top;
	memcpy(s->stack, stack, sizeof(stack));

	// Store the current parser position.
	s->input.current = p;

	// Storing r_data pointer.
	s->r_data_tail = rdata_tail - s->r_data;

	if (*wrap == WRAP_DETECTED) {
		if (set_input_string(s, "\\", 1, true) != 0) {
			return;
		}

		*wrap = WRAP_PROCESS;
		parse(s, wrap);
	} else {
		*wrap = WRAP_NONE;
	}
}

__attribute__((visibility("default")))
int zs_parse_record(
	zs_scanner_t *s)
{
	if (s == NULL) {
		return -1;
	}

	// Check if parsing is possible.
	switch (s->state) {
	case ZS_STATE_NONE:
	case ZS_STATE_DATA:
	case ZS_STATE_INCLUDE:
		break;
	case ZS_STATE_ERROR:
		if (s->error.fatal) {
			return -1;
		}
		break;
	default:
		// Return if stop or end of file.
		return 0;
	}

	// Check for the end of the input.
	if (s->input.current != s->input.end) {
		// Try to parse another item.
		s->state = ZS_STATE_NONE;
		wrap_t wrap = WRAP_NONE;
		parse(s, &wrap);

		// Finish if nothing was parsed.
		if (s->state == ZS_STATE_NONE) {
			// Parse the final block.
			if (set_input_string(s, "\n", 1, true) != 0) {
				return -1;
			}
			parse(s, &wrap);
			if (s->state == ZS_STATE_NONE) {
				s->state = ZS_STATE_EOF;
			}
		}
	} else {
		s->state = ZS_STATE_EOF;
	}

	return 0;
}

__attribute__((visibility("default")))
int zs_parse_all(
	zs_scanner_t *s)
{
	if (s == NULL) {
		return -1;
	}

	s->process.automatic = true;

	// Parse input block.
	wrap_t wrap = WRAP_NONE;
	parse(s, &wrap);

	// Parse trailing newline-char block if it makes sense.
	if (s->state != ZS_STATE_STOP && !s->error.fatal) {
		if (set_input_string(s, "\n", 1, true) != 0) {
			return -1;
		}
		parse(s, &wrap);
	}

	// Check if any errors have occurred.
	if (s->error.counter > 0) {
		return -1;
	}

	return 0;
}
